/*
 * GStreamer
 * Copyright 2005 Hannes Bistry <bistry # informatik.uni-hamburg.de>
 * Copyright 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-plugin
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch -v -m audiotestsrc ! plugin ! fakesink silent=TRUE
 * </programlisting>
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include "gstadjuster.h"


GST_DEBUG_CATEGORY_STATIC (gst_adjuster_debug);
#define GST_CAT_DEFAULT gst_adjuster_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT,
  ARG_PIXEL
};
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) {10,12,14,16}, "
     "bpp = (int) 16, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 16, 1388 ], "
     "height = (int) [ 16, 1038 ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 16, 1388 ], "
     "height = (int) [ 16, 1038 ],"
     "depth = (int) 16")
    );


static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
     GST_STATIC_CAPS("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) {10,12,14,16}, "
     "bpp = (int) 16, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 16, 1388 ], "
     "height = (int) [ 16, 1038 ];"
     "video/x-raw-bayer, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 16, 1388 ], "
     "height = (int) [ 16, 1038 ],"
     "depth = (int) 16")
    ); 


GST_BOILERPLATE (GstAdjuster, gst_adjuster, GstElement,
    GST_TYPE_ELEMENT);

static void gst_adjuster_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_adjuster_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_adjuster_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_adjuster_chain (GstPad * pad, GstBuffer * buf);
static GstFlowReturn gst_adjuster_bufferalloc (GstPad * pad, guint64 offset,
    guint size, GstCaps * caps, GstBuffer ** buf);

static void
gst_adjuster_base_init (gpointer gclass)
{
  static GstElementDetails element_details = {
    "Adjuster",
    "Gst-Plugins-Robotix",
    "This plugin tries to adjust shutter and gain",
    "Hannes Bistry <hannesb # gmx.de>"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_adjuster_class_init (GstAdjusterClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_adjuster_set_property;
  gobject_class->get_property = gst_adjuster_get_property;

  g_object_class_install_property (gobject_class, ARG_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_PIXEL,
      g_param_spec_int ("pixel", "Pixel", "Which Pixel should be evaluated ?",
          0, G_MAXINT, 0,G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_adjuster_init (GstAdjuster * filter,
    GstAdjusterClass * gclass)
{
  GstElementClass *klass = GST_ELEMENT_GET_CLASS (filter);

  filter->sinkpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "sink"), "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_adjuster_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_bufferalloc_function (filter->sinkpad,
      GST_DEBUG_FUNCPTR (gst_adjuster_bufferalloc));

  filter->srcpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "src"), "src");
  gst_pad_set_setcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_adjuster_set_caps));
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_adjuster_chain));
  filter->silent = FALSE;
  filter->initialized = FALSE;
  filter->bpp=8;
  filter->bpc=8;
  filter->depth=8;
  filter->channels=1;
  filter->pixel=0;
}

static void
gst_adjuster_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstAdjuster *filter = GST_ADJUSTER (object);

  switch (prop_id) {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case ARG_PIXEL:
      filter->pixel = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_adjuster_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstAdjuster *filter = GST_ADJUSTER (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case ARG_PIXEL:
      g_value_set_int (value, filter->pixel);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static GstFlowReturn
gst_adjuster_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  GstAdjuster *filter;
  GstFlowReturn result;
  //g_print ("Bufferalloc has been caled with Caps: %s\n",gst_caps_to_string(caps));
  filter = GST_ADJUSTER (GST_PAD_PARENT (pad));

  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  return result;
}





/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_adjuster_set_caps (GstPad * pad, GstCaps * caps)
{
  GstAdjuster *filter;
  GstPad *otherpad;
  filter = GST_ADJUSTER (gst_pad_get_parent (pad));
  gchar * capschar;
  capschar=gst_caps_to_string(caps);
  if (pad == filter->srcpad){
  g_print ("Das Zielelement will am Format drehen: Caps: %s\n",capschar);
  }
  if (pad == filter->sinkpad){
  g_print ("Das Quellelement will am Format drehen: Caps %s\n",capschar);
  }
  g_free(capschar);

  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

  return gst_pad_set_caps (otherpad, caps);
}

/* chain function
 * this function does the actual processing
 */


static GstElement* findelement(GstElement * element, const gchar* propertyname, const gchar* elementname)//finds element in chain with property "propertyname" and elementname as a prefix of objectname in not NULL
//pad specifies the search direction
{
//GstAdjuster *filter;
//filter = GST_ADJUSTER (element);
GstElement * tempel=NULL;
GstElement *result=NULL;
gchar* name;

//GstPad * temppad;
//temppad=filter->sinkpad;
GstPad * neighbour;
GParamSpec* spec=NULL;

gboolean abort=FALSE;


GList *elementlist = NULL; /*list of all known elements*/
//elementlist=g_list_alloc(); 
GList *elementlistiterator;
//initialize list with calling element
elementlist=g_list_prepend(elementlist,(gpointer)element);

GList *padlist = NULL; /*list of all known elements*/
//padlist=g_list_alloc(); 
GList *paditerator;//iterator for both the elements pad list as well as this functions pad list

do{
	abort=TRUE;
	//Get all pads
	for(elementlistiterator = elementlist; ((elementlistiterator!= NULL)&&(elementlistiterator->data!= NULL));  elementlistiterator= g_list_next(elementlistiterator)) {
		g_print("Inside Getallpads\n"); 
		GstIterator* padit=gst_element_iterate_pads ((GstElement*)elementlistiterator->data);
		gboolean done=FALSE;
		gpointer item;
		while (!done) {
			switch (gst_iterator_next (padit, (gpointer*)&item)) {
			case GST_ITERATOR_OK:
				if ((g_list_find (padlist,(gconstpointer)item))==NULL) padlist=g_list_prepend(padlist,(gpointer)item);
				gst_object_unref (item);
				break;
			case GST_ITERATOR_RESYNC:
				g_print("GstIterator Resync");
				gst_iterator_resync (padit);
				break;
			case GST_ITERATOR_ERROR:
				g_print("GstIterator Returning Unexpected Error");
				done = TRUE;
				break;
			case GST_ITERATOR_DONE:
				done = TRUE;
				break;
			default:
				g_print("GstIterator Returning Unexpected Error"); 
				done = TRUE;
				break;
			}
		}
		gst_iterator_free (padit);
	}
	//Get new elements
	for(paditerator = padlist; ((paditerator!= NULL)&&(paditerator->data!= NULL));  paditerator= g_list_next(paditerator)) {
		neighbour=gst_pad_get_peer(paditerator->data);
		g_print("Searching for new elements\n");
		//g_object_unref(G_OBJECT(temppad));//we do not need this reference
		if (!neighbour) g_print("Neighbour Pad is NULL");
		tempel= gst_pad_get_parent_element(neighbour);
		if (tempel!=NULL){	
			g_object_unref(G_OBJECT(neighbour));//we do not need an additional reference...
			if (g_list_find (elementlist,(gconstpointer)tempel)==NULL){
				name=gst_element_get_name(tempel);
				g_print("Will append Element %s",name);
				abort=FALSE;
				g_free(name);
				elementlist=g_list_prepend(elementlist,(gpointer)tempel);
			}
			g_object_unref(G_OBJECT(tempel));//we do not need an additional reference...		
		}
		else g_print("Error, Parent element null"); 
	}
	//Test if correct element was found
	for(elementlistiterator = elementlist; ((elementlistiterator!=NULL)&&(elementlistiterator->data!=NULL)); elementlistiterator = g_list_next(elementlistiterator)) {
		//g_print("Search whether %s was found in Element named %s",propertyname,gst_element_get_name(elementlistiterator->data));
		if (propertyname){
			spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(elementlistiterator->data)),propertyname);
			if (spec){
				if (elementname){//User has specified element prefix
					name=gst_element_get_name(elementlistiterator->data);
					if (g_str_has_prefix(name,elementname)){
						g_print("Quite right %s is prefix of %s, setting result\n",elementname,name);
						result=(GstElement*)elementlistiterator->data;
					}
					g_free(name);
				}
				else{
					g_print("..found:)\n");
					result=(GstElement*)elementlistiterator->data;
					}
			}
			else g_print("..not found:(\n");
		}
		else if (elementname)
			{//User has specified ONLY element prefix
			name=gst_element_get_name(elementlistiterator->data);
			if (g_str_has_prefix(name,elementname)){
				g_print("Quite right %s is prefix of %s, setting result\n",elementname,name);
				result=(GstElement*)elementlistiterator->data;
			}
			g_free(name);
		}
		else result=(GstElement*)elementlistiterator->data;//if you dont specify anything you get anything...
	}
}while ((result==NULL)&&(abort==FALSE));
if (elementlist)g_list_free(elementlist);
if (padlist)g_list_free(padlist);
//reapeat if (new elements were found and correct element has not been found yet)
return result;
}



static gboolean findsrc(GstPad * pad, const gchar* name, GstElement ** source)//finds upstream element that provides property "name"
{
GstAdjuster *filter;
filter = GST_ADJUSTER (GST_OBJECT_PARENT (pad));
GstElement * tempel=NULL;

GstPad * temppad;
temppad=filter->sinkpad;
GstPad * neighbour;
GParamSpec* spec=NULL;
*source=NULL;

g_object_ref(temppad);//to archieve unified unref in loop
//get next element	
do{
	//sleep(1);
	neighbour=gst_pad_get_peer(temppad);
	//g_print("will unref temppad\n");
	g_object_unref(G_OBJECT(temppad));
	if (!neighbour) g_print("Neighbour Pad is NULL");
	else{
		tempel= gst_pad_get_parent_element(neighbour);
		//g_print("will unref neighbour\n");
		g_object_unref(G_OBJECT(neighbour));
		if (tempel){
			//g_print("will fetch specs\n");
			//sleep(1);
			spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(tempel)),name);
			//g_print("analyzing specs\n");
			if (spec){
				//g_print("if-block analyzing specs\n");	
				*source=tempel;
			}
			else {
				//g_print("else-block analyzing specs\n");	
				temppad=gst_element_get_static_pad (tempel,"sink");
				if (!temppad){//TODO: //set up GstIterator list of avaiable srcpads
				//recursive searchtree
				//unref others
				}
			}
			//g_param_spec_unref(spec);
			//g_print("finished analyzing specs\n");
		}
	}
}while((temppad)&&(tempel)&&(!*source));

if (*source) return TRUE;
else {	g_print("No element with property named %s found",name);
	return FALSE;
	}
}

static gboolean adjust_par(GstAdjuster * filter,const gchar* name,GstElement * source,guint64 value){

GParamSpec* spec=NULL;
spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(source)),name);

if (NULL!=spec){
	switch(spec->value_type){
	case G_TYPE_BOOLEAN: /*g_print("bool\n");*/g_object_set(G_OBJECT(source),name,(gboolean)value,NULL);break; 
	case G_TYPE_ULONG:   /*g_print("ULONG\n");*/g_object_set(G_OBJECT(source),name,(gulong)value,NULL);break; 
        case G_TYPE_LONG:    /*g_print("Long\n");*/g_object_set(G_OBJECT(source),name,(glong)value,NULL);break; 
        case G_TYPE_UINT:    /*g_print("UINT\n");*/g_object_set(G_OBJECT(source),name,(guint)value,NULL);break; 
        case G_TYPE_INT:     /*g_print("INT\n");*/g_object_set(G_OBJECT(source),name,(gint)value,NULL);break; 
        case G_TYPE_UINT64:  /*g_print("UINT64\n");*/g_object_set(G_OBJECT(source),name,(guint64)value,NULL);break; 
        case G_TYPE_INT64:   /*g_print("INT64\n");*/g_object_set(G_OBJECT(source),name,(gint64)value,NULL);break; 
        case G_TYPE_FLOAT:   /*g_print("float\n");*/g_object_set(G_OBJECT(source),name,(gfloat)value,NULL);break; 
        case G_TYPE_DOUBLE:  /*g_print("DOUBLE\n");*/g_object_set(G_OBJECT(source),name,(gdouble)value,NULL);break; 
        case G_TYPE_ENUM:    /*g_print("Enum\n");*/g_object_set(G_OBJECT(source),name,(gint)value,NULL);break; 
        default: 
		//if 
		g_print("Sorry, no idea how to get this value into it %s",name);
	}
	g_print("Setting %s to %llu",name,value);
	//g_object_set(G_OBJECT(source),name,(int)value,NULL);
	//g_param_spec_unref(spec);
	return TRUE;
	}
else g_print("No property named %s",name);
//g_param_spec_unref(spec);
return FALSE;
}

static gboolean get_par(GstPad * pad,const gchar* name,GstElement * source,gint64 *value){
GstAdjuster *filter;
filter = GST_ADJUSTER (GST_OBJECT_PARENT (pad));

GParamSpec* spec=NULL;
spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(source)),name);
if (NULL!=spec){
	int shut;
	shut=42;
	g_object_get(G_OBJECT(source),name,value,NULL);
	
	g_print("Gotten %s value %lld\n",name,*value);
	//*value=(gint64)shut;
	//sleep(2);
	g_print("assignment finished\n");
	//g_param_spec_unref(spec);
	return TRUE;
	}
else g_print("No property named %s",name);
//g_param_spec_unref(spec);
return FALSE;
}




static gboolean parinfo(GstAdjuster *filter, const gchar* name,GstElement * source,gint64* now, gint64* def, gint64* min,gint64* max){
//g_print("parinfo\n");
GParamSpec* spec=NULL;
if (!source){
	g_print("Source is NULL");
	return FALSE;
}
spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(source)),name);
//g_print("analyzing specs\n");
if (NULL!=spec){
    GValue value = { 0, };	
    g_value_init (&value, spec->value_type);
    g_print("Now call get property");
    if (spec->flags & G_PARAM_READABLE) g_object_get_property (G_OBJECT (source), spec->name, &value);
    g_print("returned from get property");
    switch (G_VALUE_TYPE (&value)) {
      case G_TYPE_BOOLEAN:
      {
        GParamSpecBoolean *pboolean = G_PARAM_SPEC_BOOLEAN (spec);
	if (def) *def=(gint64)pboolean->default_value;
	if (min) *min=(gint64)FALSE;
	if (max) *max=(gint64)TRUE;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_boolean (&value);
        break;
      }
      case G_TYPE_ULONG:
      {
        GParamSpecULong *pulong = G_PARAM_SPEC_ULONG (spec);
	if (def) *def=(gint64)pulong->default_value;
	if (min) *min=(gint64)pulong->minimum;
	if (max) *max=(gint64)pulong->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now){ *now=(gint64)g_value_get_ulong (&value);
	//g_print ("Current: %lu", g_value_get_ulong (&value));}
        break;
      }
      case G_TYPE_LONG:
      {
        GParamSpecLong *plong = G_PARAM_SPEC_LONG (spec);
	if (def) *def=(gint64)plong->default_value;
	if (min) *min=(gint64)plong->minimum;
	if (max) *max=(gint64)plong->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_long (&value);
        break;
      }
      case G_TYPE_UINT:
      {
        GParamSpecUInt *puint = G_PARAM_SPEC_UINT (spec);
	if (def) *def=(gint64)puint->default_value;
	if (min) *min=(gint64)puint->minimum;
	if (max) *max=(gint64)puint->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_uint (&value);
        break;
      }
      case G_TYPE_INT:
      {
	//g_print("Setting int\n");
        GParamSpecInt *pint = G_PARAM_SPEC_INT (spec);
	//g_print("Survived G_PARAM_SPEC_INT\n");
	if (def) *def=(gint64)pint->default_value;
	if (min) *min=(gint64)pint->minimum;
	if (max) *max=(gint64)pint->maximum;
	//g_print("Survived def min max\n");
	if ((spec->flags & G_PARAM_READABLE)&& now) {
		g_print("g_value_get_int\n");
		*now=(gint64)g_value_get_int (&value);
		//g_print("Setting now int to %d",(int)*now);
		//g_print("Normaly this should be %d",g_value_get_int (&value));
		//g_print("Normaly this should be really %lld",g_value_get_int64 (&value));
	}
	//g_print("Survived now\n");
        break;
      }
      case G_TYPE_UINT64:
      {
	g_print("Warning, possible value corruption becaus of int");
        GParamSpecUInt64 *puint64 = G_PARAM_SPEC_UINT64 (spec);
	if (def) *def=(gint64)puint64->default_value;
	if (min) *min=(gint64)puint64->minimum;
	if (max) *max=(gint64)puint64->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_uint64 (&value);
        break;
      }
      case G_TYPE_INT64:
      {
        GParamSpecInt64 *pint64 = G_PARAM_SPEC_INT64 (spec);
	if (def) *def=(gint64)pint64->default_value;
	if (min) *min=(gint64)pint64->minimum;
	if (max) *max=(gint64)pint64->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now){ *now=(gint64)g_value_get_int64 (&value);
	//g_print("Value prior to manipulation %lld",*now);
	//g_object_get(G_OBJECT(source),name,now,NULL);
	//g_print("Now when readout with gobject method %lld",*now);
	//g_value_set_int64 (&value,42);
	//*now=(gint64)g_value_get_int64 (&value);
	//g_print("Value after manipulation %d",(int)*now);
	
	}
        break;
      }
      case G_TYPE_FLOAT:
      {
        GParamSpecFloat *pfloat = G_PARAM_SPEC_FLOAT (spec);
	if (def) *def=(gint64)pfloat->default_value;
	if (min) *min=(gint64)pfloat->minimum;
	if (max) *max=(gint64)pfloat->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_float (&value);
        break;
      }
      case G_TYPE_DOUBLE:
      {
        GParamSpecDouble *pdouble = G_PARAM_SPEC_DOUBLE (spec);
	if (def) *def=(gint64)pdouble->default_value;
	if (min) *min=(gint64)pdouble->minimum;
	if (max) *max=(gint64)pdouble->maximum;
	if ((spec->flags & G_PARAM_READABLE)&& now) *now=(gint64)g_value_get_double (&value);
        break;
      }
      default:
	g_print("Property %s not castable",name);
	return FALSE;
      }
      //g_print("Will return\n");
	//g_param_spec_unref(spec);
	return TRUE;
	}
   }
else g_print("No property named %s",name);
return FALSE;
}


#include "adjusterhelp.h"

static GstFlowReturn
gst_adjuster_chain (GstPad * pad, GstBuffer * buf)
{

  GstAdjuster *filter;
  filter = GST_ADJUSTER (GST_OBJECT_PARENT (pad));

  int channels, bytecounter;
  channels=3;
  static unsigned int phase;
  static unsigned int *histred,*histred8,*histgreen,*histgreen8,*histblue,*histblue8,*histbw,*histbw8;

  static unsigned int hist8divisor;
  static unsigned int pixsumred,pixsumgreen,pixsumblue,pixsumbw;
  static unsigned int avred,avgreen,avblue,avbw;
  static int shutterlevel[4096];
  static guint64 pattern;
  static gdouble motionarray[4096];
  static int counter;

	static int maxshut,brightness;

  if (!filter->initialized){
	g_print("Initializing filters\n");

	filter->motionpointer=findelement((GstElement*)filter,"pardouble2",NULL);
	if (!filter->motionpointer){
	g_print("Motionpointer is NULL");
	//sleep(5);
	}
	filter->srcpointer=findelement((GstElement*)filter,"shutter",NULL);
	if (!filter->srcpointer){
	g_print("Shutterpointer is NULL");
	sleep(5);
	}	
	//findsrc(pad, "shutter", &filter->srcpointer);
	if (filter->bpp>16){
		histred=(unsigned int*)g_malloc((pow(2,filter->bpc))*sizeof(unsigned int));//sufficient for 4 GPixel:)
		histred8=(unsigned int*)g_malloc(8*sizeof(unsigned int));
		histgreen=(unsigned int*)g_malloc((pow(2,filter->bpc))*sizeof(unsigned int));//sufficient for 4 GPixel:)
		histgreen8=(unsigned int*)g_malloc(8*sizeof(unsigned int));
		histblue=(unsigned int*)g_malloc((pow(2,filter->bpc))*sizeof(unsigned int));//sufficient for 4 GPixel:)
		histblue8=(unsigned int*)g_malloc(8*sizeof(unsigned int));
	}
	else{
		histbw=(unsigned int*)g_malloc((pow(2,filter->depth))*sizeof(unsigned int));;
		histbw8=(unsigned int*)g_malloc(8*sizeof(unsigned int));
	}
	g_print("Allocated %d for normal histogramm\n",(int)(pow(2,filter->bpc)*sizeof(unsigned int)));
	hist8divisor=(unsigned int)(pow(2,filter->bpc)/8);
  	filter->initialized=TRUE;
 	pattern=1;
	phase=0;
	maxshut=brightness=0;//for calibration of constant illumination
	int i;
	for (i=0;i<4096;i++){
		shutterlevel[i]=-1;
		motionarray[i]=0;
	}
  }
  
//   g_print("Testing buffer integrity\n");
//   for (bytecounter=0;bytecounter<(int)buf->size;bytecounter++) buf->data[bytecounter]=40;
//   g_print("Testing buffer integrity succeded\n");

  if ((filter->bpp==32)&&(filter->depth==24)){
	unsigned char* data;
  	data=(unsigned char*)buf->data;
	g_print("Here I go, bufsize= %d\n",(int)buf->size);
	memset(histred,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	g_print("survived memset red half\n");
	memset(histred8,0,8*sizeof(unsigned int));
	g_print("survived memset red\n");
	memset(histgreen,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histgreen8,0,8*sizeof(unsigned int));
	g_print("survived memset green\n");	
	memset(histblue,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histblue8,0,8*sizeof(unsigned int));
	g_print("survived memset\n");
	for (bytecounter=-1;bytecounter<(int)(buf->size-1);){
		//g_print("Here I go Byte %d, bufsize is %d, value is %u\n",bytecounter+1,(int)buf->size, data[bytecounter+1]);
		histred[data[++bytecounter]]++;	
		histred8[data[bytecounter]/hist8divisor]++;	
		pixsumred+=data[bytecounter++];
		//g_print("alive after red\n");
		histgreen[data[bytecounter]]++;	
		histgreen8[data[bytecounter]/hist8divisor]++;	
		pixsumgreen+=data[bytecounter++];
		//g_print("alive after green, now bytecounter is %d, value is %u\n",bytecounter, data[bytecounter]);
		histblue[data[bytecounter]]++;	
		//g_print("alive after blue1\n");
		histblue8[data[bytecounter]/hist8divisor]++;
		//g_print("alive after blue2\n");	
		pixsumblue+=data[bytecounter++];
		//g_print("alive after blue3\n");			
	}
  	//g_print("survived all\n");
  }
 
  else if((filter->bpp==24)&&(filter->depth==24)){
	memset(histred,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histred8,0,8*sizeof(unsigned int));
	memset(histgreen,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histgreen8,0,8*sizeof(unsigned int));
	memset(histblue,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histblue8,0,8*sizeof(unsigned int));
	for (bytecounter=0;bytecounter<(int)buf->size;){
		//g_print("Here I go Byte %d, buffersize= %d\n",bytecounter,(int)buf->size);
		histred[buf->data[bytecounter]]++;	
		histred8[(buf->data[bytecounter])/hist8divisor]++;	
		pixsumred+=buf->data[bytecounter++];
		histgreen[buf->data[bytecounter]]++;	
		histgreen8[(buf->data[bytecounter])/hist8divisor]++;	
		pixsumgreen+=buf->data[bytecounter++];
		histblue[buf->data[bytecounter]]++;	
		histblue8[(buf->data[bytecounter])/hist8divisor]++;	
		pixsumblue+=buf->data[bytecounter++];	
	}
  }
else if((pattern==4)&&(filter->bpp==8)&&(filter->depth==8)){
	g_print("Starting Black and white histogramm\n");
	memset(histbw,0,(pow(2,filter->bpc)*sizeof(unsigned int)));
	memset(histbw8,0,8*sizeof(unsigned int));
	pixsumbw=0;
	for (bytecounter=0;bytecounter<(int)(buf->size);){
		//g_print("BW pix %d\n",bytecounter);
		histbw[buf->data[bytecounter]]++;	
		histbw8[(buf->data[bytecounter])/hist8divisor]++;	
		pixsumbw+=buf->data[bytecounter++];
	}
	avbw=pixsumbw/buf->size;
	avred=avgreen=avblue=avbw;
}


if((filter->bpp==8)&&(filter->depth==8)&&(1==2)){//Abbildung Shutter-Value

	gint64 now;


	switch(phase){
		case 0://erhöhen, bis größer als
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);	
			adjust_par(filter, "shutter", filter->srcpointer,now+1);					
			if (buf->data[filter->pixel]==0){
			adjust_par(filter, "shutter", filter->srcpointer,now+1);		
			}
			else{
				g_print("Start at %lld\n",now);
				int i;
				for (i=0;i<4096;i++){
					if (shutterlevel[now]>-1){
					g_print("%d;%d\n",i,(int)shutterlevel[now]);
					}
				}
				//sleep(10);
				phase=1;
				shutterlevel[now]=buf->data[filter->pixel];
			}
			break;
		case 1:
			//TODO readout brightness into table
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);
			adjust_par(filter, "shutter", filter->srcpointer,now+5);
			g_print("%lld at the moment\n",now+5);
			phase=3;
			break;
		case 2://Toggle to 1
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);	
			shutterlevel[now]=buf->data[filter->pixel];		
			if (buf->data[filter->pixel]==255){

				g_print("End at %lld\n",now);
				int i;
				for (i=0;i<4096;i++){
					if (shutterlevel[i]>-1){
					g_print("%d %d\n",i,(int)shutterlevel[i]);
					}
				}
				sleep(10);
			}
			phase=1;
			break;
		case 3://NOOP
			phase=2;
			break;					
	}
}

else if((filter->bpp==8)&&(filter->depth==8)&&(2==3)){//Abbildung Gain-Value
	
	gint64 now;
	int start;
	switch(phase){
		case 0://adjust shutter to desired level
			start=50;
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);
			if (buf->data[filter->pixel]<(start/2)){
			adjust_par(filter, "shutter", filter->srcpointer,now+5);		
			}					
			else if (buf->data[filter->pixel]<start){
			adjust_par(filter, "shutter", filter->srcpointer,now+1);		
			}
			else{
				g_print("Start at %lld\n",now);
				phase=1;
			}
			break;
				
		case 1:
			//TODO readout brightness into table
			parinfo(filter,"gain", filter->srcpointer,&now, NULL, NULL,NULL);
			adjust_par(filter, "gain", filter->srcpointer,now+2);
			g_print("%lld at the moment\n",now+2);
			phase=3;
			break;
		case 2://Toggle to 1
			parinfo(filter,"gain", filter->srcpointer,&now, NULL, NULL,NULL);	
			shutterlevel[now]=buf->data[filter->pixel];		
			if (buf->data[filter->pixel]==255){

				g_print("End at %lld\n",now);
				int i;
				for (i=0;i<4096;i++){
					if (shutterlevel[i]>-1){
					g_print("%d %d\n",i,(int)shutterlevel[i]);
					}
				}
				sleep(10);
			}
			phase=1;
			break;
		case 3://NOOP
			phase=2;
			break;					
	}
}

else if((filter->bpp==8)&&(filter->depth==8)&&(2==3)){//Abbildung Gain-Shutter-Verschiebung Konstant
	
	gint64 now;
	int start;

	switch(phase){
		case 0://adjust shutter to desired level
			start=100;
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);
			if (buf->data[filter->pixel]<(4*(start/5))){
			adjust_par(filter, "shutter", filter->srcpointer,now+5);		
			}					
			else if (buf->data[filter->pixel]<start){
			adjust_par(filter, "shutter", filter->srcpointer,now+1);		
			}
			else{
				g_print("Start at %lld\n",now);
				brightness=maxshut=now;
				phase=1;
			}
			break;
				
		case 1:
			//parinfo(filter,"gain", filter->srcpointer,&now, NULL, NULL,NULL);
			//adjust_par(filter, "gain", filter->srcpointer,now+2);
			if (!excitesetbrightness(filter, (gint)brightness, (gint)maxshut)) maxshut=0;//to make it stop
			g_print("%lld at the moment\n",(guint64)maxshut);
			maxshut-=2;
			phase=3;
			break;
		case 2://Toggle to 1
			parinfo(filter,"shutter", filter->srcpointer,&now, NULL, NULL,NULL);	
			shutterlevel[now]=buf->data[filter->pixel];		
			if (maxshut<10){

				g_print("End at %lld\n",now);
				int i;
				for (i=0;i<4096;i++){
					if (shutterlevel[i]>-1){
					g_print("%d %d\n",i,(int)shutterlevel[i]);
					}
				}
				sleep(10);
			}
			phase=1;
			break;
		case 3://NOOP
			phase=2;
			break;					
	}
}

else if((filter->bpp==8)&&(filter->depth==8)&&(1==2)&&(pattern==4)){//Regelung Gain/Shutter/Brightness
	//<3 - 4095;, 3..10 - 2000;10-20 1000;20-30...500
	gdouble percentover=0.03;
	gdouble percentmininhigh=0.02;
	
	gint brightness;
	gdouble averagebright,motion;//1-8 
	brightness=excitegetbrightness(filter);
	//g_print("Current brightness %d\n",brightness);
	int i;
	averagebright=4;
	for (i=0;i<255;i++){
		
	}
	//if (avbw<110) brightness=(int)((111.0/(gdouble)avbw)*(gdouble)brightness);//Mindesthelligkeit
	if (histbw[255]>(int)(percentover*2*buf->size)){//hartes herabrageln um Faktor min(1/(niedrigste Helligkeit),0.8)
		g_print("Will cut off, avbrightness is %d\n",avbw);		
		for (i=0;i<255;i++){
			if (histbw[i]>0) break;
		}
		if (i>1) brightness=brightness/i;
		else brightness=(int)(0.8*(gdouble)brightness);
	}
	else if((histbw[255]>(int)(percentover*buf->size))||(histbw[254]>(int)(percentover*buf->size))||(histbw[253]>(int)(percentover*buf->size))||(histbw[252]>(int)(percentover*buf->size))||(histbw[251]>(int)(percentover*buf->size))){//langsames herabregeln
		brightness=(int)(0.98*(gdouble)brightness);
	}
	else if(histbw8[7]<=(int)(percentmininhigh*buf->size)){//set brightest pixel to 250
		for (i=223;i>0;i--){
			if (histbw[i]>0) break;
		}
		if (i>0)brightness=(int)((250.0/(gdouble)i)*(gdouble)brightness);
		else brightness=(int)((250.0)*(gdouble)brightness);
	}
	
	//Motion analysis
	//parinfo(filter,"pardouble2", filter->srcpointer,&motion, NULL, NULL,NULL);
	if (filter->motionpointer){
	motion=g_object_get_double(filter->motionpointer,"pardouble2");
	g_print("Current motion %f\n",motion);
	}
	else motion=0;
	if (motion<3) excitesetbrightness(filter, brightness, 4095);
	else if ((motion>3)&&(motion<=10)) excitesetbrightness(filter, brightness, 2000);
	else if ((motion>10)&&(motion<=20)) excitesetbrightness(filter, brightness, 1000);
	else if ((motion>20)&&(motion<=30)) excitesetbrightness(filter, brightness, 500);
	else if ((motion>20)&&(motion<=30)) excitesetbrightness(filter, brightness, 250);
	else excitesetbrightness(filter, brightness, 150);
}
else if(1==3){
	if (filter->motionpointer){
	motionarray[counter++]=g_object_get_double(filter->motionpointer,"pardouble2");
	g_print("Current motion %f\n",motionarray[counter-1]);
	if (counter>300){
		int x;
		for (x=0;x<300;x++){
			g_print("%d %.2f\n",x,motionarray[x]);
		}
		sleep(10);
		}
	}
}

else if((filter->bpp==8)&&(filter->depth==8)&&(2==2)&&(pattern==4)){//Regelung Gain/Shutter/Brightness mit Überbelichtung
	//<3 - 4095;, 3..10 - 2000;10-20 1000;20-30...500

	gint brightness;
	gdouble averagebright,motion;//1-8 
	brightness=excitegetbrightness(filter);
	//g_print("Current brightness %d\n",brightness);
	int i;
	averagebright=4;
	for (i=0;i<255;i++){
		
	}
	if (avbw<60) brightness=(int)((61.0/(gdouble)avbw)*(gdouble)brightness);//Mindesthelligkeit
	else if ((histbw[255]>0)&&(avbw>80)){//hartes herabrageln um Faktor min(1/(niedrigste Helligkeit),0.8)
		g_print("Will cut off, avbrightness is %d\n",avbw);		
		for (i=0;i<255;i++){
			if (histbw[i]>0) break;
		}
		if (i>1) brightness=brightness/i;
		else brightness=(int)(0.8*(gdouble)brightness);
	}
	else if(((histbw[254]>0)||(histbw[253]>0)||(histbw[252]>0)||(histbw[251]>0))&&(avbw>70)){//langsames herabregeln
		brightness=(int)(0.98*(gdouble)brightness);
	}
	else if(histbw8[7]==0){//set brightest pixel to 250
		for (i=223;i>0;i--){
			if (histbw[i]>0) break;
		}
		if (i>0)brightness=(int)((250.0/(gdouble)i)*(gdouble)brightness);
		else brightness=(int)((250.0)*(gdouble)brightness);
	}
	
	//Motion analysis
	//parinfo(filter,"pardouble2", filter->srcpointer,&motion, NULL, NULL,NULL);
	if (filter->motionpointer){
	motion=g_object_get_double(filter->motionpointer,"pardouble2");
	g_print("Current motion %f\n",motion);
	}
	else motion=0;
	if (motion<3) excitesetbrightness(filter, brightness, 4095);
	else if ((motion>3)&&(motion<=10)) excitesetbrightness(filter, brightness, 2000);
	else if ((motion>10)&&(motion<=20)) excitesetbrightness(filter, brightness, 1000);
	else if ((motion>20)&&(motion<=30)) excitesetbrightness(filter, brightness, 500);
	else if ((motion>20)&&(motion<=30)) excitesetbrightness(filter, brightness, 250);
	else excitesetbrightness(filter, brightness, 150);
}




else g_print("Scheiße\n");
g_print("Finished histogramms-and other stuff\n");
//if (1==1){
if (pattern==1) pattern=2;
else if (pattern==2) pattern=3;
else if (pattern==3) pattern=4;
else pattern=1;
if(0)findelement((GstElement*)filter,"hanshubert","karlaugust");
if (3==4){
	//adjust_par(pad,"pattern",pattern);
	//adjust_par(pad,"haufen",pattern);
	if(findelement((GstElement*)filter,"hanshubert","karlaugust")==0){
		g_print("So solls sein\n");
		//sleep(1);	
	}
	if(findelement((GstElement*)filter,"num-buffers",NULL)!=0){
		g_print("So solls sein, Element gefunden\n");
		//sleep(1);	
	}
	if(findelement((GstElement*)filter,"num-buffers","videotestsrc")!=0){
		g_print("So solls sein, Element bei Prefix gefunden\n");
		//sleep(1);	
	}
	if(findelement((GstElement*)filter,"num-buffers","videotestsrf")==0){
		g_print("So solls sein, Element bei falschen Prefix nicht gefunden\n");
		//sleep(1);	
	}
	if(findelement((GstElement*)0,"num-buffers","videotestsrf")==0){
		g_print("So solls sein, Element NULL nicht gefunden\n");
		//sleep(1);	
	}
	if(findelement((GstElement*)filter,NULL,"xima")==0){
		g_print("FEHLER, Element durch xima spezifiziert gefunden\n");
		//sleep(10);	
	}
	else g_print("So solls sein, Element durch xima spezifiziert gefunden\n");
	//sleep(1);


	if(findsrc(pad, "num-buffers", &filter->srcpointer)){
	gint64 def,min,max,now;def=min=max=now=0;
	parinfo(filter,"num-buffers", filter->srcpointer,&now, &def, &min,&max);
	g_print("Auswertung: Now: %lld Default: %lld Min %lld Max %lld\n",now,def,min,max);
		parinfo(filter,"blocksize", filter->srcpointer,&now, &def, &min,&max);
		g_print("Auswertung Blocksize: Now: %lld Default: %lld Min %lld Max %lld\n",now,def,min,max);	
	adjust_par(filter, "num-buffers", filter->srcpointer,(gint64)10);
	get_par(pad,"num-buffers",filter->srcpointer,&now);
	g_print("Still alive\n");
	}
}

// if ((pattern==3)&&(3==1)){
// //if (0){
// 	//adjust_par(pad,"pattern",pattern);
// 	//adjust_par(pad,"haufen",pattern);
// 	if(findsrc(pad, "shutter", &filter->srcpointer)){
// 		gint64 def,min,max,now;def=min=max=now=0;
// 		parinfo(filter,"shutter", filter->srcpointer,&now, &def, &min,&max);
// 		//get_par(pad, "shutter", filter->srcpointer,&now);
// 		g_print("Auswertung: Now: %lld Default: %lld Min %lld Max %lld\n",now,def,min,max);
// 			
// 		g_print ("Diagnose: Over: %u, last achtel %u\n",histbw[255],histbw8[7]);
// 		if (histbw[255]>0){
// 			if (histbw8[0]==0) adjust_par(pad, "shutter", filter->srcpointer,(gint)(now/2));
// 			else adjust_par(pad, "shutter", filter->srcpointer,(gint)(now-10));
// 		}
// 		else if (histbw8[7]==0) adjust_par(pad, "shutter", filter->srcpointer,(gint)(now+10));	
// 	}
// }



//}
    //g_print ("Diagnose: Red:U: %u O: %u \n.Green:U: %u O: %u \nBlue:U: %u O:%u\n",histred[0],histred[(int)(pow(2,filter->bpc)-1)],histgreen[0],histgreen[(int)(pow(2,filter->bpc)-1)],histblue[0],histblue[(int)(pow(2,filter->bpc)-1)]);
    //g_print ("Diagnose: Red8: %u %u %u %u %u %u %u %u .\n",histred8[0],histred8[1],histred8[2],histred8[3],histred8[4],histred8[5],histred8[6],histred8[7]);
    //g_print ("Diagnose: Blue8: %u %u %u %u %u %u %u %u .\n",histblue8[0],histblue8[1],histblue8[2],histblue8[3],histblue8[4],histblue8[5],histblue8[6],histblue8[7]);
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 *
 * exchange the string 'plugin' with your elemnt name
 */
static gboolean
plugin_init (GstPlugin * plugin)
{
  /* exchange the strings 'plugin' and 'This plugin saves the world' with your
   * plugin name and description */
  GST_DEBUG_CATEGORY_INIT (gst_adjuster_debug, "adjuster",
      0, "This plugin saves the world");

  return gst_element_register (plugin, "adjuster",
      GST_RANK_NONE, GST_TYPE_ADJUSTER);
}

/* this is the structure that gstreamer looks for to register plugins
 *
 * exchange the strings 'plugin' and 'This plugin saves the world' with you plugin name and
 * description
 */
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "adjuster",
    "This plugin saves the world",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
