/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-bridge
 *
 * FIXME:Describe bridge here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! bridge ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstbridge.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <error.h>
#include <string.h>
GST_DEBUG_CATEGORY_STATIC (gst_bridge_debug);
#define GST_CAT_DEFAULT gst_bridge_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_REVERSE,
  PROP_WRITEFILE,
  PROP_READFILE,
  PROP_COMMAND,
  PROP_COPYCAPS,
  PROP_CAPS
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstbridge, gst_bridge, GstElement,
    GST_TYPE_ELEMENT);

static GstPad *
gst_bridge_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name);
static void
gst_bridge_release_pad (GstElement * element, GstPad * pad);
static void gst_bridge_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_bridge_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstCaps* gst_bridge_getcaps (GstPad * pad);
static gboolean gst_bridge_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_bridge_chain (GstPad * pad, GstBuffer * buf);
static GstFlowReturn gst_bridge_bufferalloc (GstPad * pad, guint64 offset,  guint size, GstCaps * caps, GstBuffer ** buf);

/* GObject vmethod implementations */

static void
gst_bridge_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "bridge",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the bridge's class */
static void
gst_bridge_class_init (GstbridgeClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_bridge_set_property;
  gobject_class->get_property = gst_bridge_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_REVERSE,
      g_param_spec_boolean ("reverse", "Reverse", "Reverse order that filenames are fed into command? (default input,output)",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_WRITEFILE,
      g_param_spec_string ("writefile", "Writefile",
          "Path/Name where we write the temporary file", "/tmp/gstbridge.tmp", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_READFILE,
      g_param_spec_string ("readfile", "Readfile",
          "Path/Name where we read the temporary file", "/tmp/gstbridge2.tmp", G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_COMMAND,
      g_param_spec_string ("command", "Command",
          "Command to execute (first %s=input, second %s output for call)", "cp %s %s",G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_COPYCAPS,
      g_param_spec_boolean ("copycaps", "Copycaps", "Foreward the inbound caps ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CAPS,
      g_param_spec_boxed ("caps", "Caps",
          "Force Caps describing the format of the data.",
          GST_TYPE_CAPS, G_PARAM_READWRITE));
  gstelement_class->request_new_pad =
      GST_DEBUG_FUNCPTR (gst_bridge_request_new_pad);
  gstelement_class->release_pad =
      GST_DEBUG_FUNCPTR (gst_bridge_release_pad);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_bridge_init (Gstbridge * filter,
    GstbridgeClass * gclass)
{
  //filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  //gst_pad_set_setcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_bridge_set_caps));
  //gst_pad_set_getcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  //gst_pad_set_getcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_bridge_getcaps));
  //gst_pad_set_chain_function (filter->sinkpad,
  //                            GST_DEBUG_FUNCPTR(gst_bridge_chain));
  //gst_pad_set_bufferalloc_function(filter->sinkpad,GST_DEBUG_FUNCPTR(gst_bridge_bufferalloc));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_bridge_getcaps));

  //gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->reverse = FALSE;  
  filter->copycaps=TRUE;
  filter->caps=gst_caps_new_any();
  filter->offset=0;
  sprintf(filter->systemcallscheme,"cp %%s %%s");
  sprintf(filter->writefilescheme,"/tmp/gstbridge%%d.tmp");
  sprintf(filter->readfilename,"/tmp/gstbridgeout.tmp");
  filter->mutex=g_mutex_new();
  filter->states=NULL;
  filter->sinkpads=NULL;
  filter->mutexs=NULL;
}

static void
gst_bridge_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstbridge *filter = GST_BRIDGE (object);
  GstCaps *caps;
  GstCaps *new_caps;
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_REVERSE:
      filter->reverse = g_value_get_boolean (value);
      break;
    case PROP_WRITEFILE:
      printf("GSTbridge Setting writefile\n");
      strcpy(filter->writefilescheme,g_value_get_string (value));
      //sprintf(filter->writefilescheme,g_value_get_string (value));
      //filter->writefilescheme=g_strdup(g_value_get_string (value));
      break;
    case PROP_READFILE:
      printf("GSTbridge Setting readfile\n");
      strcpy(filter->readfilename,g_value_get_string (value));
      //sprintf(filter->readfilescheme,g_value_get_string (value));
      //filter->readfilescheme=g_strdup(g_value_get_string (value));
      break;
    case PROP_COMMAND:
      printf("GSTbridge Setting command\n");
      strcpy(filter->systemcallscheme,g_value_get_string (value));
      //sprintf(filter->systemcallscheme,g_value_get_string (value));
      //filter->systemcallscheme=g_strdup(g_value_get_string (value));
      break;
    case PROP_COPYCAPS:
      filter->copycaps = g_value_get_boolean (value);
      break;
    case PROP_CAPS:
      caps = gst_value_get_caps (value);
      if (caps == NULL) {
        new_caps = gst_caps_new_any ();
      } else {
        new_caps = gst_caps_copy (caps);
      }
      gst_caps_replace (&filter->caps, new_caps);
      //gst_pad_set_caps (GST_BASE_SRC_PAD (src), new_caps);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_bridge_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstbridge *filter = GST_BRIDGE (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_REVERSE:
      g_value_set_boolean (value, filter->reverse);      
      break;
    case PROP_WRITEFILE:
      g_value_set_string (value, filter->writefilescheme);
      break;
    case PROP_READFILE:
      g_value_set_string (value, filter->readfilename);
      break;
    case PROP_COMMAND:
      g_value_set_string (value, filter->systemcallscheme);
      break;
    case PROP_COPYCAPS:
      g_value_set_boolean (value, filter->copycaps);
      break;
    case PROP_CAPS:
      gst_value_set_caps (value, filter->caps);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */
static GstCaps *
gst_bridge_getcaps (GstPad * pad)
{
    return gst_caps_new_any ();
}
/* this function handles the link with other elements */
static gboolean
gst_bridge_set_caps (GstPad * pad, GstCaps * caps)
{
  //Gstbridge *filter;
  //GstPad *otherpad;
  //g_print ("Set_Caps %s.\n",gst_caps_to_string(caps));
  //filter = GST_BRIDGE (gst_pad_get_parent (pad));
  //otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  //gst_object_unref (filter);

  //return gst_pad_set_caps (otherpad, caps);
  return TRUE;
}

static GstPad *
gst_bridge_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name)
{
  gchar* tmp;
  GstPad* tempPad;
  Gstbridge *filter;
  filter = GST_BRIDGE (element);
  g_mutex_lock(filter->mutex);
  g_print("Bridge: New pad with name %s requested\n",name);

  //if (!name) tmp = g_strdup_printf ("src%d", filter->num_srcpads);
  //else tmp=g_strdup_printf ("%s", name);
  tmp = g_strdup_printf ("sink%d", filter->num_sinkpads);
  tempPad = gst_pad_new_from_static_template (&sink_factory, tmp);
  g_free (tmp);

  filter->sinkpads=g_list_append(filter->sinkpads,tempPad);
  filter->mutexs=g_list_append(filter->mutexs,g_mutex_new());
  gint* tempint=malloc(sizeof(gint));
  *tempint=0;
  filter->states=g_list_append(filter->states,tempint);
  filter->num_sinkpads++;
  //gst_videosplitter_update_overlap(filter,filter->overlap);
  //gst_videosplitter_modifycaps((GObject*)filter);
  gst_pad_set_getcaps_function (tempPad,
      GST_DEBUG_FUNCPTR (gst_bridge_getcaps));
  //gst_pad_set_getcaps_function (tempPad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_setcaps_function (tempPad,
                                GST_DEBUG_FUNCPTR(gst_bridge_set_caps));
  gst_pad_set_chain_function (tempPad,
                              GST_DEBUG_FUNCPTR(gst_bridge_chain));
  gst_element_add_pad (GST_ELEMENT (filter), tempPad);
  g_mutex_unlock(filter->mutex);
  return tempPad;
}

static void
gst_bridge_release_pad (GstElement * element, GstPad * pad)
{

g_print("releasing pads not supported yet:)\n");
for(;;);
}

static GstFlowReturn gst_bridge_bufferalloc (GstPad * pad, guint64 offset,
    guint size, GstCaps * caps, GstBuffer ** buf){
  Gstbridge *filter;
  GstFlowReturn result;
  //g_print ("Bufferalloc has been caled with Caps: %s\n",gst_caps_to_string(caps));
  filter = GST_BRIDGE (GST_PAD_PARENT (pad));

  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  return result;

}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_bridge_chain (GstPad * pad, GstBuffer * buf)
{
  Gstbridge *filter;

  filter = GST_BRIDGE (GST_OBJECT_PARENT (pad));
//vars
  GstFlowReturn ret2;
  gsize size;
  guint64 offset,timestamp,duration,offset_end;
  gchar *data;
  //gchar *filename;
  GstBuffer *newbuf;
  FILE* datei; 
//copy old metadata
  offset=buf->offset;
  timestamp=buf->timestamp;
  duration=buf->duration;
  offset_end=buf->offset_end;
  //saving to file
  int index=-1;
  int i=0;
  GList* temp;
  for(temp=filter->sinkpads;((temp)&&(index<0));temp=g_list_next(temp)){
    if (temp->data==pad) index=i;
    i++;
  }
  if ((filter->copycaps)&&(index==0)&&(filter->caps)&&(!gst_caps_is_equal(filter->caps,buf->caps))){
    filter->caps=gst_caps_copy(buf->caps);
  }
  //zwischenspeichern von Buffern
  if  (index!=-1){
      g_mutex_lock((GMutex*)g_list_nth(filter->mutexs,index)->data);
      char* writefilename=malloc(strlen(filter->writefilescheme)+100);
      sprintf(writefilename,filter->writefilescheme,index);     
    //gst_buffer_unref((GstBuffer *)g_list_nth(filter->buffers,index)->data);
      g_print("GSTBridge writing in:%s\n",writefilename);  
      datei=fopen(writefilename,"wb"); 
      free(writefilename);
      if ((!datei)||(fwrite (buf->data, buf->size, 1, datei) != 1)){
        g_print("GSTBridge Warning, could not write file\n");
	g_mutex_unlock((GMutex*)g_list_nth(filter->mutexs,index)->data);   
        return GST_FLOW_ERROR;
      }
      fclose(datei);
      gst_buffer_unref(buf);
      *((int *)g_list_nth(filter->states,index)->data)=1;
      g_mutex_unlock((GMutex*)g_list_nth(filter->mutexs,index)->data);   
  }
  else{
    g_print ("Could not find Index %d.\n",index);
    return GST_FLOW_ERROR;    
  }

  
//decide whether to do output  
int run=1;
  for(temp=filter->states;((temp)&&(index<0));temp=g_list_next(temp)){
    if (*((int *)(temp->data))==0) run=0;
  }
  if (!run){
    return GST_FLOW_OK; 
  }
  //now with mutex
  g_mutex_lock(filter->mutex);
  //Anoter thread called from another pad could have been overtaking us and been sending during last check
  for(temp=filter->states;(temp);temp=g_list_next(temp)){
    if (*((int *)(temp->data))==0) run=0;
  }
  if (!run){
    g_mutex_unlock(filter->mutex);
    return GST_FLOW_OK;
  }  
  //all mutexes are belong to us
  for(temp=filter->mutexs;((temp));temp=g_list_next(temp)){
    g_mutex_lock((GMutex*)temp->data);
  }
  char* temppointer=NULL;
  filter->writefilename[0]=0;
//compose filename of temp file
  for(i=0;i<g_list_length(filter->sinkpads);i++){
    temppointer=strchr(filter->writefilename,0);
    *(temppointer++)=32;
    sprintf(temppointer,filter->writefilescheme,i);
  }  
  if (!filter->reverse){
	sprintf(filter->systemcall,filter->systemcallscheme,filter->writefilename,filter->readfilename);
  }
  else sprintf(filter->systemcall,filter->systemcallscheme,filter->readfilename,filter->writefilename);
//syscall
  g_print("GSTBridge making Systemcall:%s\n",filter->systemcall);
  system(filter->systemcall);
  //tidy up
  for(i=0;i<g_list_length(filter->sinkpads);i++){
    *((int *)g_list_nth(filter->states,index)->data)=0;
    char* writefilename=malloc(strlen(filter->writefilescheme)+100);
    sprintf(writefilename,filter->writefilescheme,i);
    g_remove(writefilename);
    free(writefilename); 
  }  

  //g_remove(filter->writefilename);
//read temporary file
  ret2 = g_file_get_contents (filter->readfilename, &data, &size, NULL);
  
  if (!ret2) {
      g_print("GSTBridge Warning, could not read or find file\n");
      ret2=GST_FLOW_ERROR;		
  }





  newbuf = gst_buffer_new ();
  GST_BUFFER_DATA (newbuf) = (unsigned char *) data;
  GST_BUFFER_MALLOCDATA (newbuf) = GST_BUFFER_DATA (newbuf);
  newbuf->timestamp=timestamp;
  GST_BUFFER_SIZE (newbuf) = size;
  GST_BUFFER_OFFSET (newbuf) = offset;
  GST_BUFFER_OFFSET_END (newbuf) = offset_end;

  //filter->copycaps=TRUE;
  //if (filter->copycaps)gst_buffer_set_caps (newbuf,buf->caps);
  //else gst_buffer_set_caps (newbuf,gst_caps_new_any());
  //else{
  //  GST_BUFFER_OFFSET (newbuf) = filter->offset;
  //  GST_BUFFER_OFFSET_END (newbuf) = GST_BUFFER_OFFSET (newbuf)+size;  
  //  filter->offset=GST_BUFFER_OFFSET_END (newbuf);
    //gst_buffer_set_caps (newbuf,gst_caps_from_string("image/png,framerate=(fraction)30/1"));
  //  if (filter->caps == NULL || GST_CAPS_IS_SIMPLE (filter->caps))gst_buffer_set_caps (newbuf,filter->caps);
  //} 
  //GST_DEBUG_OBJECT (insrc, "read file \"%s\".", filename);
  gst_buffer_set_caps (newbuf,filter->caps);

  g_remove(filter->readfilename);

    //all mutexes are no longer belong to us
  for(temp=filter->mutexs;((temp));temp=g_list_next(temp)){
    g_mutex_unlock((GMutex*)temp->data);
  }
  g_mutex_unlock(filter->mutex);

  if (ret2==GST_FLOW_ERROR)  return GST_FLOW_ERROR;

  //g_print("GSTBridge is about to push buffer, caps %s  (old caps %s)!\n",gst_caps_to_string(newbuf->caps),gst_caps_to_string(buf->caps));
  

  return gst_pad_push (filter->srcpad, newbuf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
bridge_init (GstPlugin * bridge)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template bridge' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_bridge_debug, "bridge",
      0, "Template bridge");

  return gst_element_register (bridge, "bridge", GST_RANK_NONE,
      GST_TYPE_BRIDGE);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstbridge"
#endif

/* gstreamer looks for this structure to register bridges
 *
 * exchange the string 'Template bridge' with your bridge description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "bridge",
    "Template bridge",
    bridge_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
