/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-checker
 *
 * FIXME:Describe checker here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! checker ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>

#include "gstchecker.h"

GST_DEBUG_CATEGORY_STATIC (gst_checker_debug);
#define GST_CAT_DEFAULT gst_checker_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_X,
  PROP_Y,
  PROP_BORDER
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "video/x-raw-rgb, "
      "width=(int) [ 1, MAX ],"
      "height=(int) [ 1, MAX ],"
      "framerate = (fraction) [ 0, MAX ], " 
      "red_mask=65280,"
      "green_mask=16711680," 
      "blue_mask=-16777216,"
      "bpp = (int) 32, "
      "depth = (int) 24")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "video/x-raw-rgb, "
      "width=(int) [ 1, MAX ],"
      "height=(int) [ 1, MAX ],"
      "framerate = (fraction) [ 0, MAX ], " 
      "red_mask=65280,"
      "green_mask=16711680," 
      "blue_mask=-16777216,"
      "bpp = (int) 32, "
      "depth = (int) 24")
    );


GST_BOILERPLATE (Gstchecker, gst_checker, GstElement,
    GST_TYPE_ELEMENT);

static void gst_checker_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_checker_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstFlowReturn
gst_checker_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf);
static gboolean gst_checker_event (GstPad  *pad, GstEvent * event);

static gboolean gst_checker_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_checker_get_caps (GstPad *pad);
static GstFlowReturn gst_checker_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_checker_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "checker",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the checker's class */
static void
gst_checker_class_init (GstcheckerClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_checker_set_property;
  gobject_class->get_property = gst_checker_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_X,
      g_param_spec_int ("chx", "Chx", "Number of corners to detect in X",
          1,100,8, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_Y,
      g_param_spec_int ("chy", "Chy", "Number of corners to detect in Y",
          1,100,6, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, PROP_BORDER,
      g_param_spec_int ("minborder", "Minborder", "Minimum size of border",
          0,255,10, G_PARAM_READWRITE));

}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_checker_init (Gstchecker * filter,
    GstcheckerClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_checker_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_checker_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_checker_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_checker_get_caps));
  gst_pad_set_bufferalloc_function (filter->sinkpad,
      GST_DEBUG_FUNCPTR (gst_checker_bufferalloc));
  gst_pad_set_event_function (filter->srcpad,
                              GST_DEBUG_FUNCPTR(gst_checker_event));
  //gst_pad_set_getcaps_function (filter->srcpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->chx=8;
  filter->chy=6;
  filter->border=10;
  filter->acbx=-1;
  filter->acby=-1;
  filter->enable=FALSE;
  filter->enablebyctl=TRUE;
}

static void
gst_checker_update (GObject * object)
{
  Gstchecker *filter = GST_CHECKER (object);
  //check the maximum width of the squares
  int wwobo=filter->width-(2*filter->border);//widthwithoutborder
  while((wwobo%filter->chx!=0)&&(wwobo>0))wwobo--;
  int wwobos=wwobo/filter->chx;
  //g_print("Width without border %d \n",wwobo);
  int hwobo=filter->height-(2*filter->border);//widthwithoutborder
  while((hwobo%filter->chy!=0)&&(hwobo>0))hwobo--;
  int hwobos=hwobo/filter->chy;  
  //g_print("Height without border %d \n",hwobo);
  int min=((wwobos<hwobos) ? wwobos : hwobos);  
  filter->acw=min;
  filter->acbx=(filter->width-(min*filter->chx))/2;
  filter->acby=(filter->height-(min*filter->chy))/2;
}

 static GstFlowReturn
gst_checker_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  Gstchecker *filter;
  filter = GST_CHECKER (GST_PAD_PARENT (pad));
  GstFlowReturn result;
  if (!filter->silent){
	gchar* cs=gst_caps_to_string(caps);
	g_print ("Timestamper: Bufferalloc has been caled with Caps: %s\n",cs);
	g_free(cs);
  }


  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  if (!filter->silent){
	gchar* cs=gst_caps_to_string((*buf)->caps);
	g_print ("Timestamper: Got Buffer with Caps: %s\n",cs);
	g_free(cs);
  }
  
  
  return result;
}

static void
gst_checker_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstchecker *filter = GST_CHECKER (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_X:
      filter->chx = 1+g_value_get_int (value);
      break;
    case PROP_Y:
      filter->chy = 1+g_value_get_int (value);
      break;
    case PROP_BORDER:
      filter->border = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}



static void
gst_checker_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstchecker *filter = GST_CHECKER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_X:
      g_value_set_int (value, filter->chx-1);
      break;
    case PROP_Y:
      g_value_set_int (value, filter->chy-1);
      break;
    case PROP_BORDER:
      g_value_set_int (value, filter->border);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_checker_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstchecker *filter;
  GstPad *otherpad;

  filter = GST_CHECKER (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  //
  //const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->width)) return FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) return FALSE;
  gst_object_unref (filter);
  //return gst_pad_set_caps (otherpad, caps);
  return TRUE;
}

static GstCaps *
gst_checker_get_caps (GstPad *pad){
    g_print ("CHECKER get caps, returning %s.\n",gst_caps_to_string(gst_pad_get_pad_template_caps(pad)));
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
}
/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_checker_chain (GstPad * pad, GstBuffer * buf)
{
  Gstchecker *filter;  
  int x,y;
  filter = GST_CHECKER (GST_OBJECT_PARENT (pad));

  gst_checker_update((struct GObject *)filter);
  g_print("Width=%d height=%d chx=%d chy=%d minborder=%d acw=%d acbx=%d acby=%d\n",filter->width,filter->height,filter->chx,filter->chy,filter->border,filter->acw, filter->acbx,filter->acby);
  if (filter->enable){
   for (y=0;y<filter->height;y++){
    if ((y>=filter->acby)&&(y<(filter->height-filter->acby-1))){
      for(x=0;x<filter->width;x++){
      if ((x>=filter->acbx)&&
      (x<(filter->width-filter->acbx-1))&&
      ((((x-filter->acbx)/filter->acw)%2)==(((y-filter->acby)/filter->acw)%2))
      )
      {
        buf->data[(y*filter->width+x)*4]=0;
        buf->data[(y*filter->width+x)*4+1]=0;
        buf->data[(y*filter->width+x)*4+2]=0;
        buf->data[(y*filter->width+x)*4+3]=0;
      }
      else{
        buf->data[(y*filter->width+x)*4]=255;
        buf->data[(y*filter->width+x)*4+1]=255;
        buf->data[(y*filter->width+x)*4+2]=255;
        buf->data[(y*filter->width+x)*4+3]=255;
      }
     }
    }//if y in range....
    else memset(buf->data+(y*filter->width*4),255,(filter->width*4));
   }//for y
  }//if enabled..

  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}


static gboolean
gst_checker_event (GstPad  *pad, GstEvent * event)
{
  Gstchecker *filter;
  filter = GST_CHECKER (gst_pad_get_parent (pad));
  gboolean ret;
  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_NAVIGATION:
        //g_print("Multivis: Event Navigation! Event named %s, Structure Named %s, Content %s\n",gst_event_type_get_name(GST_EVENT_TYPE(event)),gst_structure_get_name(  gst_event_get_structure(event)),gst_structure_to_string(gst_event_get_structure(event)));
        //g_print("Eventstring: %s",g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")));
        if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"key-press")){
	  if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Control_L")){
	    //g_print("Enabling snapshot for next frame\n");
	    filter->enable=TRUE;			  
	  }
        }
        if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"event")),"key-release")){
	  if (!strcmp(g_value_get_string(gst_structure_get_value(gst_event_get_structure(event),"key")),"Control_L")){
	    //g_print("Enabling snapshot for next frame\n");
	    filter->enable=FALSE;			  
	  }
        }
      ret = gst_pad_push_event (filter->sinkpad, event);
    break;
    default:
      ret = gst_pad_push_event (filter->sinkpad, event);
   }
  return ret;

}
/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
checker_init (GstPlugin * checker)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template checker' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_checker_debug, "checker",
      0, "Template checker");

  return gst_element_register (checker, "checker", GST_RANK_NONE,
      GST_TYPE_CHECKER);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstchecker"
#endif

/* gstreamer looks for this structure to register checkers
 *
 * exchange the string 'Template checker' with your checker description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "checker",
    "Template checker",
    checker_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
