/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, inclremap4yuvuding without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-clremap4yuv
 *
 * FIXME:Describe clremap4yuv here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! clremap4yuv ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <errno.h>
#include <assert.h>
#include <string.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif

#include <opencv/cv.h>
#include "gstclremap4yuv.h"

GST_DEBUG_CATEGORY_STATIC (gst_clremap4yuv_debug);
#define GST_CAT_DEFAULT gst_clremap4yuv_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_KERNELFILE,
  PROP_DIRECTORY,
  PROP_CPU,
  PROP_CROP,
  PROP_CLCPU
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV("I420"))
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRx)
    //GST_STATIC_CAPS("video/x-raw-rgb, bpp=32, depth=24, endianness=4321, red_mask=65280, green_mask=16711680, blue_mask=-16777216")
    );

GST_BOILERPLATE (Gstclremap4yuv, gst_clremap4yuv, GstElement,
    GST_TYPE_ELEMENT);

static void gst_clremap4yuv_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_clremap4yuv_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstCaps* gst_clremap4yuv_get_caps (GstPad * pad, GstCaps * caps);
static gboolean gst_clremap4yuv_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_clremap4yuv_chain (GstPad * pad, GstBuffer * buf);
static gboolean
gst_clremap4yuv_event (GstPad  *pad, GstEvent * event);
static GstStateChangeReturn
gst_clremap4yuv_change_state (GstElement *element, GstStateChange transition);

/* GObject vmethod implementations */

static void
gst_clremap4yuv_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "clremap4yuv",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the clremap4yuv's class */
static void
gst_clremap4yuv_class_init (Gstclremap4yuvClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_clremap4yuv_set_property;
  gobject_class->get_property = gst_clremap4yuv_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_KERNELFILE,
      g_param_spec_string ("kernel", "Kernel", "Filename of kernel file",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_DIRECTORY,
      g_param_spec_string ("file", "File", "Use this XML file for the Lookup-Table.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_CPU,
      g_param_spec_boolean ("cpu", "Cpu", "Use CPU instead of GPU?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CROP,
      g_param_spec_boolean ("crop", "Crop", "Crop to match exact resolution or leave border?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CLCPU,
      g_param_spec_boolean ("clcpu", "Clcpu", "Whether to do OpenCL based CPU processing?",
          FALSE, G_PARAM_READWRITE));

  gstelement_class->change_state = gst_clremap4yuv_change_state;
}

static char *load_program_source(const char *filename)
{
    struct stat statbuf;
    FILE        *fh;
    char        *source;
	
    fh = fopen(filename, "r");
    if (fh == 0)
        return 0;
	
    stat(filename, &statbuf);
    source = (char *) malloc(statbuf.st_size + 1);
    fread(source, statbuf.st_size, 1, fh);
    source[statbuf.st_size] = '\0';
	
    return source;
}

static void gst_clremap4yuv_release_cl(Gstclremap4yuv * filter){ //release all the CLREMAP4YUV related stuff

  	clReleaseMemObject(filter->inyuv1);
  	clReleaseMemObject(filter->inyuv2);
  	clReleaseMemObject(filter->inyuv3);
  	clReleaseMemObject(filter->inyuv4);

  	clReleaseMemObject(filter->in1);
  	clReleaseMemObject(filter->in2);
  	clReleaseMemObject(filter->in3);
  	clReleaseMemObject(filter->in4);

  	clReleaseMemObject(filter->lut);

	clReleaseMemObject(filter->output);
	clReleaseProgram(filter->program);
	clReleaseKernel(filter->kernel);
	clReleaseKernel(filter->kernelyuv);
	clReleaseCommandQueue(filter->commands);
	clReleaseContext(filter->context);
}

gboolean modifycaps(Gstclremap4yuv * filter)
{
  if (filter->srccaps) gst_caps_unref (filter->srccaps);
  if ((filter->crop)&&(filter->sinkcaps)) filter->srccaps=gst_caps_new_simple ("video/x-raw-rgb",
  "width", G_TYPE_INT, filter->outwidth,
  "height", G_TYPE_INT, filter->outheight,
  "bpp",G_TYPE_INT,32, 
  "depth",G_TYPE_INT,24, 
  "endianness",G_TYPE_INT,4321,
  "red_mask",G_TYPE_INT,65280, 
  "green_mask",G_TYPE_INT,16711680, 
  "blue_mask",G_TYPE_INT,-16777216,
  "framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
  NULL);
  else if ((!filter->crop)&&(filter->sinkcaps)) filter->srccaps=gst_caps_new_simple ("video/x-raw-rgb",
  "width", G_TYPE_INT, filter->vwidth,
  "height", G_TYPE_INT, filter->vheight,
  "bpp",G_TYPE_INT,32, 
  "depth",G_TYPE_INT,24, 
  "endianness",G_TYPE_INT,4321,
  "red_mask",G_TYPE_INT,65280, 
  "green_mask",G_TYPE_INT,16711680, 
  "blue_mask",G_TYPE_INT,-16777216,
  "framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
  NULL);
  else return FALSE;
  return TRUE;
}

static gboolean gst_clremap4yuv_init_cvstuff(Gstclremap4yuv * filter){
  CvMat* lutraw = (CvMat*)cvLoad(filter->directory,NULL,NULL,NULL);
  if (!lutraw){
    g_print("Could not load %s\n",filter->directory);
    return FALSE;
  }
 // Build the undistort map used for all subsequent frames.
  filter->outwidth=(lutraw->cols)/3;
  filter->outheight=(lutraw->rows);

  filter->vwidth=filter->outwidth;
  filter->vheight=filter->outheight;
  if (filter->vwidth%filter->modulo)filter->vwidth=filter->vwidth+(filter->modulo-(filter->vwidth%filter->modulo));
  if (filter->vheight%filter->modulo)filter->vheight=filter->vheight+(filter->modulo-(filter->vheight%filter->modulo));

  if (filter->mapmod)g_free(filter->mapmod);
  filter->mapmodsize=filter->vwidth*filter->vheight*4*sizeof(float);
  filter->mapmod=(float*)g_malloc(filter->mapmodsize);
  int x,y;
  //copy mapx and mapy into our structure
  float tempx,tempy,tempi;
  for(y=0;y<filter->vheight;y++) for(x=0;x<filter->vwidth;x++){
    if ((x>=filter->outwidth)||(y>=filter->outheight)){
      tempi=0;
    }
    else{
      tempx=cvmGet(lutraw,y,3*x);
      tempy=cvmGet(lutraw,y,3*x+1);
      tempi=cvmGet(lutraw,y,3*x+2);
    }
    filter->mapmod[4*(x+y*filter->vwidth)]=tempx;
    filter->mapmod[4*(x+y*filter->vwidth)+1]=tempy;
    filter->mapmod[4*(x+y*filter->vwidth)+2]=tempi;
  }
/*
  for (i=0;i<(filter->mapmodsize/sizeof(float));i++){
    temp=0.0;
    if (i%4==0)temp=cvmGet(lutraw,((i/4)/filter->outwidth),(((i/4)%filter->outwidth)*3));
    else if (i%4==1)temp=cvmGet(lutraw,((i/4)/filter->outwidth),(1+(((i/4)%filter->outwidth)*3)));
    else if (i%4==2)temp=cvmGet(lutraw,((i/4)/filter->outwidth),(2+(((i/4)%filter->outwidth)*3)));
    filter->mapmod[i]=temp;
    //if (i%64==0)g_print("I is %d Mapmod-i: %f i-1=%f i-2=%f i-3=%f\n",i,filter->mapmod[i],filter->mapmod[i-1],filter->mapmod[i-2],filter->mapmod[i-3]);
  }
  */
  return modifycaps(filter);
 
}
static gboolean gst_clremap4yuv_init_cl(Gstclremap4yuv * filter){//initialize or update all the CL related stuff
        g_print("Init CL stuff.\n");
	//sleep(1);
        if (!(filter->width*filter->height)) {
	  g_print("Skipping, as imagesize is not recognized.\n");  
	  return FALSE;//dont do sth until it has been initialized
	}
	if (!filter->directory) {
	  g_print("Skipping, as directory is not set.\n");  
	  return FALSE;
	  }
        if (!gst_clremap4yuv_init_cvstuff(filter)){
	  g_print("Failed to init CV stuff.\n");  
	  return FALSE;
	}
	if (filter->context)gst_clremap4yuv_release_cl(filter);//delete everything if already used
	int gpu,err;
	if (!filter->clcpu)gpu=1;
	else gpu=0;
        filter->ret = clGetPlatformIDs(1, &(filter->platform_id), &(filter->ret_num_platforms));
	if((filter->ret=clGetDeviceIDs(filter->platform_id, gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &(filter->device_id), NULL)) != CL_SUCCESS){
		g_print("Failed to create a device group! Errorcode %d",err);//for CPU this message is normal
		return FALSE;
	}	
	// Create a compute context 
	if(!(filter->context = clCreateContext(0, 1, &(filter->device_id), NULL, NULL, &err))){
		g_print("Failed to create a compute context! Errorcode %d",err);
		return FALSE;
	}
	// Create a command commands
	//
	if(!(filter->commands = clCreateCommandQueue(filter->context, filter->device_id, 0, &err))){
		g_print("Failed to create a command commands! Errorcode %d",err);
		return FALSE;		
	}
	// Load Program source
	//
	char *source = load_program_source(filter->kernelfile);
        if(!source)
        {
          printf("Error: Failed to load compute program from file!\n");
          return FALSE;    
        }
	// Create the compute program from the source buffer
	//
	if(!(filter->program = clCreateProgramWithSource(filter->context, 1, (const char **) & source, NULL, &err))){
	  g_print("Failed to create compute program! Errorcode %d\n",err);
 	  free(source);
	  return FALSE;
	}
 	free(source);
	// Build the program executable
	//
	#ifdef MAC
    	char *flags = "-cl-fast-relaxed-math -DMAC";
	#else
    	char *flags = "-cl-fast-relaxed-math";
	#endif
	err = clBuildProgram(filter->program, 0, NULL, flags, NULL, NULL);
 	//free(flags);	
	if (err != CL_SUCCESS)
	{
		size_t len;
		char buffer[2048];
		
		g_print("Error: Failed to build program executable!\n");
		clGetProgramBuildInfo(filter->program, filter->device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
		g_print("%s\n", buffer);
		return FALSE;
	}
	
	// Create the compute kernel in the program we wish to run
	//
	filter->kernel = clCreateKernel(filter->program,"remap4", &err);
	if (!filter->kernel || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}

	filter->kernelyuv = clCreateKernel(filter->program,"yuvtorgba", &err);
	if (!filter->kernelyuv || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}

	filter->inyuv1 = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, (filter->width*filter->height*filter->bpp)/8 , NULL, NULL);
	filter->inyuv2 = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, (filter->width*filter->height*filter->bpp)/8 , NULL, NULL);
	filter->inyuv3 = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, (filter->width*filter->height*filter->bpp)/8 , NULL, NULL);
	filter->inyuv4 = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, (filter->width*filter->height*filter->bpp)/8 , NULL, NULL);


	filter->in1 = clCreateImage2D (filter->context,CL_MEM_READ_WRITE,&filter->format,filter->width,filter->height,0,NULL,&err);
	filter->in2 = clCreateImage2D (filter->context,CL_MEM_READ_WRITE,&filter->format,filter->width,filter->height,0,NULL,&err);
	filter->in3 = clCreateImage2D (filter->context,CL_MEM_READ_WRITE,&filter->format,filter->width,filter->height,0,NULL,&err);
	filter->in4 = clCreateImage2D (filter->context,CL_MEM_READ_WRITE,&filter->format,filter->width,filter->height,0,NULL,&err);

	filter->output =clCreateImage2D (filter->context,CL_MEM_WRITE_ONLY,&filter->format,filter->vwidth,filter->vheight,0,NULL,&err); 

	//filter->lut = clCreateImage2D (filter->context,CL_MEM_READ_ONLY||CL_MEM_COPY_HOST_PTR,&filter->lut_format,filter->width,filter->height,0,filter->mapmod,&err);
	filter->lut = clCreateImage2D (filter->context,CL_MEM_READ_ONLY,&filter->lut_format,filter->vwidth,filter->vheight,0,NULL,&err);
	

	if (!filter->in1 || !filter->in2 || !filter->in3 || !filter->in4 || !filter->output || !filter->lut){
		g_print("Failed to allocate device memory!");
		return FALSE;
	}
        size_t origin[3]={0,0,0};
        size_t region[3]={filter->vwidth,filter->vheight,1};
        err = clEnqueueWriteImage (filter->commands,filter->lut,CL_TRUE,origin,
region,0,0,filter->mapmod,0,NULL,NULL);	
	if (err != CL_SUCCESS){
	  g_print("Failed to copy map, %s\n",print_cl_errstring(err));
	  return FALSE;
	}
	return TRUE; 
}
/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_clremap4yuv_init (Gstclremap4yuv * filter,
    Gstclremap4yuvClass * gclass)
{
  filter->sink1 = gst_pad_new_from_static_template (&sink_factory, "sink1");
  gst_pad_set_setcaps_function (filter->sink1,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_set_caps));
  gst_pad_set_getcaps_function (filter->sink1,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_get_caps));
  gst_pad_set_chain_function (filter->sink1,
                              GST_DEBUG_FUNCPTR(gst_clremap4yuv_chain));
  filter->sink2 = gst_pad_new_from_static_template (&sink_factory, "sink2");
  gst_pad_set_setcaps_function (filter->sink2,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_set_caps));
  gst_pad_set_getcaps_function (filter->sink2,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_get_caps));
  gst_pad_set_chain_function (filter->sink2,
                              GST_DEBUG_FUNCPTR(gst_clremap4yuv_chain));

  filter->sink3 = gst_pad_new_from_static_template (&sink_factory, "sink3");
  gst_pad_set_setcaps_function (filter->sink3,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_set_caps));
  gst_pad_set_getcaps_function (filter->sink3,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_get_caps));
  gst_pad_set_chain_function (filter->sink3,
                              GST_DEBUG_FUNCPTR(gst_clremap4yuv_chain));

  filter->sink4 = gst_pad_new_from_static_template (&sink_factory, "sink4");
  gst_pad_set_setcaps_function (filter->sink4,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_set_caps));
  gst_pad_set_getcaps_function (filter->sink4,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_get_caps));
  gst_pad_set_chain_function (filter->sink4,
                              GST_DEBUG_FUNCPTR(gst_clremap4yuv_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_get_caps));
  gst_pad_set_event_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_clremap4yuv_event));


  gst_element_add_pad (GST_ELEMENT (filter), filter->sink4);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sink3);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sink2);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sink1);

  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->ready= FALSE;
  filter->platform_id = NULL;
  filter->context= NULL;
  filter->sinkcaps= NULL;
  filter->srccaps= NULL;
  filter->format.image_channel_order=CL_RGBA;
  filter->format.image_channel_data_type=CL_UNSIGNED_INT8;
  filter->lut_format.image_channel_order=CL_RGBA;
  filter->lut_format.image_channel_data_type=CL_FLOAT;
  filter->mapx=NULL;
  filter->mapy=NULL;
  filter->mapmod=NULL;
  filter->mapmodsize=0;
  filter->cpu=0;
  filter->crop=0;
  filter->width=filter->outwidth=filter->outheight=filter->height=filter->vheight=filter->vwidth=0;
  filter->ready1=filter->ready2=filter->ready3=filter->ready4=FALSE;
  filter->mutex=g_mutex_new();
  filter->buf1=filter->buf2=filter->buf3=filter->buf4=NULL;
  filter->modulo=8;
  filter->offset=0;
  filter->bpp=12;//for the input images
  filter->clcpu=0;
  //g_print("gst_clremap4yuv_init.\n");
  //sleep(1);
}

static void
gst_clremap4yuv_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstclremap4yuv *filter = GST_CLREMAP4YUV (object);
  g_print("gst_clremap4yuv_set_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_KERNELFILE:
      g_free (filter->kernelfile);
      filter->kernelfile = g_value_dup_string (value);
      filter->ready=gst_clremap4yuv_init_cl(filter);
      break;
   case PROP_DIRECTORY:
      g_free (filter->directory);
      filter->directory = g_value_dup_string (value);
      filter->ready=gst_clremap4yuv_init_cl(filter);      
      break;
    case PROP_CPU:
      filter->cpu = g_value_get_boolean (value);
      break;
    case PROP_CROP:
      filter->crop = g_value_get_boolean (value);
      break;
    case PROP_CLCPU:
      filter->clcpu = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_clremap4yuv_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstclremap4yuv *filter = GST_CLREMAP4YUV (object);
  g_print("gst_clremap4yuv_get_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_KERNELFILE:
      g_value_set_string (value, filter->kernelfile);
      break;
    case PROP_DIRECTORY:
      g_value_set_string (value, filter->directory);
      break;
    case PROP_CPU:
      g_value_set_boolean (value, filter->cpu);
      break;
    case PROP_CROP:
      g_value_set_boolean (value, filter->crop);
      break;
    case PROP_CLCPU:
      g_value_set_boolean (value, filter->clcpu);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_clremap4yuv_set_caps (GstPad * pad, GstCaps * caps)
{
  //g_print("gst_clremap4yuv_set_caps.\n");
  Gstclremap4yuv *filter;
  //GstPad *otherpad;
  int oldwidth,oldheight;
  gboolean ret=TRUE;
  
  filter = GST_CLREMAP4YUV (gst_pad_get_parent (pad));
  g_mutex_lock(filter->mutex);
  //otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  if (pad == filter->srcpad) return TRUE;


  oldwidth=filter->width;
  oldheight=filter->height;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->width)) ret=FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) ret=FALSE;
  gst_video_parse_caps_framerate      (caps,&filter->fps_n,&filter->fps_d);

  if (filter->sinkcaps)gst_caps_unref(filter->sinkcaps);
  /*filter->caps=gst_caps_new_simple ("video/x-raw-rgb",
      		"width", G_TYPE_INT, filter->width,
      		"height", G_TYPE_INT, filter->height,
      		"bpp",G_TYPE_INT , 32,
      		"depth",G_TYPE_INT , 8,
		"framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
		NULL);*/
  filter->sinkcaps=gst_caps_copy(caps);

  if ((oldwidth*oldheight)!=(filter->width*filter->height)){
    filter->ready=gst_clremap4yuv_init_cl(filter);
    ret=filter->ready;
    //if (ret)ret=gst_pad_set_caps (filter->srcpad, filter->srccaps);
  }
  if (ret)ret=gst_pad_set_caps (filter->srcpad, filter->srccaps);
  if (!ret) g_print ("GstCLREMAP4YUV Warning Set Caps will return false.\n");
  g_mutex_unlock(filter->mutex);
  return ret;
}

static GstCaps* gst_clremap4yuv_get_caps (GstPad * pad, GstCaps * caps){
 return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
}

/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
gst_clremap4yuv_chain (GstPad * pad, GstBuffer * buf)
{
  Gstclremap4yuv *filter;
  filter = GST_CLREMAP4YUV (GST_OBJECT_PARENT (pad));
  int err=CL_SUCCESS;



  size_t globaly[2]={0,0};                      // global domain size for our calculation
  size_t localy[2]={0,0}; 

  size_t global[2]={0,0};                      // global domain size for our calculation
  size_t local[2]={0,0}; 
  size_t origin[3]={0,0,0};
  size_t region[3]={filter->width,filter->height,1};
  
  //size_t regionout[3]={filter->outwidth,filter->outheight,1};//hier passts, weil wir nur die tatsächlichen brauchen
  size_t regionout[3];
  if (filter->crop){
	regionout[0]=filter->outwidth;
	regionout[1]=filter->outheight;
  }
  else {
	regionout[0]=filter->vwidth;
	regionout[1]=filter->vheight;
  }
  regionout[2]=1;
  GstFlowReturn ret;
  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
    //sleep(1);

  if (!filter->ready){
    g_print ("GstCLREMAP4YUV not ready for some reason/problem with kernel?.\n");
    return GST_FLOW_ERROR;
  }

  guint64 timestamp=buf->timestamp;
  guint64 offset_end=buf->offset_end;
  guint64 offset=buf->offset;
  guint64 duration=buf->duration;
  g_print ("getremapmutex.\n"); 
  g_mutex_lock(filter->mutex);
  g_print ("gotremapmutex.\n");

  if (pad==filter->sink1){
    //err = clEnqueueWriteImage (filter->commands,filter->in1,CL_FALSE,origin,region,0,0,buf->data,0,NULL,NULL);
    err = clEnqueueWriteBuffer(filter->commands, filter->inyuv1, CL_FALSE, 0, buf->size, buf->data, 0, NULL, NULL);
    filter->buf1=buf;
    filter->ready1=TRUE;
  }

  if (pad==filter->sink2){
    //err = clEnqueueWriteImage (filter->commands,filter->in2,CL_FALSE,origin,region,0,0,buf->data,0,NULL,NULL);
    err = clEnqueueWriteBuffer(filter->commands, filter->inyuv2, CL_FALSE, 0, buf->size, buf->data, 0, NULL, NULL);
    filter->buf2=buf;
    filter->ready2=TRUE;
  }

  if (pad==filter->sink3){
    //err = clEnqueueWriteImage (filter->commands,filter->in3,CL_FALSE,origin,region,0,0,buf->data,0,NULL,NULL);
    err = clEnqueueWriteBuffer(filter->commands, filter->inyuv3, CL_FALSE, 0, buf->size, buf->data, 0, NULL, NULL);
    filter->buf3=buf;
    filter->ready3=TRUE;
  }

  if (pad==filter->sink4){
    //err = clEnqueueWriteImage (filter->commands,filter->in4,CL_FALSE,origin,region,0,0,buf->data,0,NULL,NULL);
    err = clEnqueueWriteBuffer(filter->commands, filter->inyuv4, CL_FALSE, 0, buf->size, buf->data, 0, NULL, NULL);
    filter->buf4=buf;
    filter->ready4=TRUE;
  }

  //check wether we are ready

  if (err != CL_SUCCESS){
	printf("Error: Failed to write to source array! %s\n",print_cl_errstring(err));
	g_mutex_unlock(filter->mutex);
	return GST_FLOW_ERROR;
  }

  if ((!filter->ready1)||(!filter->ready2)||(!filter->ready3)||(!filter->ready4)){
    g_mutex_unlock(filter->mutex);
    return GST_FLOW_OK;
  }


  clFinish(filter->commands);//wait until async copies are done
  // Set the arguments to our compute kernel
  globaly[0] = filter->width;
  globaly[1] = filter->height; 
  localy[0] = 8;
  localy[1] = 8;
  err = 0;  
  err  |= clSetKernelArg(filter->kernelyuv, 0, sizeof(cl_mem), &(filter->inyuv1));
  err  |= clSetKernelArg(filter->kernelyuv, 1, sizeof(cl_mem), &(filter->in1));
  err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuv, 2, NULL, globaly, localy, 0, NULL, NULL);
  clFinish(filter->commands);

  err  |= clSetKernelArg(filter->kernelyuv, 0, sizeof(cl_mem), &(filter->inyuv2));
  err  |= clSetKernelArg(filter->kernelyuv, 1, sizeof(cl_mem), &(filter->in2));
  err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuv, 2, NULL, globaly, localy, 0, NULL, NULL);
  clFinish(filter->commands);

  err  |= clSetKernelArg(filter->kernelyuv, 0, sizeof(cl_mem), &(filter->inyuv3));
  err  |= clSetKernelArg(filter->kernelyuv, 1, sizeof(cl_mem), &(filter->in3));
  err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuv, 2, NULL, globaly, localy, 0, NULL, NULL);
  clFinish(filter->commands);

  err  |= clSetKernelArg(filter->kernelyuv, 0, sizeof(cl_mem), &(filter->inyuv4));
  err  |= clSetKernelArg(filter->kernelyuv, 1, sizeof(cl_mem), &(filter->in4));
  err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuv, 2, NULL, globaly, localy, 0, NULL, NULL);
  clFinish(filter->commands);

  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to transform YUV to RGB! %d %s\n", err,print_cl_errstring(err));
    return GST_FLOW_ERROR;
  }

  //
  err = 0;
  err  = clSetKernelArg(filter->kernel, 0, sizeof(cl_mem), &(filter->lut));
  err |= clSetKernelArg(filter->kernel, 1, sizeof(cl_mem), &(filter->output));
  err |= clSetKernelArg(filter->kernel, 2, sizeof(cl_mem), &(filter->in1));
  err |= clSetKernelArg(filter->kernel, 3, sizeof(cl_mem), &(filter->in2));
  err |= clSetKernelArg(filter->kernel, 4, sizeof(cl_mem), &(filter->in3));
  err |= clSetKernelArg(filter->kernel, 5, sizeof(cl_mem), &(filter->in4));
  if (err != CL_SUCCESS)
  {
	printf("Error: Failed to set kernel arguments! %d, %s\n", err,print_cl_errstring(err));
	return GST_FLOW_ERROR;
  }

  err = clGetKernelWorkGroupInfo(filter->kernel, filter->device_id, CL_KERNEL_WORK_GROUP_SIZE, sizeof(int) * 4 + sizeof(float) * 2, &local, NULL);
  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to retrieve kernel work group info! %d\n", err);
    return GST_FLOW_ERROR;
  }
  global[0] = filter->vwidth;
  global[1] = filter->vheight;
  //if (local[0]==0){
    local[0] = filter->modulo;
    local[1] = filter->modulo;
  //}

  printf("remap Work group sizes global0: %d global1: %d local0: %d local1: %d!\n",(int)global[0],(int)global[1],(int)local[0],(int)local[1]);
  //local[0] = 64;

   err = clEnqueueNDRangeKernel(filter->commands, filter->kernel, 2, NULL, global, local, 0, NULL, NULL);
  if (err)
  {
    printf("Failed to execute kernel!, code %d, %s\n",err,print_cl_errstring(err));
    return GST_FLOW_ERROR;
  }
//inbetween stuff
//unref old buffers:
  gst_buffer_unref(filter->buf1);
  gst_buffer_unref(filter->buf2);
  gst_buffer_unref(filter->buf3);
  gst_buffer_unref(filter->buf4);
  //filter->ready1=FALSE;
  //filter->ready2=FALSE;
  //filter->ready3=FALSE;
  //filter->ready4=FALSE;

  //alloc new buffer during gpu is working:)
  GstBuffer *out;
  if (filter->crop)out=gst_buffer_new_and_alloc (filter->outwidth*filter->outheight*4);
  else out=gst_buffer_new_and_alloc (filter->vwidth*filter->vheight*4);
  gst_buffer_set_caps(out,filter->srccaps);
/*  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         offset,
                                                         filter->outwidth*filter->outheight*4,
                                                         filter->srccaps,
                                                         &out);
  if (ret!=GST_FLOW_OK){
    if (!gst_pad_is_linked) g_print ("GSTCLREMAP4YUV: Error allocating buffer, becaus pad is not linked.\n");
    else g_print ("GSTCLREMAP4YUV: Error allocating buffer.\n");
    return ret;
  }*/

  out->timestamp=timestamp;
  out->offset_end=filter->offset+1;
  out->offset=filter->offset++;
  out->duration=duration;

  //sync and download
  // Wait for the command commands to get serviced before reading back results
  clFinish(filter->commands);
  err = clEnqueueReadImage (filter->commands,filter->output,CL_TRUE,origin,regionout,0,0,out->data,0,NULL,NULL);
  if (err != CL_SUCCESS)
  {
    GST_ELEMENT_ERROR(filter, LIBRARY, FAILED,
  ("Error: Failed to read output array! code."), ("%s",print_cl_errstring(err)));
    //printf("Error: Failed to read output array! code %d, %s\n", err,print_cl_errstring(err));
  }
  g_print ("remappush.\n"); 
  ret=gst_pad_push (filter->srcpad, out);
  filter->ready1=filter->ready2=filter->ready3=filter->ready4=FALSE;
  g_print ("relremapmutex.\n");
  g_mutex_unlock(filter->mutex);

  return ret;
}

static gboolean
gst_clremap4yuv_event (GstPad  *pad, GstEvent * event){
  Gstclremap4yuv* filter = GST_CLREMAP4YUV (gst_pad_get_parent (pad));
  gst_pad_push_event (filter->sink1, gst_event_ref(event));
  gst_pad_push_event (filter->sink2, gst_event_ref(event));
  gst_pad_push_event (filter->sink3, gst_event_ref(event));
  return gst_pad_push_event (filter->sink4, event);
}

static GstStateChangeReturn
gst_clremap4yuv_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstclremap4yuv *filter =  GST_CLREMAP4YUV(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      gst_clremap4yuv_release_cl(filter);
      //gst_my_filter_free_memory (filter);
      break;
    default:
      break;
  }
 return ret;
}



/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
clremap4yuv_init (GstPlugin * clremap4yuv)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template clremap4yuv' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_clremap4yuv_debug, "clremap4yuv",
      0, "Template clremap4yuv");

  return gst_element_register (clremap4yuv, "clremap4yuv", GST_RANK_NONE,
      GST_TYPE_CLREMAP4YUV);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstclremap4yuv"
#endif

/* gstreamer looks for this structure to register clremap4yuvs
 *
 * exchange the string 'Template clremap4yuv' with your clremap4yuv description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "clremap4yuv",
    "Template clremap4yuv",
    clremap4yuv_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
