/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, inclundistortuding without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-clundistort
 *
 * FIXME:Describe clundistort here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! clundistort ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <errno.h>
#include <assert.h>
#include <string.h>

#ifdef __APPLE__
#include <OpenCL/opencl.h>
#else
#include <CL/cl.h>
#endif

#include <opencv/cv.h>
#include "gstclundistort.h"

GST_DEBUG_CATEGORY_STATIC (gst_clundistort_debug);
#define GST_CAT_DEFAULT gst_clundistort_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_KERNELFILE,
  PROP_DIRECTORY,
  PROP_CPU,
  PROP_HW,
  PROP_CLCPU
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA";"GST_VIDEO_CAPS_YUV("I420")";"GST_VIDEO_CAPS_YUV("YUY2"))
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_RGBA)
    );

GST_BOILERPLATE (Gstclundistort, gst_clundistort, GstElement,
    GST_TYPE_ELEMENT);

static void gst_clundistort_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_clundistort_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps* gst_clundistort_get_caps (GstPad * pad, GstCaps * caps);
static gboolean gst_clundistort_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_clundistort_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_clundistort_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "clundistort",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the clundistort's class */
static void
gst_clundistort_class_init (GstclundistortClass * klass)
{
  GObjectClass *gobject_class;
  //GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  //gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_clundistort_set_property;
  gobject_class->get_property = gst_clundistort_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_KERNELFILE,
      g_param_spec_string ("kernel", "Kernel", "Filename of kernel file",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_DIRECTORY,
      g_param_spec_string ("directory", "Directory", "Use this directory for lens  models.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_CPU,
      g_param_spec_boolean ("cpu", "Cpu", "Use CPU instead of GPU?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_HW,
      g_param_spec_boolean ("halfwidth", "Halfwidth", "Use optimized halfwidth kernel for YUY2?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_CLCPU,
      g_param_spec_boolean ("clcpu", "Clcpu", "Whether to do OpenCL based CPU processing?",
          FALSE, G_PARAM_READWRITE));
}

static char *load_program_source(const char *filename)
{
    struct stat statbuf;
    FILE        *fh;
    char        *source;
	
    fh = fopen(filename, "r");
    if (fh == 0)
        return 0;
	
    stat(filename, &statbuf);
    source = (char *) malloc(statbuf.st_size + 1);
    if (0==fread(source, statbuf.st_size, 1, fh))return NULL;
    source[statbuf.st_size] = '\0';
	
    return source;
}

static void gst_clundistort_release_cl(Gstclundistort * filter){ //release all the CLUNDISTORT related stuff
  	clReleaseMemObject(filter->input);
	clReleaseMemObject(filter->output);
	clReleaseProgram(filter->program);
	clReleaseKernel(filter->kernel);
	clReleaseCommandQueue(filter->commands);
	clReleaseContext(filter->context);
}

static gboolean gst_clundistort_init_cvstuff(Gstclundistort * filter){

  gchar * this_object_name = (gchar *) malloc(strlen(filter->directory)+20);
  strcpy(this_object_name, filter->directory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Intrinsics.xml");
  CvMat *intrinsic = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!intrinsic){
    g_print("Could not load %s\n",this_object_name);
    return FALSE;
  }
  strcpy(this_object_name, filter->directory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Distortion.xml");  
  CvMat *distortion = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!distortion){
    g_print("Could not load %s\n",this_object_name);    
    return FALSE;
  }
 // Build the undistort map used for all subsequent frames.
  CvSize size;
  size.width=filter->width;
  size.height=filter->height;
  if (filter->mapx) cvReleaseImage(&filter->mapx);
  if (filter->mapy) cvReleaseImage(&filter->mapy);
  filter->mapx = cvCreateImage(size, IPL_DEPTH_32F, 1 );
  filter->mapy = cvCreateImage(size, IPL_DEPTH_32F, 1 );
  cvInitUndistortMap(intrinsic,distortion,filter->mapx,filter->mapy);
  if (filter->mapmod)g_free(filter->mapmod);
  filter->mapmodsize=filter->width*filter->height*4*sizeof(float);
  filter->mapmod=(float*)g_malloc(filter->mapmodsize);
  int i;
  //copy mapx and mapy into our structure
  float temp;
  for (i=0;i<(filter->mapmodsize/sizeof(float));i++){
    //g_print("I is %d, accessed object is %d Mapmodsize is %d\n",i,i/4,filter->mapmodsize);   
    temp=0.0;
    if (i%4==0)temp=((float*)(filter->mapx->imageData))[i/4];
    else if (i%4==1)temp=((float*)(filter->mapy->imageData))[i/4];
    filter->mapmod[i]=temp;
    //filter->mapmod[i]=0.0;
    //if (i%64==0)g_print("I is %d Mapmod-i: %f i-1=%f i-2=%f i-3=%f\n",i,filter->mapmod[i],filter->mapmod[i-1],filter->mapmod[i-2],filter->mapmod[i-3]);
  }
  return TRUE;
  //cvRemap( t, image, mapx, mapy ); 
}
static gboolean gst_clundistort_init_cl(Gstclundistort * filter){//initialize or update all the CL related stuff
        g_print("Init CL stuff.\n");
	//sleep(1);
        if (!(filter->width*filter->height)) {
	  g_print("Skipping, as imagesize is not recognized.\n");  
	  return FALSE;//dont do sth until it has been initialized
	}
	if (!filter->directory) {
	  g_print("Skipping, as directory is not set.\n");  
	  return FALSE;
	  }
        if (!gst_clundistort_init_cvstuff(filter)){
	  g_print("Failed to init CV stuff.\n");  
	  return FALSE;
	}
	if (filter->context)gst_clundistort_release_cl(filter);//delete everything if already used
	int gpu,err;
	if (!filter->clcpu)gpu=1;
	else gpu=0;
        filter->ret = clGetPlatformIDs(1, &(filter->platform_id), &(filter->ret_num_platforms));
	if((filter->ret=clGetDeviceIDs(filter->platform_id, gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &(filter->device_id), NULL)) != CL_SUCCESS){
		g_print("Failed to create a device group! Errorcode %d",err);//for CPU this message is normal
		return FALSE;
	}	
	// Create a compute context 
	if(!(filter->context = clCreateContext(0, 1, &(filter->device_id), NULL, NULL, &err))){
		g_print("Failed to create a compute context! Errorcode %d",err);
		return FALSE;
	}
	// Create a command commands
	//
	if(!(filter->commands = clCreateCommandQueue(filter->context, filter->device_id, 0, &err))){
		g_print("Failed to create a command commands! Errorcode %d",err);
		return FALSE;		
	}
	// Load Program source
	//
	char *source = load_program_source(filter->kernelfile);
        if(!source)
        {
          printf("Error: Failed to load compute program from file!\n");
          return FALSE;    
        }
	// Create the compute program from the source buffer
	//
	if(!(filter->program = clCreateProgramWithSource(filter->context, 1, (const char **) & source, NULL, &err))){
	  g_print("Failed to create compute program! Errorcode %d\n",err);
 	  free(source);
	  return FALSE;
	}
 	free(source);
	// Build the program executable
	//
	#ifdef MAC
    	char *flags = "-cl-fast-relaxed-math -DMAC";
	#else
    	char *flags = "-cl-fast-relaxed-math";
	#endif
	err = clBuildProgram(filter->program, 0, NULL, flags, NULL, NULL);
 	//free(flags);	
	if (err != CL_SUCCESS)
	{
		size_t len;
		char buffer[2048];
		
		g_print("Error: Failed to build program executable!\n");
		clGetProgramBuildInfo(filter->program, filter->device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
		g_print("%s\n", buffer);
		return FALSE;
	}
	
	// Create the compute kernel in the program we wish to run
	//
	filter->kernel = clCreateKernel(filter->program,"imgwarp", &err);
	if (!filter->kernel || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}
	filter->kernelyuv = clCreateKernel(filter->program,"yuvtorgba", &err);
	if (!filter->kernelyuv || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}

	filter->kernelyuy2 = clCreateKernel(filter->program,"yuy2torgba", &err);
	if (!filter->kernelyuy2 || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}
	filter->kernelyuy2hw = clCreateKernel(filter->program,"yuy2torgbahalfwidth", &err);
	if (!filter->kernelyuy2 || err != CL_SUCCESS){
		g_print("Failed to create compute kernel,  Errorcode %d, %s\n",err,print_cl_errstring(err));
	}

        if(!filter->rgbin)filter->inyuv = clCreateBuffer(filter->context,  CL_MEM_READ_ONLY, (filter->width*filter->height*filter->bpp)/8 , NULL, NULL);



	if(!filter->rgbin)filter->input = clCreateImage2D (filter->context,CL_MEM_READ_WRITE,&filter->format,filter->width,filter->height,0,NULL,&err);
	else filter->input = clCreateImage2D (filter->context,CL_MEM_READ_ONLY,&filter->format,filter->width,filter->height,0,NULL,&err);


	filter->output =clCreateImage2D (filter->context,CL_MEM_WRITE_ONLY,&filter->format,filter->width,filter->height,0,NULL,&err); 

	//filter->lut = clCreateImage2D (filter->context,CL_MEM_READ_ONLY||CL_MEM_COPY_HOST_PTR,&filter->lut_format,filter->width,filter->height,0,filter->mapmod,&err);
	filter->lut = clCreateImage2D (filter->context,CL_MEM_READ_ONLY,&filter->lut_format,filter->width,filter->height,0,NULL,&err);
	

	if (!filter->input || !filter->output || !filter->lut){
		g_print("Failed to allocate device memory!");
		return FALSE;
	}
        //size_t global[2]={0,0};                      // global domain size for our calculation
        //size_t local[2]={0,0}; 
        size_t origin[3]={0,0,0};
        size_t region[3]={filter->width,filter->height,1};
        err = clEnqueueWriteImage (filter->commands,filter->lut,CL_TRUE,origin,
region,0,0,filter->mapmod,0,NULL,NULL);	
	if (err != CL_SUCCESS){
	  g_print("Failed to copy map, %s\n",print_cl_errstring(err));
	  return FALSE;
	}
	return TRUE; 
}
/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_clundistort_init (Gstclundistort * filter,
    GstclundistortClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_clundistort_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_clundistort_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_clundistort_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_clundistort_get_caps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->ready= FALSE;
  filter->platform_id = NULL;
  filter->context= NULL;
  filter->caps= NULL;
  filter->format.image_channel_order=CL_RGBA;
  filter->format.image_channel_data_type=CL_UNSIGNED_INT8;
  filter->lut_format.image_channel_order=CL_RGBA;
  filter->lut_format.image_channel_data_type=CL_FLOAT;
  filter->mapx=NULL;
  filter->mapy=NULL;
  filter->mapmod=NULL;
  filter->mapmodsize=0;
  filter->cpu=0;
  filter->clcpu=0;
  filter->halfwidth=TRUE;
  //g_print("gst_clundistort_init.\n");
  //sleep(1);
}

static void
gst_clundistort_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstclundistort *filter = GST_CLUNDISTORT (object);
  g_print("gst_clundistort_set_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_KERNELFILE:
      g_free (filter->kernelfile);
      filter->kernelfile = g_value_dup_string (value);
      filter->ready=gst_clundistort_init_cl(filter);
      break;
   case PROP_DIRECTORY:
      g_free (filter->directory);
      filter->directory = g_value_dup_string (value);
      filter->ready=gst_clundistort_init_cl(filter);      
      break;
    case PROP_CPU:
      filter->cpu = g_value_get_boolean (value);
      break;
    case PROP_HW:
      filter->halfwidth = g_value_get_boolean (value);
      break;
    case PROP_CLCPU:
      filter->clcpu = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_clundistort_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstclundistort *filter = GST_CLUNDISTORT (object);
  g_print("gst_clundistort_get_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_KERNELFILE:
      g_value_set_string (value, filter->kernelfile);
      break;
    case PROP_DIRECTORY:
      g_value_set_string (value, filter->directory);
      break;
    case PROP_CPU:
      g_value_set_boolean (value, filter->cpu);
      break;
    case PROP_HW:
      g_value_set_boolean (value, filter->halfwidth);
      break;
    case PROP_CLCPU:
      g_value_set_boolean (value, filter->clcpu);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_clundistort_set_caps (GstPad * pad, GstCaps * caps)
{
  //g_print("gst_clundistort_set_caps. %s\n",gst_caps_to_string(caps));
  Gstclundistort *filter;
  GstPad *otherpad;
  int oldwidth,oldheight;
  gboolean ret;
  
  filter = GST_CLUNDISTORT (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  oldwidth=filter->width;
  oldheight=filter->height;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  const gchar *mime;
  mime = gst_structure_get_name (structure); 
  if (0 == strcmp (mime, "video/x-raw-rgb"))filter->rgbin=TRUE;
  else filter->rgbin=FALSE;
  if (0 == strcmp (mime, "video/x-raw-yuv")){
    guint32 fourccval;
    gst_structure_get_fourcc(structure,"format",&fourccval);
    if (fourccval==GST_MAKE_FOURCC ('Y', 'U', 'Y', '2'))filter->bpp=16;
    else if (fourccval==GST_MAKE_FOURCC ('I', '4', '2', '0'))filter->bpp=12;
  }

  if (!gst_structure_get_int(structure,"width",&filter->width)) return FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->height)) return FALSE;
  gst_video_parse_caps_framerate      (caps,&filter->fps_n,&filter->fps_d);

  if (filter->caps)gst_caps_unref(filter->caps);
  filter->caps=gst_caps_new_simple ("video/x-raw-rgb",
  "width", G_TYPE_INT, filter->width,
  "height", G_TYPE_INT, filter->height,
  "bpp",G_TYPE_INT,32, 
  "depth",G_TYPE_INT,24, 
  "endianness",G_TYPE_INT,4321,
  "red_mask",G_TYPE_INT,65280, 
  "green_mask",G_TYPE_INT,16711680, 
  "blue_mask",G_TYPE_INT,-16777216,
  "framerate", GST_TYPE_FRACTION, filter->fps_n,filter->fps_d,
  NULL);

  //filter->caps=gst_caps_copy(caps);
   //g_print("gst_clundistort_set_ other caps. %s\n",gst_caps_to_string(filter->caps));
  if ((oldwidth*oldheight)!=(filter->width*filter->height)){
    filter->ready=gst_clundistort_init_cl(filter);
  }
  ret=gst_pad_set_caps (otherpad, filter->caps);
  if (!ret) g_print ("GstCLUNDISTORT Warning Set Caps will return false.\n");
  return ret;
}

static GstCaps* gst_clundistort_get_caps (GstPad * pad, GstCaps * caps){
 return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_clundistort_chain_cpu (GstPad * pad, GstBuffer * buf)
{
  Gstclundistort *filter;
  filter = GST_CLUNDISTORT (GST_OBJECT_PARENT (pad));
  if (!filter->rgbin){
    g_print("WARNING CPU Mode only supports RGB, please use ffmpegcolorspace!\n");
    return GST_FLOW_ERROR;
  }
  GstFlowReturn ret;
  IplImage *process, *ipout;	
  CvSize size;
  size.width=filter->width;
  size.height=filter->height;
  process=cvCreateImageHeader(size, IPL_DEPTH_8U,4 );
  process->imageData=(char*)buf->data;
  GstBuffer* outbuf;
  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         buf->offset,
                                                         filter->width*filter->height*4,
                                                         filter->caps,
                                                         &outbuf);
  if (ret!=GST_FLOW_OK)return ret;
  ipout=cvCreateImageHeader( size, IPL_DEPTH_8U,4);
  ipout->imageData=(char*)outbuf->data;
  cvRemap(process,ipout,filter->mapx,filter->mapy,CV_INTER_LINEAR+CV_WARP_FILL_OUTLIERS, cvScalarAll(0));
  outbuf->timestamp=buf->timestamp;
  outbuf->duration=buf->duration;
  outbuf->offset=buf->offset;
  outbuf->offset_end=buf->offset_end;
  cvReleaseImageHeader(&process);
  cvReleaseImageHeader(&ipout);
  gst_buffer_unref(buf);  
  return gst_pad_push (filter->srcpad, outbuf);
}
static GstFlowReturn
gst_clundistort_chain (GstPad * pad, GstBuffer * buf)
{
  Gstclundistort *filter;
  filter = GST_CLUNDISTORT (GST_OBJECT_PARENT (pad));
  if (filter->cpu){
     return gst_clundistort_chain_cpu(pad,buf);
  }
  int err;

  size_t globaly[2]={0,0};                      // global domain size for our yuv calculation
  size_t localy[2]={0,0}; 


  size_t global[2]={0,0};                      // global domain size for our calculation
  size_t local[2]={0,0}; 
  size_t origin[3]={0,0,0};
  size_t region[3]={filter->width,filter->height,1};
  GstFlowReturn ret;
  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
    //sleep(1);


  //check wether we are ready



  if (!filter->ready){
    g_print ("GstCLUNDISTORT not ready for some reason/problem with kernel?.\n");
    return GST_FLOW_ERROR;
  }

  if (!filter->rgbin){
    err = clEnqueueWriteBuffer(filter->commands, filter->inyuv, CL_TRUE, 0, buf->size, buf->data, 0, NULL, NULL);
    if (err != CL_SUCCESS)
    {
      printf("Error: Failed to upload YUV buffer! %s\n",print_cl_errstring(err));
      return GST_FLOW_ERROR;
    }
    //clFinish(filter->commands);//kann ja nie schaden:)
    globaly[1] = filter->height; 
    localy[0] = 8;
    localy[1] = 8;
    err = 0; 
    if (filter->bpp==12){
      globaly[0] = filter->width;
      err  |= clSetKernelArg(filter->kernelyuv, 0, sizeof(cl_mem), &(filter->inyuv));
      err  |= clSetKernelArg(filter->kernelyuv, 1, sizeof(cl_mem), &(filter->input));
      err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuv, 2, NULL, globaly, localy, 0, NULL, NULL);
    }
    else if ((filter->bpp==16)&&(!filter->halfwidth)){ 
      globaly[0] = filter->width;
      err  |= clSetKernelArg(filter->kernelyuy2, 0, sizeof(cl_mem), &(filter->inyuv));
      err  |= clSetKernelArg(filter->kernelyuy2, 1, sizeof(cl_mem), &(filter->input));
      err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuy2, 2, NULL, globaly, localy, 0, NULL, NULL);
    }
    else if ((filter->bpp==16)&&(filter->halfwidth)){ 
      globaly[0] = filter->width/2;
      err  |= clSetKernelArg(filter->kernelyuy2hw, 0, sizeof(cl_mem), &(filter->inyuv));
      err  |= clSetKernelArg(filter->kernelyuy2hw, 1, sizeof(cl_mem), &(filter->input));
      err |= clEnqueueNDRangeKernel(filter->commands, filter->kernelyuy2hw, 2, NULL, globaly, localy, 0, NULL, NULL);
    }
  clFinish(filter->commands);   
  }
  //upload and launch
  // Write our data set into the input array in device memory 
  else{//upload as rbg
    err = clEnqueueWriteImage (filter->commands,filter->input,CL_TRUE,origin,
region,0,0,buf->data,0,NULL,NULL);
    if (err != CL_SUCCESS)
    {
	printf("Error: Failed to write to source array! %s\n",print_cl_errstring(err));
	return GST_FLOW_ERROR;
    }
  }
  // Set the arguments to our compute kernel
  //
  err = 0;
  err  = clSetKernelArg(filter->kernel, 0, sizeof(cl_mem), &(filter->input));
  err |= clSetKernelArg(filter->kernel, 1, sizeof(cl_mem), &(filter->output));
  err |= clSetKernelArg(filter->kernel, 2, sizeof(cl_mem), &(filter->lut));
  //err |= clSetKernelArg(filter->kernel, 3, sizeof(unsigned int), &(filter->height));
  if (err != CL_SUCCESS)
  {
	printf("Error: Failed to set kernel arguments! %d, %s\n", err,print_cl_errstring(err));
	return GST_FLOW_ERROR;
  }

  err = clGetKernelWorkGroupInfo(filter->kernel, filter->device_id, CL_KERNEL_WORK_GROUP_SIZE, sizeof(int) * 4 + sizeof(float) * 2, &local, NULL);
  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to retrieve kernel work group info! %d\n", err);
    return GST_FLOW_ERROR;
  }
  global[0] = filter->width;
  global[1] = filter->height;
  //if (local[0]==0){
    local[0] = 8;
    local[1] = 8;
  //}

  if (!filter->silent)printf("Work group sizes global0: %d global1: %d local0: %d local1: %d!\n",(int)global[0],(int)global[1],(int)local[0],(int)local[1]);
  //local[0] = 64;
  //err = clEnqueueNDRangeKernel(filter->commands, filter->kernel, 2, NULL, &global, &local, 0, NULL, NULL);
   err = clEnqueueNDRangeKernel(filter->commands, filter->kernel, 2, NULL, global, local, 0, NULL, NULL);
  if (err)
  {
    printf("Failed to execute kernel!, code %d, %s\n",err,print_cl_errstring(err));
    return GST_FLOW_ERROR;
  }
//inbetween stuff
  //alloc new buffer during gpu is working:)
  //g_print("gst_clundistort_ before pushing with caps. %s\n",gst_caps_to_string(filter->caps));
  GstBuffer *out;
  ret=gst_pad_alloc_buffer_and_set_caps   (filter->srcpad,
                                                         buf->offset,
                                                         filter->width*filter->height*4,
                                                         filter->caps,
                                                         &out);
  if (ret!=GST_FLOW_OK){
    g_print ("GSTCLUNDISTORT: Error allocating buffer.\n");
    return ret;
  }
  out->timestamp=buf->timestamp;
  out->offset_end=buf->offset_end;
  out->duration=buf->duration;
  gst_buffer_unref(buf);
  //sync and download
  // Wait for the command commands to get serviced before reading back results
//
  clFinish(filter->commands);
  //err = clEnqueueReadBuffer(filter->commands, filter->output, CL_TRUE, 0, filter->width*filter->height, out->data, 0, NULL, NULL );  
  err = clEnqueueReadImage (filter->commands,filter->output,CL_TRUE,origin,region,0,0,out->data,0,NULL,NULL);
  if (err != CL_SUCCESS)
  {
    printf("Error: Failed to read output array! code %d, %s\n", err,print_cl_errstring(err));
    exit(1);
  }
  //g_print ("Pushing buffer in 1 s.\n");
  //sleep(1);
  /* just push out the incoming buffer without touching it */
  //ret=gst_pad_push (filter->srcpad, out);
  ret=gst_pad_push (filter->srcpad, out);
  //g_print ("returning %d in 5 s.\n",ret);
  return ret;
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
clundistort_init (GstPlugin * clundistort)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template clundistort' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_clundistort_debug, "clundistort",
      0, "Template clundistort");

  return gst_element_register (clundistort, "clundistort", GST_RANK_NONE,
      GST_TYPE_CLUNDISTORT);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstclundistort"
#endif

/* gstreamer looks for this structure to register clundistorts
 *
 * exchange the string 'Template clundistort' with your clundistort description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "clundistort",
    "Template clundistort",
    clundistort_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
