/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-fusion
 *
 * FIXME:Describe fusion here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! fusion ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <sys/time.h>
#include "gstfusion.h"

GST_DEBUG_CATEGORY_STATIC (gst_fusion_debug);
#define GST_CAT_DEFAULT gst_fusion_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TIMEDIFF,
  PROP_TOONEW,
  PROP_TOOOLD,
  PROP_MAXLAME
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_SOMETIMES,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstfusion, gst_fusion, GstElement,
    GST_TYPE_ELEMENT);

static void gst_fusion_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_fusion_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstPad *gst_fusion_request_new_pad (GstElement * element,
    GstPadTemplate * temp, const gchar * name);
static void gst_fusion_release_pad (GstElement * element, GstPad * pad);

static gboolean gst_fusion_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_fusion_chain (GstPad * pad, GstBuffer * buf);
static gboolean
gst_fusion_event (GstPad  *pad, GstEvent * event);
/* GObject vmethod implementations */

static void
gst_fusion_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "fusion",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the fusion's class */
static void
gst_fusion_class_init (GstfusionClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_fusion_set_property;
  gobject_class->get_property = gst_fusion_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_TIMEDIFF,
      g_param_spec_uint64 ("timediff", "Timediff", "Max timediff of incoming buffers in ns",
          0,G_MAXUINT64, 100000000000ULL, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_TOONEW,
      g_param_spec_uint64 ("toonew", "Toonew", "delete buffer if it comes before x ns after old one, use to lower framerate",
          0,G_MAXUINT64, 0, G_PARAM_READWRITE));   
   g_object_class_install_property (gobject_class, PROP_TOOOLD,
      g_param_spec_uint64 ("tooold", "Tooold", "delete buffer if it is older than x us, 0 = egal",
          0,G_MAXUINT64, 0, G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, PROP_MAXLAME,
      g_param_spec_uint64 ("maxlame", "Maxlame", "How many times an element is allowed to block the others, reset after each incomming buffer(0=ignore)",
          0,G_MAXUINT64, 200, G_PARAM_READWRITE));      
   
  gstelement_class->request_new_pad =
      GST_DEBUG_FUNCPTR (gst_fusion_request_new_pad);
  gstelement_class->release_pad =
      GST_DEBUG_FUNCPTR (gst_fusion_release_pad);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_fusion_init (Gstfusion * filter,
    GstfusionClass * gclass)
{
//   filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
//   gst_pad_set_setcaps_function (filter->sinkpad,
//                                 GST_DEBUG_FUNCPTR(gst_fusion_set_caps));
//   gst_pad_set_getcaps_function (filter->sinkpad,
//                                 GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
//   gst_pad_set_chain_function (filter->sinkpad,
//                               GST_DEBUG_FUNCPTR(gst_fusion_chain));
// 
//   filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
//   gst_pad_set_getcaps_function (filter->srcpad,
//                                 GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
// 
//   gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
//   gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->timediff = 100000000000ULL;
  filter->toonew = 0;
  filter->tooold = 0; 
  filter->maxlame = 200; 
  filter->lastsend = 0;    
  filter->mutex = g_mutex_new();  
}

static void
gst_fusion_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstfusion *filter = GST_FUSION (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_TIMEDIFF:
      filter->timediff = g_value_get_uint64 (value);
      break;      
    case PROP_TOONEW:
      filter->toonew = g_value_get_uint64 (value);
      break;
     case PROP_TOOOLD:
      filter->tooold = g_value_get_uint64 (value);
      break; 
      case PROP_MAXLAME:
      filter->maxlame = g_value_get_uint64 (value);
      break;    
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_fusion_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstfusion *filter = GST_FUSION (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TIMEDIFF:
      g_value_set_uint64 (value, filter->timediff);
      break;       
    case PROP_TOONEW:
      g_value_set_uint64 (value, filter->toonew);
      break;  
     case PROP_TOOOLD:
      g_value_set_uint64 (value, filter->tooold);
      break;  
     case PROP_MAXLAME:  
      g_value_set_uint64 (value, filter->maxlame);
      break;    
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_fusion_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstfusion *filter;
  GstPad *otherpad=NULL;

  filter = GST_FUSION (gst_pad_get_parent (pad));
  GList* temp;
  for(temp=filter->padsets;(temp);temp=g_list_next(temp)){
    if (pad==((struct padset*)(temp->data))->sinkpad){
      otherpad=((struct padset*)(temp->data))->srcpad;
    }
  }  
  gst_object_unref (filter);
  if (otherpad)return gst_pad_set_caps (otherpad, caps);
  else return FALSE;
}

static GstPad *gst_fusion_request_new_pad (GstElement * element,
    GstPadTemplate * temp, const gchar * name){
    gchar* tmp;
  Gstfusion *filter;
  struct padset* temppadset;
  temppadset=malloc(sizeof(struct padset));
  temppadset->buffer=NULL;
  temppadset->timestamp=0;
  temppadset->lame=0;
  filter = GST_FUSION (element);
  g_mutex_lock(filter->mutex);
  if (filter->silent == FALSE) g_print("Fusion: New pad with name %s requested\n",name);

  //if (!name) tmp = g_strdup_printf ("src%d", g_list_length(filter->padsets));
  //else tmp=g_strdup_printf ("%s", name);
  tmp = g_strdup_printf ("sink%d", g_list_length(filter->padsets));
  temppadset->sinkpad = gst_pad_new_from_static_template (&sink_factory, tmp);
  gst_pad_set_setcaps_function (temppadset->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_fusion_set_caps));
  gst_pad_set_getcaps_function (temppadset->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (temppadset->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_fusion_chain));  
  gst_pad_set_event_function (temppadset->sinkpad,GST_DEBUG_FUNCPTR(gst_fusion_event)); 

  g_free (tmp);
  tmp = g_strdup_printf ("src%d", g_list_length(filter->padsets));
  temppadset->srcpad = gst_pad_new_from_static_template (&src_factory, tmp); 
  gst_pad_set_getcaps_function (temppadset->srcpad,
                                 GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));  
  
  gst_pad_set_event_function (temppadset->srcpad,GST_DEBUG_FUNCPTR(gst_fusion_event));
  filter->padsets=g_list_append(filter->padsets,temppadset);
  
  gst_element_add_pad (GST_ELEMENT (filter), temppadset->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), temppadset->srcpad);
  
  g_mutex_unlock(filter->mutex);
  return temppadset->sinkpad;
}


static void gst_fusion_release_pad (GstElement * element, GstPad * pad){
  
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_fusion_chain (GstPad * pad, GstBuffer * buf)
{
  Gstfusion *filter;
  GstFlowReturn ret;
  filter = GST_FUSION (GST_OBJECT_PARENT (pad));
  
  //check whether buffer is too new
  if ((filter->toonew)&&(buf->timestamp<filter->lastsend+filter->toonew)){
    if (filter->silent == FALSE)
      g_print ("GstFusion dropping, because buffer too new!! Timestamp: %llu Filter->Lastsend %llu.\n",(long long unsigned int)buf->timestamp,(long long unsigned int)filter->lastsend);
    gst_buffer_unref(buf);      
    return GST_FLOW_OK;
  }
 
  
  //if (filter->silent == FALSE)
  //  g_print ("GstFusion getting mutex, pad=%p.\n",(void*)pad);
  g_mutex_lock(filter->mutex);
  struct timeval now;
  gettimeofday(&now, NULL);
  guint64 timenow;
  timenow=((now.tv_sec*1000000000ULL)+(now.tv_usec*1000ULL));
  if ((filter->tooold>0)&&(buf->timestamp+filter->tooold<timenow)){//we had to wait too long for our mutex:(
    if (filter->silent == FALSE)
      g_print ("GstFusion dropping, because buffer too new!! Timestamp: %llu Filter->Lastsend %llu.\n",(long long unsigned int)buf->timestamp,(long long unsigned int)filter->lastsend);
    g_mutex_unlock(filter->mutex);   
    gst_buffer_unref(buf);      
    return GST_FLOW_OK;
  }  
  int x=0;
  GList* temp;
  struct padset* thispadset=NULL;
  for(temp=filter->padsets;(temp);temp=g_list_next(temp)){
    if (pad==((struct padset*)(temp->data))->sinkpad){
      if (filter->silent == FALSE)g_print ("GstFusion is on pad %d.",x);
      thispadset=(struct padset*)(temp->data);
    }
    x++;
  }
  if (!thispadset){
    if (filter->silent == FALSE)g_print ("Fusion Critical Error.\n");
    g_mutex_unlock(filter->mutex);    
    return GST_FLOW_ERROR;
  }
  thispadset->lame=0;
  //Zwischenspeichern des buffers, löschen alter
  if (thispadset->buffer){
    gst_buffer_unref(thispadset->buffer);  
  }
  thispadset->buffer=buf;
  //prüfen ob senden 
  gboolean send=TRUE;
  x=0;
  guint64 timemin=G_MAXUINT64,timemax=0;
  GstBuffer *tempbuf;
  if (filter->silent == FALSE) g_print ("GstFusion checking if buffers are available, Pad ");
  for(temp=filter->padsets;(temp);temp=g_list_next(temp)){
    if (filter->silent == FALSE) g_print ("%d: ",x++);
    if ((tempbuf=((struct padset*)(temp->data))->buffer)!=NULL){
      if (tempbuf->timestamp<timemin)timemin=tempbuf->timestamp;
      if (tempbuf->timestamp>timemax)timemax=tempbuf->timestamp;
      if (filter->silent == FALSE) g_print ("%llu: ",(long long unsigned int)tempbuf->timestamp);
    }
    else {
      send=FALSE;
      if ((filter->maxlame>0)&&(filter->maxlame<((struct padset*)(temp->data))->lame++)){
	GST_ELEMENT_ERROR (filter, RESOURCE,READ,
        ("Ran out of data from %s.",GST_ELEMENT_NAME(gst_pad_get_parent_element(gst_pad_get_peer(((struct padset*)(temp->data))->sinkpad)))), ("Element GstFusion"));
	}

      if (filter->silent == FALSE)g_print("NULL ");
    }
  }  
  if (filter->silent == FALSE) g_print ("\n");
  if ((timemax-timemin)>filter->timediff){//kick all late buffers
        if (filter->silent == FALSE) g_print ("GstFusion: Timediff too big: %llu, dropping.\n",(long long unsigned int)(timemax-timemin));
    send=FALSE;
    for(temp=filter->padsets;(temp && send);temp=g_list_next(temp)){
      if ((tempbuf=((struct padset*)(temp->data))->buffer)!=NULL){
        if (tempbuf->timestamp<(timemax-filter->timediff)){
	  gst_buffer_unref(tempbuf);
	  ((struct padset*)(temp->data))->buffer=NULL;
	}
      }
    }  
  }   
    
  if (!send) {
    g_mutex_unlock(filter->mutex); 
    return GST_FLOW_OK;
  }
  //senden
  filter->lastsend=timemax;
  if (filter->silent == FALSE) g_print ("GstFusion is about to send. Timestamp %llu.\n",(long long unsigned int)timemax);
  for(temp=filter->padsets;(temp);temp=g_list_next(temp)){
    thispadset=(struct padset*)(temp->data);
    ret=gst_pad_push (thispadset->srcpad, thispadset->buffer);
    thispadset->buffer=NULL;
    if (ret!=GST_FLOW_OK){
      g_mutex_unlock(filter->mutex); 
      return ret;
    }
  } 
  
  /* just push out the incoming buffer without touching it */
  g_mutex_unlock(filter->mutex);  
  return GST_FLOW_OK;
  //return gst_pad_push (filter->srcpad, buf);
}

static gboolean
gst_fusion_event (GstPad  *pad, GstEvent * event){
  Gstfusion* filter = GST_FUSION (gst_pad_get_parent (pad));
  GstPad* otherpad=NULL;
  GList* temp;
  struct padset* thispadset;
  for(temp=filter->padsets;(temp);temp=g_list_next(temp)){
    thispadset=(struct padset*)(temp->data);
    if (thispadset->srcpad==pad)otherpad=thispadset->sinkpad;
    if (thispadset->sinkpad==pad)otherpad=thispadset->srcpad;
  } 
  if (otherpad)return gst_pad_push_event (otherpad, event);
  else return FALSE;
}
/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
fusion_init (GstPlugin * fusion)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template fusion' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_fusion_debug, "fusion",
      0, "Template fusion");

  return gst_element_register (fusion, "fusion", GST_RANK_NONE,
      GST_TYPE_FUSION);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstfusion"
#endif

/* gstreamer looks for this structure to register fusions
 *
 * exchange the string 'Template fusion' with your fusion description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "fusion",
    "Template fusion",
    fusion_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
