/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-adder
 *
 * FIXME:Describe adder here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! adder ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>

#include "gstadder.h"

GST_DEBUG_CATEGORY_STATIC (gst_adder_debug);
#define GST_CAT_DEFAULT gst_adder_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_OPMODE,
  PROP_SCALEUP
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(GST_VIDEO_CAPS_GRAY8)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("value/uint64")
    );

static GstStaticPadTemplate sink_factory_all = GST_STATIC_PAD_TEMPLATE ("sink-coord",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "coordinates-list"
    )
    );

GST_BOILERPLATE (Gstadder, gst_adder, GstElement,
    GST_TYPE_ELEMENT);

static void gst_adder_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_adder_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_adder_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps *
gst_adder_get_caps (GstPad *pad);
static GstFlowReturn gst_adder_chain (GstPad * pad, GstBuffer * buf);
static gboolean little_endian;
#define GST_TYPE_ADDER_OP_MODE (gst_type_adder_op_mode ())
static GType
gst_type_adder_op_mode (void)
{
  static GType op_mode = 0;
  static const GEnumValue functions[] = {
    {ADD, "Add all pixel values", "add"},
    {ADD_NOR, "Add all pixel values normalized", "add_nor"},
    {TAKE_MAX, "Search for the Maximum", "take_max"},
    {TAKE_MAX10, "Search for the 10 Maximum values", "take_max10"},
    {0, NULL, NULL}
  };

  if (!op_mode) {
    op_mode =
        g_enum_register_static ("GstAdderOpMode", functions);
  }
  return op_mode;
}

/* GObject vmethod implementations */

static void
gst_adder_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "adder",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the adder's class */
static void
gst_adder_class_init (GstadderClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_adder_set_property;
  gobject_class->get_property = gst_adder_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OPMODE,
      g_param_spec_enum ("opmode", "Opmode",
          "How will we treat the input?", GST_TYPE_ADDER_OP_MODE,
          ADD, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SCALEUP,
      g_param_spec_double ("scaleup", "Scaleup",
          "Multiply coordinate info?", 0, G_MAXDOUBLE,1.0, (GParamFlags)G_PARAM_READWRITE));


}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_adder_init (Gstadder * filter,
    GstadderClass * gclass)
{
  int x=1;
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_adder_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_adder_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_adder_chain));

  filter->sinkcoord = gst_pad_new_from_static_template (&sink_factory_all, "sink-coord");
  gst_pad_set_setcaps_function (filter->sinkcoord,
                                GST_DEBUG_FUNCPTR(gst_adder_set_caps));
  gst_pad_set_getcaps_function (filter->sinkcoord,
                                GST_DEBUG_FUNCPTR(gst_adder_get_caps));
  gst_pad_set_chain_function (filter->sinkcoord,
                              GST_DEBUG_FUNCPTR(gst_adder_chain));


  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_adder_get_caps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkcoord);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  little_endian=(gboolean) (*(char*)&x); 
  filter->opmode=ADD;
  filter->mutex=g_mutex_new();
  filter->coordbuf=NULL;
  filter->width=-1;
  filter->height=-1;
  filter->scaleup=1.0;
}

static void
gst_adder_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstadder *filter = GST_ADDER (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_OPMODE:
      filter->opmode=(AdderOpMode)g_value_get_enum(value);
      break;
    case PROP_SCALEUP:
      filter->scaleup=g_value_get_double(value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_adder_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstadder *filter = GST_ADDER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_OPMODE:
      g_value_set_enum(value, filter->opmode);
      break;
    case PROP_SCALEUP:
      g_value_set_double(value, filter->scaleup);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_adder_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstadder *filter;
  //GstPad *otherpad;

  filter = GST_ADDER (gst_pad_get_parent (pad));
  //otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gboolean ret=TRUE;
  if (pad == filter->sinkpad){
    const gchar *mime;
    GstStructure *structure=gst_caps_get_structure(caps,0);
    mime = gst_structure_get_name (structure);  
    if (!gst_structure_get_int(structure,"width",&(filter->width))) ret=FALSE;
    if (!gst_structure_get_int(structure,"height",&(filter->height))) ret=FALSE;
    g_print("Adder caps to parse: %s Extracted width: %d height: %d!!\n",gst_caps_to_string (caps),filter->width,filter->height);
  }

  gst_object_unref (filter);

  return ret;
}

static GstCaps *
gst_adder_get_caps (GstPad *pad){
  g_print("Adder: Start get caps!!\n");
  //Gstadder *filter;

  //filter = GST_ADDER (GST_OBJECT_PARENT (pad));

  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  
}

guint64 gst_adder_faktor (Gstadder *filter,int pos){
  if (!filter->coordbuf)return 1;
  if (filter->width <0)return 1;
  if (filter->height <0)return 1;
  int numelements=filter->coordbuf->size/sizeof(struct coordinates);
  struct coordinates* temp;
  gboolean found=FALSE;
  int i=0;//index
  int w=pos % filter->width;
  int h=pos / filter->width;
  while ((!found)&&(i<numelements)){
    temp=&(((struct coordinates*)filter->coordbuf->data)[i]);
    if ((w>(filter->scaleup*temp->x1))&&(w<(filter->scaleup*temp->x2))&&(h>(filter->scaleup*temp->y1))&&(h<(filter->scaleup*temp->y2))){
      found=TRUE;
    }
    i++;
  }
  if (found)return 100;
  else return 1;
}
/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_adder_chain (GstPad * pad, GstBuffer * buf)
{
  Gstadder *filter;
  GstBuffer *out;
  guint64 result=0;
  guchar max=0;
  filter = GST_ADDER (GST_OBJECT_PARENT (pad));
  g_mutex_lock(filter->mutex); 
  if (pad==filter->sinkcoord){
    if (filter->coordbuf){
	gst_buffer_unref(filter->coordbuf);
        filter->coordbuf=NULL;
	}
    filter->coordbuf=buf;
    g_mutex_unlock(filter->mutex); 
    return GST_FLOW_OK;
  }
  out=gst_buffer_new_and_alloc (sizeof(guint64));


  int i;
  for (i=0;i<buf->size;i++){

  if(buf->data[i] > 10){		//threshold    
	result+=gst_adder_faktor(filter,i)*buf->data[i];
	if ((buf->data[i])>max)max=buf->data[i];}
  }
  if (filter->opmode==ADD_NOR)result=result/(guint64)buf->size;
  else if (filter->opmode==TAKE_MAX)result=max;
  else if (filter->opmode==TAKE_MAX10)result=max;

  if (!little_endian)result=Swap64(result);
  *((guint64*)out->data)=result;
  if (filter->silent == FALSE)
    g_print ("I'm ready, calculated %lu.\n",result);
  

  out->timestamp=buf->timestamp;
  out->offset_end=buf->offset_end;
  out->duration=buf->duration;
  gst_buffer_unref(buf);
  /* just push out the incoming buffer without touching it */
  g_mutex_unlock(filter->mutex); 
  return gst_pad_push (filter->srcpad, out);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
adder_init (GstPlugin * adder)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template adder' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_adder_debug, "adder",
      0, "Template adder");

  return gst_element_register (adder, "adder", GST_RANK_NONE,
      GST_TYPE_ADDER);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstadder"
#endif

/* gstreamer looks for this structure to register adders
 *
 * exchange the string 'Template adder' with your adder description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "adder",
    "Template adder",
    adder_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
