/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2010 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_MULTIVIS_H__
#define __GST_MULTIVIS_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_MULTIVIS \
  (gst_multivis_get_type())
#define GST_MULTIVIS(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_MULTIVIS,Gstmultivis))
#define GST_MULTIVIS_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_MULTIVIS,GstmultivisClass))
#define GST_IS_MULTIVIS(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_MULTIVIS))
#define GST_IS_MULTIVIS_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_MULTIVIS))
#define FEATURE_MAX_D 128

/*
GOOD
you look great today


BAD:
get outa my way
dont stand around here, you'd better work
oh no, not you again
dont touch me, I am afraid you rip off my cables again
better run for your life
you look dämn wasted today, did you drink the whole night
*/

typedef struct _Gstmultivis      Gstmultivis;
typedef struct _GstmultivisClass GstmultivisClass;

struct feature{
  double x;                      /**< x coord */
  double y;                      /**< y coord */
  double a;                      /**< Oxford-type affine region parameter */
  double b;                      /**< Oxford-type affine region parameter */
  double c;                      /**< Oxford-type affine region parameter */
  double scl;                    /**< scale of a Lowe-style feature */
  double ori;                    /**< orientation of a Lowe-style feature */
  int d;                         /**< descriptor length */
  double descr[FEATURE_MAX_D];   /**< descriptor */
  int type;                      /**< feature type, OXFD or LOWE */
  int category;                  /**< all-purpose feature category */
  struct feature* fwd_match;     /**< matching feature from forward image */
  struct feature* bck_match;     /**< matching feature from backmward image */
  struct feature* mdl_match;     /**< matching feature from model */
  CvPoint2D64f img_pt;           /**< location in image */
  CvPoint2D64f mdl_pt;           /**< location in model */
  void* feature_data;            /**< user-definable data */
};

struct camconfig
{
  int active;//whether the camera is active or not, use this to recalibrate PTU if any
  int relative;//currently only absolute values are supported. this way, currently always 0 
  float hfov;//how to set the zoom
  float vfov;//how to set the zoom, <0 does not matter... 
  float yaw;//turn left or right, 0 to 360 degrees
  float pitch;//up and down, depending on input image
  float roll;//for future use
  gdouble hastranslation;
  float x;//for future use
  float y;//for future use
  float z;//for future use
  float shutter;//shutter speed in ms, <0 is AUTO
  float iris;//iris <0 is auto
  float focus;//manually set focus value -1
  int limit;//whether current conf hits the limit
};

struct coordinates
{
  gdouble x1;
  gdouble y1;
  gdouble x2;
  gdouble y2;
};

struct uber_coordinates
{
  int imagewidth;//size of the image where we found this 
  int imageheight;//
  int mwidth;//size of the model that we used to find this this   
  int mheight;//
  gdouble scale;//Rectangular Region where this is matched into
  gdouble angle;  
  gdouble x1;//Rectangular Region where this is matched into
  gdouble y1;
  gdouble x2;
  gdouble y2;
  gdouble a00;//transformation matrix 
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
  guint64 timestamp;
  char name[30];//name of found element
};

struct grasp
{
int x; //in mm
int y;
int z;
int pan;//reserved, zunächst nur horizontaler approach
int tilt;//reserved
int yaw;//reserved
int action;//reserved, later for release
};

int swapint(int ui)
{
    int result;
    result = (ui >> 24) |
         ((ui<<8) & 0x00FF0000) |
         ((ui>>8) & 0x0000FF00) |
         (ui << 24);
    return result;
}
gdouble DoubleSwap( double f )
{
  union
  {
    double f;
    unsigned char b[8];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[7];
  dat2.b[1] = dat1.b[6];
  dat2.b[2] = dat1.b[5];
  dat2.b[3] = dat1.b[4];
  dat2.b[4] = dat1.b[3];
  dat2.b[5] = dat1.b[2];
  dat2.b[6] = dat1.b[1];
  dat2.b[7] = dat1.b[0];
  return dat2.f;
}

guint64 uint64Swap( guint64 f )
{
  union
  {
    guint64 f;
    unsigned char b[8];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[7];
  dat2.b[1] = dat1.b[6];
  dat2.b[2] = dat1.b[5];
  dat2.b[3] = dat1.b[4];
  dat2.b[4] = dat1.b[3];
  dat2.b[5] = dat1.b[2];
  dat2.b[6] = dat1.b[1];
  dat2.b[7] = dat1.b[0];
  return dat2.f;
}

float FloatSwap( float f )
{
  union
  {
    float f;
    unsigned char b[4];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[3];
  dat2.b[1] = dat1.b[2];
  dat2.b[2] = dat1.b[1];
  dat2.b[3] = dat1.b[0];
  return dat2.f;
}
void swapgrasp(struct grasp* grasp){
int x=1;
gboolean little_endian;
little_endian=(gboolean) (*(char*)&x); 
if (little_endian) return;
grasp->x=swapint(grasp->x);
grasp->y=swapint(grasp->y);
grasp->z=swapint(grasp->z);
grasp->pan=swapint(grasp->pan);
grasp->tilt=swapint(grasp->tilt);
grasp->yaw=swapint(grasp->yaw);
grasp->action=swapint(grasp->action);
return;
}

void swapcamconfig(struct camconfig* camconfig){
int x=1;
gboolean little_endian;
little_endian=(gboolean) (*(char*)&x); 
if (little_endian) return;
camconfig->active=swapint(camconfig->active);
camconfig->relative=swapint(camconfig->relative);
camconfig->hfov=FloatSwap(camconfig->hfov);
camconfig->vfov=FloatSwap(camconfig->vfov);
camconfig->yaw=FloatSwap(camconfig->yaw);
camconfig->pitch=FloatSwap(camconfig->pitch);
camconfig->roll=FloatSwap(camconfig->roll);
camconfig->hastranslation=DoubleSwap(camconfig->hastranslation);
camconfig->x=FloatSwap(camconfig->x);
camconfig->y=FloatSwap(camconfig->y);
camconfig->z=FloatSwap(camconfig->z);
camconfig->shutter=FloatSwap(camconfig->shutter);
camconfig->iris=FloatSwap(camconfig->iris);
camconfig->focus=FloatSwap(camconfig->focus);
camconfig->limit=swapint(camconfig->limit);
return;
}
void swap_coordinates(struct coordinates* c){
int x=1;
gboolean little_endian;
little_endian=(gboolean) (*(char*)&x); 
if (little_endian) return;
c->x1=DoubleSwap(c->x1);
c->y1=DoubleSwap(c->y1);
c->x2=DoubleSwap(c->x2);
c->y2=DoubleSwap(c->y2);
return;
}
void swapuber_coordinates(struct uber_coordinates* c){
int x=1;
gboolean little_endian;
little_endian=(gboolean) (*(char*)&x); 
if (little_endian) return;
c->imagewidth=swapint(c->imagewidth);
c->imageheight=swapint(c->imageheight);
c->mwidth=swapint(c->mwidth);
c->mheight=swapint(c->mheight);
c->scale=DoubleSwap(c->scale);
c->angle=DoubleSwap(c->angle);
c->x1=DoubleSwap(c->x1);
c->y1=DoubleSwap(c->y1);
c->x2=DoubleSwap(c->x2);
c->y2=DoubleSwap(c->y2);
c->a00=DoubleSwap(c->a00);
c->a01=DoubleSwap(c->a01);
c->a02=DoubleSwap(c->a02);
c->a10=DoubleSwap(c->a10);
c->a11=DoubleSwap(c->a11);
c->a12=DoubleSwap(c->a12);
c->a20=DoubleSwap(c->a20);
c->a21=DoubleSwap(c->a21);
c->a22=DoubleSwap(c->a22);
c->timestamp=uint64Swap(c->timestamp);
return;
}
//average of n numbers, looses precision or risks overflow!!! use only for measurements in ns!!
struct avuint64{
int n;
int shift;//shift to 
guint64* numbers;
guint64 sum;
int nextdata;
int ready;
};

void initavuint64(struct avuint64* data,int n, int shift){
data->n=n;
data->shift=shift;
data->numbers=g_malloc0(n*sizeof(guint64));
int i;
for (i=0;i<n;i++) data->numbers[i]=0;
data->nextdata=0;
data->sum=0;
data->ready=0;
}

void avuint64app(struct avuint64* data, guint64 value){
//printf("append\n");
data->sum-=data->numbers[data->nextdata];
//printf("quarter way\n");
data->numbers[data->nextdata]=value>>data->shift;
//printf("half way\n");
data->sum+=data->numbers[data->nextdata];
data->nextdata=(data->nextdata+1);
if (data->nextdata==data->n)data->ready=1;
data->nextdata=data->nextdata%data->n;
}

inline guint64 getavuint64(struct avuint64* data){//dont call untill you gave n values to it
if (data->ready) return ((data->sum/data->n)<<data->shift);
else if (data->nextdata>0) return ((data->sum/data->nextdata)<<data->shift);
else return G_MAXUINT64;
}

struct padset{
  GstPad *sinkpad;
  GstBuffer *buffer;
  int mode;//0=uber-coordinates, 1=uber-coordinates-list, 2=coordinates-list
  int width,height;
  gchar* name; //get the name from the src-element-name 
  struct avuint64 *latency,*avg,*between;//things for evaluation
  guint64 timestamp;
};

struct _Gstmultivis
{
  GstElement element;
  GMutex* mutex;
  GstPad *sinkpad, *sinksift,*sinkpip, *srcpad, *srccamconfig,*sinkcamconfig,*srcselection,*srcsnapshot,*srcmap,*srcgrasp;
  int siftwidth,siftheight;
  GList *objectbuf;
  struct avuint64 *mainlatency,*mainbetween,*siftlatency,*siftavg,*siftbetween,*objlatency,*objavg,*objbetween;
  GstBuffer *siftbuf,*pipbuf;
  GList *padsets;
  gboolean silent;
  gint bpp,fps_n,fps_d;
  gint thickness;
  guint64 timeout,lastbuffer,starttime,firstbufts;//teme that image information may be old
  CvSize imagesize;
  CvFont font;
  gdouble fontsize;
  char text[1024];
  //horizontal fov is always 360 degrees, 0 degrees is middle point of image
  //vertical 
  //stuff for PIP
  gboolean pipoverlay;//wether we generally want to draw overlay
  gboolean pipactive;
  CvSize pipimagesize;//negotiated size of the pip-stream
  //float pipfocalmin,pipfocalmax,mainfocal;//configured in main, later in 
  int seloriginx,seloriginy,seloriginx_tmp,seloriginy_tmp,seloriginx_raw,seloriginy_raw,relposx,relposy,selwidth,selheight,minbox;//currently selected region, minimal boxsize=50
  int ovoriginx,ovoriginy,ovwidth,ovheight;
  gint64 x_pending, y_pending,yawoffset;//currently pending coordinates
  float mainhfov,currentlyorderedhfov,orderedyaw,orderedpitch,orderedfocus;
  gboolean resendcamconfig;
  struct camconfig* currentconf;
  GstBuffer *ccbuf;//camconfig Buffer
  //Stuff for selection
  gboolean selectionnext,snapshotnext;
  //debug
  int numcsent,numcrecv;
  gboolean resendmap,drag,alwaysdisplay,manualfocus;//wether to generate and resend transformation LUT in next step, whether we treat mouse movements as drag for yawoffset, whether to always display PIP at user defined positopn 

  int acwidth,acheight;
  int asx,asy;
  float asscale;

  gboolean graspmode;
  int graspx,graspy;

  gboolean docc;//hackaround to make ptu turn 
  int fixz;
  gboolean alwaysgrasp;
};

struct _GstmultivisClass 
{
  GstElementClass parent_class;
};

char* deserialize_int64 (gint64* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  while (!isdigit(*s)&&(*s!=0)&&(*s!=45))s++;
  if (*s==0){
    //*result=NAN;
    return NULL;
  }
  char *end;
  *result = g_ascii_strtoll (s, &end, 0);
  return end;
}

GType gst_multivis_get_type (void);

G_END_DECLS

#endif /* __GST_MULTIVIS_H__ */
