/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-pipgen
 *
 * FIXME:Describe pipgen here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! pipgen ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <string.h>
#include <unistd.h>

#include "gstpipgen.h"

GST_DEBUG_CATEGORY_STATIC (gst_pipgen_debug);
#define GST_CAT_DEFAULT gst_pipgen_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_SYNC
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
/*static GstCaps* baslerstreamheader=GST_STATIC_CAPS ("video/x-raw-yuv, "
     "framerate = (fraction) [ 18, 1 ], "
     "format = (fourcc) { UYVY }, " // Bayer8 u. a. later 
     "width = (int) 1388, "
     "height = (int) 1038,"
     "streamheader = (guint64) 0");

static GstCaps* baslerstreamheader=gst_caps_from_string("video/x-raw-yuv, "
     "framerate = (fraction) [ 18, 1 ], "
     "format = (fourcc) { UYVY }, " // Bayer8 u. a. later 
     "width = (int) 1388, "
     "height = (int) 1038,"
     "streamheader = (guint64) 0"); */

static GstStaticPadTemplate sink_factory_coord = GST_STATIC_PAD_TEMPLATE ("sink-coord",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "coordinates-list"
    )
    );

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory_coord = GST_STATIC_PAD_TEMPLATE ("src-coord",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("coordinates-list")
    );

GST_BOILERPLATE (GstPIPgen, gst_pipgen, GstElement,
    GST_TYPE_ELEMENT);

static void gst_pipgen_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_pipgen_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static void gst_pipgen_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps * gst_pipgen_get_caps (GstPad *pad);
static gboolean gst_pipgen_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_pipgen_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn gst_pipgen_change_state (GstElement *element, GstStateChange transition);
static GstFlowReturn
gst_pipgen_chain_sync (GstPad * pad, GstBuffer * buf);
static gboolean little_endian;
/* GObject vmethod implementations */

static void
gst_pipgen_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "PIPgen",
    "Gst-Plugins-Robotix",
    "Generate subpictures of a given frame",
    "Hannes Bistry <<hannesb@gmx.de>>");


  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the pipgen's class */
static void
gst_pipgen_class_init (GstPIPgenClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_pipgen_set_property;
  gobject_class->get_property = gst_pipgen_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SYNC,
      g_param_spec_boolean ("sync", "Sync", "Sync small frames to coordinates ?",
          FALSE, G_PARAM_READWRITE));
  gstelement_class->change_state = gst_pipgen_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_pipgen_init (GstPIPgen * filter,
    GstPIPgenClass * gclass)
{
  int x=1;//for endian test, as testing is not supportet while crosscompiling
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pipgen_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pipgen_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_pipgen_chain));


  filter->sinkcoord = gst_pad_new_from_static_template (&sink_factory_coord, "sink-coord");
  gst_pad_set_setcaps_function (filter->sinkcoord,
                                GST_DEBUG_FUNCPTR(gst_pipgen_set_caps));
  gst_pad_set_getcaps_function (filter->sinkcoord,
                                GST_DEBUG_FUNCPTR(gst_pipgen_get_caps));
  gst_pad_set_chain_function (filter->sinkcoord,
                              GST_DEBUG_FUNCPTR(gst_pipgen_chain));



  filter->srccoord = gst_pad_new_from_static_template (&src_factory_coord, "src-coord");
  gst_pad_set_getcaps_function (filter->srccoord,
                                GST_DEBUG_FUNCPTR(gst_pipgen_get_caps));




  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pipgen_get_caps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkcoord);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srccoord);

  filter->mutex=g_mutex_new();
  filter->silent = FALSE;
  filter->sync = FALSE;
  filter->preroll=TRUE;
  filter->coordbu=NULL;
  filter->coord=NULL;
  filter->hardcut=10;
  filter->lbig=NULL;
  little_endian=(gboolean) (*(char*)&x); 
}

static void
gst_pipgen_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstPIPgen *filter = GST_PIPGEN (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_SYNC:
      filter->sync = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_pipgen_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstPIPgen *filter = GST_PIPGEN (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_SYNC:
      g_value_set_boolean (value, filter->sync);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static GstCaps *
gst_pipgen_get_caps (GstPad *pad){
  g_print("PIPgen: Start get caps!!\n");
  GstPIPgen *filter;

  filter = GST_PIPGEN (GST_OBJECT_PARENT (pad));

  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}






/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_pipgen_set_caps (GstPad * pad, GstCaps * caps)
{
  GstPIPgen *filter;
  GstPad *otherpad;

  filter = GST_PIPGEN (gst_pad_get_parent (pad));
  if (pad==filter->sinkpad){
    filter->tomodify=gst_caps_copy(caps);
    const gchar *mime;
    GstStructure *structure=gst_caps_get_structure(caps,0);
    mime = gst_structure_get_name (structure);//get datatype of caps
    if (strstr(mime,"video")){//get borders
      gst_structure_get_int(structure,"width",&(filter->orgwidth));
      gst_structure_get_int(structure,"height",&(filter->orgheight));
      gst_structure_get_int(structure,"bpp",&(filter->bpp)); 
    }
    if (0==strcmp(mime,"video/x-raw-yuv")){
	guint32 fourccval;
	if (!gst_structure_has_field(structure,"format")){
		g_print ("No format field");
		return FALSE;
	}  
	if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
		g_print ("FourCC could not be read");
		return FALSE;
	}
	if (fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
	  filter->bpp=8;
	}
	//else if (fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
	//  filter->bpp=8;
	//}
	}
  }

  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  return gst_pad_set_caps (otherpad, caps);
}




static gboolean modifiy_caps(GstCaps * caps, int width, int height){//modify format of image
  //g_print ("Modify aufgerufen.\n");
  GstStructure *structure;
  gboolean result=TRUE;
  structure = gst_caps_get_structure (caps, 0);
  gst_structure_set (structure,
        "width", G_TYPE_INT, width,
        "height", G_TYPE_INT, height,
        //"streamheader", G_TYPE_UINT64, ((((guint64)height)<<32)+(guint64)width), 
	NULL);
  g_print ("Modify finished.\n");
  //result=gst_structure_fixate_field_nearest_int (structure, "width", width);
  //result=result&&gst_structure_fixate_field_nearest_int (structure, "height", height);
  return result;
}
/* chain function
 * this function does the actual processing
 */









static GstFlowReturn
gst_pipgen_chain (GstPad * pad, GstBuffer * buf)
{

  GstPIPgen *filter;
  GstFlowReturn ret=GST_FLOW_OK;
  filter = GST_PIPGEN (GST_OBJECT_PARENT (pad));
  if (filter->sync) return gst_pipgen_chain_sync (pad,buf);
  GstBuffer * outbuf;	


  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in. Getting mutex\n");

  if (pad== filter->sinkcoord){
      if (!g_mutex_trylock(filter->mutex)){
        g_print ("Did not get mutex in sinkcoord mode coord pointer %p , coordbu pointer %p\n",filter->coord,filter->coordbu);
	if ((filter->coordbu)&&(!filter->coord)){//if there is some old buffer at the end of one stream(preventing preroll)	  
	  //filter->coord=buf;
	  gst_buffer_unref(filter->coordbu);
	  filter->coordbu=buf;
	  return GST_FLOW_OK;
	  }
	else if ((filter->coordbu)&&(filter->coord)){//will hapen at the end of the stream
	  gst_buffer_unref(buf);
	  return GST_FLOW_OK;
	}
	filter->coordbu=buf;
	return GST_FLOW_OK;	
	}
      if (filter->coord)gst_buffer_unref(filter->coord);
      filter->coord=buf;
  }
  else if ((pad== filter->sinkpad)&&(filter->coord)){	
    //foreward current coordinates
    if (filter->silent == FALSE)g_print("PIPgen Getting mutex for normal operation\n");
    g_mutex_lock(filter->mutex);
    if (filter->silent == FALSE)g_print("PIPgen ...and got it\n");    
    outbuf=gst_buffer_copy(filter->coord);
    outbuf->timestamp=buf->timestamp;
    outbuf->duration=buf->duration;
    outbuf->offset=buf->offset;
    outbuf->offset_end=buf->offset_end;    
    ret=gst_pad_push (filter->srccoord, outbuf);
    if (ret!=GST_FLOW_OK) {
	if (filter->silent == FALSE)g_print("PIPgen: Warning!!! Buffer on srccoord went wrong\n");  
        g_mutex_unlock(filter->mutex);	
	return ret;
    }
    //generate list from recent coordinates
    int numregions=filter->coord->size/sizeof(struct coordinates);
    if (filter->silent == FALSE)g_print("checking %d regions\n",numregions);
    int i,j,totalsize,lines,rows,linesize,x1,x2,y1,y2;

    for (i=0;i<numregions;i++){
	if (little_endian){
	x1=(int)(unsigned int)((struct coordinates*)(filter->coord->data))[i].x1;
	x2=(int)(unsigned int)((struct coordinates*)(filter->coord->data))[i].x2;
	y1=(int)(unsigned int)((struct coordinates*)(filter->coord->data))[i].y1;
	y2=(int)(unsigned int)((struct coordinates*)(filter->coord->data))[i].y2;
	}
	else{	
        x1=(int)(unsigned int)DoubleSwap(((struct coordinates*)(filter->coord->data))[i].x1);
        x2=(int)(unsigned int)DoubleSwap(((struct coordinates*)(filter->coord->data))[i].x2);
        y1=(int)(unsigned int)DoubleSwap(((struct coordinates*)(filter->coord->data))[i].y1);
        y2=(int)(unsigned int)DoubleSwap(((struct coordinates*)(filter->coord->data))[i].y2);
	}

	if ((x1>=filter->orgwidth)||(x2>=filter->orgwidth)||(y1>=filter->orgheight)||(y2>=filter->orgheight)){
	  if (filter->silent == FALSE)g_print("GstPip:Warning, coordinates outside image bounds x1: %d x2 %d y1 %d y2 %d Widht: %d Height %d!!\n",x1,x2,y1,y2,filter->orgwidth,filter->orgheight);
	  return GST_FLOW_ERROR;	
	}
        //g_print("PIPGEN Float values %f,%f, %f, %f\n",((struct coordinates*)(filter->coord->data))[i].x1,((struct coordinates*)(filter->coord->data))[i].x2,((struct coordinates*)(filter->coord->data))[i].y1,((struct coordinates*)(filter->coord->data))[i].y2);
//	if (x1>=filter->orgwidth)x1=filter->orgwidth-1;
//	if (x2>=filter->orgwidth)x2=filter->orgwidth-1;
//	if (y1>=filter->orgheight)y1=filter->orgheight-1;
//	if (y2>=filter->orgheight)y2=filter->orgheight-1;
	lines=1+y2-y1;
	rows=1+x2-x1;
	linesize=(1+x2-x1)*(filter->bpp/8);
	totalsize=linesize*lines;
	outbuf=gst_buffer_new_and_alloc (totalsize);
 	GstCaps *tempcaps=gst_caps_copy (filter->tomodify);
        //GstCaps *tempcaps=gst_caps_copy (baslerstreamheader);
//         GstCaps *tempcaps=gst_caps_from_string("video/x-raw-yuv, "
//         "framerate = (fraction) [ 18, 1 ], "
//         "format = (fourcc) { UYVY }, " // Bayer8 u. a. later 
//         "width = (int) 1388, "
//         "height = (int) 1038,"
//         "streamheader = (guint64) 0"); 
        //GstCaps *tempcaps=gst_caps_from_string("video/x-raw-yuv, format=(fourcc)Y800, width=(int)1388, height=(int)1038,streamheader=(guint64)0, framerate=(fraction)18/1");
	if (FALSE==modifiy_caps(tempcaps,rows,lines)){
	 if (filter->silent == FALSE)g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps);//ref has been "transferred" to buffer
	for (j=0;j<lines;j++){
	  memcpy(outbuf->data+(j*linesize),buf->data+((filter->bpp/8)*((y1+j)*filter->orgwidth+x1)),linesize);
	}
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	if (filter->silent == FALSE)g_print("Pushing buffer (totalsize %d) size %d x %d that comes from %d and %d with %d bits each pixel!!\n",totalsize,rows,lines,x1,y1,filter->bpp);
	ret=gst_pad_push (filter->srcpad, outbuf);
	if (ret!=GST_FLOW_OK){
	 if (filter->silent == FALSE)g_print("GstPip:Sourcepad push went wrong!!");
	 return ret;
	}
    }
    if ((numregions==0)&&(filter->preroll)){
	outbuf=gst_buffer_new_and_alloc (16*filter->bpp/8);
	if (filter->silent == FALSE)g_print("GstPip:Alloceted %d for preroll!!\n",(16*filter->bpp/8));
 	GstCaps *tempcaps=gst_caps_copy (filter->tomodify);
	if (FALSE==modifiy_caps(tempcaps,4,4)){
	 if (filter->silent == FALSE)g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps);
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	ret=gst_pad_push (filter->srcpad, outbuf);
	}
    gst_buffer_unref(buf);
  }
  else if ((pad== filter->sinkpad)&&(!filter->coord)){
	//check if PAUSED, send empty buffer instead
	if (filter->silent == FALSE)g_print("PIPgen Getting mutex for Preroll\n");
        g_mutex_lock(filter->mutex);
	//outbuf=gst_buffer_new_and_alloc (16*filter->bpp/8);
        outbuf=gst_buffer_new_and_alloc (0);
	GstCaps *tempcaps=gst_caps_copy (gst_pad_get_pad_template_caps(filter->srccoord));
        gst_buffer_set_caps(outbuf,tempcaps);
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	gst_caps_unref(tempcaps);
        ret=gst_pad_push (filter->srccoord, outbuf);
// 	g_print("GstPip:Alloceted %d for preroll!!",(16*filter->bpp/8));
//  	GstCaps *tempcaps=gst_caps_copy (filter->tomodify);
// 	if (FALSE==modifiy_caps(tempcaps,4,4)){
// 	 g_print("GstPip:Warning, caps modification failed!!");
// 	 g_mutex_unlock(filter->mutex);
// 	 return GST_FLOW_ERROR;
// 	}
//         gst_buffer_set_caps(outbuf,tempcaps);
// 	gst_caps_unref(tempcaps);
// 	outbuf->timestamp=buf->timestamp;
//         outbuf->duration=buf->duration;
//         outbuf->offset=buf->offset;
//         outbuf->offset_end=buf->offset_end;
// 	ret=gst_pad_push (filter->srcpad, outbuf);
	ret=gst_pad_push (filter->srcpad, buf);
        //gst_buffer_unref(buf);
  }
//see whether another thread left us a buffer
  if (filter->coordbu){
    gst_buffer_unref(filter->coord);
    filter->coord=filter->coordbu;
    filter->coordbu=NULL;
  }
  g_mutex_unlock(filter->mutex);
  return ret;;
}

static GstStateChangeReturn
gst_pipgen_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  GstPIPgen *filter;
  filter = GST_PIPGEN (element);
  //g_mutex_lock(filter->mutex);  
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      if (filter->coord)gst_buffer_unref(filter->coord);
      if (filter->coordbu)gst_buffer_unref(filter->coordbu);
      if (filter->bigbu)gst_buffer_unref(filter->bigbu);
      if (filter->big)gst_buffer_unref(filter->big);
      filter->big=filter->bigbu=filter->coord=filter->coordbu=NULL;
      filter->prerollcnt=3;
      filter->sentyet=FALSE;

      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      if (filter->coord)gst_buffer_unref(filter->coord);
      if (filter->coordbu)gst_buffer_unref(filter->coordbu);
      filter->coord=filter->coordbu=NULL;
      g_list_foreach(filter->lbig,(GFunc)gst_mini_object_unref,NULL);//bug gst_buffer_unref will not compile
      g_list_free(filter->lbig);
      filter->lbig=NULL;
      g_mutex_unlock(filter->mutex);
      //gst_my_filter_free_memory (filter);
      break;
    default:
      break;
  }
 //g_mutex_unlock(filter->mutex);
 return ret;
}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
pipgen_init (GstPlugin * pipgen)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template pipgen' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_pipgen_debug, "pipgen",
      0, "Template pipgen");

  return gst_element_register (pipgen, "pipgen", GST_RANK_NONE,
      GST_TYPE_PIPGEN);
}

/* gstreamer looks for this structure to register pipgens
 *
 * exchange the string 'Template pipgen' with your pipgen description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "pipgen",
    "Template pipgen",
    pipgen_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)

static GstFlowReturn
gst_pipgen_chain_sync (GstPad * pad, GstBuffer * buf)
{
  GstPIPgen *filter;
  GstFlowReturn ret=GST_FLOW_OK;
  filter = GST_PIPGEN (GST_OBJECT_PARENT (pad));
  GstBuffer * outbuf=NULL;
  GstBuffer * outbufcoord=NULL;	


  if (filter->silent == FALSE)
    g_print ("PIPgen: I'm plugged, therefore I'm in. Getting mutex\n");
//enqueue big picture
	//check wehter list is too long 
	//enqueue
  if (pad== filter->sinkpad){
    g_mutex_lock(filter->mutex);
    if (filter->silent == FALSE)g_print ("PIPgen: gotten mutex for normal sinkpad\n");	

    if ((!filter->sentyet)&&(filter->prerollcnt>0)){//then we need to preroll
        if (filter->silent == FALSE)g_print("GstPip:Prerolling becaus not send yet, prerollcnt=%d and offset %llu!!\n",filter->prerollcnt, buf->offset);
        outbufcoord=gst_buffer_new_and_alloc (0);
	GstCaps *tempcaps=gst_caps_copy (gst_pad_get_pad_template_caps(filter->srccoord));
        gst_buffer_set_caps(outbufcoord,tempcaps);
	outbufcoord->timestamp=buf->timestamp;
        outbufcoord->duration=buf->duration;
        outbufcoord->offset=buf->offset;
        outbufcoord->offset_end=buf->offset_end;
	gst_caps_unref(tempcaps);
        ret=gst_pad_push (filter->srccoord, outbufcoord);
        //dummy image buffer
	outbuf=gst_buffer_new_and_alloc (16*filter->bpp/8);
        tempcaps=gst_caps_copy (filter->tomodify);
	if (FALSE==modifiy_caps(tempcaps,4,4)){
	 if (filter->silent == FALSE)g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps);
	outbuf->timestamp=buf->timestamp;;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	ret=gst_pad_push (filter->srcpad, outbuf);
	//filter->prerollcnt--;	
        //if (filter->prerollcnt==0)filter->sentyet=TRUE;	
    }
    if ((filter->lbig)&&(g_list_length(filter->lbig)>(filter->hardcut))) gst_buffer_unref(buf);
    else filter->lbig=g_list_append(filter->lbig,buf); 
    g_mutex_unlock(filter->mutex);
    if (g_list_length(filter->lbig)>(filter->hardcut/2)){
        if (filter->silent == FALSE)g_print("PIPgen going to sleep\n");
	usleep(300000);
    }
    return ret;
  }
//position 
  if (pad== filter->sinkcoord){
    if (filter->silent == FALSE)g_print("PIPgen Getting mutex for synced coordination usage\n");
    g_mutex_lock(filter->mutex);
    if (filter->silent == FALSE)g_print("PIPgen Got mutex for synced coordination usage and will start searching:");
//unref all older and find new
    guint64 delete;
    delete=buf->offset;
    GstBuffer * big=NULL;
    GList* temp=filter->lbig;
    int numregions=0;
    while (temp){
      g_print("Found Nr. %llu,",((GstBuffer*)temp->data)->offset);
      if (((GstBuffer*)temp->data)->offset==delete){
	 big=(GstBuffer*)temp->data;
	 temp=NULL;
      }
      else if (((GstBuffer*)temp->data)->offset<delete){
	gst_buffer_unref((GstBuffer*)temp->data);
	filter->lbig=g_list_remove(filter->lbig,temp->data);
	temp=filter->lbig;
      }
      else temp=g_list_next(temp);
    }
    if (filter->silent == FALSE)g_print("PIPgen: Finished searching for corresponding buffer %llu\n",buf->offset);
    if (big){
//send both, first coords
      g_print("PIPgen: found corresponding buffers %llu\n",buf->offset);
      gst_buffer_ref(buf);//because it will be unrefed
      ret=gst_pad_push (filter->srccoord, buf);
      if (ret!=GST_FLOW_OK) {
	g_print("PIPgen: Warning!!! Buffer on srccoord went wrong\n");  
        g_mutex_unlock(filter->mutex);	
	return ret;
      }
      numregions=buf->size/sizeof(struct coordinates);
      g_print("checking %d regions",numregions);
      int i,j,totalsize,lines,rows,linesize,x1,x2,y1,y2;

      for (i=0;i<numregions;i++){
        filter->sentyet=TRUE;	
	if (little_endian){
	x1=(int)(unsigned int)((struct coordinates*)(buf->data))[i].x1;
	x2=(int)(unsigned int)((struct coordinates*)(buf->data))[i].x2;
	y1=(int)(unsigned int)((struct coordinates*)(buf->data))[i].y1;
	y2=(int)(unsigned int)((struct coordinates*)(buf->data))[i].y2;}
	else{
        x1=(int)(unsigned int)DoubleSwap(((struct coordinates*)(buf->data))[i].x1);
        x2=(int)(unsigned int)DoubleSwap(((struct coordinates*)(buf->data))[i].x2);
        y1=(int)(unsigned int)DoubleSwap(((struct coordinates*)(buf->data))[i].y1);
        y2=(int)(unsigned int)DoubleSwap(((struct coordinates*)(buf->data))[i].y2);}
      
	if ((x1>=filter->orgwidth)||(x2>=filter->orgwidth)||(y1>=filter->orgheight)||(y2>=filter->orgheight)){
	  g_print("GstPip:Warning, coordinates outside image bounds x1: %d x2 %d y1 %d y2 %d Widht: %d Height %d!!\n",x1,x2,y1,y2,filter->orgwidth,filter->orgheight);
	  return GST_FLOW_ERROR;	
	}
        //g_print("PIPGEN Float values %f,%f, %f, %f\n",((struct coordinates*)(buf->data))[i].x1,((struct coordinates*)(buf->data))[i].x2,((struct coordinates*)(buf->data))[i].y1,((struct coordinates*)(buf->data))[i].y2);
//	if (x1>=filter->orgwidth)x1=filter->orgwidth-1;
//	if (x2>=filter->orgwidth)x2=filter->orgwidth-1;
//	if (y1>=filter->orgheight)y1=filter->orgheight-1;
//	if (y2>=filter->orgheight)y2=filter->orgheight-1;
	lines=1+y2-y1;
	rows=1+x2-x1;
	linesize=(1+x2-x1)*(filter->bpp/8);
	totalsize=linesize*lines;
	outbuf=gst_buffer_new_and_alloc (totalsize);
 	GstCaps *tempcaps=gst_caps_copy (filter->tomodify);

	if (FALSE==modifiy_caps(tempcaps,rows,lines)){
	 g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps); //ref has been "transferred" to buffer
	for (j=0;j<lines;j++){
	  memcpy(outbuf->data+(j*linesize),big->data+((filter->bpp/8)*((y1+j)*filter->orgwidth+x1)),linesize);
	}
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	if (filter->silent == FALSE)g_print("Pushing buffer (totalsize %d) size %d x %d that comes from %d and %d with %d bits each pixel!!\n",totalsize,rows,lines,x1,y1,filter->bpp);
	ret=gst_pad_push (filter->srcpad, outbuf);
	if (ret!=GST_FLOW_OK){
	 if (filter->silent == FALSE)g_print("GstPip:Sourcepad push went wrong!!");
	 return ret;
	}
      }
      if ((filter->prerollcnt>0)&&(!filter->sentyet)&&(numregions==0)){
        if (filter->silent == FALSE)g_print("GstPip:Prerolling becaus numregions=0 and not send yet and prerollcnt= %d with offset %llu!!\n",filter->prerollcnt,buf->offset);
	outbuf=gst_buffer_new_and_alloc (16*filter->bpp/8);
        GstCaps *tempcaps;
        tempcaps=gst_caps_copy (filter->tomodify);
	if (FALSE==modifiy_caps(tempcaps,4,4)){
	 if (filter->silent == FALSE)g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps);
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	ret=gst_pad_push (filter->srcpad, outbuf);
	//filter->prerollcnt--;
        //if (filter->prerollcnt==0)filter->sentyet=TRUE;	
      }
    }
    else if ((!filter->sentyet)&&(!big)){//then we need to preroll
        if (filter->silent == FALSE)g_print("GstPip:Prerolling becaus bigbuffer is null and not send yet, prerollcnt=%d, offset=%llu!!\n",filter->prerollcnt,buf->offset);
        outbufcoord=gst_buffer_new_and_alloc (0);
	GstCaps *tempcaps=gst_caps_copy (gst_pad_get_pad_template_caps(filter->srccoord));
        gst_buffer_set_caps(outbufcoord,tempcaps);
	outbufcoord->timestamp=buf->timestamp;
        outbufcoord->duration=buf->duration;
        outbufcoord->offset=buf->offset;
        outbufcoord->offset_end=buf->offset_end;
	gst_caps_unref(tempcaps);
        ret=gst_pad_push (filter->srccoord, outbufcoord);
        //dummy image buffer
	outbuf=gst_buffer_new_and_alloc (16*filter->bpp/8);
        tempcaps=gst_caps_copy (filter->tomodify);
	if (FALSE==modifiy_caps(tempcaps,4,4)){
	 if (filter->silent == FALSE)g_print("GstPip:Warning, caps modification failed!!");
	 g_mutex_unlock(filter->mutex);
	 return GST_FLOW_ERROR;
	}
        gst_buffer_set_caps(outbuf,tempcaps);
	gst_caps_unref(tempcaps);
	outbuf->timestamp=buf->timestamp;
        outbuf->duration=buf->duration;
        outbuf->offset=buf->offset;
        outbuf->offset_end=buf->offset_end;
	ret=gst_pad_push (filter->srcpad, outbuf);
	//filter->prerollcnt--;	
        //if (filter->prerollcnt==0)filter->sentyet=TRUE;	
    }

    gst_buffer_unref(buf);
//see whether another thread left us a buffer
  }
  g_mutex_unlock(filter->mutex);
  return ret;
}

