/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_PIPTRANS_H__
#define __GST_PIPTRANS_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_PIPTRANS \
  (gst_piptranstrans_get_type())
#define GST_PIPTRANS(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_PIPTRANS,Gstpiptranstrans))
#define GST_PIPTRANS_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_PIPTRANS,GstpiptranstransClass))
#define GST_IS_PIPTRANS(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_PIPTRANS))
#define GST_IS_PIPTRANS_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_PIPTRANS))

typedef struct _Gstpiptranstrans      Gstpiptranstrans;
typedef struct _GstpiptranstransClass GstpiptranstransClass;

struct coordinates
{
  gdouble x1;
  gdouble y1;
  gdouble x2;
  gdouble y2;
};


struct transmat
{
  gdouble a00;
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
};



struct _Gstpiptranstrans
{
  GstElement element;

  GstPad *sinkpad,*sinksmall,*sinkpos, *srcpad;

  gboolean silent,trigger,trigger_small,trigger_pos, trigger_num,stopped1,stopped2,stopped3;

  GMutex* mutex;

  GstBuffer *b1,*b2,*b3;

  gint off_x,off_y;
  guint siftoff_x,siftoff_y;
  gint w,h,ws,hs,ws_neg,hs_neg;//width and height of main image
  gint bpp,bppsmall;
  gboolean bigneg,smallneg;//wether big and small input have been negotiated
  gboolean diff,fweos,justtrans;//wether to visualise sift(TRUE) or normal PIPTRANS, justtrans will only transform image 2 to 1 without underlaying 
  guint64 offbig,offsmall,offpos;
  gdouble scale;//amount of pixels that the small picture has been downscaled
  CvSize imagesize,smallsize;

  GList *lbig,*lsmall,*lpos;//list of buffers
  guint64 offbigold, offsmallold, offposold, offbignew, offsmallnew, offposnew,penalty,valve_buf; 
  gdouble difff;
  gboolean avg;
};

struct _GstpiptranstransClass 
{
  GstElementClass parent_class;
};

GType gst_piptranstrans_get_type (void);

G_END_DECLS

#endif /* __GST_PIPTRANS_H__ */
