#define _PELCO_D
#define POINTMOVE_TIME 50
#define STOPCOMMAND_TIME 50
#define PTU_ADDR 0x01

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>

#ifdef _PELCO_D
enum _Direct
{
    STOP=0x00 ,
    CRIGHT=0x02 ,
    CLEFT=0x04 ,
    CUP=0x08 ,
    CDOWN=0x10
};
#else
enum _Direct
{
    STOP=0x00 ,
    CRIGHT=0x01 ,
    CLEFT=0x02 ,
    CUP=0x08 ,
    CDOWN=0x04
};
#endif
typedef enum _Direct Direct;

int  speed_arr[]={B38400,B19200,B9600,B4800,B2400,B1200,B300,B38400,B19200,B9600,B4800,B2400,B1200,B300,};
int  name_arr[]={38400,19200,9600,4800,2400,1200,300,38400,19200,9600,4800,2400,1200,300,};

void  SetSpeed(int fd, int speed)
{
    int    i;
    int    status;
    struct termios   Opt;
    tcgetattr(fd,  & Opt);
    for (i=0;i<sizeof(speed_arr)/sizeof(int);i++)
    {
        if (speed==name_arr[i])
        {
            tcflush(fd, TCIOFLUSH);
            cfsetispeed( & Opt, speed_arr[i]);
            cfsetospeed( & Opt, speed_arr[i]);
            status=tcsetattr(fd,TCSANOW, &Opt);
            if (status  !=   0)
                printf( " set_speed [Failue]\n " );
            return ;
        }
        tcflush(fd,TCIOFLUSH);
    }
}

int  SetParity(int  fd, int  databits, int  stopbits, int  parity)
{
    struct termios options;
    if ( tcgetattr( fd, & options)  !=    0 )
    {
        printf( " set_Parity [Failue]\n " );
        return FALSE;
    }
    options.c_cflag  &=   ~ CSIZE;
    switch  (databits)
    {
        case   7 :
            options.c_cflag  |=  CS7;
        break ;
        case   8 :
            options.c_cflag  |=  CS8;
        break ;
        default :
            fprintf(stderr, " Unsupported data size\n " );
        return  FALSE;
    }
    switch  (parity)
    {
        case   'n' :
        case   'N' :
            options.c_cflag  &=   ~ PARENB;     /**/ /*  Clear parity enable  */
            options.c_iflag  &=   ~ INPCK;      /**/ /*  Enable parity checking  */
        break ;
        case   'o' :
        case   'O' :
            options.c_cflag  |=  (PARODD  |  PARENB);
            options.c_iflag  |=  INPCK;               /**/ /*  Disnable parity checking  */
        break ;
        case   'e' :
        case   'E' :
            options.c_cflag  |=  PARENB;       /**/ /*  Enable parity  */
            options.c_cflag  &=   ~ PARODD;
            options.c_iflag  |=  INPCK;        /**/ /*  Disnable parity checking  */
        break ;
        case   'S' :
        case   's' :   /**/ /* as no parity */
            options.c_cflag  &=   ~ PARENB;
            options.c_cflag  &=   ~ CSTOPB;
        break ;
        default :
            fprintf(stderr, " Unsupported parity\n " );
            return  FALSE;
    }
    switch  (stopbits)
    {
        case   1 :
            options.c_cflag  &=   ~ CSTOPB;
        break ;
        case   2 :
            options.c_cflag  |=  CSTOPB;
        break ;
        default :
            fprintf(stderr, " Unsupported stop bits\n " );
        return FALSE;
    }

    if  (parity  !=   'n' || parity !='N')
        options.c_iflag  |=  INPCK;
    options.c_cc[VTIME]  =   150 ;  //  15 seconds
    options.c_cc[VMIN]  =   0 ;

    tcflush(fd,TCIFLUSH);  /**/ /*  Update the options and do it NOW  */
    if  (tcsetattr(fd,TCSANOW, & options)  !=   0 )
    {
        perror( " SetupSerial 3 " );
        return  FALSE;
    }
    return  TRUE;
}

void do_delay(long msec)
{
  struct timespec ts, rem;
  int retc;

  ts.tv_sec = msec / 1000;
  ts.tv_nsec = (msec % 1000) * 1000000;
  do
  {
    retc = nanosleep(&ts, &rem);
    ts = rem;
  } while (retc == -1);
}

int OpenDev(char *Dev)
{
    int  fd=open(Dev,O_RDWR);          // | O_NOCTTY | O_NDELAY

    if (-1 == fd)
    {
        return -1;
    }
    else
    {
        return fd;
    }
}

inline  int  CheckSum(char *strCommOut)
{
    int  sum = 0 ;
    int  i;
    for (i = 1 ; i < 6 ; i ++ )
    {
        sum += strCommOut[i];
    }
    if (sum  >   0xFF )
        strCommOut[ 6 ] = sum % 0x100 ;
    else
        strCommOut[ 6 ] = sum;
    return sum;
}

char *MakeDirect(char *strCommOut, char chDirrect, char chCammerID)
{
#ifdef _PELCO_D
    strCommOut[0]=0xFF ;
    strCommOut[1]=chCammerID;
    strCommOut[2]=0x0 ;
    strCommOut[3]=chDirrect;
    strCommOut[4]=0x0;
    strCommOut[5]=0x0;
#else
    strCommOut[0]=0xA5 ;
    strCommOut[1]=chCammerID;
    strCommOut[2]=chDirrect ;
    strCommOut[3]=0x00;
    strCommOut[4]=0x00;//xff;// 0x20 ;
    strCommOut[5]=0x00 ;
#endif
    CheckSum(strCommOut);
    return strCommOut;
}

int Running(int fd,char chCammerID,char chDir)
{
    char strCommOut[7];
    int t;
    MakeDirect(strCommOut,chDir,chCammerID);
    t=write(fd,strCommOut,sizeof(strCommOut));
    return t;
}

void PointMove(int fd,Direct d)
{
    Running(fd,PTU_ADDR,d);
    do_delay(POINTMOVE_TIME);
    Running(fd,PTU_ADDR,STOP);
    do_delay(STOPCOMMAND_TIME);
}

void ContinuousMoveWithTime(int fd,Direct d,int msec)
{
    Running(fd,PTU_ADDR,d);
    do_delay(msec);
    Running(fd,PTU_ADDR,STOP);
    do_delay(STOPCOMMAND_TIME);
}

void ContinuousMoveWithRelatedDegree(int fd,Direct d,double degree)
{
    int bb=TRUE;
    int msec;
    switch(d)
    {
        case CUP:
            msec=(int)(degree*17400/91);
            break;
        case CDOWN:
            msec=(int)(degree*17400/91);
            break;
        case CLEFT:
            msec=(int)(degree*43400/355);
            break;
        case CRIGHT:
            msec=(int)(degree*43400/355);
            break;
        default:
            bb=FALSE;
            break;
    }
    if (bb) ContinuousMoveWithTime(fd,d,msec);
}

void Initialization(int fd)
{
    // Put the camera at the center position
    ContinuousMoveWithRelatedDegree(fd,CLEFT,360);
    ContinuousMoveWithRelatedDegree(fd,CDOWN,95);
    ContinuousMoveWithRelatedDegree(fd,CRIGHT,177.5);
    ContinuousMoveWithRelatedDegree(fd,CUP,40.5);
}
/*
int  main()
{
    char  *dev2  = "/dev/ttyUSB0";
    //char *dev2="/dev/ttyS0";
    int  fd  =  OpenDev(dev2);
    if  (fd > 0 )
    {
        printf("Serial Port '%s' is opened\n ",dev2);
        SetSpeed(fd, 2400 );
    }
    else
    {
        printf( " Can't Open Serial Port '%s'\n " ,dev2);
        return   -1 ;
    }
    if  (SetParity(fd,8,1,'n') == -1 )
    {
        printf( " Set Parity Error\n " );
        return   -1 ;
    }

    // following codes is just for example // delete?

    printf("Position Initializing\n");
    Initialization(fd);

    int di;
    double rt;
    Direct d;
    printf("Now it is command waiting.\n");
    while (true)
    {
        printf("Please enter the direction (1-4 for left right up down) and degree (deg), other to exit.\n");
        scanf("%d%lf",&di,&rt);
        if (di==1)
            d=LEFT;
        else if (di==2)
            d=RIGHT;
        else if (di==3)
            d=UP;
        else if (di==4)
            d=DOWN;
        else break;
        ContinuousMoveWithRelatedDegree(fd,d,rt);
    }

    // above is just for example // delete?

    return 0;
}*/
