/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-rosparam
 *
 * FIXME:Describe rosparam here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! rosparam ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */
#define MAXSTRING 10000

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif
//#include <gtk/gtk.h>
#include <ros/ros.h>
#include "roscommon.h"
#include <gio/gio.h>
#include <gst/gst.h>
#include <gdk/gdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gstrosparam.h"

GST_DEBUG_CATEGORY_STATIC (gst_rosparam_debug);
#define GST_CAT_DEFAULT gst_rosparam_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_SPORT,
  PROP_RPORT,
  PROP_INTERVAL,
  PROP_HOSTNAME,
  PROP_DESC,
  PROP_CONF,
  PROP_DELAY,
  PROP_EXTRADELAY,
  PROP_JHOSTNAME,
  PROP_JPORT,
  PROP_TOPIC
};


/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstrosparam, gst_rosparam, GstElement,
    GST_TYPE_ELEMENT);

static void gst_rosparam_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_rosparam_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstStateChangeReturn gst_rosparam_change_state (GstElement *element, GstStateChange transition);
 static GstFlowReturn
gst_rosparam_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf);

static gboolean gst_rosparam_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_rosparam_chain (GstPad * pad, GstBuffer * buf);
void msgtoall(Gstrosparam *filter, char* msg);
void genspinoffs(Gstrosparam *filter);
void setupsockets(Gstrosparam *filter);
/* GObject vmethod implementations */

static void
gst_rosparam_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "rosparam",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the rosparam's class */
static void
gst_rosparam_class_init (GstrosparamClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_rosparam_set_property;
  gobject_class->get_property = gst_rosparam_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SPORT,
      g_param_spec_int ("sendport", "Sendport", "To what port should we send Packets",0,65535,4444, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_RPORT,
      g_param_spec_int ("rcvport", "Rcvport", "To what port should we listen",0,65535,4445, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_INTERVAL,
      g_param_spec_int ("interval", "Interval", "Minimum interval (ms) between checking the states of properties",-1,G_MAXINT,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_HOSTNAME,
      g_param_spec_string ("host", "Host", "Host adress of GUI-Program",
          "127.0.0.1", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_DESC,
      g_param_spec_string ("exportdesc", "Exportdesc", "Export description, Elementname;Propertyname;Elementname2;Pr...",
          NULL, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_CONF,
      g_param_spec_string ("conffiles", "Conffiles", "a comma separated list of files that will be read to set up remote systems",
          NULL, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_DELAY,
      g_param_spec_int ("delay", "Delay", "Seconds to block pipeline construction after Play was sent to downstream",0,G_MAXINT,5, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DELAY,
      g_param_spec_int ("extradelay", "Extradelay", "Seconds to immedeately block pipeline construction",0,G_MAXINT,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_JHOSTNAME,
      g_param_spec_string ("jhost", "Jhost", "Host adress to send join request to",
          NULL, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_JPORT,
      g_param_spec_int ("jport", "Jport", "On what port should we send our joinrequest, if adress given",0,65535,4445, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TOPIC,
      g_param_spec_string ("topic", "Topic", "On which topic-prefix we should publish /image_raw will be appended",
          "gstros", G_PARAM_READWRITE)); 

  gstelement_class->change_state = gst_rosparam_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_rosparam_init (Gstrosparam * filter,
    GstrosparamClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_rosparam_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_rosparam_chain));
  gst_pad_set_bufferalloc_function (filter->sinkpad,
      GST_DEBUG_FUNCPTR (gst_rosparam_bufferalloc));


  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->remoteadress=g_strdup("127.0.0.1");
  //filter->exportdsc=g_strdup("videotestsrc0;pattern;videotestsrc0;is-live");
  filter->exportdsc=NULL;
  filter->rport=4445;
  filter->sport=4444;
  filter->interval=-1;
  filter->firsttime=TRUE;
  filter->propertylist=NULL;
  filter->width=-1;
  filter->height=-1;
  filter->rsock=NULL;
  filter->ssock=NULL;
  filter->source=0;
  filter->conffiles=NULL;
  filter->spinoffs=NULL;
  filter->controllers=NULL;
  filter->uidesc=NULL;
  filter->socketopen=FALSE;
  filter->delay=5;
  filter->jport=4445;
  filter->jadress=NULL;
  filter->jaddr=NULL;
  filter->jiaddr=NULL;
  filter->sendj=FALSE;
  filter->joinui[0]=0;
  filter->topic=g_strdup_printf("gstros");
}
 
static void
gst_rosparam_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstrosparam *filter = GST_ROSPARAM (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_SPORT:
      filter->sport = g_value_get_int (value);
      break;
    case PROP_RPORT:
      filter->rport = g_value_get_int (value);
      break;
    case PROP_INTERVAL:
      filter->interval = g_value_get_int (value);
      break;
    case PROP_DESC:
      if (filter->exportdsc)g_free (filter->exportdsc);
      filter->exportdsc = g_value_dup_string (value);
      filter->firsttime=TRUE;
      break;
    case PROP_HOSTNAME:
      if (filter->remoteadress)g_free (filter->remoteadress);
      filter->remoteadress = g_value_dup_string (value);
      filter->firsttime=TRUE;
      break;
    case PROP_CONF:
      if (filter->conffiles)g_free (filter->conffiles);
      filter->conffiles = g_value_dup_string (value);
      if (!filter->socketopen)setupsockets(filter);
      filter->socketopen=TRUE;
      genspinoffs(filter);
      msgtoall(filter,"PLAY;");
      sleep(filter->delay);
      break;
    case PROP_DELAY:
      filter->delay = g_value_get_int (value);
      break;
    case PROP_EXTRADELAY:
      filter->extradelay = g_value_get_int (value);
      sleep(filter->extradelay);
      break;
    case PROP_JHOSTNAME:
      if (filter->jadress)g_free (filter->jadress);
      filter->jadress = g_value_dup_string (value);
      filter->sendj=TRUE;
      break;
    case PROP_JPORT:
      filter->jport = g_value_get_int (value);
      break;
    case PROP_TOPIC:
      if (filter->topic)g_free (filter->topic);
      filter->topic = g_value_dup_string (value);
	//todo mutex,unadvertise,advertise,unlock
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
  //if (prop_id==PROP_SPORT)

}

static void
gst_rosparam_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstrosparam *filter = GST_ROSPARAM (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_SPORT:
      g_value_set_int (value, filter->sport);
      break;
    case PROP_RPORT:
      g_value_set_int (value, filter->rport);
      break;
    case PROP_INTERVAL:
      g_value_set_int (value, filter->interval);
      break;
    case PROP_HOSTNAME:
      g_value_set_string (value, filter->remoteadress);
      break;
    case PROP_DESC:
      g_value_set_string (value, filter->exportdsc);
      break;
    case PROP_CONF:
      g_value_set_string (value, filter->conffiles);
      break;
    case PROP_DELAY:
      g_value_set_int (value, filter->delay);
      break;
    case PROP_EXTRADELAY:
      g_value_set_int (value, filter->extradelay);
      break;
    case PROP_JPORT:
      g_value_set_int(value, filter->jport);
      break;
    case PROP_JHOSTNAME:
      g_value_set_string (value, filter->jadress);
      break;
    case PROP_TOPIC:
      g_value_set_string (value, filter->topic);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

 static GstFlowReturn
gst_rosparam_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  Gstrosparam *filter;
  GstFlowReturn result;
  //g_print ("Bufferalloc has been caled with Caps: %s\n",gst_caps_to_string(caps));
  filter = GST_ROSPARAM (GST_PAD_PARENT (pad));

  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  return result;
}

static gchar* setuppropertylist(Gstrosparam *filter){
  char* temp=filter->exportdsc;//the char that we use to parse
  gchar* outputstring=g_malloc0(MAXSTRING);
  //if (filter->firsttime)
  strcat(outputstring,"UI;");//User interface
  gboolean valid=1;
  gboolean finished=0;
  while (valid && !finished && temp && strlen(temp)){
    gdouble definedmin;
    gdouble definedmax;
    gboolean defined=FALSE;
    struct property* thisproperty=g_malloc0(sizeof(struct property));
    char* thiselement=g_malloc0(MAXSTRING);
    char* thispropertyname=g_malloc0(MAXSTRING);
    char* thistempstr=g_malloc0(MAXSTRING);
    temp=deserialize_string(thiselement,temp);
    temp=deserialize_string(thispropertyname,temp);  
    if(*temp==91){
      temp=deserialize_double(&definedmin,temp+1);
      temp=deserialize_double(&definedmax,temp);
      defined=TRUE;
      temp++;
    } 
    else g_print("Dont do user definition because rest starts with %u is %s\n",(unsigned int)*temp,temp); 
    if (strlen(thispropertyname)){
      thisproperty->elpointer=gst_bin_get_by_name_recurse_up (GST_BIN(GST_ELEMENT_PARENT(filter)),thiselement);/*
      if (!thisproperty->elpointer){
	GstBin* binbin=GST_BIN(GST_ELEMENT_PARENT(filter));
	if (!binbin) g_print("Bin Failed");
	GstElementFactory *factory=gst_element_factory_find(thiselement);
	if (!factory) g_print("Factory Failed");
	if (!gst_element_factory_get_element_type)g_print("Type Failed");
	thisproperty->elpointer=gst_bin_get_by_interface(GST_BIN(GST_ELEMENT_PARENT(filter)),gst_element_factory_get_element_type(gst_element_factory_find(thiselement)));
	if (thisproperty->elpointer)g_print("and no element:(");
	sleep(10);  funzt nicht der scheiß
      }*/
      if (thisproperty->elpointer)     
	thisproperty->spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(thisproperty->elpointer)),thispropertyname);
      thisproperty->elname=g_strdup(thiselement);
      thisproperty->propertyname=g_strdup(thispropertyname);
      thisproperty->alias=g_strdup_printf("%s:%s",thiselement,thispropertyname);
      if (thisproperty->spec==NULL){
      }
      else if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)&&(thisproperty->spec->flags & G_PARAM_READABLE)){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&tempbool,NULL);
	//thisproperty->type=BOOLEAN;
	thisproperty->active=tempbool;
	if (tempbool)sprintf(thistempstr,"%s:%s;BOOLEAN;1;",thiselement,thispropertyname);
	else sprintf(thistempstr,"%s:%s;BOOLEAN;0;",thiselement,thispropertyname);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&tempbool,NULL);
	//thisproperty->type=BOOLEAN_RO;
	thisproperty->active=tempbool;
	if (tempbool)sprintf(thistempstr,"%s:%s;BOOLEAN_RO;1;",thiselement,thispropertyname);
	else sprintf(thistempstr,"%s:%s;BOOLEAN_RO;0;",thiselement,thispropertyname);
      }	
      else if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&!(thisproperty->spec->flags & G_PARAM_READABLE)){
        //gboolean tempbool;
	//g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&tempbool,NULL);
	//thisproperty->type=BOOLEAN_WO;
	//thisproperty->active=tempbool;
	sprintf(thistempstr,"%s:%s;BOOLEAN_WO;",thiselement,thispropertyname);
      }	
				
      else if ((thisproperty->spec->value_type==G_TYPE_INT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	int min,max,cur;					
	min=G_PARAM_SPEC_INT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT(thisproperty->spec)->maximum;
	//thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(int)definedmin;
	if (defined)max=(int)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%d;%d;%d;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	int min,max,cur;					
	min=G_PARAM_SPEC_INT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT(thisproperty->spec)->maximum;
	//thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(int)definedmin;
	if (defined)max=(int)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%d;%d;%d;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	unsigned int min,max,cur;					
	min=G_PARAM_SPEC_UINT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT(thisproperty->spec)->maximum;
	//thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(unsigned int)definedmin;
	if (defined)max=(unsigned int)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%u;%u;%u;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	unsigned int min,max,cur;					
	min=G_PARAM_SPEC_UINT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT(thisproperty->spec)->maximum;
	//thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(unsigned int)definedmin;
	if (defined)max=(unsigned int)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%u;%u;%u;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT64)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gint64 min,max,cur;					
	min=G_PARAM_SPEC_INT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT64(thisproperty->spec)->maximum;
	//thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(gint64)definedmin;
	if (defined)max=(gint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%lld;%lld;%lld;",thiselement,thispropertyname,(long long int)min,(long long int)max,(long long int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_INT64)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gint64 min,max,cur;					
	min=G_PARAM_SPEC_INT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_INT64(thisproperty->spec)->maximum;
	//thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(gint64)definedmin;
	if (defined)max=(gint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%lld;%lld;%lld;",thiselement,thispropertyname,(long long int)min,(long long int)max,(long long int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT64)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	guint64 min,max,cur;					
	min=G_PARAM_SPEC_UINT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT64(thisproperty->spec)->maximum;
	//thisproperty->type=INT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(guint64)definedmin;
	if (defined)max=(guint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT;%llu;%llu;%llu;",thiselement,thispropertyname,(long long unsigned int)min,(long long unsigned int)max,(long long unsigned int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_UINT64)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	guint64 min,max,cur;					
	min=G_PARAM_SPEC_UINT64(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_UINT64(thisproperty->spec)->maximum;
	//thisproperty->type=INT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnumi=cur;
	if (defined)min=(guint64)definedmin;
	if (defined)max=(guint64)definedmax;
	sprintf(thistempstr,"%s:%s;INT_RO;%llu;%llu;%llu;",thiselement,thispropertyname,(long long unsigned int)min,(long long unsigned int)max,(long long unsigned int)cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_DOUBLE)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gdouble min,max,cur;					
	min=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->maximum;
	//thisproperty->type=FLOAT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=definedmin;
	if (defined)max=definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_DOUBLE)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gdouble min,max,cur;					
	min=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_DOUBLE(thisproperty->spec)->maximum;
	//thisproperty->type=FLOAT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=definedmin;
	if (defined)max=definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT_RO;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_FLOAT)&&(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gfloat min,max,cur;					
	min=G_PARAM_SPEC_FLOAT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_FLOAT(thisproperty->spec)->maximum;
	//thisproperty->type=FLOAT;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=(float)definedmin;
	if (defined)max=(float)definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
      else if ((thisproperty->spec->value_type==G_TYPE_FLOAT)&&!(thisproperty->spec->flags & G_PARAM_WRITABLE)){						
	gfloat min,max,cur;					
	min=G_PARAM_SPEC_FLOAT(thisproperty->spec)->minimum;		
	max=G_PARAM_SPEC_FLOAT(thisproperty->spec)->maximum;
	//thisproperty->type=FLOAT_RO;	
	g_object_get(G_OBJECT(thisproperty->elpointer),thispropertyname,&cur,NULL);
	thisproperty->currentnum=cur;
	if (defined)min=(float)definedmin;
	if (defined)max=(float)definedmax;
	sprintf(thistempstr,"%s:%s;FLOAT_RO;%f;%f;%f;",thiselement,thispropertyname,min,max,cur);
      }
	
      else if (G_IS_PARAM_SPEC_ENUM (thisproperty->spec)){
	gint i;
        gchar *str=NULL;
        GEnumClass *cclass = G_ENUM_CLASS (g_type_class_ref (thisproperty->spec->value_type));
        GValue value = { 0, };	
        g_value_init (&value, thisproperty->spec->value_type);
	g_object_get_property(G_OBJECT(thisproperty->elpointer),thispropertyname,&value);
	int current=g_value_get_enum(&value);
	sprintf(thistempstr,"%s:%s;MENU;%d;",thiselement,thispropertyname,cclass->n_values);
	for (i = 0; i < cclass->n_values; i++) {
          GEnumValue *value = &cclass->values[i];
	  strcat(thistempstr,value->value_nick);
	  strcat(thistempstr,";");
	  if (i==current)str=g_strdup(value->value_nick);
        }
	strcat(thistempstr,str);
	strcat(thistempstr,";");
	thisproperty->charval=str;
        g_type_class_unref(cclass);
        g_value_unset(&value);
      }
    else if (!strcmp("GstPad",g_type_name (thisproperty->spec->value_type))){

        gint i=0;//iterator
        gchar *str=g_malloc0(MAXSTRING);
	GObject* current;
	GstIterator* padit=gst_element_iterate_pads (thisproperty->elpointer);
        g_object_get (G_OBJECT (thisproperty->elpointer),thisproperty->propertyname,&current, NULL);
	if (thisproperty->charval)g_free(thisproperty->charval);
	thisproperty->charval=gst_pad_get_pe_name((GstPad*)current);
	gboolean done=FALSE;
	gpointer item;
        GstPadTemplate *ctemplate;
        gchar *temppadn;
	while (!done) {
	  switch (gst_iterator_next (padit, (gpointer*)&item)) {
	  case GST_ITERATOR_OK:
		//pview->enum_pointer[i++]=item;
                ctemplate=gst_pad_get_pad_template(((GstPad*)item));
		if (ctemplate && GST_PAD_REQUEST==GST_PAD_TEMPLATE_PRESENCE(ctemplate)){
		  i++;
		  temppadn=gst_pad_get_pe_name((GstPad*)item);
		  strcat (str,temppadn);
		  g_free(temppadn);
		  strcat (str,";");
		}
		//gst_object_unref (template);
		gst_object_unref (item);
		break;
	  case GST_ITERATOR_RESYNC:
		gst_iterator_free (padit);
		padit=gst_element_iterate_pads (thisproperty->elpointer);
		str[0]=0;
		i=0;
		break;
	  case GST_ITERATOR_ERROR:
		//g_print("GstIterator Returning Unexpected Error");
		done = TRUE;
		break;
	  case GST_ITERATOR_DONE:
		//g_print("GstIterator Done");
		done = TRUE;
		break;
	  default:
		//g_print("GstIterator Returning Unexpected Error"); 
		done = TRUE;
		break;
	  }
	}
	gst_iterator_free (padit);
	sprintf(thistempstr,"%s:%s;MENU;%d;",thiselement,thispropertyname,i);
	strcat (thistempstr,str);
	strcat (thistempstr,thisproperty->charval);//append current selection
	strcat (thistempstr,";");
        g_free(str);
      }



    }
    if (strlen(thistempstr)){
	strcat(outputstring,thistempstr);
        filter->propertylist=g_list_append(filter->propertylist,thisproperty);
    }
    else g_free(thisproperty);
    g_free(thistempstr);
    g_free(thiselement);
    g_free(thispropertyname);

  }
//cat all 



 return outputstring;
}

static gchar* chupdate(Gstrosparam *filter,gboolean sendall){
  gchar* outputstring=g_malloc0(MAXSTRING);
  GList *templ=filter->propertylist;
  struct property* thisproperty;
  while (templ){
    char* thistempstr=g_malloc0(MAXSTRING);
    char* thistempstros=g_malloc0(MAXSTRING);
    thisproperty=(struct property*)templ->data;
    if (!filter->silent)g_print("Checking listelement %s for updates\n",thisproperty->elname);
    sprintf(thistempstros,"%s/%s/%s",filter->topic,thisproperty->elname,thisproperty->propertyname);//set the ros name
    if ((thisproperty->spec->value_type==G_TYPE_BOOLEAN)&&(thisproperty->spec->flags & G_PARAM_READABLE)){
        gboolean tempbool;
	g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&tempbool,NULL);
        if (!(thisproperty->active==tempbool)||sendall){
          if (tempbool)sprintf(thistempstr,"%s:%s;1;",thisproperty->elname,thisproperty->propertyname);
 	  else sprintf(thistempstr,"%s:%s;0;",thisproperty->elname,thisproperty->propertyname);
	  thisproperty->active=tempbool;
	  bool cppbool=((tempbool==FALSE) ? false : true);
	  filter->nh->setParam(thistempstros, cppbool);
        }
      }
    else if (thisproperty->spec->value_type==G_TYPE_INT){						
      int cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%d;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnumi=cur;
	filter->nh->setParam(thistempstros, cur);

      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_UINT){						
      unsigned int cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%u;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnumi=cur;
	int curi=cur;
	filter->nh->setParam(thistempstros, curi);
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_INT64){						
      gint64 cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%lld;",thisproperty->elname,thisproperty->propertyname,(long long int)cur);
	thisproperty->currentnumi=cur;
	int curi=cur;
	filter->nh->setParam(thistempstros, curi);
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_UINT64){						
      guint64 cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnumi==cur)||sendall){
        sprintf(thistempstr,"%s:%s;%llu;",thisproperty->elname,thisproperty->propertyname,(long long unsigned int)cur);
	thisproperty->currentnumi=cur;
	int curi=cur;
	filter->nh->setParam(thistempstros, curi);
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_DOUBLE){						
      gdouble cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!(thisproperty->currentnum==cur)||sendall){
	//g_print("Resending double, because old %f, new %f\n",thisproperty->currentnum,cur);
        sprintf(thistempstr,"%s:%s;%f;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnum=cur;
	filter->nh->setParam(thistempstros, cur);
      }
    }
    else if (thisproperty->spec->value_type==G_TYPE_FLOAT){						
      gfloat cur;					
      g_object_get(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&cur,NULL);
      if (!((float)thisproperty->currentnum==cur)||sendall){
	//g_print("Resending float, because old %f, new %f\n",thisproperty->currentnum,cur);
        sprintf(thistempstr,"%s:%s;%f;",thisproperty->elname,thisproperty->propertyname,cur);
	thisproperty->currentnum=cur;
	filter->nh->setParam(thistempstros, (double)cur);
      }
    }
    else if (G_IS_PARAM_SPEC_ENUM (thisproperty->spec)){	
      gint i;
      gchar *str=NULL;
      GEnumClass *cclass = G_ENUM_CLASS (g_type_class_ref (thisproperty->spec->value_type));
      GValue value = { 0, };	
      g_value_init (&value, thisproperty->spec->value_type);
      g_object_get_property(G_OBJECT(thisproperty->elpointer),thisproperty->propertyname,&value);
      int current=g_value_get_enum(&value);
      for (i = 0; i < cclass->n_values; i++) {
          GEnumValue *value = &cclass->values[i];
	  if (i==current)str=g_strdup(value->value_nick);
        }
      if (!strcmp(str,thisproperty->charval)){
     	if (sendall)sprintf(thistempstr,"%s:%s;%s;",thisproperty->elname,thisproperty->propertyname,str);
	g_free(str);
      }
      else{
	g_free(thisproperty->charval);
	thisproperty->charval=str;
	sprintf(thistempstr,"%s:%s;%s;",thisproperty->elname,thisproperty->propertyname,str);
      }
      g_type_class_unref(cclass);
      g_value_unset(&value);
    }
    else if (!strcmp("GstPad",g_type_name (thisproperty->spec->value_type))){
      gchar *str=NULL;
      GObject* current;
      g_object_get (G_OBJECT (thisproperty->elpointer),thisproperty->propertyname,&current, NULL);
      str=gst_pad_get_pe_name((GstPad*)current);
      if (sendall || strcmp(str,thisproperty->charval)){
	sprintf(thistempstr,"%s:%s;%s;",thisproperty->elname,thisproperty->propertyname,str);
	if (thisproperty->charval)g_free(thisproperty->charval);
	thisproperty->charval=str;	
	}
      else g_free(str);
      }
    if (strlen(thistempstr))strcat(outputstring,thistempstr);
    templ=g_list_next(templ);
    g_free(thistempstr);
    g_free(thistempstros);
  }
  return outputstring;
}


/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_rosparam_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstrosparam *filter;
  GstPad *otherpad;

  filter = GST_ROSPARAM (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);
  
  GstStructure *structure = gst_caps_get_structure (caps, 0);  
  if (!gst_structure_get_int(structure,"width",&(filter->width)))filter->width=-1 ;
  if (!gst_structure_get_int(structure,"height",&(filter->height)))filter->height=-1 ;
  return gst_pad_set_caps (otherpad, caps);
}

/* chain function
 * this function does the actual processing
 */
struct property* rpluproperty(GList* list,gchar* name){
  if (!name) return NULL;
  GList* temp=list;
  struct property* result=NULL;
  while(temp){
    if(!strcmp(((struct property*)(temp->data))->alias,name))result=(struct property*)(temp->data);
    temp=g_list_next(temp);
  }
  return result;
}


static gboolean chupdatew(Gstrosparam *filter,gchar* description){
  gboolean finished=FALSE,valid=TRUE;
  gchar* temp=description;
  struct property* thisproperty;
  while(!finished&&valid){  
    char* thiscontrolname=g_malloc0(MAXSTRING);
    temp=deserialize_string(thiscontrolname,temp);
    thisproperty=rpluproperty(filter->propertylist,thiscontrolname);
    GParamSpec* spec=NULL;
    GstElement* elpointer=NULL;
    gchar* propertyname=NULL;
    //g_print("Now updating eventually %s with poiner %p\n",thiscontrolname,thisproperty);
    if (!thisproperty){//it seems to be an unknown , then maybe its an element
	elpointer=gst_bin_get_by_name_recurse_up (GST_BIN(GST_ELEMENT_PARENT(filter)),thiscontrolname);
	if (elpointer){
		propertyname=g_malloc0(MAXSTRING);
		temp=deserialize_string(propertyname,temp);
		spec=g_object_class_find_property (G_OBJECT_GET_CLASS(G_OBJECT(elpointer)),propertyname);
	}
    }
    else{
      spec=thisproperty->spec;
      elpointer=thisproperty->elpointer;
      propertyname=thisproperty->propertyname;
    }
    if (filter->silent ==FALSE)g_print("Now setting %s;%s Elpointer %p",thiscontrolname,propertyname,elpointer);
    if (!spec){//maybe in one of our spinnoffs?
      g_print("Searching for Spinnoff to handle %s\n",thiscontrolname);
      GList *temp=filter->spinoffs;
      struct spinoff* found=NULL;
      struct spinoff* current;
      while (temp){
        current=(struct spinoff*)temp->data;
        g_print("Checking Spinnoff with %s.\n",current->ui);
        if (strstr(current->ui,thiscontrolname))found=current;
        temp=g_list_next(temp);
      }
      if (found){
        GError *err = NULL;
        g_print("Found Spinnoff to handle %s, therefore sending\n",thiscontrolname);	
	g_socket_send_to(filter->ssock,found->spxpaddr,description,strlen(description),NULL,&err);
	return TRUE;
	}
    } 
    //finished=TRUE;//could not find it anywhere
    else if (spec->value_type==G_TYPE_BOOLEAN){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gboolean)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_INT){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gint)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_UINT){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(guint)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_INT64){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gint64)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_UINT64){
	gint64 cur;
	temp=deserialize_int64(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(guint64)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_DOUBLE){
	gdouble cur;
	temp=deserialize_double(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gdouble)cur,NULL);
    }
    else if (spec->value_type==G_TYPE_FLOAT){
	gdouble cur;
	temp=deserialize_double(&cur,temp);
	g_object_set(G_OBJECT(elpointer),propertyname,(gfloat)cur,NULL);
    }
    else if (G_IS_PARAM_SPEC_ENUM (spec)){
	int i;
	gchar* selection=g_malloc0(MAXSTRING);
	temp=deserialize_string(selection,temp);	
	GEnumClass *cclass = G_ENUM_CLASS (g_type_class_ref (spec->value_type));
	gint selpointer=0;
        for (i = 0; i < cclass->n_values; i++) {
          GEnumValue *value = &cclass->values[i];
	  if (!strcmp(selection,value->value_nick))selpointer=value->value;
        }
        //g_print("Updating Enum to %d\n",selpointer);
	g_object_set(G_OBJECT(elpointer),propertyname,selpointer,NULL);
	g_free(selection);
	g_type_class_unref(cclass);
    }
    else if (!strcmp("GstPad",g_type_name (thisproperty->spec->value_type))){
	gchar* selection=g_malloc0(MAXSTRING);
	temp=deserialize_string(selection,temp);//danach suchen wir
	GObject* current=NULL;
	GstIterator* padit=gst_element_iterate_pads (thisproperty->elpointer);
	gboolean done=FALSE;
	gchar* temppadn;
	gpointer item;
	while (!done) {
	  switch (gst_iterator_next (padit, (gpointer*)&item)) {
	  case GST_ITERATOR_OK:
		temppadn=gst_pad_get_pe_name((GstPad*)item);
		if (!strcmp(selection,temppadn)){
		  current=item; 
		}
		g_free(temppadn);
		gst_object_unref (item);
		break;
	  case GST_ITERATOR_RESYNC:
		gst_iterator_free (padit);
		padit=gst_element_iterate_pads (thisproperty->elpointer);
		break;
	  case GST_ITERATOR_ERROR:
		//g_print("GstIterator Returning Unexpected Error");
		done = TRUE;
		break;
	  case GST_ITERATOR_DONE:
		//g_print("GstIterator Done");
		done = TRUE;
		break;
	  default:
		//g_print("GstIterator Returning Unexpected Error"); 
		done = TRUE;
		break;
	  }
	}
	gst_iterator_free (padit);
	g_free(selection);
	if (current)g_object_set(G_OBJECT(elpointer),propertyname,current,NULL);
	else g_print("Propexport critical error, cannot find desired pad\n");
    }
    if (!thisproperty){
      if (propertyname)g_free(propertyname);//then we have malloced it
      if (elpointer)gst_object_unref(GST_OBJECT (elpointer)); 
    }
    if (!temp)finished=TRUE;   
    g_free(thiscontrolname); 
  }
  return TRUE;
}

void genbrevent(Gstrosparam * filter, gchar* desc){
//parse description
  int buttoni;
  gint64 button,startx_raw,starty_raw,endx_raw,endy_raw,width_raw,height_raw;
  gdouble startx,starty,endx,endy;
  gchar* temp=desc;
  temp=deserialize_int64(&button,temp);  
  temp=deserialize_int64(&endx_raw,temp);  
  temp=deserialize_int64(&endy_raw,temp);
  temp=deserialize_int64(&startx_raw,temp);
  temp=deserialize_int64(&starty_raw,temp);
  temp=deserialize_int64(&width_raw,temp);
  temp=deserialize_int64(&height_raw,temp);
  //calculate accurate positions..
  buttoni=(int)button;
  if (42){//not scaled, just placed in the middle
    if ((filter->width>-1)&&(width_raw>(gdouble)filter->width)) startx=(gdouble)startx_raw-(0.5*(width_raw-(gdouble)filter->width));
    else startx=(gdouble)startx_raw;
    if ((filter->height>-1)&&(height_raw>(gdouble)filter->height))starty=(gdouble)starty_raw-(0.5*(height_raw-(gdouble)filter->height));
    else starty=(gdouble)starty_raw;
    endx=startx+(endx_raw-startx_raw);
    endy=starty+(endy_raw-starty_raw);
  }
  else if (0){//scaled, changed aspect ratio
    if ((filter->width>-1)){
      startx=((gdouble)startx_raw/(gdouble)width_raw)*(gdouble)filter->width;
      endx=((gdouble)endx_raw/(gdouble)width_raw)*(gdouble)filter->width; 
      }     
    else{ 
      startx=(gdouble)startx_raw;
      endx=(gdouble)endx_raw;   
      } 
    if ((filter->height>-1)){
      starty=((gdouble)starty_raw/(gdouble)height_raw)*(gdouble)filter->height;
      endy=((gdouble)endy_raw/(gdouble)height_raw)*(gdouble)filter->height;  
      }    
    else{ 
      starty=(gdouble)starty_raw;
      endy=(gdouble)endy_raw;
      }    
  }
  else if (0){//scaled, original aspect ratio
  }
  //generate pressed event
  GstStructure *structure1 =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "mouse-button-press",
                            "pointer_x",      G_TYPE_DOUBLE, startx,
                            "pointer_y",      G_TYPE_DOUBLE, starty,
                            "button",      G_TYPE_INT, buttoni,
                             NULL);
  GstEvent* event1=gst_event_new_navigation(structure1);
  gst_pad_push_event(filter->sinkpad,event1);
  //generate released event
  GstStructure *structure2 =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "mouse-button-release",
                            "pointer_x",      G_TYPE_DOUBLE, endx,
                            "pointer_y",      G_TYPE_DOUBLE, endy,
                            "button",      G_TYPE_INT, buttoni,
                             NULL);
  GstEvent* event2=gst_event_new_navigation(structure2);
  gst_pad_push_event(filter->sinkpad,event2);
  return;
}

void genkeyevent(Gstrosparam * filter, gchar* desc){
  gint64 key;
  deserialize_int64(&key,desc);  
  //*ks=(gchar)key;
  GstStructure *structure =gst_structure_new(  "application/x-gst-navigation",
                            "event",    G_TYPE_STRING,  "key-press",
                            //"key",      G_TYPE_STRING,  ks,
			     "key",      G_TYPE_STRING,  gdk_keyval_name((uint)key),
                             NULL);
  GstEvent* event=gst_event_new_navigation(structure);
  gst_pad_push_event(filter->sinkpad,event);
}


gboolean rprcvcallback (GIOChannel *source,GIOCondition condition,gpointer data){
  if (!GST_IS_ROSPARAM(data))return FALSE;
  Gstrosparam *filter=GST_ROSPARAM(data);
  if(!filter->silent)g_print("Propexport Received something\n");
  GError *err = NULL;
  gchar* recv=g_malloc0(MAXSTRING);
  gsize bytes_read=0;
  //g_io_channel_read_chars(source,recv,MAXSTRING,&bytes_read,&err);
  GSocketAddress *recfrom;
  gboolean fromjoin=FALSE;
  bytes_read=g_socket_receive_from (filter->rsock,&recfrom,recv,MAXSTRING,NULL,&err);
  struct spinoff* found=NULL;
  if ((G_SOCKET_FAMILY_IPV4==g_socket_address_get_family(recfrom))||(G_SOCKET_FAMILY_IPV4==g_socket_address_get_family(recfrom))){
    GInetAddress *thisaddr=g_inet_socket_address_get_address((GInetSocketAddress *)recfrom);
    GList *temp=filter->spinoffs;
    struct spinoff* current;
    while (temp){
      current=(struct spinoff*)temp->data;
      if (g_inet_address_equal(current->iaddr,thisaddr))found=current;
      temp=g_list_next(temp);
    }
    if (g_inet_address_equal(filter->jiaddr,thisaddr)){
	fromjoin=TRUE;
	if(!filter->silent)g_print("Received message from the pipeline we are joining: %s\n",recv); 
    }
  }
  if ((!found)&&(recv==strstr(recv,"JOIN;"))){
    if(!filter->silent)g_print("Someone likes to join us: %s\n",recv); 
    gint64 sport;
    deserialize_int64 (&sport,(recv+5));
    struct controller* newcontroller;
    newcontroller=(struct controller*)g_malloc0(sizeof(struct controller));
     
    newcontroller->saddr=g_inet_socket_address_new(g_inet_socket_address_get_address((GInetSocketAddress *)recfrom),(guint16)sport);
    filter->firsttime=TRUE;
    filter->controllers=g_list_append(filter->controllers,newcontroller);
  }  
  
  g_object_unref(recfrom);
  if(!filter->silent)g_print("Received %d bytes: %s\n",(int)bytes_read,recv);  
  if ((found)&&(recv==strstr(recv,"UI;"))){
    g_print("Received UI description and found pointer number of spinnoff %p\n",found); 
	strcpy(found->ui,(recv+3));
	filter->firsttime=TRUE; 
  }
  else if ((!found)&&(recv==strstr(recv,"UI;"))){//then its from joinui
 	strcpy(filter->joinui,(recv+3));
	filter->firsttime=TRUE;    
  }
  //else if (fromjoin){
  //  g_socket_send_to(filter->ssock,filter->saddr,recv,strlen(recv),NULL,&err);
  //}
  else if ((fromjoin)||(found)){//was anderes empfangen, einfach mal weiterleiten:)
    int numsent;
    GList *temp=filter->controllers;
    struct controller* curcont=NULL;
    while(temp){
      curcont=(struct controller*)temp->data;
      numsent=g_socket_send_to(filter->ssock,curcont->saddr,recv,strlen(recv),NULL,&err);
      if (numsent!=strlen(recv))g_print ("Error sending !\n");
      temp=g_list_next(temp);      
    }
    //numsent=g_socket_send_to(filter->ssock,filter->saddr,recv,strlen(recv),NULL,&err);
    //if (numsent!=strlen(recv))g_print ("Forwarding in callback: Error sending !\n");
  }

  else if (recv==strstr(recv,"EVENTBR"))genbrevent(filter,recv);
  else if (recv==strstr(recv,"EVENTK"))genkeyevent(filter,recv);
  else {
	chupdatew(filter,recv);
	if (filter->jaddr)g_socket_send_to(filter->ssock,filter->jaddr,recv,bytes_read,NULL,&err);  
  }
  /*if (!varsp->saddr){
    varsp->saddr=recfrom;
    gint checkport=0;
    g_print("Will Set port: %d\n",checkport);
    g_object_set(G_OBJECT(varsp->saddr),"port",varsp->sport,NULL);
    g_print("Will Checkport: %d\n",checkport);
    g_object_get(G_OBJECT(varsp->saddr),"port",&checkport,NULL);
    g_print("Checkport: %d\n",checkport);
    g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
  }*/
  //if brevent then foreward
  if ((recv==strstr(recv,"EVENTBR"))||(recv==strstr(recv,"EVENTK"))){
    GList *temp=filter->spinoffs;
    struct spinoff* current;
    while (temp){
      current=(struct spinoff*)temp->data;
      g_print("Sending Event %s to Spinnoff with %s.\n",recv,current->ui);
      g_socket_send_to(filter->ssock,current->spxpaddr,recv,bytes_read,NULL,&err);
      temp=g_list_next(temp);
    }
    if (filter->jaddr)g_socket_send_to(filter->ssock,filter->jaddr,recv,bytes_read,NULL,&err);   
  }
  g_free(recv);
  return TRUE;
}

void msgtoall(Gstrosparam *filter, char* msg){
  GList *temp=filter->spinoffs;
  struct spinoff* current;
  while (temp){
    current=(struct spinoff*)temp->data;
    GError* err;
    int numsent=0;
    numsent=g_socket_send_to(filter->ssock,current->saddr,msg,strlen(msg),NULL,&err);
    if (numsent!=strlen(msg))g_print("Propexport, spinoff Error sending !\n");
    if (!strcmp(msg,"PLAY"))current->state=7;
    temp=g_list_next(temp);
  }
}
void genspinoffs(Gstrosparam *filter){
  gboolean finished=FALSE;
  char filename[1023];
  char hostname[1023];
  gint64 port=0;
  char* pipeline=g_malloc0(10000);
  char* tosend=g_malloc0(10100);  
  char* tempchar=filter->conffiles;
  char* tempchar2=NULL;
  GError* err=NULL;
  struct spinoff* newspinnoff;
  while (!finished && (tempchar!=NULL) &&  (strlen(tempchar)>0)){
     pipeline[0]=0;
     tosend[0]=0;
     port=0;//reset everything
     if (err)g_error_free(err);
     tempchar=deserialize_string2 (filename,tempchar,';');
     g_print("Propexport reading spinnoff file: %s\n",filename); 
     gboolean isfile=FALSE;
     gchar* content=NULL;
     isfile=g_file_get_contents(filename,&content,NULL,NULL);
     if (!isfile){//send the whole stuff to one adress
       finished=TRUE;
       content=g_malloc0(strlen(filename)+1);
       strcpy(content,filename);
     }
     //now, in content
     tempchar2=content;
     tempchar2=deserialize_string2 (hostname,tempchar2,':');
     tempchar2=deserialize_int64 (&port,tempchar2);
     //tempchar2=deserialize_string (pipeline,tempchar2);//eigentlich nur zum wegschneiden des Semikolons
     while ((tempchar2!=NULL) && !isalpha(*tempchar2)&&(*tempchar2!=0)) tempchar2++;//segfault ;)
     if ((tempchar2)&&(*tempchar2!=0))strcpy(pipeline,tempchar2);
     g_print("Propexport parsing results: Host:%s Port:%d Pipeline:%s\n",hostname,(int)port,pipeline); 
     newspinnoff=g_malloc0(sizeof(struct spinoff));
     //setting up Socket Adress
     GResolver *wozuderscheiss=g_resolver_get_default();
     if (hostname[0]=='$'){
	char* hostnametemp=g_malloc(strlen(hostname));
	strcpy(hostnametemp,hostname);
	strcpy(hostname, getenv(hostnametemp));
	g_free(hostnametemp);
     }
     if (isalpha(hostname[0])){
   
       GList* result=NULL;
       result=g_resolver_lookup_by_name(wozuderscheiss,hostname,NULL,&err);
       if (err)g_error_free(err);
       if(result){
        newspinnoff->saddr=g_inet_socket_address_new((GInetAddress*)result->data,(int)port);
        g_object_ref((GInetAddress*)result->data);
	newspinnoff->iaddr=(GInetAddress*)result->data;
        gchar* lookedup=g_resolver_lookup_by_address(wozuderscheiss,
                                                         (GInetAddress*)result->data,
                                                         NULL,
                                                         &err);
        if (err)g_error_free(err);
        if (lookedup)strcpy(newspinnoff->name,lookedup);
	}
       g_resolver_free_addresses(result);
     }
     else{
       GInetAddress* inaddr=g_inet_address_new_from_string(hostname);
       newspinnoff->iaddr=inaddr;
       newspinnoff->saddr=g_inet_socket_address_new(inaddr,(int)port);
       gchar* lookedup=g_resolver_lookup_by_address(wozuderscheiss,
                                                         inaddr,
                                                         NULL,
                                                         &err);
       strcpy(newspinnoff->name,lookedup);
     }
     g_object_unref(wozuderscheiss);
     g_print("Looked up Hostname %s\n",newspinnoff->name);
     //now we send these things
     if (newspinnoff->saddr){
       gint64 pxpport=4445;
       gchar *parseport=strstr(pipeline,"rcvport");
       if (parseport){
	 deserialize_int64 (&pxpport,parseport);
       }
       newspinnoff->spxpaddr=g_inet_socket_address_new(newspinnoff->iaddr,(int)pxpport);
       sprintf(tosend,"PIPELINE;%s",pipeline);
       if (strlen(tosend)){
         int numsent=0;
         numsent=g_socket_send_to(filter->ssock,newspinnoff->saddr,tosend,strlen(tosend),NULL,&err);
         if (numsent!=strlen(tosend))g_print("Propexport, spinoff Error sending !\n");
       }
       filter->spinoffs=g_list_append(filter->spinoffs,newspinnoff);
     }
     else{
       g_free(newspinnoff);
       finished=TRUE;	
     }
     g_free(content);	
  }
  g_free(pipeline);
  g_free(tosend);	
  return;
}


void setupsockets(Gstrosparam *filter){
    GError *err = NULL;
    filter->rsock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    g_assert(err == NULL);
    filter->raddr=g_inet_socket_address_new(g_inet_address_new_any(G_SOCKET_FAMILY_IPV6),filter->rport);
    g_socket_bind(filter->rsock,G_SOCKET_ADDRESS(filter->raddr),TRUE,&err);
    g_assert(err == NULL);

    int fd = g_socket_get_fd(filter->rsock);
    filter->rchannel = g_io_channel_unix_new(fd);
    filter->source = g_io_add_watch(filter->rchannel, G_IO_IN,(GIOFunc) rprcvcallback, filter);
    g_print("Set up source %d\n",filter->source);
    g_io_channel_unref(filter->rchannel);




    //now set up send socket...
    filter->ssock = g_socket_new(G_SOCKET_FAMILY_IPV6,
	                    G_SOCKET_TYPE_DATAGRAM,
	                    G_SOCKET_PROTOCOL_UDP,
	                    &err);
    struct controller* newcontroller;
    newcontroller=(struct controller*)g_malloc0(sizeof(struct controller));
     
    newcontroller->saddr=g_inet_socket_address_new(g_inet_address_new_from_string(filter->remoteadress),filter->sport);
    filter->controllers=g_list_append(filter->controllers,newcontroller);
    //filter->saddr=NULL;
    //filter->saddr=g_inet_socket_address_new(g_inet_address_new_from_string(filter->remoteadress),filter->sport);
}


static GstFlowReturn
gst_rosparam_chain (GstPad * pad, GstBuffer * buf)
{
  if (!ros::ok()) return GST_FLOW_UNEXPECTED;
  Gstrosparam *filter;
  GError *err = NULL;
  filter = GST_ROSPARAM (GST_OBJECT_PARENT (pad));
  gchar* tosend=NULL;
  if (filter->sendj){
    tosend=g_malloc0(128);
    sprintf(tosend,"JOIN;%d",filter->rport);
    GResolver *wozuderscheiss=g_resolver_get_default(); 
    gchar* hostname=filter->jadress; 
    //struct spinoff* newspinnoff;
    //newspinnoff=g_malloc0(sizeof(struct spinoff));
    if (isalpha(hostname[0])){  
       GList* result=NULL;
       result=g_resolver_lookup_by_name(wozuderscheiss,hostname,NULL,&err);
       if (err)g_error_free(err);
       if(result){
        g_object_ref((GInetAddress*)result->data);
        filter->jiaddr=(GInetAddress*)result->data;
        filter->jaddr=g_inet_socket_address_new((GInetAddress*)result->data,(int)filter->jport);
        if (err)g_error_free(err);
        //if (lookedup)strcpy(newspinnoff->name,lookedup);
	}
       g_resolver_free_addresses(result);
     }
     else{
       filter->jiaddr=g_inet_address_new_from_string(hostname);
       filter->jaddr=g_inet_socket_address_new(filter->jiaddr,(int)filter->jport);
     }
     g_object_unref(wozuderscheiss);
     if (filter->jaddr)g_socket_send_to(filter->ssock,filter->jaddr,tosend,strlen(tosend),NULL,&err);
     g_free(tosend);
     filter->sendj=FALSE;
  }
  if (filter->firsttime){
    tosend=g_malloc0(MAXSTRING);
    //setupsockets(Gstrosparam *filter);
    //g_socket_connect(varsp->ssock, varsp->saddr, NULL, NULL);
    //g_socket_bind(filter->ssock,G_SOCKET_ADDRESS(filter->saddr),TRUE,&err);
    if (!filter->uidesc)filter->uidesc=setuppropertylist(filter);
    strcpy(tosend,filter->uidesc);
    //set up other 
    GList *temp=filter->spinoffs;
    struct spinoff* current;
    while (temp){
      current=(struct spinoff*)temp->data;
      if (current->ui)strcat(tosend,current->ui);
      temp=g_list_next(temp);
    }
    if (strlen(filter->joinui)>0)strcat(tosend,filter->joinui);
  } 

  //Generate and send property
  else tosend=chupdate(filter,FALSE);
  if (strlen(tosend)){
    GError* err;
    int numsent=0;
    g_print ("List_length: %d Tosend: %s\n",g_list_length(filter->propertylist),tosend);

    GList *temp=filter->controllers;
    struct controller* curcont=NULL;
    while(temp){
      curcont=(struct controller*)temp->data;
      numsent=g_socket_send_to(filter->ssock,curcont->saddr,tosend,strlen(tosend),NULL,&err);
      if (numsent!=strlen(tosend))g_print ("Error sending !\n");
      temp=g_list_next(temp);
    }
  }
  if (filter->firsttime){//initialize ROS parameters
    chupdate(filter,TRUE);
    filter->firsttime=FALSE;
  }	
  g_free(tosend);

  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
  filter->firsttime=FALSE;
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, buf);
}

static GstStateChangeReturn
gst_rosparam_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstrosparam *filter;
  GError* err;
  filter =  GST_ROSPARAM(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:{
      if (!filter->socketopen)setupsockets(filter);

      filter->socketopen=TRUE;
      int schrott=0;
      ros::init(schrott, NULL, "rossink", ros::init_options::AnonymousName);
      filter->nh = new ros::NodeHandle();
      break;}
    default:
      break;
  }

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      msgtoall(filter,"STOP;");
      //gst_my_filter_free_memory (filter);
      g_print ("Propexport Dropping Sockets.\n");
      if (filter->source) g_source_remove (filter->source);
      filter->source=0;
      if (filter->rsock)g_socket_close(filter->rsock,&err);
      filter->rsock=0;
      if (filter->ssock)g_socket_close(filter->ssock,&err);
      filter->ssock=0;
      filter->firsttime=TRUE;
      g_free(filter->uidesc);
      filter->uidesc=NULL;
      filter->socketopen=FALSE;

      ros::shutdown();
      //g_io_channel_close (filter->rchannel);
      //g_io_channel_unref(filter->rchannel);
      break;
    default:
      break;
  }
  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

 return ret;
}




/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
/*
static gboolean
rosparam_init (GstPlugin * rosparam)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template rosparam' with your description

  GST_DEBUG_CATEGORY_INIT (gst_rosparam_debug, "rosparam",
      0, "Template rosparam");

  return gst_element_register (rosparam, "rosparam", GST_RANK_NONE,
      GST_TYPE_ROSPARAM);
}
   */
/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstrosparam"
#endif

/* gstreamer looks for this structure to register rosparams
 *
 * exchange the string 'Template rosparam' with your rosparam description

GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "rosparam",
    "Template rosparam",
    rosparam_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
) */
