/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-rossift3d
 *
 * FIXME:Describe rossift3d here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v rossift3d trigger1=1 name=sift3d ! pip trigger=0 trigger-small=0 sift=1 name=pip siftoffx=640 \
v4l2src device=/dev/video0 ! tee name=v0 ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! siftextractor ! sift3d.sink1 \
v4l2src device=/dev/video1 ! tee name=v1 ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! siftextractor ! sift3d.sink2 \
v0. ! ffmpegcolorspace ! videobox right=-640 ! ffmpegcolorspace ! pip.sink \
v1. ! ffmpegcolorspace ! pip.small \
pip. ! ffmpegcolorspace ! ximagesink sync=0 async=0 
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif



//#include "sift.h"
//#include "imgfeatures.h"
//#include "kdtree.h"
//#include "utils.h"
//#include "xform.h"
#include "math.h"

#include <sys/time.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>


#include <gst/gst.h>

#include <gsl/gsl_sf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

#include "cvsift.h"
#include "gstrossift3d.h"

#ifndef NAN
#define NAN (0.0/0.0)
#endif
GST_DEBUG_CATEGORY_STATIC (gst_rossift3d_debug);
#define GST_CAT_DEFAULT gst_rossift3d_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TRIGGER1,
  PROP_TRIGGER2,
  PROP_OPMODE,
  PROP_OFFX,
  PROP_OFFY,
  PROP_MAXSIM,
  PROP_MAXSIM2,
  PROP_GOODDIST,
  PROP_THRESHOLD,
  PROP_REMATCHDOUBLE,
  PROP_SHOWINLIERS,
  PROP_MODE,
  PROP_LOCK  
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("sift/raw,"
        "framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];")
    );

/*
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );
*/
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "coordinates-list;"
      "matrix-list;"
	"sift/raw,"
        //"framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];")
    );
GST_BOILERPLATE (Gstrossift3d, gst_rossift3d, GstElement,
    GST_TYPE_ELEMENT);

static void gst_rossift3d_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_rossift3d_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static gboolean gst_rossift3d_event (GstPad  *pad, GstEvent * event);
static gboolean gst_rossift3d_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_rossift3d_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn
gst_rossift3d_change_state (GstElement *element, GstStateChange transition);
static GstCaps *
gst_sift_extractor_get_src_caps (GstPad *pad);
static GstCaps *
gst_sift_extractor_get_sink_caps (GstPad *pad);


#define GST_SIFTTOOL_TYPE_OP_MODE (gst_type_sifttool_op_mode ())


//GST-Stuff
static GType
gst_type_sifttool_op_mode (void)
{
  static GType sifttool_op_mode = 0;
  static const GEnumValue functions[] = {
    {MATCH, "matches two sets of features", "match"},
    {MERGE, "merges two sets of features, set offset values!", "merge"},
    {TRANSFORMPOINT,"transforms the point given by the offset values from the small image(2) to the big(1)!", "transformpoint"},
    {TRANSFORM,"gives the transformation equation", "transform"}, 
    {0, NULL, NULL}
  };

  if (!sifttool_op_mode) {
    sifttool_op_mode =
        g_enum_register_static ("SiftToolOpMode", functions);
  }
  return sifttool_op_mode;
}


/* GObject vmethod implementations */

static void
gst_rossift3d_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "rossift3d",
    "Gst-Plugins-Robotix",
    "Match Sift-Features",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the rossift3d's class */
static void
gst_rossift3d_class_init (Gstrossift3dClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_rossift3d_set_property;
  gobject_class->get_property = gst_rossift3d_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER1,
      g_param_spec_boolean ("trigger1", "Trigger1", "Trigger on new data on Input 1 ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TRIGGER2,
      g_param_spec_boolean ("trigger2", "Trigger2", "Trigger on new data on Input 2 ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OPMODE,
      g_param_spec_enum ("opmode", "Opmode",
          "What should be done, match(default) or merge, or calculate transformation ?", GST_SIFTTOOL_TYPE_OP_MODE,
          MATCH, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OFFX,
      g_param_spec_int ("overlapx", "Overlapx", "X overlapping of second compared to first image (-1 is full)",
          -1,4000,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_OFFY,
      g_param_spec_int ("overlapy", "Overlapy", "Y overlapping of second compared to first image (-1 is full)",
          -1,4000,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM,
      g_param_spec_double ("maxsim", "Maxsim", "The second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.49, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM2,
      g_param_spec_double ("maxsim2", "Maxsim2", "In case of a really good match (distance<gooddist), the second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.9, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_GOODDIST,
      g_param_spec_double ("gooddist", "Gooddist", "What is a good match, is square distance < than this value",
          0,1,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_THRESHOLD,
      g_param_spec_double ("threshold", "Threshold", "Threshold for Ransac",
          0,1,0.01, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_REMATCHDOUBLE,
      g_param_spec_boolean ("rematchdouble", "Rematchdouble", "Whether to kill double matches and rematch them in one direction.",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SHOWINLIERS,
      g_param_spec_boolean ("showinliers", "Showinliers", "Show only valid matches after Ransac.",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MODE,
      g_param_spec_int ("mode", "Mode", "0=raw, 1=lowe, 2=oxford 3=lowelight",
          0,3,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_REMATCHDOUBLE,
      g_param_spec_boolean ("lock", "Lock", "A lock that disables updating the feature vectors (in order to try different parameters for detection).",
          FALSE, G_PARAM_READWRITE));
  gstelement_class->change_state = gst_rossift3d_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
struct ThreeDmatch* gennewmatch(Gstrossift3d * filter,struct feature* feat1, struct feature* feat2, int rank){
  g_print("Checking 3d match x1 %f y1 %f x2 %f y2 %f Rank %d\n",feat1->x,feat1->y,feat2->x,feat2->y,rank);
  g_print("Width 1 %d Height1 %d Width2 %d Height2 %d\n",filter->width1,filter->height1,filter->width2,filter->height2);
  struct ThreeDmatch* newmatch=(struct ThreeDmatch*)g_malloc0(sizeof(struct ThreeDmatch));
  



  newmatch->feat1=feat1;
  newmatch->feat2=feat2;
  newmatch->rank=rank;

  feat1->x=((filter->width1-1)/2.0)+0;
  feat2->x=((filter->width2-1)/2.0)-10;
  feat1->y=(filter->height1-1)/2.0;
  feat2->y=(filter->height1-1)/2.0;
  CvMat* v1  = cvCreateMat(3,1,CV_32FC1);
  CV_MAT_ELEM(*v1, float,0,0) = (feat1->x-((filter->width1-1)/2))/filter->fx1;
  CV_MAT_ELEM(*v1, float,1,0) = (feat1->y-((filter->height1-1)/2))/filter->fy1;
  CV_MAT_ELEM(*v1, float,2,0) = 1;//normalized as the upper values got devided by f
  g_print("We found v1 with x %f y %f z %f\n",CV_MAT_ELEM(*v1, float,0,0),CV_MAT_ELEM(*v1, float,1,0),CV_MAT_ELEM(*v1, float,2,0));
  CvMat* v2pre  = cvCreateMat(3,1,CV_32FC1);
  CV_MAT_ELEM(*v2pre, float,0,0) = (feat2->x-((filter->width2-1)/2))/filter->fx2;
  CV_MAT_ELEM(*v2pre, float,1,0) = (feat2->y-((filter->height2-1)/2))/filter->fy2;
  CV_MAT_ELEM(*v2pre, float,2,0) = 1;//normalized as the upper values got devided by f
  g_print("We found v2pre with x %f y %f z %f\n",CV_MAT_ELEM(*v2pre, float,0,0),CV_MAT_ELEM(*v2pre, float,1,0),CV_MAT_ELEM(*v2pre, float,2,0));
  CvMat* v2= cvCreateMat(3,1,CV_32FC1);
  //v2=filter->R*v2pre;
  cvMatMul(filter->R,v2pre,v2);
  g_print("We found v2 with x %f y %f z %f\n",CV_MAT_ELEM(*v2, float,0,0),CV_MAT_ELEM(*v2, float,1,0),CV_MAT_ELEM(*v2, float,2,0));
  CvMat* v1y=cvCreateMat(3,1,CV_32FC1);
  cvCrossProduct(v1,v2,v1y);//the vector that is perpendiculat to v1 and v2
g_print("We found v1y with x %f y %f z %f\n",CV_MAT_ELEM(*v1y, float,0,0),CV_MAT_ELEM(*v1y, float,1,0),CV_MAT_ELEM(*v1y, float,2,0));
  //now we have two vectors defining the plane, but we want a 
  //Ax + By + Cz + D =0 
  //representation, therefore we get the normal vector of the plane n
  CvMat* n1=cvCreateMat(3,1,CV_32FC1);
  cvCrossProduct(v1,v1y,n1);  
  //n_x*(x-x_0)+  n_y*(y-y_0) +  n_z*(z-z_0) = 0 
  //as x_0,y_0,z_0 is 0,0,0
  //A is n1x , B is n1y , C is n1z D is 0
  //now we have a vector (g*v2_x+t_x,g*v2_y+t_y,g*v2_z+t_z)
  //that intersects with the plane
  double g=(-CV_MAT_ELEM(*n1, float,0,0)*CV_MAT_ELEM(*filter->T, float,0,0)-CV_MAT_ELEM(*n1, float,1,0)*CV_MAT_ELEM(*filter->T, float,1,0)-CV_MAT_ELEM(*n1, float,2,0)*CV_MAT_ELEM(*filter->T, float,2,0))/(CV_MAT_ELEM(*n1, float,0,0)*CV_MAT_ELEM(*v2, float,0,0)+CV_MAT_ELEM(*n1, float,1,0)*CV_MAT_ELEM(*v2, float,1,0)+CV_MAT_ELEM(*n1, float,2,0)*CV_MAT_ELEM(*v2, float,2,0));
  CvMat* p2= cvCreateMat(3,1,CV_32FC1); 
  CV_MAT_ELEM(*p2, float,0,0) = g * CV_MAT_ELEM(*v2, float,0,0) + CV_MAT_ELEM(*filter->T, float,0,0);
  CV_MAT_ELEM(*p2, float,1,0) = g * CV_MAT_ELEM(*v2, float,1,0) + CV_MAT_ELEM(*filter->T, float,1,0);
  CV_MAT_ELEM(*p2, float,2,0) = g * CV_MAT_ELEM(*v2, float,2,0) + CV_MAT_ELEM(*filter->T, float,2,0);
  g_print("We found p2 with x %f y %f z %f\n",CV_MAT_ELEM(*p2, float,0,0),CV_MAT_ELEM(*p2, float,1,0),CV_MAT_ELEM(*p2, float,2,0));
  //now we want to express the point p2 in terms of the coordinates v1 and v1y, defining a and b 
  double a,b;
  b=(CV_MAT_ELEM(*p2, float,1,0)-((CV_MAT_ELEM(*v1, float,1,0)*CV_MAT_ELEM(*p2, float,0,0))/CV_MAT_ELEM(*v1, float,0,0)))/(CV_MAT_ELEM(*v1y, float,1,0)-((CV_MAT_ELEM(*v1y, float,0,0)*CV_MAT_ELEM(*v1, float,1,0))/(CV_MAT_ELEM(*v1, float,0,0))));
  a=(CV_MAT_ELEM(*p2, float,0,0)-(b*CV_MAT_ELEM(*v1y, float,0,0)))/(CV_MAT_ELEM(*v1, float,0,0));
  //now we can define point p1 that is on the one axis of the plane
  CvMat* p1  = cvCreateMat(3,1,CV_32FC1);
  CV_MAT_ELEM(*p1, float,0,0) = a*CV_MAT_ELEM(*v1, float,0,0);
  CV_MAT_ELEM(*p1, float,1,0) = a*CV_MAT_ELEM(*v1, float,1,0);
  CV_MAT_ELEM(*p1, float,2,0) = a*CV_MAT_ELEM(*v1, float,2,0);
  g_print("We found p1 with x %f y %f z %f\n",CV_MAT_ELEM(*p1, float,0,0),CV_MAT_ELEM(*p1, float,1,0),CV_MAT_ELEM(*p1, float,2,0));
  //we can calculate the final point pf
  CvMat* pf  = cvCreateMat(3,1,CV_32FC1); 
  //cv::Mat pf(3, 1, DataType<float>::type);
  CV_MAT_ELEM(*pf, float,0,0) = 0.5*(CV_MAT_ELEM(*p1, float,0,0)+CV_MAT_ELEM(*p2, float,0,0));
  CV_MAT_ELEM(*pf, float,1,0) = 0.5*(CV_MAT_ELEM(*p1, float,1,0)+CV_MAT_ELEM(*p2, float,1,0));
  CV_MAT_ELEM(*pf, float,2,0) = 0.5*(CV_MAT_ELEM(*p1, float,2,0)+CV_MAT_ELEM(*p2, float,2,0));
  g_print("We found pf with x %f y %f z %f\n",CV_MAT_ELEM(*pf, float,0,0),CV_MAT_ELEM(*pf, float,1,0),CV_MAT_ELEM(*pf, float,2,0));
  gdouble deviation=sqrt((CV_MAT_ELEM(*p1, float,0,0)-CV_MAT_ELEM(*p2, float,0,0))*(CV_MAT_ELEM(*p1, float,0,0)-CV_MAT_ELEM(*p2, float,0,0))+(CV_MAT_ELEM(*p1, float,1,0)-CV_MAT_ELEM(*p2, float,1,0))*(CV_MAT_ELEM(*p1, float,1,0)-CV_MAT_ELEM(*p2, float,1,0))+(CV_MAT_ELEM(*p1, float,2,0)-CV_MAT_ELEM(*p2, float,2,0))*(CV_MAT_ELEM(*p1, float,2,0)-CV_MAT_ELEM(*p2, float,2,0)));
  g_print("We found a deviation of %f, therefore the match is valid or invalid\n",deviation);
  g_free(newmatch);
  cvReleaseMat( &v1 );
  cvReleaseMat( &v2pre );
  cvReleaseMat( &v2 );
  cvReleaseMat( &v1y );
  cvReleaseMat( &n1 );
  cvReleaseMat( &p1 );
  cvReleaseMat( &p2 );
  cvReleaseMat( &pf );

  return newmatch;
}

static void
gst_rossift3d_init (Gstrossift3d * filter,
    Gstrossift3dClass * gclass)
{
  filter->sinkpad1 = gst_pad_new_from_static_template (&sink_factory, "sink1");
  gst_pad_set_setcaps_function (filter->sinkpad1,
                                GST_DEBUG_FUNCPTR(gst_rossift3d_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad1,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_sink_caps));
  gst_pad_set_chain_function (filter->sinkpad1,
                              GST_DEBUG_FUNCPTR(gst_rossift3d_chain));
  gst_pad_set_event_function (filter->sinkpad1,GST_DEBUG_FUNCPTR(gst_rossift3d_event));

  filter->sinkpad2 = gst_pad_new_from_static_template (&sink_factory, "sink2");
  gst_pad_set_setcaps_function (filter->sinkpad2,
                                GST_DEBUG_FUNCPTR(gst_rossift3d_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad2,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_sink_caps));
  gst_pad_set_chain_function (filter->sinkpad2,
                              GST_DEBUG_FUNCPTR(gst_rossift3d_chain));
  gst_pad_set_event_function (filter->sinkpad2,GST_DEBUG_FUNCPTR(gst_rossift3d_event));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_src_caps));
  gst_pad_set_setcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_rossift3d_set_caps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad1);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad2);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->offsetx=-1;
  filter->offsety=-1;
  filter->mutex=g_mutex_new();
  filter->maxsim=0.49;
  filter->maxsim2=0.9;
  filter->threshold=0.01;
  filter->showinliers=FALSE;
  filter->mode=0;
  filter->lock=0;
  filter->maxmatches=5;
  filter->fx1=538;
  filter->fy1=538;
  filter->fx2=538;
  filter->fy2=538;
  filter->T=cvCreateMat(3,1,CV_32FC1);
  cvSet(filter->T, cvScalar(0));
  CV_MAT_ELEM(*filter->T, float,0,0) = 0.08;//two sony cameras
  filter->R=cvCreateMat(3,3,CV_32FC1);
  cvSet(filter->R, cvScalar(0));
  CV_MAT_ELEM(*filter->R, float,0,0) = 1;
  CV_MAT_ELEM(*filter->R, float,1,1) = 1;
  CV_MAT_ELEM(*filter->R, float,2,2) = 1;
}

static void
gst_rossift3d_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstrossift3d *filter = GST_ROSSIFT3D (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER1:
      filter->trigger1 = g_value_get_boolean (value);
      break;
    case PROP_TRIGGER2:
      filter->trigger2 = g_value_get_boolean (value);
      break;
    case PROP_OPMODE:
      filter->opmode=(SiftToolOpMode)g_value_get_enum(value);
      break;
    case PROP_OFFX:
      filter->offsetx=g_value_get_int (value);
      break;
    case PROP_OFFY:
      filter->offsety=g_value_get_int (value);
      break;
    case PROP_MAXSIM:
      filter->maxsim=g_value_get_double (value);
      break;
    case PROP_MAXSIM2:
      filter->maxsim2=g_value_get_double (value);
      break;
    case PROP_GOODDIST:
      filter->gooddist=g_value_get_double (value);
      break;
    case PROP_THRESHOLD:
      filter->threshold=g_value_get_double (value);
      break;
   case PROP_REMATCHDOUBLE:
      filter->rematchdouble = g_value_get_boolean (value);
      break;
   case PROP_SHOWINLIERS:
      filter->showinliers = g_value_get_boolean (value);
      break;
   case PROP_MODE:
      filter->mode = g_value_get_int (value);
      break;
   case PROP_LOCK:
      filter->lock = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_rossift3d_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstrossift3d *filter = GST_ROSSIFT3D (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TRIGGER1:
      g_value_set_boolean (value, filter->trigger1);
      break;
    case PROP_TRIGGER2:
      g_value_set_boolean (value, filter->trigger2);
      break;
    case PROP_OPMODE:
      g_value_set_enum(value, filter->opmode);
      break;
    case PROP_OFFX:
      g_value_set_int(value, filter->offsetx);
      break;
    case PROP_OFFY:
      g_value_set_int(value, filter->offsety);
      break;
    case PROP_MAXSIM:
      g_value_set_double(value, filter->maxsim);
      break;
    case PROP_MAXSIM2:
      g_value_set_double(value, filter->maxsim2);
      break;
    case PROP_GOODDIST:
      g_value_set_double(value, filter->gooddist);
      break;
    case PROP_THRESHOLD:
      g_value_set_double(value, filter->threshold);
      break;
    case PROP_REMATCHDOUBLE:
      g_value_set_boolean (value, filter->rematchdouble);
      break;
   case PROP_SHOWINLIERS:
      g_value_set_boolean (value, filter->showinliers);
      break;
   case PROP_MODE:
      g_value_set_int (value, filter->mode);
      break;
   case PROP_LOCK:
      g_value_set_boolean (value, filter->lock);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}
//

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_rossift3d_set_caps (GstPad * pad, GstCaps * caps)
{
  g_print("Sifttools set caps with %s!!\n",gst_caps_to_string (caps));
  Gstrossift3d *filter;
  GstPad *otherpad;

  filter = GST_ROSSIFT3D (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad1 : filter->srcpad;
  gst_object_unref (filter);
  gboolean ret=TRUE;

  if (pad == filter->sinkpad1){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     mime = gst_structure_get_name (structure);
     if (strcmp (mime, "sift/raw") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));
	filter->modes1=0;
     }
     else if (strcmp (mime, "sift/lowe") == 0) filter->modes1=1;
     else if (strcmp (mime, "sift/oxf") == 0) filter->modes1=2;
     else if (strcmp (mime, "sift/lowelight") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));	
	filter->modes1=3;
     }
     else ret=FALSE;
  }
  if (pad == filter->sinkpad2){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     mime = gst_structure_get_name (structure);
     if (strcmp (mime, "sift/raw") == 0) {
	gst_structure_get_int(structure,"width",&(filter->width2));
	gst_structure_get_int(structure,"height",&(filter->height2));
	gst_structure_get_int(structure,"overlap_v",&(filter->offsetx));
	gst_structure_get_int(structure,"overlap_h",&(filter->offsety));
	filter->modes2=0;
     }
     else if (strcmp (mime, "sift/lowe") == 0) filter->modes2=1;
     else if (strcmp (mime, "sift/oxf") == 0) filter->modes2=2;
     else if (strcmp (mime, "sift/lowelight") == 0){
	gst_structure_get_int(structure,"width",&(filter->width2));
	gst_structure_get_int(structure,"height",&(filter->height2));
	gst_structure_get_int(structure,"overlap_v",&(filter->offsetx));
	gst_structure_get_int(structure,"overlap_h",&(filter->offsety));
	filter->modes2=3;
     }
     else ret=FALSE;     
  }
  else ret=TRUE;
  //return gst_pad_set_caps (otherpad, caps);
  return ret;
}



static GstCaps *
gst_sift_extractor_get_sink_caps (GstPad *pad){
  g_print("Start get sink caps!!\n");
  Gstrossift3d *filter;

  filter = GST_ROSSIFT3D (GST_OBJECT_PARENT (pad));
  gint mode;
  if (pad== filter->sinkpad1) mode=filter->modes1;
  else mode=filter->modes2;
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  

}

static GstCaps *
gst_sift_extractor_get_src_caps (GstPad *pad){
  g_print("Start get src caps!!\n");
  Gstrossift3d *filter;

  filter = GST_ROSSIFT3D (GST_OBJECT_PARENT (pad));



  GstCaps* outcaps;
  switch(filter->opmode){
  case MATCH:
    outcaps=gst_caps_new_simple ("coordinates-list",
		NULL);
    break;
  case MERGE:
    if (filter->mode==0){
      if (filter->width1==0) outcaps=gst_caps_new_simple ("sift/raw",
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"overlap_v", GST_TYPE_INT_RANGE, -1, G_MAXINT,
      		"overlap_h", GST_TYPE_INT_RANGE, -1, G_MAXINT,
		NULL);

      else outcaps=gst_caps_new_simple ("sift/raw",
      		"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
      		"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
      		"overlap_v", G_TYPE_INT, filter->offsetx,
      		"overlap_h", G_TYPE_INT, filter->offsety,
		NULL);
    }
    else if (filter->mode==3){
      if (filter->width1==0) outcaps=gst_caps_new_simple ("sift/lowelight",
      		"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		"overlap_v", GST_TYPE_INT_RANGE, -1, G_MAXINT,
      		"overlap_h", GST_TYPE_INT_RANGE, -1, G_MAXINT,
		NULL);

      else outcaps=gst_caps_new_simple ("sift/lowelight",
      		"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
      		"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
      		"overlap_v", G_TYPE_INT, filter->offsetx,
      		"overlap_h", G_TYPE_INT, filter->offsety,
		NULL);
    }   
    break;
  case TRANSFORM:
    outcaps=gst_caps_new_simple ("matrix-list",
		NULL);
    break;
  case TRANSFORMPOINT:
    outcaps=gst_caps_new_simple ("coordinates-list",
		NULL);
    break; 
  default:
    outcaps=NULL;
  }


  //outcaps=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  g_print("Returning %s!!\n",gst_caps_to_string(outcaps));
  return outcaps;
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_rossift3d_chain (GstPad * pad, GstBuffer * buf)
{
  g_print("Start matcher!!\n");
  Gstrossift3d *filter;

  filter = GST_ROSSIFT3D (GST_OBJECT_PARENT (pad));
  gboolean trigger=FALSE;
 
  //get mutex
  g_mutex_lock(filter->mutex);
  g_print("Got mutex!!\n");
  //exchange buffer(unref old)
  if (filter->lock){
	gst_buffer_unref(buf);
	trigger=TRUE;
  }
  else if (pad== filter->sinkpad1){
    trigger=filter->trigger1;
    switch(filter->modes1){
    case 0:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=buf;
      filter->feat1=(struct feature*)buf->data;
      filter->n1=(buf->size)/(sizeof(struct feature));
      break;
    case 1:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_buf_from_lowe((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));     
      gst_buffer_unref(buf);  
      break;
    case 2:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_buf_from_oxfd((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));   
      gst_buffer_unref(buf);   
      break;
    case 3:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_buf_from_lowelight(buf);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));    
      gst_buffer_unref(buf);   
      break;
    default:
      if (!filter->silent)g_print("Serious Error!!\n");
      return GST_FLOW_ERROR; 
    }
  }
  else if (pad == filter->sinkpad2){
    trigger=filter->trigger2;
    switch(filter->modes2){
    case 0:
      if (filter->s2) gst_buffer_unref(filter->s2);
      filter->s2=buf;
      filter->feat2=(struct feature*)buf->data;
      filter->n2=(buf->size)/(sizeof(struct feature));
      break;
    case 1:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_buf_from_lowe((char*)buf->data);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature)); 
      gst_buffer_unref(buf);  
      break;
    case 2:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_buf_from_oxfd((char*)buf->data);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature));  
      gst_buffer_unref(buf);  
      break;
    case 3:
      if (filter->s2)gst_buffer_unref(filter->s2);
      filter->s2=gst_buf_from_lowelight(buf);
      filter->s2->timestamp=buf->timestamp;
      filter->s2->duration=buf->duration;
      filter->s2->offset=buf->offset;
      filter->s2->offset_end=buf->offset_end;
      filter->feat2=(struct feature*)filter->s2->data;  
      filter->n2=(filter->s2->size)/(sizeof(struct feature)); 
      gst_buffer_unref(buf);     
      break;
    default:
      if (!filter->silent)g_print("Serious Error!!\n");
      return GST_FLOW_ERROR; 
    }
    if ((filter->opmode==MATCH)||(filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT))filter->kd_root = kdtree_build( filter->feat2, filter->n2 );
  }
  //decide to generate output?
  if ((filter->s1)&&(filter->s2)&&(trigger)&&((filter->opmode==MATCH)||(filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT))){
    if (!filter->silent)g_print("Start Eval!!\n");
    GstBuffer *outbuf;
    struct feature* feat;
    gint i,m=0,k;
    double d0, d1;
    struct feature** nbrs;
    CvMat* H=NULL;
    struct feature invalid;
    outbuf=gst_buffer_new_and_alloc(filter->n1*sizeof(struct coordinates));
    for( i = 0; i < filter->n2; i++ ){
	filter->feat2[i].bck_match=NULL;
	filter->feat2[i].fwd_match=NULL;	
      }
    //g_print("allocated %d for a maximum number of %d features!!\n",filter->n1*sizeof(struct coordinates),filter->n1);
    for( i = 0; i < filter->n1; i++ )
      {
      //g_print("In for!!\n");
      feat = filter->feat1 + i;
      // Generating new list of matches
      k = kdtree_bbf_knn( filter->kd_root, feat, filter->maxmatches, &nbrs, 200 );//todo 200 to parameter
      int z;
      for (z=0;z<k;z++){
	gennewmatch(filter, feat,nbrs[z], z);
	
      }
      if (filter->kd_root)free( nbrs );
      // 
      k = kdtree_bbf_knn( filter->kd_root, feat, 2, &nbrs, 200 );//todo 200 to parameter
      if( k == 2 )
	{
	  //g_print("found maybe sth!!\n");
	  d0 = descr_dist_sq( feat, nbrs[0] );
	  d1 = descr_dist_sq( feat, nbrs[1] );
	  if(( d0 < d1 * filter->maxsim )||((d0<filter->gooddist)&&(d0 < d1 * filter->maxsim2)))//0.49 to parameter
	    {
	      //pt1 = cvPoint( cvRound( feat->x ), cvRound( feat->y ) );
	      //pt2 = cvPoint( cvRound( nbrs[0]->x ), cvRound( nbrs[0]->y ) );
	      //pt2.y += img1->height;
	      //cvLine( stacked, pt1, pt2, CV_RGB(255,0,255), 1, 8, 0 );
  	      //g_print("found a feature, will write to %p!!\n",outbuf->data);
	      //struct coordinates* base=(struct coordinates*)(outbuf->data+(m*32));
	      //base->x1=1.00;
		//g_print("survived new adressing..!!\n");
	      //((struct coordinates*)(outbuf->data))[m].x1=feat->x;

	      //g_print("and was allowed to write it!!\n");
	      m++;
	      filter->feat1[i].fwd_match = nbrs[0];
              if (nbrs[0]->bck_match==NULL) nbrs[0]->bck_match = &filter->feat1[i];
	      else nbrs[0]->bck_match=&invalid;
	      if (!filter->silent)g_print("feature x=%d y=%d corresponds to x=%d y=%d!!\n",(int)filter->feat1[i].x,(int)filter->feat1[i].y,(int)nbrs[0]->x,(int)nbrs[0]->y);
	    }
	}
      if (filter->kd_root)free( nbrs );
    }
  m=0;

  if (filter->rematchdouble){//now the masterpiece, first the normal ones
    //delete
    for( i = 0; i < filter->n1; i++ ){
      if ((filter->feat1[i].fwd_match)&&(filter->feat1[i].fwd_match->bck_match!=&filter->feat1[i])){
	if (!filter->silent)g_print("deleting double\n");
	filter->feat1[i].fwd_match=NULL;
	}
    }
    //rematch
    struct kd_node* kdback;
    kdback = kdtree_build( filter->feat1, filter->n1 );
    for( i = 0; i < filter->n2; i++ ){
      if (filter->feat2[i].bck_match==&invalid){
        k = kdtree_bbf_knn( kdback,&filter->feat2[i], 2, &nbrs, 200 );
	if (k==2){
	  filter->feat2[i].bck_match=nbrs[0];
	  nbrs[0]->fwd_match=&filter->feat2[i];
	} 
      if (kdback) free( nbrs );	
      }
    }    
  kdtree_release( kdback );
  }


  if (filter->showinliers){
    struct feature** inliers;
    int n_in; 
    
    H = ransac_xform( filter->feat1, filter->n1, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, &inliers, &n_in );
    if (!filter->silent)g_print("Found %d inliers!\n",n_in);
    for( i = 0; i < n_in; i++ ){
      if (!filter->silent)g_print("Copying inlier No. %d !\n",i);	

      ((struct coordinates*)(outbuf->data))[m].x1=inliers[i]->x;
      ((struct coordinates*)(outbuf->data))[m].y1=inliers[i]->y;
      ((struct coordinates*)(outbuf->data))[m].x2=inliers[i]->fwd_match->x;
      ((struct coordinates*)(outbuf->data))[m].y2=inliers[i]->fwd_match->y;
      m++;
    }


  	
  }
  else for( i = 0; i < filter->n1; i++ ){
    if (filter->feat1[i].fwd_match){
      ((struct coordinates*)(outbuf->data))[m].x1=filter->feat1[i].x;
      ((struct coordinates*)(outbuf->data))[m].y1=filter->feat1[i].y;
      ((struct coordinates*)(outbuf->data))[m].x2=filter->feat1[i].fwd_match->x;
      ((struct coordinates*)(outbuf->data))[m].y2=filter->feat1[i].fwd_match->y;
      m++;
    }
  }




  if (!filter->silent)g_print("found %d matches!",m);
  GST_BUFFER_SIZE(outbuf)=m*sizeof(struct coordinates);
  //set up outbuf
  //set up 2 feature-pointers
  //decide to parse
  outbuf->timestamp=filter->s1->timestamp;
  outbuf->duration=filter->s1->duration;
  outbuf->offset=filter->s1->offset;
  outbuf->offset_end=filter->s1->offset_end;
  g_mutex_unlock(filter->mutex);
  if ((filter->opmode==TRANSFORM)||(filter->opmode==TRANSFORMPOINT)){
    if (!H) H = ransac_xform( filter->feat1, filter->n1, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, NULL, NULL );
    gst_buffer_unref(outbuf);
    outbuf=gst_buffer_new_and_alloc(sizeof(struct transmat));
    gdouble bufx=NAN,bufy=NAN,bufangle=NAN,bufscale=NAN;
    if (H){
      gint x,y;
      gdouble scale;
        if (!filter->silent)g_print("The point is x=%d and y=%d \n",(int)(filter->width2/2.0),(int)(filter->height2*0.9));
 	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,0,0),cvmGet(H,0,1),cvmGet(H,0,2));
	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,1,0),cvmGet(H,1,1),cvmGet(H,1,2));
	if (!filter->silent)printf("matrix data %f %f %f\n",cvmGet(H,2,0),cvmGet(H,2,1),cvmGet(H,2,2));

        double src_matrix[9], dst_matrix[9];
        CvMat A = cvMat( 3, 3, CV_64F, src_matrix ),
        invA = cvMat( 3, 3, CV_64F, dst_matrix );
        cvConvertScale( H, &A ,1,0);
        cvInvert( &A, &invA, CV_SVD );

 	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,0,0),cvmGet(&invA,0,1),cvmGet(&invA,0,2));
	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,1,0),cvmGet(&invA,1,1),cvmGet(&invA,1,2));
	if (!filter->silent)printf("inverted matrix data %f %f %f\n",cvmGet(&invA,2,0),cvmGet(&invA,2,1),cvmGet(&invA,2,2));

	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,0,0)/cvmGet(&invA,2,2),cvmGet(&invA,0,1)/cvmGet(&invA,2,2),cvmGet(&invA,0,2)/cvmGet(&invA,2,2));
	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,1,0)/cvmGet(&invA,2,2),cvmGet(&invA,1,1)/cvmGet(&invA,2,2),cvmGet(&invA,1,2)/cvmGet(&invA,2,2));
	if (!filter->silent)printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,2,0)/cvmGet(&invA,2,2),cvmGet(&invA,2,1)/cvmGet(&invA,2,2),cvmGet(&invA,2,2)/cvmGet(&invA,2,2));

	((struct transmat*)(outbuf->data))->a00=cvmGet(&invA,0,0);
	((struct transmat*)(outbuf->data))->a01=cvmGet(&invA,0,1);
	((struct transmat*)(outbuf->data))->a02=cvmGet(&invA,0,2);
	((struct transmat*)(outbuf->data))->a10=cvmGet(&invA,1,0);
	((struct transmat*)(outbuf->data))->a11=cvmGet(&invA,1,1);
	((struct transmat*)(outbuf->data))->a12=cvmGet(&invA,1,2);
	((struct transmat*)(outbuf->data))->a20=cvmGet(&invA,2,0);
	((struct transmat*)(outbuf->data))->a21=cvmGet(&invA,2,1);
	((struct transmat*)(outbuf->data))->a22=cvmGet(&invA,2,2);


        scale=(cvmGet(&invA,2,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,2,1)*((float)filter->height2*0.9)+cvmGet(&invA,2,2));
	x=(int)((cvmGet(&invA,0,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,0,1)*((float)filter->height2*0.9)+cvmGet(&invA,0,2))/scale);
	y=(int)((cvmGet(&invA,1,0)*((gdouble)filter->width2/2.0)+cvmGet(&invA,1,1)*((float)filter->height2*0.9)+cvmGet(&invA,1,2))/scale);
	
        bufscale=(cvmGet(&invA,2,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,2,1)*((float)filter->height2*0.5)+cvmGet(&invA,2,2));
	bufx=(cvmGet(&invA,0,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,0,1)*((float)filter->height2*0.5)+cvmGet(&invA,0,2))/bufscale;
	bufy=(cvmGet(&invA,1,0)*((gdouble)filter->width2*0.5)+cvmGet(&invA,1,1)*((float)filter->height2*0.5)+cvmGet(&invA,1,2))/bufscale;	
	bufangle=atan2((cvmGet(&invA,0,0)/cvmGet(&invA,2,2)),(cvmGet(&invA,0,0)/cvmGet(&invA,2,2)));
	bufscale=(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))+(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2));

	g_print("The point is mapped to x=%d, y=%d, scale= %f! that leads to x= %d y= %d\n",x,y,scale,(int)((float)x/scale),(int)((float)y/scale));
        //sprintf((char*)outbuf->data,"1\n");
	//sprintf((char*)outbuf->data+2,"%d %d\n",x,y);
      //cvReleaseMat( &H );
    }
    else{
        if (!filter->silent)g_print("Sorry, could not find transformation!");	
        memset (outbuf->data,0,sizeof(struct transmat));
	}
    gst_buffer_set_caps(outbuf,gst_caps_new_simple ("matrix-list",NULL));
    if (filter->opmode==TRANSFORMPOINT){
      gst_buffer_unref(outbuf);
      outbuf=gst_buffer_new_and_alloc(sizeof(struct coordinates));	
      gst_buffer_set_caps(outbuf,gst_caps_new_simple ("coordinates-list",NULL));
      if (!little_endian){
	((struct coordinates*)(outbuf->data))[0].x1=DoubleSwap(bufx);
	((struct coordinates*)(outbuf->data))[0].y1=DoubleSwap(bufy);
	((struct coordinates*)(outbuf->data))[0].x2=DoubleSwap(bufscale);
	((struct coordinates*)(outbuf->data))[0].y2=DoubleSwap(bufangle);
      }
      else{
	((struct coordinates*)(outbuf->data))[0].x1=bufx;
	((struct coordinates*)(outbuf->data))[0].y1=bufy;
	((struct coordinates*)(outbuf->data))[0].x2=bufscale;
	((struct coordinates*)(outbuf->data))[0].y2=bufangle;
	}
    }
    outbuf->timestamp=filter->s1->timestamp;
    outbuf->duration=filter->s1->duration;
    outbuf->offset=filter->s1->offset;
    outbuf->offset_end=filter->s1->offset_end;
    return gst_pad_push (filter->srcpad, outbuf);
  }
  else{
      //gst_buffer_set_caps(outbuf,gst_pad_get_pad_template_caps(filter->srcpad));
      gst_buffer_set_caps(outbuf,gst_caps_new_simple ("coordinates-list",
		NULL));
      return gst_pad_push (filter->srcpad, outbuf);
    }
  }
  else if ((filter->s1)&&(filter->s2)&&(trigger)&&(filter->opmode==MERGE)){
    if (!filter->silent)g_print("Start Merge\n");
    GstBuffer *outbuf;
    struct feature* outmem;
    int used=0;//number of used features
    int bytesperdesc=0;
    int i;
    //if (1) bytesperdesc=sizeof(struct lowelight);
    if (1) bytesperdesc=sizeof(struct feature);
    outmem=(struct feature*)g_malloc((filter->n1+filter->n2)*bytesperdesc);
    GList* incut=NULL;
    struct feature* feat;
    //if (filter->offsetx==-1)filter->offsetx=filter->width1;
    //if (filter->offsety==-1)filter->offsety=filter->height1;
    if (filter->offsetx==-1)filter->offsetx=0;
    if (filter->offsety==-1)filter->offsety=0;
    if (!filter->silent)g_print("Start Merge with overlap X %d and Y %d\n",filter->offsetx,filter->offsety);
    //copy all non overlapping features to result, and put questionable into a buffer
    for( i = 0; i < filter->n1; i++ ){
	feat=filter->feat1+i;
	if (!filter->silent)g_print("Sifttool merge analyzing %d. Feature X: %f Y: %f \n",i,feat->x,feat->y);
	if ((feat->x>((double)filter->width1-(double)filter->offsetx))||(feat->y>((double)filter->height1-(double)filter->offsety))) incut=g_list_append (incut, feat);
	else {
	  if (!filter->silent)g_print("Sifttool doing copy\n");
	  memcpy(outmem+used,feat,bytesperdesc);
	  used++;
	}
    }
    if (!filter->silent)g_print("After taking features of first part having %d final features\n",used);
    for( i = 0; i < filter->n2; i++ ){
	feat=filter->feat2+i;
	if (!filter->silent)g_print("Sifttool merge analyzing %d, with %f and %f\n",i,feat->x,feat->y);
	if (filter->offsetx>0){
          feat->x=feat->x+filter->width1-filter->offsetx;
	  feat->img_pt.x=feat->img_pt.x+filter->width1-filter->offsetx;
	}
	if (filter->offsety>0){
	  feat->y=feat->y+filter->height1-filter->offsety;
 	  feat->img_pt.y=feat->img_pt.y+filter->height1-filter->offsety;
	}
	if (!filter->silent)g_print("Sifttool after modifying number %d, with %f and %f\n",i,feat->x,feat->y);
	if ((feat->x<filter->width1)&&(feat->y<filter->height1)) incut=g_list_append (incut, feat);//if in the same region...
	else {
	  if (!filter->silent)g_print("Sifttool doing copy, out adress %p in adress %p\n",outmem+used,feat);
	  memcpy(outmem+used,feat,bytesperdesc);
	  used++;
	}
    }
    if (!filter->silent)g_print("After taking features of second part having %d final features\n",used);
    if (!filter->silent)g_print("We have to analyze %d features wether they are unique\n",g_list_length(incut));
    while(incut){

	struct feature* current=incut->data;
	struct feature* tempfeat;
	if (!filter->silent)g_print("Entering list, %d left, X %f, Y %f!!\n",g_list_length(incut),current->x,current->y);
	GList* temp=g_list_next(incut);
	gboolean found=FALSE;
	while ((temp)&&(!found)){
	  tempfeat=temp->data;
	  if (((tempfeat->x-current->x)*(tempfeat->x-current->x)+(tempfeat->x-current->x)*(tempfeat->x-current->x))<2.0){
	    if (!filter->silent)g_print("Found same feature, not including!!\n");
	    if (1) found=TRUE;
	  }
	  temp=g_list_next(temp);
	}
	if (!found){
	  if (!filter->silent)g_print("Not found same feature, including!!\n");
	  memcpy(outmem+used,current,bytesperdesc);
	  used++;
	}
	incut=g_list_delete_link (incut,incut);
	//incut=g_list_next(incut)
    }
    if (!filter->silent)g_print("After checking for duplicates we have %d final features\n",used);
    outmem=g_realloc(outmem,used*bytesperdesc);
    if (filter->mode==0){
      outbuf=gst_buffer_new_and_alloc(0);
      int outwidth=0;
      if (filter->offsetx>0)outwidth=filter->width1+filter->width2-filter->offsetx;
      else outwidth=filter->width1;

      int outheight=0;
      if (filter->offsety>0)outheight=filter->height1+filter->height2-filter->offsety;
      else outheight=filter->height1;

      gst_buffer_set_data(outbuf,(unsigned char*)outmem,used*bytesperdesc);
      GstCaps* outcaps=gst_caps_new_simple ("sift/raw",
		  //"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
		  //"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
		  "width", G_TYPE_INT, outwidth,
		  "height", G_TYPE_INT, outheight,
		  "overlap_v", G_TYPE_INT, filter->offsetx,
		  "overlap_h", G_TYPE_INT, filter->offsety,
		  NULL);
      gst_buffer_set_caps(outbuf,outcaps);
      gst_caps_unref(outcaps);
      //g_print("%s Pushing with caps %s height1 %d height2 %d!!\n",GST_ELEMENT_NAME(filter),gst_caps_to_string(outbuf->caps),filter->height1,filter->height2);
      GST_BUFFER_MALLOCDATA(outbuf)=(unsigned char*)outmem;
    }
    else if (filter->mode==3){
      //gchar* outbufmem;
      gint offset=0;
      int i,j,d;
      struct feature* feat=outmem;
    
      if (used) d=feat[0].d;
      outbuf=gst_buffer_new_and_alloc(used*sizeof(struct lowelight));
      struct lowelight *resll=(struct lowelight*)outbuf->data;

      for( i = 0; i < used; i++ )
        {
        if (little_endian){
          resll[i].x=feat[i].x;
          resll[i].y=feat[i].y;
          resll[i].scl=feat[i].scl;
          resll[i].ori=feat[i].ori;
        }
        else{
          resll[i].x=DoubleSwap(feat[i].x);
          resll[i].y=DoubleSwap(feat[i].y);
          resll[i].scl=DoubleSwap(feat[i].scl);
          resll[i].ori=DoubleSwap(feat[i].ori);
        }	
        for( j = 0; j < d; j++ )
	  {
	   resll[i].data[j]=(int)(feat[i].descr[j]);
	  }
        }
    g_print ("Allocated %d\n",used*sizeof(struct lowelight));
    free(outmem);
      int outwidth=0;
      if (filter->offsetx>0)outwidth=filter->width1+filter->width2-filter->offsetx;
      else outwidth=filter->width1;

      int outheight=0;
      if (filter->offsety>0)outheight=filter->height1+filter->height2-filter->offsety;
      else outheight=filter->height1;

      GstCaps* outcaps=gst_caps_new_simple ("sift/lowelight",
		  //"width", G_TYPE_INT, (filter->width1+filter->width2-filter->offsetx),
		  //"height", G_TYPE_INT, (filter->height1+filter->height2-filter->offsety),
		  "width", G_TYPE_INT, outwidth,
		  "height", G_TYPE_INT, outheight,
		  "overlap_v", G_TYPE_INT, filter->offsetx,
		  "overlap_h", G_TYPE_INT, filter->offsety,
		  NULL);
      gst_buffer_set_caps(outbuf,outcaps);
      gst_caps_unref(outcaps);
      //g_print("%s Pushing with caps %s height1 %d height2 %d!!\n",GST_ELEMENT_NAME(filter),gst_caps_to_string(outbuf->caps),filter->height1,filter->height2);
    }

      outbuf->timestamp=filter->s1->timestamp;
      outbuf->duration=filter->s1->duration;
      outbuf->offset=filter->s1->offset;
      outbuf->offset_end=filter->s1->offset_end;
      if (filter->trigger1){
	  gst_buffer_unref(filter->s1);
	  filter->s1=NULL;
      }
      if (filter->trigger2){
	  gst_buffer_unref(filter->s2);
	  filter->s2=NULL;
      }
      g_mutex_unlock(filter->mutex);
      return gst_pad_push (filter->srcpad, outbuf);

  }
  else{
    g_mutex_unlock(filter->mutex);
    return GST_FLOW_OK;
  }
}

static GstStateChangeReturn
gst_rossift3d_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstrossift3d *filter;
  filter = GST_ROSSIFT3D (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      filter->stopped1=!filter->trigger1;
      filter->stopped2=!filter->trigger2;
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      if (filter->s2) gst_buffer_unref(filter->s2);
      filter->s2=NULL;
      if (filter->s1) gst_buffer_unref(filter->s1);
      filter->s1=NULL;
      break;
    default:
      break;
  }
 return ret;
}





static gboolean
gst_rossift3d_event (GstPad  *pad, GstEvent * event)
{
  Gstrossift3d *filter;
  gboolean ret=TRUE;

  filter = GST_ROSSIFT3D (gst_pad_get_parent (pad));

  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_EOS:
      if (pad == filter->sinkpad1) filter->stopped1=TRUE;
      if (pad == filter->sinkpad2) filter->stopped2=TRUE;
      if ((filter->stopped1)&&(filter->stopped2)) ret = gst_pad_push_event (filter->srcpad, event);
      g_print("Event EOS!\n");
      break;
    default:
      ret = gst_pad_push_event (filter->srcpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}

