/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_ROSSIFT3D_H__
#define __GST_ROSSIFT3D_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_ROSSIFT3D \
  (gst_rossift3d_get_type())
#define GST_ROSSIFT3D(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ROSSIFT3D,Gstrossift3d))
#define GST_ROSSIFT3D_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ROSSIFT3D,Gstrossift3dClass))
#define GST_IS_ROSSIFT3D(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ROSSIFT3D))
#define GST_IS_ROSSIFT3D_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ROSSIFT3D))

#define FEATURE_MAX_D 128
#define MINPQ_INIT_NALLOCD 512

/* RANSAC error tolerance in pixels */
#define RANSAC_ERR_TOL 3

/** pessimistic estimate of fraction of inlers for RANSAC */
#define RANSAC_INLIER_FRAC_EST 0.25

/** estimate of the probability that a correspondence supports a bad model */
#define RANSAC_PROB_BAD_SUPP 0.10

/* extracts a feature's RANSAC data */
#define feat_ransac_data( feat ) ( (struct ransac_data*) (feat)->feature_data )

typedef struct _Gstrossift3d      Gstrossift3d;
typedef struct _Gstrossift3dClass Gstrossift3dClass;

typedef enum {
  MATCH,
  MERGE,  //merge two sets of sift-features
  TRANSFORM,
  TRANSFORMPOINT  //calculate transformation of one poinnt
} SiftToolOpMode;

typedef CvMat* (*ransac_xform_fn)( CvPoint2D64f* pts, CvPoint2D64f* mpts,
				   int n );
typedef double (*ransac_err_fn)( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* T );


typedef struct ThreeDmatch{
  struct feature* feat1;
  struct feature* feat2;
  double x;//x position in camera 1 coordinate space
  double y;//y position in camera 1 coordinate space
  double z;//z position in camera 1 coordinate space
  double error;//The error between the two reprojection lines 
  double matchdist;//distance to match
  double nextbestmatch;//distance ratio to next best match 
  int rank;//1 best choice ...
};

struct _Gstrossift3d
{
  GstElement element;

  GstPad *sinkpad1,*sinkpad2, *srcpad;
  SiftToolOpMode opmode;
  gboolean silent,trigger1,trigger2,stopped1,stopped2;
  gint width1,height1,width2,height2;
  gboolean rematchdouble,showinliers;//whether to put out coordinates in booka-style instead of matchlist

  GstBuffer *s1,*s2;
  gint modes1,modes2;//mode of each input (to parse)
  gint suggestedmodes1,suggestedmodes2;//suggestedmode from user input
  GMutex* mutex;
  struct feature* feat1;
  struct feature* feat2;
  struct kd_node* kd_root;
  gint n1,n2;
  gint offsetx,offsety;//overlap of second image compared to first
  gdouble maxsim,maxsim2,gooddist,threshold;
  int mode;
  gboolean lock;//a lock that disables updating the feature vectors (in order to try different parameters for )
  GList* ThreeDmatches;
  int maxmatches;

  //Transformation from camera one to camera two
  CvMat* R;
  CvMat* T;
  double fx1,fx2,fy1,fy2;
};

struct _Gstrossift3dClass 
{
  GstElementClass parent_class;
};

GType gst_rossift3d_get_type (void);



G_END_DECLS

#endif /* __GST_ROSSIFT3D_H__ */
