/* GStreamer
 * Copyright (C) 2006 David A. Schleef <ds@schleef.org>
 *
 * gstrossrc.c:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_MULTIFILESRC_H__
#define __GST_MULTIFILESRC_H__

#include <gst/gst.h>
#include <gst/base/gstpushsrc.h>

G_BEGIN_DECLS

#define GST_TYPE_ROSSRC \
  (gst_rossrc_get_type())
#define GST_ROSSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ROSSRC,Gstrossrc))
#define GST_ROSSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ROSSRC,GstrossrcClass))
#define GST_IS_ROSSRC(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ROSSRC))
#define GST_IS_ROSSRC_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ROSSRC))


typedef struct _Gstrossrc Gstrossrc;
typedef struct _GstrossrcClass GstrossrcClass;



struct _Gstrossrc
{
  GstPushSrc parent;
  gboolean silent;

  gchar *dirname;
  int index;
  int i,length;//running through event buffer, length of current event return
  int fd,wd,childpid;
  int offset,width,height;
  char filename[1024];
  char *topic;
  char *transport;
  gboolean sigusr,deletefile;//send sigusr to child each run? delete sent files?
  GstCaps *caps;
  int timeout,sleep;//timeout for select calls
  gboolean successful_read;
  fd_set rfds;
  struct timeval tv;
  int zaehler,nenner;
  GCond* data_cond; /* Must be initialized somewhere */
  GMutex* data_mutex;
 

  GstBuffer* current;
  GstCaps* newcaps;

  //sensor_msgs::ImageConstPtr last1;
  //sensor_msgs::ImageConstPtr last2;
  //sensor_msgs::ImageConstPtr last3;
  std::string* format; 

  ros::NodeHandle* nh;
  image_transport::Subscriber sub;

  GList* sentbuffers;
  std::vector < sensor_msgs::ImageConstPtr > msglist;
  pthread_t rosspin,watchdog;
  gboolean end;

  //void (*GFreeFunc) freerosmsg(gpointer data){
  //void freerosmsg(gpointer data){
  //void freerosmsg(gpointer data){
  //  g_mutex_lock (data_mutex); 
  //	g_print("Freeing Buffer to Adress %p\n",data);
  //  g_mutex_unlock (data_mutex);    
  //}

  void imageCallback(const sensor_msgs::ImageConstPtr& msg)
  {
  //boost::lock_guard<boost::mutex> guard(image_mutex_);
	   
// Hang on to message pointer for sake of mouse_cb
//last_msg_ = msg;

// May want to view raw bayer data
// NB: This is hacky, but should be OK since we have only one image CB.
//  if (msg->encoding.find("bayer") != std::string::npos) 
    if (!silent)g_print("Got Image Transport Message\n");
    g_mutex_lock (data_mutex);
    current = gst_buffer_new ();
    GST_BUFFER_DATA (current) = (unsigned char *) &(msg->data[0]);//scheiß c++
    //GST_BUFFER_MALLOCDATA(current)  = (unsigned char *) &(msg->data[0]);//scheiß c++

    if (!silent)g_print("Set Buffer to Adress %p\n",GST_BUFFER_DATA (current));
    GST_BUFFER_SIZE (current) = msg->step*msg->height;
    //boost::function<void (gpointer)> freeRosMsgCb = std::bind1st(
    //  std::mem_fun(&_Gstrossrc::freerosmsg), this);
    //GST_BUFFER_FREE_FUNC(current)=/*(*GFreeFunc)*/(void (*)(void*))freeRosMsgCb;
    //touching everything...
    GST_BUFFER_TIMESTAMP(current)=(guint64)msg->header.stamp.sec*1000000000ULL+(guint64)msg->header.stamp.nsec;
    guint64 sum=0;
    msglist.push_back(msg);
    //for (int i=1;i<GST_BUFFER_SIZE (current);i++){
	//sum+=current->data[i];
	//g_print("P%d%d",i,current->data[i]);
    //}
    //g_print("Sum is %llu\n",sum);
    GST_BUFFER_OFFSET (current) = (guint64)msg->header.seq;
    GST_BUFFER_OFFSET_END (current) = GST_BUFFER_OFFSET (current)+1;
    if ((!format->compare(msg->encoding))||(width!=msg->width)||(height!=msg->height)){
	width=msg->width;
	height=msg->height;
	*format=msg->encoding;
 	if (newcaps)gst_caps_unref(newcaps);
	int endianess=((msg->is_bigendian) ? 1234 : 4321);
	if (!msg->encoding.compare("rgb8")) newcaps=gst_caps_new_simple ("video/x-raw-rgb",
  		"framerate", GST_TYPE_FRACTION, (int)zaehler,(int)nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
                "red_mask",G_TYPE_INT,16711680, 
  		"green_mask",G_TYPE_INT,65280, 
		"blue_mask",G_TYPE_INT,255,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("rgba8")) newcaps=gst_caps_new_simple ("video/x-raw-rgb",
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
                "red_mask",G_TYPE_INT,(int)0xff000000, 
                "green_mask",G_TYPE_INT,16711680, 
  		"blue_mask",G_TYPE_INT,65280, 
		"alpha_mask",G_TYPE_INT,255,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	if (!msg->encoding.compare("bgr8")) newcaps=gst_caps_new_simple ("video/x-raw-rgb",
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
		"red_mask",G_TYPE_INT,255,
  		"green_mask",G_TYPE_INT,65280, 
                "blue_mask",G_TYPE_INT,16711680, 
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("bgra8")) newcaps=gst_caps_new_simple ("video/x-raw-rgb",
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"blue_mask",G_TYPE_INT,(int)0xff000000, 
                "green_mask",G_TYPE_INT,16711680, 
  		"red_mask",G_TYPE_INT,65280, 
		"alpha_mask",G_TYPE_INT,255,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("mono8")||!msg->encoding.compare("mono16")) newcaps=gst_caps_new_simple ("video/x-raw-gray",
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("yuv422")) newcaps=gst_caps_new_simple ("video/x-raw-yuv",
 		"format", GST_TYPE_FOURCC,GST_MAKE_FOURCC ('U', 'Y', 'V', 'Y') ,
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("bayer_rggb8")||!msg->encoding.compare("bayer_rggb16")) newcaps=gst_caps_new_simple ("video/x-raw-bayer",
 		"format", G_TYPE_STRING,"rggb" ,
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("bayer_bggr8")||!msg->encoding.compare("bayer_bggr16")) newcaps=gst_caps_new_simple ("video/x-raw-bayer",
 		"format", G_TYPE_STRING,"bggr" ,
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("bayer_gbrg8")||!msg->encoding.compare("bayer_gbrg16")) newcaps=gst_caps_new_simple ("video/x-raw-bayer",
 		"format", G_TYPE_STRING,"gbrg" ,
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	else if (!msg->encoding.compare("bayer_grbg8")||!msg->encoding.compare("bayer_grbg16")) newcaps=gst_caps_new_simple ("video/x-raw-bayer",
 		"format", G_TYPE_STRING,"grbg",
  		"framerate", GST_TYPE_FRACTION, zaehler,nenner,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,endianess,
  		"pixel-aspect-ratio", GST_TYPE_FRACTION, 1,1,
		NULL);
	}

    gst_buffer_set_caps (current,newcaps);
    /*gst_buffer_set_caps (current, gst_caps_new_simple ("video/x-raw-rgb",
  		"framerate", GST_TYPE_FRACTION, 30,1,
      		"width", G_TYPE_INT, msg->width,
      		"height", G_TYPE_INT, msg->height,
		"bpp", G_TYPE_INT,(msg->step/msg->width)*8 ,
		"depth", G_TYPE_INT, (msg->step/msg->width)*8,
  		"endianness",G_TYPE_INT,4321,
                "red_mask",G_TYPE_INT,16711680, 
  		"green_mask",G_TYPE_INT,65280, 
		"blue_mask",G_TYPE_INT,255,
  		//"green_mask",G_TYPE_INT,16711680, 
  		//"blue_mask",G_TYPE_INT,0,
		NULL));*/
    //buffers=g_list_append(buffers,msg);
    //if ((index%3)==0)last1=msg;
    //if ((index%3)==1)last2=msg;
    //if ((index%3)==2)last3=msg;   

    g_cond_signal (data_cond);
    g_mutex_unlock (data_mutex);
    if (!silent)g_print("Finished processing message\n");
  }
};

struct _GstrossrcClass
{
  GstPushSrcClass parent_class;
};

GType gst_rossrc_get_type (void);

G_END_DECLS

#endif /* __GST_MULTIFILESRC_H__ */
