/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2013 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, inrosTFPUBuding without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-rosTFPUB
 *
 * FIXME:Describe rosTFPUB here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m v4l2src ! ffmpegcolorspace ! rosTFPUB sqaresize=17.5 ! ffmpegcolorspace ! ximagesink sync=0 qos=0
 gst-launch -v v4l2src ! tee name=t ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! rosTFPUB silent=0 ! ffmpegcolorspace ! ximagesink sync=0 qos=0 t. ! queue ! ffmpegcolorspace !
 ximagesink sync=0 qos=0

gst-launch -v -m v4l2src ! propexport exportdesc="un;takenext" ! queue leaky=2 max-size-buffers=1 ! ffmpegcolorspace ! rosTFPUB name=un skip=1000 nboards=6 ! ffmpegcolorspace ! ximagesink sync=0 qos=0

Simulating the full thing
Start RemoteUI

videotestsrc ! video/x-raw-rgb, width=1, height=1 ! identity sync=1 ! timestamper t1=-1 t2=1 ! propexport exportdesc="un;inverse;un;manr;un;manp;un;many;un;manpx;un;manpy;un;manpz" ! rostfpub name=un ! fakesink




 * ]|
 * </refsect2>
 */
#ifndef PI
#define PI 3.14159265359
#endif

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>

#include <errno.h>
#include <assert.h>
#include <string.h>


//#include <opencv/cv.h>



#include <ros/ros.h>
#include <visualization_msgs/Marker.h>



#include <geometry_msgs/PoseStamped.h>
#include <sensor_msgs/CameraInfo.h>
#include "tf/tf.h"
#include "tf/tfMessage.h"
#include "tf/transform_broadcaster.h"
#include "tf/transform_listener.h"

#include "roscommon.h"
#include "gstrostfpub.h"

GST_DEBUG_CATEGORY_STATIC (gst_rostfpub_debug);
#define GST_CAT_DEFAULT gst_rostfpub_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_FRAMEID,
  PROP_CHILDFRAMEID,
  PROP_INV,
  PROP_MANR,
  PROP_MANP,
  PROP_MANY,
  PROP_MANPX,
  PROP_MANPY,
  PROP_MANPZ,
  PROP_DHTHETA,
  PROP_DHALPHA,
  PROP_DHA,
  PROP_DHD
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );


GST_BOILERPLATE (Gstrostfpub, gst_rostfpub, GstElement,
    GST_TYPE_ELEMENT);

static void gst_rostfpub_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_rostfpub_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_rostfpub_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_rostfpub_get_caps (GstPad *pad);

static GstFlowReturn gst_rostfpub_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_rostfpub_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "rosTFPUB",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the rosTFPUB's class */
static void
gst_rostfpub_class_init (GstrostfpubClass * klass)
{
  GObjectClass *gobject_class;
  //GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  //gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_rostfpub_set_property;
  gobject_class->get_property = gst_rostfpub_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_FRAMEID,
      g_param_spec_string ("frame_id", "Frame_id", "What frame_id will be the parent frame",
          "parent", G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, PROP_CHILDFRAMEID,
      g_param_spec_string ("childframe_id", "Childframe_id", "What frame_id will be the child frame ","child", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_INV,
      g_param_spec_boolean ("inverse", "Inverse", "Try inverse operation",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANR,
      g_param_spec_double ("manr", "Manr", "Manual Roll adjustment",
          -3.14,3.14,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANP,
      g_param_spec_double ("manp", "Manp", "Manual Pitch adjustment",
          -3.14,3.14,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANY,
      g_param_spec_double ("many", "Many", "Manual Yaw adjustment",
          -3.14,3.14,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANPX,
      g_param_spec_double ("manpx", "Manpx", "Manual Position X adjustment",
          -2,2,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANPY,
      g_param_spec_double ("manpy", "Manpy", "Manual Position Y adjustment",
          -2,2,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MANPZ,
      g_param_spec_double ("manpz", "Manpz", "Manual Position Z adjustment",
          -2,2,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DHTHETA,
      g_param_spec_double ("dhtheta", "Dhtheta", "Adjust DH-Parameter Theta (switches to DH mode)",
          -3.14,3.14,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DHALPHA,
      g_param_spec_double ("dhalpha", "Dhalpha", "Adjust DH-Parameter Alpha (switches to DH mode)",
          -3.14,3.14,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DHA,
      g_param_spec_double ("dha", "Dha", "Adjust DH-Parameter A (switches to DH mode)",
          -2,2,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DHD,
      g_param_spec_double ("dhd", "Dhd", "Adjust DH-Parameter D (switches to DH mode)",
          -2,2,0, G_PARAM_READWRITE));
}



/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_rostfpub_init (Gstrostfpub * filter,
    GstrostfpubClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_rostfpub_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_rostfpub_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));





  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);

  filter->silent = FALSE;
  filter->ready= FALSE;
  filter->caps= NULL;



  filter->firsttime=TRUE;
  filter->frame_id=g_strdup_printf("parent");
  filter->childframe_id=g_strdup_printf("child");

  filter->dha=0;
  filter->dhd=0;
  filter->dhtheta=0;
  filter->dhalpha=0;

}

static void
gst_rostfpub_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstrostfpub *filter = GST_ROSTFPUB (object);
  g_print("gst_rostfpub_set_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_FRAMEID:
      if (filter->frame_id)g_free (filter->frame_id);
      filter->frame_id = g_value_dup_string (value);
	//todo mutex,unadvertise,advertise,unlock
      break; 
    case PROP_CHILDFRAMEID:
      if (filter->childframe_id)g_free (filter->childframe_id);
      filter->childframe_id = g_value_dup_string (value);
      break; 
    case PROP_INV:
      filter->inverse = g_value_get_boolean (value);
      break;
   case PROP_MANR:
      filter->fromdh=FALSE; 
      filter->manr = g_value_get_double (value);
      break;
   case PROP_MANP:
      filter->fromdh=FALSE; 
      filter->manp = g_value_get_double (value);
      break;
   case PROP_MANY:
      filter->fromdh=FALSE; 
      filter->many = g_value_get_double (value);
      break;
   case PROP_MANPX:
      filter->fromdh=FALSE; 
      filter->manpx = g_value_get_double (value);
      break;
   case PROP_MANPY:
      filter->fromdh=FALSE; 
      filter->manpy = g_value_get_double (value);
      break;
   case PROP_MANPZ:
      filter->fromdh=FALSE;
      filter->manpz = g_value_get_double (value);
      break;
   case PROP_DHTHETA:
      filter->fromdh=TRUE;    
      filter->dhtheta= g_value_get_double (value);
   break;
   case PROP_DHALPHA:
      filter->fromdh=TRUE;    
      filter->dhalpha= g_value_get_double (value);
   break;
   case PROP_DHA:
      filter->fromdh=TRUE;    
      filter->dha= g_value_get_double (value);
   break;
   case PROP_DHD:
      filter->fromdh=TRUE;    
      filter->dhd= g_value_get_double (value);
   break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }

}

static void
gst_rostfpub_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstrostfpub *filter = GST_ROSTFPUB (object);
  g_print("gst_rostfpub_get_property.\n");
  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
   case PROP_FRAMEID:
      g_value_set_string (value, filter->frame_id);     
   break;
   case PROP_CHILDFRAMEID:
      g_value_set_string (value, filter->childframe_id);     
   break;
   case PROP_INV:
      g_value_set_boolean (value, filter->inverse);
   break;
   case PROP_MANR:
      g_value_set_double (value,filter->manr);
   break;
   case PROP_MANP:
      g_value_set_double (value,filter->manp);
   break;
   case PROP_MANY:
      g_value_set_double (value,filter->many);
   break;
   case PROP_MANPX:
      g_value_set_double (value,filter->manpx);
   break;
   case PROP_MANPY:
      g_value_set_double (value,filter->manpy);
   break;
   case PROP_MANPZ:
      g_value_set_double (value,filter->manpz);
   break;
   case PROP_DHTHETA:
      g_value_set_double (value,filter->dhtheta);
   break;
   case PROP_DHALPHA:
      g_value_set_double (value,filter->dhalpha);
   break;
   case PROP_DHA:
      g_value_set_double (value,filter->dha);
   break;
   case PROP_DHD:
      g_value_set_double (value,filter->dhd);
   break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

//gboolean gettransform(Gstrostfpub *filter){
//  
//  
//
//}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_rostfpub_set_caps (GstPad * pad, GstCaps * caps)
{
  //g_print("gst_rostfpub_set_caps.\n");
  Gstrostfpub *filter;
  GstPad *otherpad;

  gboolean ret;
  
  filter = GST_ROSTFPUB (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

  ret=gst_pad_set_caps (otherpad, caps);
  if (!ret) g_print ("GstROSTFPUB Warning Set Caps will return false.\n");
  return ret;
}

static GstCaps *
gst_rostfpub_get_caps (GstPad *pad){
  GstCaps * ret;
  ret=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  //g_print("rosTFPUB get caps, PadName %s, returning %s",gst_pad_get_name(pad),gst_caps_to_string(ret));
  return ret; 
}

/* chain function
 * this function does the actual processing
 */
static gint globalrosinit=FALSE;
static gint initdone=0;

static GstFlowReturn
gst_rostfpub_chain (GstPad * pad, GstBuffer * buf)
{
  Gstrostfpub *filter;
  filter = GST_ROSTFPUB (GST_OBJECT_PARENT (pad));
  g_print ("Rosstfpub chain function.\n");
  if (filter->firsttime){

      int schrott=0;
      if (g_atomic_int_compare_and_exchange (&globalrosinit, FALSE, TRUE)){
	g_print ("GstROSTFPUB Doing RosInit.\n");
	ros::init(schrott, NULL, "rossink", ros::init_options::AnonymousName);
	initdone=1;
      }
      while (initdone==0){ usleep(10000); };
      filter->nh = new ros::NodeHandle();
      
      filter->rpub= new  ros::Publisher();

      gchar* topicname=g_strdup_printf("%s","unused");
      *filter->rpub = filter->nh->advertise<visualization_msgs::Marker>(topicname, 1);
      g_free(topicname);

      filter->br=new tf::TransformBroadcaster();


      filter->firsttime=FALSE;
  }





  GstFlowReturn ret;




	//publish the TF

      tf::StampedTransform trst;
      tf::StampedTransform trst2;

      if(filter->fromdh){

	CvMat* rota=cvCreateMat(3,3,CV_32FC1);
        CV_MAT_ELEM(*rota, float,0,0) = cos(filter->dhtheta);
        CV_MAT_ELEM(*rota, float,0,1) = -sin(filter->dhtheta)*cos(filter->dhalpha);
        CV_MAT_ELEM(*rota, float,0,2) = sin(filter->dhtheta)*sin(filter->dhalpha);
        CV_MAT_ELEM(*rota, float,1,0) = sin(filter->dhtheta);
        CV_MAT_ELEM(*rota, float,1,1) = cos(filter->dhtheta)*cos(filter->dhalpha);
        CV_MAT_ELEM(*rota, float,1,2) = -cos(filter->dhtheta)*sin(filter->dhalpha);
        CV_MAT_ELEM(*rota, float,2,0) = 0;
        CV_MAT_ELEM(*rota, float,2,1) = sin(filter->dhalpha);
        CV_MAT_ELEM(*rota, float,2,2) = cos(filter->dhalpha);
        double w,x,y,z;
        genquat(&w,&x,&y,&z,rota);
        trst.setOrigin( tf::Vector3((filter->dha*cos(filter->dhtheta)),(filter->dha*sin(filter->dhtheta)),filter->dhd) );      
        trst.setRotation(tf::Quaternion(x,y,z,w)); 
      }
      else{
        tf::Quaternion manquat=tf::createQuaternionFromRPY(filter->manr,filter->manp,filter->many);
        tf::Quaternion quat=manquat;
        trst.setOrigin( tf::Vector3(filter->manpx,filter->manpy,filter->manpz) );
        trst.setRotation(quat);
      }



      
      trst.stamp_.sec=buf->timestamp/1000000000ULL;
      trst.stamp_.nsec=buf->timestamp%1000000000ULL;



      trst.frame_id_.assign(filter->frame_id);
      trst.child_frame_id_.assign(filter->childframe_id);
      if (filter->inverse){
        g_print("Doing Inverse\n");
	ros::Time now = ros::Time::now();
	trst2=tf::StampedTransform(trst.inverse(),now,
				 filter->childframe_id, filter->frame_id);
      }
      if (filter->inverse)filter->br->sendTransform(trst2);
      else filter->br->sendTransform(trst);
      if (!filter->silent) g_print ("Rosstfpub start spinning after sending tf stamped.\n");
      ros::spinOnce();

  return gst_pad_push (filter->srcpad, buf);
}



/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
//static gboolean
//rosTFPUB_init (GstPlugin * rosTFPUB)
//{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template rosTFPUB' with your description
   */
//  GST_DEBUG_CATEGORY_INIT (gst_rostfpub_debug, "rosTFPUB",
//      0, "Template rosTFPUB");

//  return gst_element_register (rosTFPUB, "rosTFPUB", GST_RANK_NONE,
//      GST_TYPE_ROSTFPUB);
//}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstrosTFPUB"
#endif

/* gstreamer looks for this structure to register rosTFPUBs
 *
 * exchange the string 'Template rosTFPUB' with your rosTFPUB description

GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "rosTFPUB",
    "Template rosTFPUB",
    rosTFPUB_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
) */
