/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2013 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_ROSUNDISTORT_H__
#define __GST_ROSUNDISTORT_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_ROSUNDISTORT \
  (gst_rosrosundistort_get_type())
#define GST_ROSUNDISTORT(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_ROSUNDISTORT,Gstrosrosundistort))
#define GST_ROSUNDISTORT_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_ROSUNDISTORT,GstrosrosundistortClass))
#define GST_IS_ROSUNDISTORT(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_ROSUNDISTORT))
#define GST_IS_ROSUNDISTORT_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_ROSUNDISTORT))

#define COEFS_SIZE sizeof(float)*9

typedef struct _Gstrosrosundistort      Gstrosrosundistort;
typedef struct _GstrosrosundistortClass GstrosrosundistortClass;

struct grasp
{
int x; //in mm
int y;
int z;
int pan;//reserved, zunächst nur horizontaler approach
int tilt;//reserved
int yaw;//reserved
int action;//reserved, later for release
};



struct _Gstrosrosundistort
{
  GstElement element;

  GstPad *sinkpad, *srcpad, *srccalimg,*srcgrasp;

  gboolean silent;
  int width, height,fps_n,fps_d;

  gchar* directory;
  gboolean ready;
  GstCaps* caps;

  IplImage* mapx;
  IplImage* mapy;
  float* mapmod;
  int mapmodsize;

  //stuff for learning
  CvMat* image_points;
  CvMat* object_points;
  CvMat* point_counts;
  CvMat* intrinsic_matrix;
  CvMat* distortion_coeffs;

  int board_w,board_h,n_boards,board_total;
  CvSize board_sz;
  gboolean firstcal;
  CvPoint2D32f* corners;
  int corner_count,successes;
  int step, frame;
  IplImage *gray_image;
  int frame_step;
  guint64 lasttaken;
  gboolean takenext;

  CvFont font;
  gboolean calws;
  gdouble squaresize;
  gboolean alwayscalws;  

  ros::NodeHandle* nh;
  ros::Publisher* rpub;
  tf::TransformBroadcaster* br;

  gchar* topic;
  gchar* frame_id;
  gchar* boardframe_id;
  gboolean firsttime;
  gchar* tipframe; 

  gboolean setbase;
  gboolean setx;
  gboolean sety;

  tf::StampedTransform* transformbase; 
  tf::StampedTransform* transformx; 
  tf::StampedTransform* transformy; 

  gboolean inverse;
  gboolean toworld;

  double manr;
  double manp;
  double many;

  double manpx;
  double manpy;
  double manpz;

  //tf::TransformListener* listener;

  gchar* caminfotopic; 
};

struct _GstrosrosundistortClass 
{
  GstElementClass parent_class;
};
 
GType gst_rosrosundistort_get_type (void);

G_END_DECLS

#endif /* __GST_ROSUNDISTORT_H__ */
