/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2006-2007  Rob Hess <hess@eecs.oregonstate.edu>
 * Copyright (C) 2010 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-siftextractor
 *
 * FIXME:Describe siftextractor here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! siftextractor ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

/*
#include "sift.h"
#include "imgfeatures.h"
#include "kdtree.h"
#include "utils.h"
#include "xform.h"
*/

#include <sys/time.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

#include <stdio.h>


#include "gstsiftextractor.h"

GST_DEBUG_CATEGORY_STATIC (gst_sift_extractor_debug);
#define GST_CAT_DEFAULT gst_sift_extractor_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_MODE,
  PROP_DOUBLE
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
     GST_PAD_SINK,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );


static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("sift/raw,"
        "framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ];"
)
    );//whether to pass sift information in raw/lowe/or oxf format

/*
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("video/sift,"
	"height = (int) [ 1, MAX ]")
    );
*/
/*
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
     GST_PAD_SRC,
     GST_PAD_ALWAYS,
     GST_STATIC_CAPS ("video/x-raw-gray, "
     "framerate = (fraction) [ 0, MAX ], "
     "depth = (int) 8, "
     "bpp = (int) 8, "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-rgb, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ];"
     "video/x-raw-yuv, "
     "format = (fourcc)Y800, "
     "framerate = (fraction) [ 0, MAX ], "
     "width = (int) [ 1, MAX ], "
     "height = (int) [ 1, MAX ]")
    );

*/


GST_BOILERPLATE (GstSiftExtractor, gst_sift_extractor, GstElement,
    GST_TYPE_ELEMENT);

static void gst_sift_extractor_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_sift_extractor_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_sift_extractor_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps *gst_sift_extractor_get_caps (GstPad *pad);
static GstFlowReturn gst_sift_extractor_chain (GstPad * pad, GstBuffer * buf);
static GstFlowReturn gst_sift_extractor_bufferalloc (GstPad * pad, guint64 offset,
    guint size, GstCaps * caps, GstBuffer ** buf);
//static void gst_sift_extractor_src_fixate (GstPad * pad, GstCaps * caps);
/* GObject vmethod implementations */
//first all SIFT related stuff:
void create_init_img( IplImage*, struct siftset*,int, double );
void convert_to_gray32( IplImage*, struct siftset*);
void build_pyr( struct siftset*, double );
CvSeq* scale_space_extrema( IplImage***, int, int, double, int, CvMemStorage*);
int is_extremum( IplImage***, int, int, int, int );
struct feature* interp_extremum( IplImage***, int, int, int, int, int, double);
void interp_step( IplImage***, int, int, int, int, double*, double*, double* );
CvMat* deriv_3D( IplImage***, int, int, int, int );
CvMat* hessian_3D( IplImage***, int, int, int, int );
double interp_contr( IplImage***, int, int, int, int, double, double, double );
struct feature* new_feature( void );
int is_too_edge_like( IplImage*, int, int, int );
void calc_feature_scales( CvSeq*, double, int );
void adjust_for_img_dbl( CvSeq* );
void calc_feature_oris( CvSeq*, IplImage*** );
double* ori_hist( IplImage*, int, int, int, int, double );
int calc_grad_mag_ori( IplImage*, int, int, double*, double* );
void smooth_ori_hist( double*, int );
double dominant_ori( double*, int );
void add_good_ori_features( CvSeq*, double*, int, double, struct feature* );
struct feature* clone_feature( struct feature* );
void compute_descriptors( CvSeq*, IplImage***, int, int );
double*** descr_hist( IplImage*, int, int, double, double, int, int );
void interp_hist_entry( double***, double, double, double, double, int, int);
void hist_to_descr( double***, int, int, struct feature* );
void normalize_descr( struct feature* );
int feature_cmp( void*, void*, void* );
void release_descr_hist( double****, int );
void release_pyr( IplImage****, int, int );

static inline float pixval32f( IplImage* img, int r, int c )
{
  return ( (float*)(img->imageData + img->widthStep*r) )[c];
}

static gboolean little_endian;


struct siftset* make_siftset(IplImage* img,int intvls,int img_dbl)
{
struct siftset* result=(struct siftset*)malloc(sizeof(struct siftset));
if( img->nChannels > 1 )result->gray8 = cvCreateImage( cvGetSize(img), IPL_DEPTH_8U, 1 );
else result->gray8 = NULL;
result->gray32 = cvCreateImage( cvGetSize(img), IPL_DEPTH_32F, 1 );
if (img_dbl) result->init_img=cvCreateImage( cvSize( img->width*2, img->height*2 ),IPL_DEPTH_32F, 1 );
else result->init_img=result->gray32;
//Oktavenzahl berechnen
result->octvs = log( MIN( result->init_img->width, result->init_img->height ) ) / log(2) - 2;
result->intvls=intvls;
result->dbl_img=img_dbl;
result->channels=img->nChannels;
//gauss-pyramide bauen
int i,o;
result->gauss_pyr = calloc( result->octvs, sizeof( IplImage** ) );
  for( i = 0; i < result->octvs; i++ )
    result->gauss_pyr[i] = calloc( result->intvls + 3, sizeof( IplImage *) );
  for( o = 0; o < result->octvs; o++ )
    for( i = 0; i < result->intvls + 3; i++ )
      {
	if( o == 0  &&  i == 0 )
	  result->gauss_pyr[o][i] = result->init_img;

	/* base of new octvave is halved image from end of previous octave */
	else if( i == 0 )
	  result->gauss_pyr[o][i] = cvCreateImage( cvSize((result->gauss_pyr[o-1][intvls] )->width / 2, (result->gauss_pyr[o-1][intvls])->height / 2),
				     IPL_DEPTH_32F, 1 );	  
	/* blur the current octave's last image to create the next one */
	else
	  {
	    result->gauss_pyr[o][i] = cvCreateImage( cvGetSize(result->gauss_pyr[o][i-1]),IPL_DEPTH_32F, 1 );
	  }
      }
//dog pyr bauen
  result->dog_pyr = calloc( result->octvs, sizeof( IplImage** ) );
  for( i = 0; i < result->octvs; i++ )
    result->dog_pyr[i] = calloc( result->intvls + 2, sizeof(IplImage*) );
  for( o = 0; o < result->octvs; o++ )
    for( i = 0; i < result->intvls + 2; i++ )
      {
	result->dog_pyr[o][i] = cvCreateImage( cvGetSize(result->gauss_pyr[o][i]),
				       IPL_DEPTH_32F, 1 );
	}
return result;
}

void killsiftset (struct siftset* siftset){
  release_pyr( &(siftset->gauss_pyr), siftset->octvs, siftset->intvls + 3 );
  release_pyr( &(siftset->dog_pyr), siftset->octvs, siftset->intvls + 2 );
if (siftset->dbl_img) cvReleaseImage( &siftset->gray32);//normalerweise schon weg weils in der gausspyr drinne war
if( siftset->channels > 1 ) cvReleaseImage( &siftset->gray8);
free(siftset);
}

int _sift_features( IplImage* img, struct feature** feat, struct siftset* siftset, int intvls,
		    double sigma, double contr_thr, int curv_thr,
		    int img_dbl, int descr_width, int descr_hist_bins )
{
  //IplImage*** gauss_pyr, *** dog_pyr;
  CvMemStorage* storage;
  CvSeq* features;
  int octvs, i, n = 0;
  /* check arguments */
  if( ! img )
    g_print( "IMG is NULL\n");
  if( ! feat )
    g_print("Feat is NULL\n");

  /* build scale space pyramid; smallest dimension of top level is ~4 pixels */

  create_init_img( img,siftset, img_dbl, sigma );
  octvs = log( MIN( siftset->init_img->width, siftset->init_img->height ) ) / log(2) - 2;
  build_pyr( siftset, sigma );
  
  
  storage = cvCreateMemStorage( 0 );
  features = scale_space_extrema( siftset->dog_pyr, siftset->octvs, siftset->intvls, contr_thr,
				  curv_thr, storage );
  calc_feature_scales( features, sigma, siftset->intvls );

  if( img_dbl )
    adjust_for_img_dbl( features );
  calc_feature_oris( features, siftset->gauss_pyr );
  compute_descriptors( features, siftset->gauss_pyr, descr_width, descr_hist_bins );

  /* sort features by decreasing scale and move from CvSeq to array */
  cvSeqSort( features, (CvCmpFunc)feature_cmp, NULL );
  n = features->total;
  //g_print("Feature init %p",*feat);
  *feat=NULL;
  if(n){
  *feat = calloc( n, sizeof(struct feature) );
  //g_print("Feature after calloc %p sets: %d",*feat,n);
  *feat = cvCvtSeqToArray( features, *feat, CV_WHOLE_SEQ );
  //g_print("Feature after cvCvtSeqToArray %p",*feat);
  }
  
  for( i = 0; i < n; i++ )
    {
      free( (*feat)[i].feature_data );
      (*feat)[i].feature_data = NULL;
    } 
  

  cvReleaseMemStorage( &storage );
  return n;
}

void create_init_img( IplImage* img, struct siftset* siftset, int img_dbl, double sigma )
{
  //IplImage* gray, * dbl;
  double sig_diff;

  convert_to_gray32( img, siftset );
  if( img_dbl )
    {
      sig_diff = sqrt( sigma * sigma - SIFT_INIT_SIGMA * SIFT_INIT_SIGMA * 4 );
      //dbl = cvCreateImage( cvSize( img->width*2, img->height*2 ),
	//		   IPL_DEPTH_32F, 1 );
      cvResize( siftset->gray32, siftset->init_img, CV_INTER_CUBIC );
      cvSmooth( siftset->init_img, siftset->init_img, CV_GAUSSIAN, 0, 0, sig_diff, sig_diff );
      //cvReleaseImage( &gray );
      //return dbl;
    }
  else
    {
      sig_diff = sqrt( sigma * sigma - SIFT_INIT_SIGMA * SIFT_INIT_SIGMA );
      cvSmooth( siftset->gray32, siftset->gray32, CV_GAUSSIAN, 0, 0, sig_diff, sig_diff );
      //return gray;
    }
return;
}



/*
  Converts an image to 32-bit grayscale

  @param img a 3-channel 8-bit color (BGR) or 8-bit gray image

  @return nothing as it writes the first 32-bit pic into the pyramid
*/
void convert_to_gray32( IplImage* img, struct siftset* siftset )
{
  if( img->nChannels == 1 )
    cvConvertScale( img, siftset->gray32, 1.0 / 255.0, 0 );
  else
    {
      //gray8 = cvCreateImage( cvGetSize(img), IPL_DEPTH_8U, 1 );
      cvCvtColor( img, siftset->gray8, CV_BGR2GRAY );
      cvConvertScale( siftset->gray8, siftset->gray32, 1.0 / 255.0, 0 );
    }
return;
}

void build_pyr( struct siftset* siftset, double sigma )
{
  //IplImage*** gauss_pyr;
  const int _intvls = siftset->intvls;
  double sig[_intvls+3], sig_total, sig_prev, k;
  int i, o;

  sig[0] = sigma;
  k = pow( 2.0, 1.0 / siftset->intvls );
  for( i = 1; i < siftset->intvls + 3; i++ )
    {
      sig_prev = pow( k, i - 1 ) * sigma;
      sig_total = sig_prev * k;
      sig[i] = sqrt( sig_total * sig_total - sig_prev * sig_prev );
    }
  for( o = 0; o < siftset->octvs; o++ )
    for( i = 0; i < siftset->intvls + 3; i++ )
      {
	if( o == 0  &&  i == 0 ){}

	/* base of new octvave is halved image from end of previous octave */
	else if( i == 0 )
	  //gauss_pyr[o][i] = downsample( gauss_pyr[o-1][intvls] );
	  cvResize(siftset->gauss_pyr[o-1][siftset->intvls],siftset->gauss_pyr[o][i],CV_INTER_NN );
	/* blur the current octave's last image to create the next one */
	else
	  {
	    cvSmooth( siftset->gauss_pyr[o][i-1], siftset->gauss_pyr[o][i],
		      CV_GAUSSIAN, 0, 0, sig[i], sig[i] );
	  }
      }

  //return gauss_pyr;
  //build dog pyr
  for( o = 0; o < siftset->octvs; o++ )
    for( i = 0; i < siftset->intvls + 2; i++ )
      {
	cvSub( siftset->gauss_pyr[o][i+1], siftset->gauss_pyr[o][i], siftset->dog_pyr[o][i], NULL );
      }
  return;
}


/*
  Detects features at extrema in DoG scale space.  Bad features are discarded
  based on contrast and ratio of principal curvatures.

  @param dog_pyr DoG scale space pyramid
  @param octvs octaves of scale space represented by dog_pyr
  @param intvls intervals per octave
  @param contr_thr low threshold on feature contrast
  @param curv_thr high threshold on feature ratio of principal curvatures
  @param storage memory storage in which to store detected features

  @return Returns an array of detected features whose scales, orientations,
    and descriptors are yet to be determined.
*/
CvSeq* scale_space_extrema( IplImage*** dog_pyr, int octvs, int intvls,
			    double contr_thr, int curv_thr,
			    CvMemStorage* storage )
{
  CvSeq* features;
  double prelim_contr_thr = 0.5 * contr_thr / intvls;
  struct feature* feat;
  struct detection_data* ddata;
  int o, i, r, c;

  features = cvCreateSeq( 0, sizeof(CvSeq), sizeof(struct feature), storage );
  for( o = 0; o < octvs; o++ )
    for( i = 1; i <= intvls; i++ )
      for(r = SIFT_IMG_BORDER; r < dog_pyr[o][0]->height-SIFT_IMG_BORDER; r++)
	for(c = SIFT_IMG_BORDER; c < dog_pyr[o][0]->width-SIFT_IMG_BORDER; c++)
	  /* perform preliminary check on contrast */
	  if( ABS( pixval32f( dog_pyr[o][i], r, c ) ) > prelim_contr_thr )
	    if( is_extremum( dog_pyr, o, i, r, c ) )
	      {
		feat = interp_extremum(dog_pyr, o, i, r, c, intvls, contr_thr);
		if( feat )
		  {
		    ddata = feat_detection_data( feat );
		    if( ! is_too_edge_like( dog_pyr[ddata->octv][ddata->intvl],
					    ddata->r, ddata->c, curv_thr ) )
		      {
			cvSeqPush( features, feat );
		      }
		    else
		      free( ddata );
		    free( feat );
		  }
	      }
  
  return features;
}



/*
  Determines whether a pixel is a scale-space extremum by comparing it to it's
  3x3x3 pixel neighborhood.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's scale space octave
  @param intvl pixel's within-octave interval
  @param r pixel's image row
  @param c pixel's image col

  @return Returns 1 if the specified pixel is an extremum (max or min) among
    it's 3x3x3 pixel neighborhood.
*/
int is_extremum( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  double val = pixval32f( dog_pyr[octv][intvl], r, c );
  int i, j, k;

  /* check for maximum */
  if( val > 0 )
    {
      for( i = -1; i <= 1; i++ )
	for( j = -1; j <= 1; j++ )
	  for( k = -1; k <= 1; k++ )
	    if( val < pixval32f( dog_pyr[octv][intvl+i], r + j, c + k ) )
	      return 0;
    }

  /* check for minimum */
  else
    {
      for( i = -1; i <= 1; i++ )
	for( j = -1; j <= 1; j++ )
	  for( k = -1; k <= 1; k++ )
	    if( val > pixval32f( dog_pyr[octv][intvl+i], r + j, c + k ) )
	      return 0;
    }

  return 1;
}



/*
  Interpolates a scale-space extremum's location and scale to subpixel
  accuracy to form an image feature.  Rejects features with low contrast.
  Based on Section 4 of Lowe's paper.  

  @param dog_pyr DoG scale space pyramid
  @param octv feature's octave of scale space
  @param intvl feature's within-octave interval
  @param r feature's image row
  @param c feature's image column
  @param intvls total intervals per octave
  @param contr_thr threshold on feature contrast

  @return Returns the feature resulting from interpolation of the given
    parameters or NULL if the given location could not be interpolated or
    if contrast at the interpolated loation was too low.  If a feature is
    returned, its scale, orientation, and descriptor are yet to be determined.
*/
struct feature* interp_extremum( IplImage*** dog_pyr, int octv, int intvl,
				 int r, int c, int intvls, double contr_thr )
{
  struct feature* feat;
  struct detection_data* ddata;
  double xi, xr, xc, contr;
  int i = 0;
  
  while( i < SIFT_MAX_INTERP_STEPS )
    {
      interp_step( dog_pyr, octv, intvl, r, c, &xi, &xr, &xc );
      if( ABS( xi ) < 0.5  &&  ABS( xr ) < 0.5  &&  ABS( xc ) < 0.5 )
	break;
      
      c += cvRound( xc );
      r += cvRound( xr );
      intvl += cvRound( xi );
      
      if( intvl < 1  ||
	  intvl > intvls  ||
	  c < SIFT_IMG_BORDER  ||
	  r < SIFT_IMG_BORDER  ||
	  c >= dog_pyr[octv][0]->width - SIFT_IMG_BORDER  ||
	  r >= dog_pyr[octv][0]->height - SIFT_IMG_BORDER )
	{
	  return NULL;
	}
      
      i++;
    }
  
  /* ensure convergence of interpolation */
  if( i >= SIFT_MAX_INTERP_STEPS )
    return NULL;
  
  contr = interp_contr( dog_pyr, octv, intvl, r, c, xi, xr, xc );
  if( ABS( contr ) < contr_thr / intvls )
    return NULL;

  feat = new_feature();
  ddata = feat_detection_data( feat );
  feat->img_pt.x = feat->x = ( c + xc ) * pow( 2.0, octv );
  feat->img_pt.y = feat->y = ( r + xr ) * pow( 2.0, octv );
  ddata->r = r;
  ddata->c = c;
  ddata->octv = octv;
  ddata->intvl = intvl;
  ddata->subintvl = xi;

  return feat;
}



/*
  Performs one step of extremum interpolation.  Based on Eqn. (3) in Lowe's
  paper.

  @param dog_pyr difference of Gaussians scale space pyramid
  @param octv octave of scale space
  @param intvl interval being interpolated
  @param r row being interpolated
  @param c column being interpolated
  @param xi output as interpolated subpixel increment to interval
  @param xr output as interpolated subpixel increment to row
  @param xc output as interpolated subpixel increment to col
*/

void interp_step( IplImage*** dog_pyr, int octv, int intvl, int r, int c,
		  double* xi, double* xr, double* xc )
{
  CvMat* dD, * H, * H_inv, X;
  double x[3] = { 0 };
  
  dD = deriv_3D( dog_pyr, octv, intvl, r, c );
  H = hessian_3D( dog_pyr, octv, intvl, r, c );
  H_inv = cvCreateMat( 3, 3, CV_64FC1 );
  cvInvert( H, H_inv, CV_SVD );
  cvInitMatHeader( &X, 3, 1, CV_64FC1, x, CV_AUTOSTEP );
  cvGEMM( H_inv, dD, -1, NULL, 0, &X, 0 );
  
  cvReleaseMat( &dD );
  cvReleaseMat( &H );
  cvReleaseMat( &H_inv );

  *xi = x[2];
  *xr = x[1];
  *xc = x[0];
}



/*
  Computes the partial derivatives in x, y, and scale of a pixel in the DoG
  scale space pyramid.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's octave in dog_pyr
  @param intvl pixel's interval in octv
  @param r pixel's image row
  @param c pixel's image col

  @return Returns the vector of partial derivatives for pixel I
    { dI/dx, dI/dy, dI/ds }^T as a CvMat*
*/
CvMat* deriv_3D( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  CvMat* dI;
  double dx, dy, ds;

  dx = ( pixval32f( dog_pyr[octv][intvl], r, c+1 ) -
	 pixval32f( dog_pyr[octv][intvl], r, c-1 ) ) / 2.0;
  dy = ( pixval32f( dog_pyr[octv][intvl], r+1, c ) -
	 pixval32f( dog_pyr[octv][intvl], r-1, c ) ) / 2.0;
  ds = ( pixval32f( dog_pyr[octv][intvl+1], r, c ) -
	 pixval32f( dog_pyr[octv][intvl-1], r, c ) ) / 2.0;
  
  dI = cvCreateMat( 3, 1, CV_64FC1 );
  cvmSet( dI, 0, 0, dx );
  cvmSet( dI, 1, 0, dy );
  cvmSet( dI, 2, 0, ds );

  return dI;
}



/*
  Computes the 3D Hessian matrix for a pixel in the DoG scale space pyramid.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's octave in dog_pyr
  @param intvl pixel's interval in octv
  @param r pixel's image row
  @param c pixel's image col

  @return Returns the Hessian matrix (below) for pixel I as a CvMat*

  / Ixx  Ixy  Ixs \ <BR>
  | Ixy  Iyy  Iys | <BR>
  \ Ixs  Iys  Iss /
*/
CvMat* hessian_3D( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  CvMat* H;
  double v, dxx, dyy, dss, dxy, dxs, dys;
  
  v = pixval32f( dog_pyr[octv][intvl], r, c );
  dxx = ( pixval32f( dog_pyr[octv][intvl], r, c+1 ) + 
	  pixval32f( dog_pyr[octv][intvl], r, c-1 ) - 2 * v );
  dyy = ( pixval32f( dog_pyr[octv][intvl], r+1, c ) +
	  pixval32f( dog_pyr[octv][intvl], r-1, c ) - 2 * v );
  dss = ( pixval32f( dog_pyr[octv][intvl+1], r, c ) +
	  pixval32f( dog_pyr[octv][intvl-1], r, c ) - 2 * v );
  dxy = ( pixval32f( dog_pyr[octv][intvl], r+1, c+1 ) -
	  pixval32f( dog_pyr[octv][intvl], r+1, c-1 ) -
	  pixval32f( dog_pyr[octv][intvl], r-1, c+1 ) +
	  pixval32f( dog_pyr[octv][intvl], r-1, c-1 ) ) / 4.0;
  dxs = ( pixval32f( dog_pyr[octv][intvl+1], r, c+1 ) -
	  pixval32f( dog_pyr[octv][intvl+1], r, c-1 ) -
	  pixval32f( dog_pyr[octv][intvl-1], r, c+1 ) +
	  pixval32f( dog_pyr[octv][intvl-1], r, c-1 ) ) / 4.0;
  dys = ( pixval32f( dog_pyr[octv][intvl+1], r+1, c ) -
	  pixval32f( dog_pyr[octv][intvl+1], r-1, c ) -
	  pixval32f( dog_pyr[octv][intvl-1], r+1, c ) +
	  pixval32f( dog_pyr[octv][intvl-1], r-1, c ) ) / 4.0;
  
  H = cvCreateMat( 3, 3, CV_64FC1 );
  cvmSet( H, 0, 0, dxx );
  cvmSet( H, 0, 1, dxy );
  cvmSet( H, 0, 2, dxs );
  cvmSet( H, 1, 0, dxy );
  cvmSet( H, 1, 1, dyy );
  cvmSet( H, 1, 2, dys );
  cvmSet( H, 2, 0, dxs );
  cvmSet( H, 2, 1, dys );
  cvmSet( H, 2, 2, dss );

  return H;
}



/*
  Calculates interpolated pixel contrast.  Based on Eqn. (3) in Lowe's
  paper.

  @param dog_pyr difference of Gaussians scale space pyramid
  @param octv octave of scale space
  @param intvl within-octave interval
  @param r pixel row
  @param c pixel column
  @param xi interpolated subpixel increment to interval
  @param xr interpolated subpixel increment to row
  @param xc interpolated subpixel increment to col

  @param Returns interpolated contrast.
*/
double interp_contr( IplImage*** dog_pyr, int octv, int intvl, int r,
		     int c, double xi, double xr, double xc )
{
  CvMat* dD, X, T;
  double t[1], x[3] = { xc, xr, xi };

  cvInitMatHeader( &X, 3, 1, CV_64FC1, x, CV_AUTOSTEP );
  cvInitMatHeader( &T, 1, 1, CV_64FC1, t, CV_AUTOSTEP );
  dD = deriv_3D( dog_pyr, octv, intvl, r, c );
  cvGEMM( dD, &X, 1, NULL, 0, &T,  CV_GEMM_A_T );
  cvReleaseMat( &dD );

  return pixval32f( dog_pyr[octv][intvl], r, c ) + t[0] * 0.5;
}



/*
  Allocates and initializes a new feature

  @return Returns a pointer to the new feature
*/
struct feature* new_feature( void )
{
  struct feature* feat;
  struct detection_data* ddata;

  feat = malloc( sizeof( struct feature ) );
  memset( feat, 0, sizeof( struct feature ) );
  ddata = malloc( sizeof( struct detection_data ) );
  memset( ddata, 0, sizeof( struct detection_data ) );
  feat->feature_data = ddata;
  feat->type = FEATURE_LOWE;

  return feat;
}



/*
  Determines whether a feature is too edge like to be stable by computing the
  ratio of principal curvatures at that feature.  Based on Section 4.1 of
  Lowe's paper.

  @param dog_img image from the DoG pyramid in which feature was detected
  @param r feature row
  @param c feature col
  @param curv_thr high threshold on ratio of principal curvatures

  @return Returns 0 if the feature at (r,c) in dog_img is sufficiently
    corner-like or 1 otherwise.
*/
int is_too_edge_like( IplImage* dog_img, int r, int c, int curv_thr )
{
  double d, dxx, dyy, dxy, tr, det;

  /* principal curvatures are computed using the trace and det of Hessian */
  d = pixval32f(dog_img, r, c);
  dxx = pixval32f( dog_img, r, c+1 ) + pixval32f( dog_img, r, c-1 ) - 2 * d;
  dyy = pixval32f( dog_img, r+1, c ) + pixval32f( dog_img, r-1, c ) - 2 * d;
  dxy = ( pixval32f(dog_img, r+1, c+1) - pixval32f(dog_img, r+1, c-1) -
	  pixval32f(dog_img, r-1, c+1) + pixval32f(dog_img, r-1, c-1) ) / 4.0;
  tr = dxx + dyy;
  det = dxx * dyy - dxy * dxy;

  /* negative determinant -> curvatures have different signs; reject feature */
  if( det <= 0 )
    return 1;

  if( tr * tr / det < ( curv_thr + 1.0 )*( curv_thr + 1.0 ) / curv_thr )
    return 0;
  return 1;
}



/*
  Calculates characteristic scale for each feature in an array.

  @param features array of features
  @param sigma amount of Gaussian smoothing per octave of scale space
  @param intvls intervals per octave of scale space
*/
void calc_feature_scales( CvSeq* features, double sigma, int intvls )
{
  struct feature* feat;
  struct detection_data* ddata;
  double intvl;
  int i, n;

  n = features->total;
  for( i = 0; i < n; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      ddata = feat_detection_data( feat );
      intvl = ddata->intvl + ddata->subintvl;
      feat->scl = sigma * pow( 2.0, ddata->octv + intvl / intvls );
      ddata->scl_octv = sigma * pow( 2.0, intvl / intvls );
    }
}



/*
  Halves feature coordinates and scale in case the input image was doubled
  prior to scale space construction.

  @param features array of features
*/
void adjust_for_img_dbl( CvSeq* features )
{
  struct feature* feat;
  int i, n;

  n = features->total;
  for( i = 0; i < n; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      feat->x /= 2.0;
      feat->y /= 2.0;
      feat->scl /= 2.0;
      feat->img_pt.x /= 2.0;
      feat->img_pt.y /= 2.0;
    }
}



/*
  Computes a canonical orientation for each image feature in an array.  Based
  on Section 5 of Lowe's paper.  This function adds features to the array when
  there is more than one dominant orientation at a given feature location.

  @param features an array of image features
  @param gauss_pyr Gaussian scale space pyramid
*/
void calc_feature_oris( CvSeq* features, IplImage*** gauss_pyr )
{
  struct feature* feat;
  struct detection_data* ddata;
  double* hist;
  double omax;
  int i, j, n = features->total;

  for( i = 0; i < n; i++ )
    {
      feat = malloc( sizeof( struct feature ) );
      cvSeqPopFront( features, feat );
      ddata = feat_detection_data( feat );
      hist = ori_hist( gauss_pyr[ddata->octv][ddata->intvl],
		       ddata->r, ddata->c, SIFT_ORI_HIST_BINS,
		       cvRound( SIFT_ORI_RADIUS * ddata->scl_octv ),
		       SIFT_ORI_SIG_FCTR * ddata->scl_octv );
      for( j = 0; j < SIFT_ORI_SMOOTH_PASSES; j++ )
	smooth_ori_hist( hist, SIFT_ORI_HIST_BINS );
      omax = dominant_ori( hist, SIFT_ORI_HIST_BINS );
      add_good_ori_features( features, hist, SIFT_ORI_HIST_BINS,
			     omax * SIFT_ORI_PEAK_RATIO, feat );
      free( ddata );
      free( feat );
      free( hist );
    }
}



/*
  Computes a gradient orientation histogram at a specified pixel.

  @param img image
  @param r pixel row
  @param c pixel col
  @param n number of histogram bins
  @param rad radius of region over which histogram is computed
  @param sigma std for Gaussian weighting of histogram entries

  @return Returns an n-element array containing an orientation histogram
    representing orientations between 0 and 2 PI.
*/
double* ori_hist( IplImage* img, int r, int c, int n, int rad, double sigma)
{
  double* hist;
  double mag, ori, w, exp_denom, PI2 = CV_PI * 2.0;
  int bin, i, j;

  hist = calloc( n, sizeof( double ) );
  exp_denom = 2.0 * sigma * sigma;
  for( i = -rad; i <= rad; i++ )
    for( j = -rad; j <= rad; j++ )
      if( calc_grad_mag_ori( img, r + i, c + j, &mag, &ori ) )
	{
	  w = exp( -( i*i + j*j ) / exp_denom );
	  bin = cvRound( n * ( ori + CV_PI ) / PI2 );
	  bin = ( bin < n )? bin : 0;
	  hist[bin] += w * mag;
	}

  return hist;
}



/*
  Calculates the gradient magnitude and orientation at a given pixel.

  @param img image
  @param r pixel row
  @param c pixel col
  @param mag output as gradient magnitude at pixel (r,c)
  @param ori output as gradient orientation at pixel (r,c)

  @return Returns 1 if the specified pixel is a valid one and sets mag and
    ori accordingly; otherwise returns 0
*/
int calc_grad_mag_ori( IplImage* img, int r, int c, double* mag, double* ori )
{
  double dx, dy;

  if( r > 0  &&  r < img->height - 1  &&  c > 0  &&  c < img->width - 1 )
    {
      dx = pixval32f( img, r, c+1 ) - pixval32f( img, r, c-1 );
      dy = pixval32f( img, r-1, c ) - pixval32f( img, r+1, c );
      *mag = sqrt( dx*dx + dy*dy );
      *ori = atan2( dy, dx );
      return 1;
    }

  else
    return 0;
}



/*
  Gaussian smooths an orientation histogram.

  @param hist an orientation histogram
  @param n number of bins
*/
void smooth_ori_hist( double* hist, int n )
{
  double prev, tmp, h0 = hist[0];
  int i;

  prev = hist[n-1];
  for( i = 0; i < n; i++ )
    {
      tmp = hist[i];
      hist[i] = 0.25 * prev + 0.5 * hist[i] + 
	0.25 * ( ( i+1 == n )? h0 : hist[i+1] );
      prev = tmp;
    }
}



/*
  Finds the magnitude of the dominant orientation in a histogram

  @param hist an orientation histogram
  @param n number of bins

  @return Returns the value of the largest bin in hist
*/
double dominant_ori( double* hist, int n )
{
  double omax;
  int maxbin, i;

  omax = hist[0];
  maxbin = 0;
  for( i = 1; i < n; i++ )
    if( hist[i] > omax )
      {
	omax = hist[i];
	maxbin = i;
      }
  return omax;
}



/*
  Interpolates a histogram peak from left, center, and right values
*/
#define interp_hist_peak( l, c, r ) ( 0.5 * ((l)-(r)) / ((l) - 2.0*(c) + (r)) )



/*
  Adds features to an array for every orientation in a histogram greater than
  a specified threshold.

  @param features new features are added to the end of this array
  @param hist orientation histogram
  @param n number of bins in hist
  @param mag_thr new features are added for entries in hist greater than this
  @param feat new features are clones of this with different orientations
*/
void add_good_ori_features( CvSeq* features, double* hist, int n,
			    double mag_thr, struct feature* feat )
{
  struct feature* new_feat;
  double bin, PI2 = CV_PI * 2.0;
  int l, r, i;

  for( i = 0; i < n; i++ )
    {
      l = ( i == 0 )? n - 1 : i-1;
      r = ( i + 1 ) % n;
      
      if( hist[i] > hist[l]  &&  hist[i] > hist[r]  &&  hist[i] >= mag_thr )
	{
	  bin = i + interp_hist_peak( hist[l], hist[i], hist[r] );
	  bin = ( bin < 0 )? n + bin : ( bin >= n )? bin - n : bin;
	  new_feat = clone_feature( feat );
	  new_feat->ori = ( ( PI2 * bin ) / n ) - CV_PI;
	  cvSeqPush( features, new_feat );
	  free( new_feat );
	}
    }
}



/*
  Makes a deep copy of a feature

  @param feat feature to be cloned

  @return Returns a deep copy of feat
*/
struct feature* clone_feature( struct feature* feat )
{
  struct feature* new_feat;
  struct detection_data* ddata;

  new_feat = new_feature();
  ddata = feat_detection_data( new_feat );
  memcpy( new_feat, feat, sizeof( struct feature ) );
  memcpy( ddata, feat_detection_data(feat), sizeof( struct detection_data ) );
  new_feat->feature_data = ddata;

  return new_feat;
}



/*
  Computes feature descriptors for features in an array.  Based on Section 6
  of Lowe's paper.

  @param features array of features
  @param gauss_pyr Gaussian scale space pyramid
  @param d width of 2D array of orientation histograms
  @param n number of bins per orientation histogram
*/
void compute_descriptors( CvSeq* features, IplImage*** gauss_pyr, int d, int n)
{
  struct feature* feat;
  struct detection_data* ddata;
  double*** hist;
  int i, k = features->total;

  for( i = 0; i < k; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      ddata = feat_detection_data( feat );
      hist = descr_hist( gauss_pyr[ddata->octv][ddata->intvl], ddata->r,
			 ddata->c, feat->ori, ddata->scl_octv, d, n );
      hist_to_descr( hist, d, n, feat );
      release_descr_hist( &hist, d );
    }
}



/*
  Computes the 2D array of orientation histograms that form the feature
  descriptor.  Based on Section 6.1 of Lowe's paper.

  @param img image used in descriptor computation
  @param r row coord of center of orientation histogram array
  @param c column coord of center of orientation histogram array
  @param ori canonical orientation of feature whose descr is being computed
  @param scl scale relative to img of feature whose descr is being computed
  @param d width of 2d array of orientation histograms
  @param n bins per orientation histogram

  @return Returns a d x d array of n-bin orientation histograms.
*/
double*** descr_hist( IplImage* img, int r, int c, double ori,
		      double scl, int d, int n )
{
  double*** hist;
  double cos_t, sin_t, hist_width, exp_denom, r_rot, c_rot, grad_mag,
    grad_ori, w, rbin, cbin, obin, bins_per_rad, PI2 = 2.0 * CV_PI;
  int radius, i, j;

  hist = calloc( d, sizeof( double** ) );
  for( i = 0; i < d; i++ )
    {
      hist[i] = calloc( d, sizeof( double* ) );
      for( j = 0; j < d; j++ )
	hist[i][j] = calloc( n, sizeof( double ) );
    }
  
  cos_t = cos( ori );
  sin_t = sin( ori );
  bins_per_rad = n / PI2;
  exp_denom = d * d * 0.5;
  hist_width = SIFT_DESCR_SCL_FCTR * scl;
  radius = hist_width * sqrt(2) * ( d + 1.0 ) * 0.5 + 0.5;
  for( i = -radius; i <= radius; i++ )
    for( j = -radius; j <= radius; j++ )
      {
	/*
	  Calculate sample's histogram array coords rotated relative to ori.
	  Subtract 0.5 so samples that fall e.g. in the center of row 1 (i.e.
	  r_rot = 1.5) have full weight placed in row 1 after interpolation.
	*/
	c_rot = ( j * cos_t - i * sin_t ) / hist_width;
	r_rot = ( j * sin_t + i * cos_t ) / hist_width;
	rbin = r_rot + d / 2 - 0.5;
	cbin = c_rot + d / 2 - 0.5;
	
	if( rbin > -1.0  &&  rbin < d  &&  cbin > -1.0  &&  cbin < d )
	  if( calc_grad_mag_ori( img, r + i, c + j, &grad_mag, &grad_ori ))
	    {
	      grad_ori -= ori;
	      while( grad_ori < 0.0 )
		grad_ori += PI2;
	      while( grad_ori >= PI2 )
		grad_ori -= PI2;
	      
	      obin = grad_ori * bins_per_rad;
	      w = exp( -(c_rot * c_rot + r_rot * r_rot) / exp_denom );
	      interp_hist_entry( hist, rbin, cbin, obin, grad_mag * w, d, n );
	    }
      }

  return hist;
}



/*
  Interpolates an entry into the array of orientation histograms that form
  the feature descriptor.

  @param hist 2D array of orientation histograms
  @param rbin sub-bin row coordinate of entry
  @param cbin sub-bin column coordinate of entry
  @param obin sub-bin orientation coordinate of entry
  @param mag size of entry
  @param d width of 2D array of orientation histograms
  @param n number of bins per orientation histogram
*/
void interp_hist_entry( double*** hist, double rbin, double cbin,
			double obin, double mag, int d, int n )
{
  double d_r, d_c, d_o, v_r, v_c, v_o;
  double** row, * h;
  int r0, c0, o0, rb, cb, ob, r, c, o;

  r0 = cvFloor( rbin );
  c0 = cvFloor( cbin );
  o0 = cvFloor( obin );
  d_r = rbin - r0;
  d_c = cbin - c0;
  d_o = obin - o0;

  /*
    The entry is distributed into up to 8 bins.  Each entry into a bin
    is multiplied by a weight of 1 - d for each dimension, where d is the
    distance from the center value of the bin measured in bin units.
  */
  for( r = 0; r <= 1; r++ )
    {
      rb = r0 + r;
      if( rb >= 0  &&  rb < d )
	{
	  v_r = mag * ( ( r == 0 )? 1.0 - d_r : d_r );
	  row = hist[rb];
	  for( c = 0; c <= 1; c++ )
	    {
	      cb = c0 + c;
	      if( cb >= 0  &&  cb < d )
		{
		  v_c = v_r * ( ( c == 0 )? 1.0 - d_c : d_c );
		  h = row[cb];
		  for( o = 0; o <= 1; o++ )
		    {
		      ob = ( o0 + o ) % n;
		      v_o = v_c * ( ( o == 0 )? 1.0 - d_o : d_o );
		      h[ob] += v_o;
		    }
		}
	    }
	}
    }
}



/*
  Converts the 2D array of orientation histograms into a feature's descriptor
  vector.
  
  @param hist 2D array of orientation histograms
  @param d width of hist
  @param n bins per histogram
  @param feat feature into which to store descriptor
*/
void hist_to_descr( double*** hist, int d, int n, struct feature* feat )
{
  int int_val, i, r, c, o, k = 0;

  for( r = 0; r < d; r++ )
    for( c = 0; c < d; c++ )
      for( o = 0; o < n; o++ )
	feat->descr[k++] = hist[r][c][o];

  feat->d = k;
  normalize_descr( feat );
  for( i = 0; i < k; i++ )
    if( feat->descr[i] > SIFT_DESCR_MAG_THR )
      feat->descr[i] = SIFT_DESCR_MAG_THR;
  normalize_descr( feat );

  /* convert floating-point descriptor to integer valued descriptor */
  for( i = 0; i < k; i++ )
    {
      int_val = SIFT_INT_DESCR_FCTR * feat->descr[i];
      feat->descr[i] = MIN( 255, int_val );
    }
}



/*
  Normalizes a feature's descriptor vector to unitl length

  @param feat feature
*/
void normalize_descr( struct feature* feat )
{
  double cur, len_inv, len_sq = 0.0;
  int i, d = feat->d;

  for( i = 0; i < d; i++ )
    {
      cur = feat->descr[i];
      len_sq += cur*cur;
    }
  len_inv = 1.0 / sqrt( len_sq );
  for( i = 0; i < d; i++ )
    feat->descr[i] *= len_inv;
}



/*
  Compares features for a decreasing-scale ordering.  Intended for use with
  CvSeqSort

  @param feat1 first feature
  @param feat2 second feature
  @param param unused

  @return Returns 1 if feat1's scale is greater than feat2's, -1 if vice versa,
    and 0 if their scales are equal
*/
int feature_cmp( void* feat1, void* feat2, void* param )
{
  struct feature* f1 = (struct feature*) feat1;
  struct feature* f2 = (struct feature*) feat2;

  if( f1->scl < f2->scl )
    return 1;
  if( f1->scl > f2->scl )
    return -1;
  return 0;
}



/*
  De-allocates memory held by a descriptor histogram

  @param hist pointer to a 2D array of orientation histograms
  @param d width of hist
*/
void release_descr_hist( double**** hist, int d )
{
  int i, j;

  for( i = 0; i < d; i++)
    {
      for( j = 0; j < d; j++ )
	free( (*hist)[i][j] );
      free( (*hist)[i] );
    }
  free( *hist );
  *hist = NULL;
}

/*
  De-allocates memory held by a scale space pyramid

  @param pyr scale space pyramid
  @param octvs number of octaves of scale space
  @param n number of images per octave
*/
void release_pyr( IplImage**** pyr, int octvs, int n )
{
  int i, j;
  for( i = 0; i < octvs; i++ )
    {
      for( j = 0; j < n; j++ )
	cvReleaseImage( &(*pyr)[i][j] );
      free( (*pyr)[i] );
    }
  free( *pyr );
  *pyr = NULL;
}
//now Gstreamer stuff
static void
gst_sift_extractor_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "SiftExtractor",
    "Gst-Plugins-Robotix",
    "Extract Sift-Vector from images",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the siftextractor's class */
static void
gst_sift_extractor_class_init (GstSiftExtractorClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_sift_extractor_set_property;
  gobject_class->get_property = gst_sift_extractor_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MODE,
      g_param_spec_int ("mode", "Mode", "0=raw, 1=lowe, 2=oxford 3=lowelight",
          0,3,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DOUBLE,
      g_param_spec_boolean ("double", "Double", "Double image size? (slow but more features)",
          FALSE, G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_sift_extractor_init (GstSiftExtractor * filter,
    GstSiftExtractorClass * gclass)
{
  int x=1;
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_sift_extractor_get_caps));
  //gst_pad_set_bufferalloc_function (filter->sinkpad,
  //    GST_DEBUG_FUNCPTR (gst_sift_extractor_bufferalloc));
  //gst_pad_set_getcaps_function (filter->sinkpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));  
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_sift_extractor_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_setcaps_function (filter->srcpad,
                               GST_DEBUG_FUNCPTR(gst_sift_extractor_set_caps));
  gst_pad_set_getcaps_function (filter->srcpad,
                               GST_DEBUG_FUNCPTR(gst_sift_extractor_get_caps));
  //gst_pad_set_getcaps_function (filter->srcpad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));  
  //gst_pad_set_fixatecaps_function (filter->srcpad, gst_sift_extractor_src_fixate);
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->imagesize.height=1;
  filter->imagesize.width=1;
  filter->rate_numerator=1;
  filter->rate_denominator=1;
  filter->overlap_x=-1;
  filter->overlap_y=-1;  
  filter->mutex=g_mutex_new();
  little_endian=(gboolean) (*(char*)&x); 
}

static void
gst_sift_extractor_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstSiftExtractor *filter = GST_SIFTEXTRACTOR (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_MODE:
      filter->mode = g_value_get_int (value);
      break;
    case PROP_DOUBLE:
      g_mutex_lock(filter->mutex);//otherwise dangerous
      filter->double_image = g_value_get_boolean (value);
      if (filter->siftset)killsiftset(filter->siftset);
      filter->siftset=NULL;
      g_mutex_unlock(filter->mutex);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_sift_extractor_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstSiftExtractor *filter = GST_SIFTEXTRACTOR (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_MODE:
      g_value_set_int (value, filter->mode);
      break;
    case PROP_DOUBLE:
      g_value_set_boolean (value, filter->double_image);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}


static gboolean
gst_cv_parse_caps (GstPad * pad, GstCaps * caps){//copy framerate, width and height
  GstSiftExtractor *filter;
  GstPad *otherpad;

  filter = GST_SIFTEXTRACTOR (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  GstStructure *structure = gst_caps_get_structure (caps, 0);

  if (!gst_structure_get_int(structure,"width",&(filter->imagesize.width))){
	g_print ("Width not set!");
	return FALSE;
  }
  if (!gst_structure_get_int(structure,"height",&(filter->imagesize.height))){
	g_print ("Height not set!");
	return FALSE;
  }


  const GValue *framerate;
  framerate = gst_structure_get_value(structure, "framerate");
  if (framerate) {
  filter->rate_numerator = gst_value_get_fraction_numerator (framerate);
  filter->rate_denominator = gst_value_get_fraction_denominator (framerate);
  } else{ 
	g_print ("No framerate specified!");
	return FALSE;
  }
  return TRUE;
}





static gboolean acceptonechannel(GstPad * pad, GstCaps * caps){
  GstSiftExtractor *filter;
  GstPad *otherpad;

  filter = GST_SIFTEXTRACTOR (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  const gchar *mime;
  gint temp;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  

if (0 == strcmp (mime, "video/x-raw-yuv")){ 
	guint32 fourccval;
	if (!gst_structure_has_field(structure,"format")){
		g_print ("No format field");
		return FALSE;
	}  
	if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
		g_print ("FourCC could not be read");
		return FALSE;
	}
	if (!fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
		g_print ("Wrong format suggested: %" GST_FOURCC_FORMAT "\n", GST_FOURCC_ARGS (fourccval));
		return FALSE;
	}
	filter->channels=1;

	if (!gst_cv_parse_caps(pad,caps)) return FALSE; 

	g_print ("Caps negotiation success!");
	return TRUE;
}
else if (0 == strcmp (mime, "video/x-raw-gray")){
	filter->channels=1;
	if (!gst_structure_get_int(structure,"depth",&temp)) return FALSE;
	if (temp!=8) return FALSE;
	if (!gst_structure_get_int(structure,"bpp",&temp)) return FALSE;
	if (temp!=8) return FALSE;
	if (gst_structure_get_int(structure,"overlap_v",&temp)) filter->overlap_x=temp;
	if (gst_structure_get_int(structure,"overlap_h",&temp)) filter->overlap_y=temp; 
	return gst_cv_parse_caps(pad,caps);
 }
else {			
	g_print ("Wrong mimetype %s provided, we only support %s",mime, "video/x-raw-yuv");
	return FALSE;
}
return FALSE;
}

static gboolean acceptrgb(GstPad * pad, GstCaps * caps){
  GstSiftExtractor *filter;
  GstPad *otherpad;

  filter = GST_SIFTEXTRACTOR (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  const gchar *mime;
  GstStructure *structure=gst_caps_get_structure(caps,0);
  mime = gst_structure_get_name (structure);  
	if (0 == strcmp (mime, "video/x-raw-rgb")){
		int temp;
		if (!gst_structure_get_int(structure,"depth",&temp)) return FALSE;
		if ((temp!=24)&&(temp!=32)) return FALSE;
		//filter->indepth=8;
		if (!gst_structure_get_int(structure,"bpp",&temp)) return FALSE;
		if ((temp!=24)&&(temp!=32)) return FALSE;
		filter->inbpp=temp;
		if (gst_structure_get_int(structure,"overlap_v",&temp)) filter->overlap_x=temp;
		if (gst_structure_get_int(structure,"overlap_h",&temp)) filter->overlap_y=temp; 		

		if (!gst_cv_parse_caps(pad,caps)) return FALSE; 
		//if (!(gst_pad_set_caps (otherpad, caps))) {g_print ("Failed, next element does not support RGB!");return FALSE;}
			//same input and outpuf format for canny
		g_print ("Caps negotiation success with RGB!");
		return TRUE;
	}
  return FALSE;
}




/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_sift_extractor_set_caps (GstPad * pad, GstCaps * caps)
{
  gchar* capschar=gst_caps_to_string (caps);
  g_print ("SiftExtractor SetCaps aufgerufen with %s\n",capschar);
  g_free(capschar);
  GstSiftExtractor *filter;
  GstPad *otherpad;
  gboolean ret=FALSE;
  filter = GST_SIFTEXTRACTOR (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

  if (pad == filter->sinkpad){
     if (filter->siftset)killsiftset(filter->siftset); 
     filter->siftset=NULL;  
     ret=acceptonechannel(pad, caps);
     if (!ret) ret=acceptrgb(pad, caps);
     if (ret){
        if(filter->mode==0) ret=gst_pad_set_caps (otherpad,gst_caps_new_simple ("sift/raw",
      		//"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		//"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		//"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
      		"width", G_TYPE_INT, filter->imagesize.width,
      		"height", G_TYPE_INT, filter->imagesize.height,
      		"framerate", GST_TYPE_FRACTION, filter->rate_numerator,filter->rate_denominator,
      		"overlap_v", G_TYPE_INT, filter->overlap_x,
      		"overlap_h", G_TYPE_INT, filter->overlap_y,
		NULL));
	else if (filter->mode==1) ret=gst_pad_set_caps (otherpad,gst_caps_new_simple ("sift/lowe",
		NULL));
        else if (filter->mode==2) ret=gst_pad_set_caps (otherpad,gst_caps_new_simple ("sift/oxf",
		NULL));
        else if (filter->mode==3) ret=gst_pad_set_caps (otherpad,gst_caps_new_simple ("sift/lowelight",
      		"width", G_TYPE_INT, filter->imagesize.width,
      		"height", G_TYPE_INT, filter->imagesize.height,
      		"overlap_v", G_TYPE_INT, filter->overlap_x,
      		"overlap_h", G_TYPE_INT, filter->overlap_y,
		NULL));
     }
  }
  else if (pad == filter->srcpad){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     ret=FALSE;
     mime = gst_structure_get_name (structure);
     if(filter->mode==0) if (strcmp (mime, "sift/raw") == 0) ret=TRUE;
     if(filter->mode==1) if (strcmp (mime, "sift/lowe") == 0) ret=TRUE;
     if(filter->mode==2) if (strcmp (mime, "sift/oxf") == 0) ret=TRUE;
     if(filter->mode==3) if (strcmp (mime, "sift/lowelight") == 0) ret=TRUE;
     capschar=gst_caps_to_string (caps);
     g_print ("SetCaps bei %s aufgerufen vom Zielelement mit %s\n",GST_ELEMENT_NAME((GstElement*)filter),capschar);
     g_free(capschar);
  }
  gst_object_unref (filter);
  g_print ("DiftExtractor DetCaps finished, read out overlap h:%d v:%d\n",filter->overlap_y,filter->overlap_x);
  return ret;
}

static GstCaps *
gst_sift_extractor_get_caps (GstPad *pad)
{
  g_print ("GetCaps aufgerufen\n");
  GstSiftExtractor *filter;

  filter = GST_SIFTEXTRACTOR (GST_OBJECT_PARENT (pad));
  GstCaps * caps;
  //caps=gst_pad_get_pad_template_caps(pad);
  if (pad == filter->sinkpad){
    g_print ("Das Quellelement macht getcaps\n");
    if (1) caps=gst_caps_from_string ("video/x-raw-gray, framerate=(fraction)[ 0/1, 2147483647/1 ], depth=(int)8, bpp=(int)8, width=(int)[ 1, 2147483647 ], height=(int)[ 1, 2147483647 ]; video/x-raw-rgb, framerate=(fraction)[ 0/1, 2147483647/1 ], width=(int)[ 1, 2147483647 ], height=(int)[ 1, 2147483647 ]; video/x-raw-yuv, format=(fourcc)Y800, framerate=(fraction)[ 0/1, 2147483647/1 ], width=(int)[ 1, 2147483647 ], height=(int)[ 1, 2147483647 ]");
  }
  if (pad == filter->srcpad){
  g_print ("Das Zielelement macht getcaps\n");
  	//caps=gst_caps_from_string("")
  	//if (!GST_IS_CAPS(caps))caps=gst_caps_from_string("video/sift, height = (int) [ 1, 2147483647 ]");
	if (filter->mode==0){
	  caps=gst_caps_new_simple ("sift/raw",
      		//"width", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		//"height", GST_TYPE_INT_RANGE, 1, G_MAXINT,
      		//"framerate", GST_TYPE_FRACTION_RANGE, 0, 1,G_MAXINT,1,
      		"width", G_TYPE_INT, filter->imagesize.width,
      		"height", G_TYPE_INT, filter->imagesize.height,
      		"framerate", GST_TYPE_FRACTION, filter->rate_numerator,filter->rate_denominator,
      		"overlap_v", G_TYPE_INT, filter->overlap_x,
      		"overlap_h", G_TYPE_INT, filter->overlap_y,
		NULL);
	}
	else if (filter->mode==1){
	  caps=gst_caps_new_simple ("sift/lowe",
		NULL);
	}
	else if (filter->mode==2){
	  caps=gst_caps_new_simple ("sift/oxf",
		NULL);
	}
	else if (filter->mode==3){
	  caps=gst_caps_new_simple ("sift/lowelight",
      		"width", G_TYPE_INT, filter->imagesize.width,
      		"height", G_TYPE_INT, filter->imagesize.height,
      		"overlap_v", G_TYPE_INT, filter->overlap_x,
      		"overlap_h", G_TYPE_INT, filter->overlap_y,
		NULL);
	}
  }
  //g_print ("Will return %s\n",gst_caps_to_string (caps));
  return gst_caps_copy(caps);
  //return caps;
}

static GstFlowReturn
gst_sift_extractor_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  g_print ("Buffer alloc!!\n");  
  GstSiftExtractor *filter;
  GstFlowReturn result;
  //g_print ("Bufferalloc has been caled with Caps: %s\n",gst_caps_to_string(caps));
  filter = GST_SIFTEXTRACTOR (GST_PAD_PARENT (pad));

  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  return result;
}

/*
static void gst_sift_extractor_src_fixate (GstPad * pad, GstCaps * caps)
{
  g_print ("Fixate aufgerufen.\n");
  GstStructure *structure;

  structure = gst_caps_get_structure (caps, 0);

  //gst_structure_field_nearest_int (structure, "width", 1388);
  //gst_structure_fixate_field_nearest_int (structure, "height", 1038);
  //gst_structure_fixate_field_nearest_fraction (structure, "framerate", 18, 1);
}
*/

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_sift_extractor_chain (GstPad * pad, GstBuffer * buf)
{
  GstSiftExtractor *filter;

  filter = GST_SIFTEXTRACTOR (GST_OBJECT_PARENT (pad));

  IplImage *process;
  gint num_features;
  


//clear buffers, in case any downstream element is stockpiling

  GList *templ=filter->buffers_in_use;
  
  while (templ){
	g_print("Im in while, checking %d\n",g_list_length(filter->buffers_in_use));
  	if (GST_MINI_OBJECT_REFCOUNT((GstBuffer*)templ->data)<2){
		g_print("will unref - newcode\n");
		free(((GstBuffer*)templ->data)->data);
		gst_buffer_unref((GstBuffer*)templ->data);		
		filter->buffers_in_use=g_list_remove(filter->buffers_in_use,templ->data);	
		templ=filter->buffers_in_use;

	}
	else templ=g_list_next(templ);
  }





  GstBuffer* out;


  process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->inbpp/8));
  process->imageData=(char*)buf->data;  

  //num_features=sift_features(process,&(filter->feat));
  g_mutex_lock(filter->mutex);  
  if (!filter->siftset) filter->siftset=make_siftset(process,SIFT_INTVLS,filter->double_image);
  //while(42){
  num_features=_sift_features(process,&(filter->feat) ,filter->siftset, SIFT_INTVLS, SIFT_SIGMA, SIFT_CONTR_THR,
 			 SIFT_CURV_THR, filter->double_image, SIFT_DESCR_WIDTH,
			 SIFT_DESCR_HIST_BINS );
  //free(filter->feat);
  //}
  if (filter->silent == FALSE){
    g_print ("%s: Found %d Features killing siftset\n",GST_ELEMENT_NAME(filter),num_features);
    killsiftset(filter->siftset);
    filter->siftset=NULL;
  }
  else g_print ("%s: Found %d Features not killing siftset\n",GST_ELEMENT_NAME(filter),num_features);
  g_mutex_unlock(filter->mutex);  

  cvReleaseImageHeader(&process);


  if (filter->mode==0){//raw
    out=gst_buffer_new_and_alloc(0);
    gst_buffer_set_data(out,filter->feat, num_features*sizeof(struct feature));
    GST_BUFFER_MALLOCDATA(out)=NULL;
    gst_buffer_ref(out);//keep reference on buffers to unref them manually
    //g_print ("Pushing buffer\n");
    filter->buffers_in_use=g_list_append(filter->buffers_in_use,out);
  }

  else if (filter->mode==1){//lowe
    gchar* outbuf;
    gint offset=0;
    int i,j,d;
    struct feature* feat=filter->feat;
    d=feat[0].d;
    outbuf=g_malloc(12+(num_features*50)+(num_features*d*4));//just guessing output size 
    offset+=sprintf( outbuf, "%d %d\n", num_features, d );
  
    for( i = 0; i < num_features; i++ )
      {
      offset+=sprintf( outbuf+offset, "%f %f %f %f", feat[i].y, feat[i].x,
	       feat[i].scl, feat[i].ori );
      for( j = 0; j < d; j++ )
	{
	  /* write 20 descriptor values per line */
	  if( j % 20 == 0 )
	     offset+=sprintf( outbuf+offset,"\n" );
	  offset+=sprintf( outbuf+offset, " %d", (int)(feat[i].descr[j]) );
	}
      offset+=sprintf( outbuf+offset, "\n" );
      }
    g_print ("Allocated %d, used %d \n",(12+(num_features*50)+(num_features*d*4)),offset );
    outbuf=g_realloc(outbuf,offset+1);
    out=gst_buffer_new_and_alloc(0);
    gst_buffer_set_data(out,outbuf,offset+1);
    GST_BUFFER_MALLOCDATA(out)=outbuf;
    free(filter->feat);
  }
  else if (filter->mode==2){
    gchar* outbuf;
    gint offset=0;
    int i,j,d;
    struct feature* feat=filter->feat;

    d = feat[0].d;
    outbuf=g_malloc(12+(num_features*60)+(num_features*d*11));
    offset+=sprintf( outbuf+offset, "%d\n", d);
    offset+=sprintf( outbuf+offset, "%d\n", num_features );
    for( i = 0; i < num_features; i++ )
      {
        offset+=sprintf( outbuf+offset, "%f %f %f %f %f", feat[i].x, feat[i].y, feat[i].a,feat[i].b, feat[i].c );
        for( j = 0; j < d; j++ )
	  offset+=sprintf( outbuf+offset, " %f", feat[i].descr[j] );
        offset+=sprintf( outbuf+offset, "\n" );
      }

    g_print ("Allocated %d, used %d \n",(12+(num_features*60)+(num_features*d*11)),offset );
    outbuf=g_realloc(outbuf,offset+1);
    out=gst_buffer_new_and_alloc(0);
    gst_buffer_set_data(out,outbuf,offset+1);
    GST_BUFFER_MALLOCDATA(out)=outbuf;
    free(filter->feat);
  }
  else if (filter->mode==3){//new lowelight mode of Hannes
    gchar* outbuf;
    gint offset=0;
    int i,j,d;
    struct feature* feat=filter->feat;
    
    if (num_features) d=feat[0].d;
    out=gst_buffer_new_and_alloc(num_features*sizeof(struct lowelight));
    struct lowelight *resll=(struct lowelight*)out->data;

    for( i = 0; i < num_features; i++ )
      {
      if (little_endian){
        resll[i].x=feat[i].x;
        resll[i].y=feat[i].y;
        resll[i].scl=feat[i].scl;
        resll[i].ori=feat[i].ori;
      }
      else{
        resll[i].x=DoubleSwap(feat[i].x);
        resll[i].y=DoubleSwap(feat[i].y);
        resll[i].scl=DoubleSwap(feat[i].scl);
        resll[i].ori=DoubleSwap(feat[i].ori);
      }	
      for( j = 0; j < d; j++ )
	{
	 resll[i].data[j]=(int)(feat[i].descr[j]);
	}
      }
    g_print ("Allocated %d\n",num_features*sizeof(struct lowelight));
    free(filter->feat);
  }


  else return FALSE;
  gst_buffer_set_caps(out,GST_PAD_CAPS(filter->srcpad));
  //g_print ("Pushed buffer with %s\n",gst_caps_to_string (GST_PAD_CAPS(filter->srcpad)));
  out->timestamp=buf->timestamp;
  out->duration=buf->duration;
  out->offset=buf->offset;
  out->offset_end=buf->offset_end;
  gst_buffer_unref(buf);
  return gst_pad_push (filter->srcpad, out);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
siftextractor_init (GstPlugin * siftextractor)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template siftextractor' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_sift_extractor_debug, "siftextractor",
      0, "Template siftextractor");

  return gst_element_register (siftextractor, "siftextractor", GST_RANK_NONE,
      GST_TYPE_SIFTEXTRACTOR);
}

/* gstreamer looks for this structure to register siftextractors
 *
 * exchange the string 'Template siftextractor' with your siftextractor description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "siftextractor",
    "Template siftextractor",
    siftextractor_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
