/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-siftfolder
 *
 * FIXME:Describe siftfolder here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! siftfolder ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif



//#include "sift.h"
//#include "imgfeatures.h"
//#include "kdtree.h"
//#include "utils.h"
//#include "xform.h"
#include "math.h"

#include <sys/time.h>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>


#include <gst/gst.h>

#include <gsl/gsl_sf.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

//stuff for inotifysrc
#include <sys/inotify.h>
#include <sys/select.h>

#include <sys/types.h>
#include  <signal.h>
#include <sys/time.h>
#include <unistd.h>

//#include <linux/inotify.h>
#include <pthread.h>

#include "gstsiftfolder.h"

GST_DEBUG_CATEGORY_STATIC (gst_siftfolder_debug);
#define GST_CAT_DEFAULT gst_siftfolder_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_MAXSIM,
  PROP_MAXSIM2,
  PROP_GOODDIST,
  PROP_THRESHOLD,
  PROP_REMATCHDOUBLE,
  PROP_SHOWINLIERS,
  PROP_DIRECTORY,
  PROP_NTOTAL,
  PROP_THISN,
  PROP_THRESH,
  PROP_HFOV,
  PROP_UNDDIRECTORY,
  PROP_OTHERWAY
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("sift/raw,"
        "framerate = (fraction) [ 0, MAX ], "
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];"
	"sift/lowe;"
	"sift/oxf;"
	"sift/lowelight,"
        "width = (int) [ 1, MAX ], "
        "height = (int) [ 1, MAX ],"
        "overlap_x = (int) [ -1, MAX ], "
        "overlap_y = (int) [ -1, MAX ];")
    );

/*
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );
*/
static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "uber-coordinates")
    );
GST_BOILERPLATE (Gstsiftfolder, gst_siftfolder, GstElement,
    GST_TYPE_ELEMENT);

void * watchthread (void* data);
static void gst_siftfolder_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_siftfolder_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static gboolean gst_siftfolder_event (GstPad  *pad, GstEvent * event);
static gboolean gst_siftfolder_set_caps (GstPad * pad, GstCaps * caps);
static gboolean gst_siftfolder_load_gallery(Gstsiftfolder * filter);
static GstFlowReturn gst_siftfolder_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn
gst_siftfolder_change_state (GstElement *element, GstStateChange transition);
static GstCaps *
gst_siftfolder_get_src_caps (GstPad *pad);
static GstCaps *
gst_siftfolder_get_sink_caps (GstPad *pad);


static inline struct feature* get_match( struct feature*, int );
int get_matched_features( struct feature*, int, int, struct feature*** );
int calc_min_inliers( int, int, double, double );
struct feature** draw_ransac_sample( struct feature**, int, int, gsl_rng* );
void extract_corresp_pts( struct feature**, int, int, CvPoint2D64f**,
			  CvPoint2D64f** );
int find_consensus( struct feature**, int, int, CvMat*, ransac_err_fn,
		    double, struct feature*** );
static inline void release_mem( CvPoint2D64f*, CvPoint2D64f*,
				struct feature** );
CvMat* lsq_homog( CvPoint2D64f* pts, CvPoint2D64f* mpts, int n );
double homog_xfer_err( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* H );
CvPoint2D64f persp_xform_pt( CvPoint2D64f pt, CvMat* T );

struct kd_node* kd_node_init( struct feature*, int );
void expand_kd_node_subtree( struct kd_node* );
void assign_part_key( struct kd_node* );
double median_select( double*, int );
double rank_select( double*, int, int );
void insertion_sort( double*, int );
int partition_array( double*, int, double );
void partition_features( struct kd_node* );
struct kd_node* explore_to_leaf( struct kd_node*, struct feature*,
				 struct min_pq* );
int insert_into_nbr_array( struct feature*, struct feature**, int, int );
int within_rect( CvPoint2D64f, CvRect );

//SIFT 
void create_init_img( IplImage*, struct siftset*,int, double );
void convert_to_gray32( IplImage*, struct siftset*);
void build_pyr( struct siftset*, double );
CvSeq* scale_space_extrema( IplImage***, int, int, double, int, CvMemStorage*);
int is_extremum( IplImage***, int, int, int, int );
struct feature* interp_extremum( IplImage***, int, int, int, int, int, double);
void interp_step( IplImage***, int, int, int, int, double*, double*, double* );
CvMat* deriv_3D( IplImage***, int, int, int, int );
CvMat* hessian_3D( IplImage***, int, int, int, int );
double interp_contr( IplImage***, int, int, int, int, double, double, double );
struct feature* new_feature( void );
int is_too_edge_like( IplImage*, int, int, int );
void calc_feature_scales( CvSeq*, double, int );
void adjust_for_img_dbl( CvSeq* );
void calc_feature_oris( CvSeq*, IplImage*** );
double* ori_hist( IplImage*, int, int, int, int, double );
int calc_grad_mag_ori( IplImage*, int, int, double*, double* );
void smooth_ori_hist( double*, int );
double dominant_ori( double*, int );
void add_good_ori_features( CvSeq*, double*, int, double, struct feature* );
struct feature* clone_feature( struct feature* );
void compute_descriptors( CvSeq*, IplImage***, int, int );
double*** descr_hist( IplImage*, int, int, double, double, int, int );
void interp_hist_entry( double***, double, double, double, double, int, int);
void hist_to_descr( double***, int, int, struct feature* );
void normalize_descr( struct feature* );
int feature_cmp( void*, void*, void* );
void release_descr_hist( double****, int );
void release_pyr( IplImage****, int, int );
static inline float pixval32f( IplImage* img, int r, int c )
{
  return ( (float*)(img->imageData + img->widthStep*r) )[c];
}
//#define GST_SIFTTOOL_TYPE_OP_MODE (gst_type_sifttool_op_mode ())

double descr_dist_sq( struct feature* f1, struct feature* f2 )
{
  double diff, dsq = 0;
  double* descr1, * descr2;
  int i, d;

  d = f1->d;
  if( f2->d != d )
    return DBL_MAX;
  descr1 = f1->descr;
  descr2 = f2->descr;

  for( i = 0; i < d; i++ )
    {
      diff = descr1[i] - descr2[i];
      dsq += diff*diff;
    }
  return dsq;
}

CvMat* ransac_xform( struct feature* features, int n, int mtype,
		     ransac_xform_fn xform_fn, int m, double p_badxform,
		     ransac_err_fn err_fn, double err_tol,
		     struct feature*** inliers, int* n_in )
{
  struct feature** matched, ** sample, ** consensus, ** consensus_max = NULL;
  struct ransac_data* rdata;
  CvPoint2D64f* pts, * mpts;
  CvMat* M = NULL;
  gsl_rng* rng;
  double p, in_frac = RANSAC_INLIER_FRAC_EST;
  int i, nm, in, in_min, in_max = 0, k = 0;

  nm = get_matched_features( features, n, mtype, &matched );
  if( nm < m )
    {
      //fprintf( stderr, "Warning: not enough matches to compute xform, %s" \
	       " line %d\n", __FILE__, __LINE__ );
      goto end;
    }

  /* initialize random number generator */
  rng = gsl_rng_alloc( gsl_rng_mt19937 );
  gsl_rng_set( rng, time(NULL) );

  in_min = calc_min_inliers( nm, m, RANSAC_PROB_BAD_SUPP, p_badxform );
  p = pow( 1.0 - pow( in_frac, m ), k );
  while( p > p_badxform )
    {
      sample = draw_ransac_sample( matched, nm, m, rng );
      extract_corresp_pts( sample, m, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, m );
      if( ! M )
	goto iteration_end;
      in = find_consensus( matched, nm, mtype, M, err_fn, err_tol, &consensus);
      if( in > in_max )
	{
	  if( consensus_max )
	    free( consensus_max );
	  consensus_max = consensus;
	  in_max = in;
	  in_frac = (double)in_max / nm;
	}
      else
	free( consensus );
      cvReleaseMat( &M );

    iteration_end:
      release_mem( pts, mpts, sample );
      p = pow( 1.0 - pow( in_frac, m ), ++k );
    }

  /* calculate final transform based on best consensus set */
  if( in_max >= in_min )
    {
      extract_corresp_pts( consensus_max, in_max, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, in_max );
      in = find_consensus( matched, nm, mtype, M, err_fn, err_tol, &consensus);
      cvReleaseMat( &M );
      release_mem( pts, mpts, consensus_max );
      extract_corresp_pts( consensus, in, mtype, &pts, &mpts );
      M = xform_fn( pts, mpts, in );
      if( inliers )
	{
	  free(pts);
	  free(mpts);  
	  *inliers = consensus;
	  consensus = NULL;
	}
      else release_mem( pts, mpts, consensus );
      if( n_in )
	*n_in = in;

    }
  else if( consensus_max )
    {
      if( inliers )
	*inliers = NULL;
      if( n_in )
	*n_in = 0;
      free( consensus_max );
    }

  gsl_rng_free( rng );
 end:
  for( i = 0; i < nm; i++ )
    {
      rdata = feat_ransac_data( matched[i] );
      matched[i]->feature_data = rdata->orig_feat_data;
      free( rdata );
    }
  free( matched );
  return M;
}

CvMat* lsq_homog( CvPoint2D64f* pts, CvPoint2D64f* mpts, int n )
{
  CvMat* H, * A, * B, X;
  double x[9];
  int i;

  if( n < 4 )
    {
      fprintf( stderr, "Warning: too few points in lsq_homog(), %s line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }

  /* set up matrices so we can unstack homography into X; AX = B */
  A = cvCreateMat( 2*n, 8, CV_64FC1 );
  B = cvCreateMat( 2*n, 1, CV_64FC1 );
  X = cvMat( 8, 1, CV_64FC1, x );
  H = cvCreateMat(3, 3, CV_64FC1);
  cvZero( A );
  for( i = 0; i < n; i++ )
    {
      cvmSet( A, i, 0, pts[i].x );
      cvmSet( A, i+n, 3, pts[i].x );
      cvmSet( A, i, 1, pts[i].y );
      cvmSet( A, i+n, 4, pts[i].y );
      cvmSet( A, i, 2, 1.0 );
      cvmSet( A, i+n, 5, 1.0 );
      cvmSet( A, i, 6, -pts[i].x * mpts[i].x );
      cvmSet( A, i, 7, -pts[i].y * mpts[i].x );
      cvmSet( A, i+n, 6, -pts[i].x * mpts[i].y );
      cvmSet( A, i+n, 7, -pts[i].y * mpts[i].y );
      cvmSet( B, i, 0, mpts[i].x );
      cvmSet( B, i+n, 0, mpts[i].y );
    }
  cvSolve( A, B, &X, CV_SVD );
  x[8] = 1.0;
  X = cvMat( 3, 3, CV_64FC1, x );
  cvConvert( &X, H );

  cvReleaseMat( &A );
  cvReleaseMat( &B );
  return H;
}



/*
  Calculates the transfer error between a point and its correspondence for
  a given homography, i.e. for a point x, it's correspondence x', and
  homography H, computes d(x', Hx)^2.
  
  @param pt a point
  @param mpt pt's correspondence
  @param H a homography matrix
  
  @return Returns the transfer error between pt and mpt given H
*/
double homog_xfer_err( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* H )
{
  CvPoint2D64f xpt = persp_xform_pt( pt, H );
  
  return sqrt( dist_sq_2D( xpt, mpt ) );
}



/*
  Performs a perspective transformation on a single point.  That is, for a
  point (x, y) and a 3 x 3 matrix T this function returns the point
  (u, v), where
  
  [x' y' w']^T = T * [x y 1]^T,
  
  and
  
  (u, v) = (x'/w', y'/w').

  Note that affine transforms are a subset of perspective transforms.
  
  @param pt a 2D point
  @param T a perspective transformation matrix
  
  @return Returns the point (u, v) as above.
*/
CvPoint2D64f persp_xform_pt( CvPoint2D64f pt, CvMat* T )
{
  CvMat XY, UV;
  double xy[3] = { pt.x, pt.y, 1.0 }, uv[3] = { 0 };
  CvPoint2D64f rslt;

  cvInitMatHeader( &XY, 3, 1, CV_64FC1, xy, CV_AUTOSTEP );
  cvInitMatHeader( &UV, 3, 1, CV_64FC1, uv, CV_AUTOSTEP );
  cvMatMul( T, &XY, &UV );
  rslt = cvPoint2D64f( uv[0] / uv[2], uv[1] / uv[2] );

  return rslt;
}


static inline struct feature* get_match( struct feature* feat, int mtype )
{
  if( mtype == FEATURE_MDL_MATCH )
    return feat->mdl_match;
  if( mtype == FEATURE_BCK_MATCH )
    return feat->bck_match;
  if( mtype == FEATURE_FWD_MATCH )
    return feat->fwd_match;
  return NULL;
}




int get_matched_features( struct feature* features, int n, int mtype,
			  struct feature*** matched )
{
  struct feature** _matched;
  struct ransac_data* rdata;
  int i, m = 0;

  _matched = calloc( n, sizeof( struct feature* ) );
  for( i = 0; i < n; i++ )
    if( get_match( features + i, mtype ) )
      {
	rdata = malloc( sizeof( struct ransac_data ) );
	memset( rdata, 0, sizeof( struct ransac_data ) );
	rdata->orig_feat_data = features[i].feature_data;
	_matched[m] = features + i;
	_matched[m]->feature_data = rdata;
	m++;
      }
  *matched = _matched;
  return m;
}


int calc_min_inliers( int n, int m, double p_badsupp, double p_badxform )
{
  double pi, sum;
  int i, j;

  for( j = m+1; j <= n; j++ )
    {
      sum = 0;
      for( i = j; i <= n; i++ )
	{
	  pi = (i-m) * log( p_badsupp ) + (n-i+m) * log( 1.0 - p_badsupp ) +
	    gsl_sf_lnchoose( n - m, i - m );
	  sum += exp( pi );
	}
      if( sum < p_badxform )
	break;
    }
  return j;
}

struct feature** draw_ransac_sample( struct feature** features, int n,
				     int m, gsl_rng* rng )
{
  struct feature** sample, * feat;
  struct ransac_data* rdata;
  int i, x;

  for( i = 0; i < n; i++ )
    {
      rdata = feat_ransac_data( features[i] );
      rdata->sampled = 0;
    }

  sample = calloc( m, sizeof( struct feature* ) );
  for( i = 0; i < m; i++ )
    {
      do
	{
	  x = gsl_rng_uniform_int( rng, n );
	  feat = features[x];
	  rdata = feat_ransac_data( feat );
	}
      while( rdata->sampled );
      sample[i] = feat;
      rdata->sampled = 1;
    }

  return sample;
}


void extract_corresp_pts( struct feature** features, int n, int mtype,
			  CvPoint2D64f** pts, CvPoint2D64f** mpts )
{
  struct feature* match;
  CvPoint2D64f* _pts, * _mpts;
  int i;

  _pts = calloc( n, sizeof( CvPoint2D64f ) );
  _mpts = calloc( n, sizeof( CvPoint2D64f ) );

  if( mtype == FEATURE_MDL_MATCH )
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of correct type\n");
	_pts[i] = features[i]->img_pt;
	_mpts[i] = match->mdl_pt;
      }

  else
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	_pts[i] = features[i]->img_pt;
	_mpts[i] = match->img_pt;
      }

  *pts = _pts;
  *mpts = _mpts;
}




int find_consensus( struct feature** features, int n, int mtype,
		    CvMat* M, ransac_err_fn err_fn, double err_tol,
		    struct feature*** consensus )
{
  struct feature** _consensus;
  struct feature* match;
  CvPoint2D64f pt, mpt;
  double err;
  int i, in = 0;

  _consensus = calloc( n, sizeof( struct feature* ) );

  if( mtype == FEATURE_MDL_MATCH )
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	pt = features[i]->img_pt;
	mpt = match->mdl_pt;
	err = err_fn( pt, mpt, M );
	if( err <= err_tol )
	  _consensus[in++] = features[i];
      }

  else
    for( i = 0; i < n; i++ )
      {
	match = get_match( features[i], mtype );
	if( ! match )
	  g_print( "feature does not have match of type\n");
	pt = features[i]->img_pt;
	mpt = match->img_pt;
	err = err_fn( pt, mpt, M );
	if( err <= err_tol )
	  _consensus[in++] = features[i];
      }
  *consensus = _consensus;
  return in;
}


static inline void release_mem( CvPoint2D64f* pts1, CvPoint2D64f* pts2,
				struct feature** features )
{
  free( pts1 );
  free( pts2 );
  if( features )
    free( features );
}


struct kd_node* kdtree_build( struct feature* features, int n )
{
  struct kd_node* kd_root;

  if( ! features  ||  n <= 0 )
    {
      fprintf( stderr, "Warning: kdtree_build(): no features, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }

  kd_root = kd_node_init( features, n );
  expand_kd_node_subtree( kd_root );

  return kd_root;
}

int kdtree_bbf_knn( struct kd_node* kd_root, struct feature* feat, int k,
		    struct feature*** nbrs, int max_nn_chks )
{
  struct kd_node* expl;
  struct min_pq* min_pq;
  struct feature* tree_feat, ** _nbrs;
  struct bbf_data* bbf_data;
  int i, t = 0, n = 0;

  if( ! nbrs  ||  ! feat  ||  ! kd_root )
    {
      fprintf( stderr, "Warning: NULL pointer error, %s, line %d\n",
	       __FILE__, __LINE__ );
      return -1;
    }

  _nbrs = calloc( k, sizeof( struct feature* ) );
  min_pq = minpq_init();
  minpq_insert( min_pq, kd_root, 0 );
  while( min_pq->n > 0  &&  t < max_nn_chks )
    {
      expl = (struct kd_node*)minpq_extract_min( min_pq );
      if( ! expl )
	{
	  fprintf( stderr, "Warning: PQ unexpectedly empty, %s line %d\n",
		   __FILE__, __LINE__ );
	  goto fail;
	}

      expl = explore_to_leaf( expl, feat, min_pq );
      if( ! expl )
	{
	  fprintf( stderr, "Warning: PQ unexpectedly empty, %s line %d\n",
		   __FILE__, __LINE__ );
	  goto fail;
	}

      for( i = 0; i < expl->n; i++ )
	{
	  tree_feat = &expl->features[i];
	  bbf_data = malloc( sizeof( struct bbf_data ) );
	  if( ! bbf_data )
	    {
	      fprintf( stderr, "Warning: unable to allocate memory,"
		       " %s line %d\n", __FILE__, __LINE__ );
	      goto fail;
	    }
	  bbf_data->old_data = tree_feat->feature_data;
	  bbf_data->d = descr_dist_sq(feat, tree_feat);
	  tree_feat->feature_data = bbf_data;
	  n += insert_into_nbr_array( tree_feat, _nbrs, n, k );
	}
      t++;
    }

  minpq_release( &min_pq );
  for( i = 0; i < n; i++ )
    {
      bbf_data = _nbrs[i]->feature_data;
      _nbrs[i]->feature_data = bbf_data->old_data;
      free( bbf_data );
    }
  *nbrs = _nbrs;
  return n;

 fail:
  minpq_release( &min_pq );
  for( i = 0; i < n; i++ )
    {
      bbf_data = _nbrs[i]->feature_data;
      _nbrs[i]->feature_data = bbf_data->old_data;
      free( bbf_data );
    }
  free( _nbrs );
  *nbrs = NULL;
  return -1;
}

void kdtree_release( struct kd_node* kd_root )
{
  if( ! kd_root )
    return;
  kdtree_release( kd_root->kd_left );
  kdtree_release( kd_root->kd_right );
  free( kd_root );
}




struct kd_node* kd_node_init( struct feature* features, int n )
{
  struct kd_node* kd_node;

  kd_node = malloc( sizeof( struct kd_node ) );
  memset( kd_node, 0, sizeof( struct kd_node ) );
  kd_node->ki = -1;
  kd_node->features = features;
  kd_node->n = n;

  return kd_node;
}


void expand_kd_node_subtree( struct kd_node* kd_node )
{
  /* base case: leaf node */
  if( kd_node->n == 1  ||  kd_node->n == 0 )
    {
      kd_node->leaf = 1;
      return;
    }

  assign_part_key( kd_node );
  partition_features( kd_node );

  if( kd_node->kd_left )
    expand_kd_node_subtree( kd_node->kd_left );
  if( kd_node->kd_right )
    expand_kd_node_subtree( kd_node->kd_right );
}


void assign_part_key( struct kd_node* kd_node )
{
  struct feature* features;
  double kv, x, mean, var, var_max = 0;
  double* tmp;
  int d, n, i, j, ki = 0;

  features = kd_node->features;
  n = kd_node->n;
  d = features[0].d;

  /* partition key index is that along which descriptors have most variance */
  for( j = 0; j < d; j++ )
    {
      mean = var = 0;
      for( i = 0; i < n; i++ )
	mean += features[i].descr[j];
      mean /= n;
      for( i = 0; i < n; i++ )
	{
	  x = features[i].descr[j] - mean;
	  var += x * x;
	}
      var /= n;

      if( var > var_max )
	{
	  ki = j;
	  var_max = var;
	}
    }

  /* partition key value is median of descriptor values at ki */
  tmp = calloc( n, sizeof( double ) );
  for( i = 0; i < n; i++ )
    tmp[i] = features[i].descr[ki];
  kv = median_select( tmp, n );
  free( tmp );

  kd_node->ki = ki;
  kd_node->kv = kv;
}


double median_select( double* array, int n )
{
  return rank_select( array, n, (n - 1) / 2 );
}



double rank_select( double* array, int n, int r )
{
  double* tmp, med;
  int gr_5, gr_tot, rem_elts, i, j;

  /* base case */
  if( n == 1 )
    return array[0];

  /* divide array into groups of 5 and sort them */
  gr_5 = n / 5;
  gr_tot = cvCeil( n / 5.0 );
  rem_elts = n % 5;
  tmp = array;
  for( i = 0; i < gr_5; i++ )
    {
      insertion_sort( tmp, 5 );
      tmp += 5;
    }
  insertion_sort( tmp, rem_elts );

  /* recursively find the median of the medians of the groups of 5 */
  tmp = calloc( gr_tot, sizeof( double ) );
  for( i = 0, j = 2; i < gr_5; i++, j += 5 )
    tmp[i] = array[j];
  if( rem_elts )
    tmp[i++] = array[n - 1 - rem_elts/2];
  med = rank_select( tmp, i, ( i - 1 ) / 2 );
  free( tmp );
  
  /* partition around median of medians and recursively select if necessary */
  j = partition_array( array, n, med );
  if( r == j )
    return med;
  else if( r < j )
    return rank_select( array, j, r );
  else
    {
      array += j+1;
      return rank_select( array, ( n - j - 1 ), ( r - j - 1 ) );
    }
}


void insertion_sort( double* array, int n )
{
  double k;
  int i, j;

  for( i = 1; i < n; i++ )
    {
      k = array[i];
      j = i-1;
      while( j >= 0  &&  array[j] > k )
	{
	  array[j+1] = array[j];
	  j -= 1;
	}
      array[j+1] = k;
    }
}


int partition_array( double* array, int n, double pivot )
{
  double tmp;
  int p=0, i, j;

  i = -1;
  for( j = 0; j < n; j++ )
    if( array[j] <= pivot )
      {
	tmp = array[++i];
	array[i] = array[j];
	array[j] = tmp;
	if( array[i] == pivot )
	  p = i;
      }
  array[p] = array[i];
  array[i] = pivot;
  
  return i;
}


void partition_features( struct kd_node* kd_node )
{
  struct feature* features, tmp;
  double kv;
  int n, ki, p=0, i, j = -1;

  features = kd_node->features;
  n = kd_node->n;
  ki = kd_node->ki;
  kv = kd_node->kv;
  for( i = 0; i < n; i++ )
    if( features[i].descr[ki] <= kv )
      {
	tmp = features[++j];
	features[j] = features[i];
	features[i] = tmp;
	if( features[j].descr[ki] == kv )
	  p = j;
      }
  tmp = features[p];
  features[p] = features[j];
  features[j] = tmp;

  /* if all records fall on same side of partition, make node a leaf */
  if( j == n - 1 )
    {
      kd_node->leaf = 1;
      return;
    }

  kd_node->kd_left = kd_node_init( features, j + 1 );
  kd_node->kd_right = kd_node_init( features + ( j + 1 ), ( n - j - 1 ) );
}




struct kd_node* explore_to_leaf( struct kd_node* kd_node, struct feature* feat,
				 struct min_pq* min_pq )
{
  struct kd_node* unexpl, * expl = kd_node;
  double kv;
  int ki;

  while( expl  &&  ! expl->leaf )
    {
      ki = expl->ki;
      kv = expl->kv;
      
      if( ki >= feat->d )
	{
	  fprintf( stderr, "Warning: comparing imcompatible descriptors, %s" \
		   " line %d\n", __FILE__, __LINE__ );
	  return NULL;
	}
      if( feat->descr[ki] <= kv )
	{
	  unexpl = expl->kd_right;
	  expl = expl->kd_left;
	}
      else
	{
	  unexpl = expl->kd_left;
	  expl = expl->kd_right;
	}
      
      if( minpq_insert( min_pq, unexpl, ABS( kv - feat->descr[ki] ) ) )
	{
	  fprintf( stderr, "Warning: unable to insert into PQ, %s, line %d\n",
		   __FILE__, __LINE__ );
	  return NULL;
	}
    }

  return expl;
}


int insert_into_nbr_array( struct feature* feat, struct feature** nbrs,
			   int n, int k )
{
  struct bbf_data* fdata, * ndata;
  double dn, df;
  int i, ret = 0;

  if( n == 0 )
    {
      nbrs[0] = feat;
      return 1;
    }

  /* check at end of array */
  fdata = (struct bbf_data*)feat->feature_data;
  df = fdata->d;
  ndata = (struct bbf_data*)nbrs[n-1]->feature_data;
  dn = ndata->d;
  if( df >= dn )
    {
      if( n == k )
	{
	  feat->feature_data = fdata->old_data;
	  free( fdata );
	  return 0;
	}
      nbrs[n] = feat;
      return 1;
    }

  /* find the right place in the array */
  if( n < k )
    {
      nbrs[n] = nbrs[n-1];
      ret = 1;
    }
  else
    {
      nbrs[n-1]->feature_data = ndata->old_data;
      free( ndata );
    }
  i = n-2;
  while( i >= 0 )
    {
      ndata = (struct bbf_data*)nbrs[i]->feature_data;
      dn = ndata->d;
      if( dn <= df )
	break;
      nbrs[i+1] = nbrs[i];
      i--;
    }
  i++;
  nbrs[i] = feat;

  return ret;
}

int within_rect( CvPoint2D64f pt, CvRect rect )
{
  if( pt.x < rect.x  ||  pt.y < rect.y )
    return 0;
  if( pt.x > rect.x + rect.width  ||  pt.y > rect.y + rect.height )
    return 0;
  return 1;
}

void release_pyr( IplImage**** pyr, int octvs, int n )
{
  int i, j;
  for( i = 0; i < octvs; i++ )
    {
      for( j = 0; j < n; j++ )
	cvReleaseImage( &(*pyr)[i][j] );
      free( (*pyr)[i] );
    }
  free( *pyr );
  *pyr = NULL;
}

struct siftset* make_siftset(IplImage* img,int intvls,int img_dbl)
{
struct siftset* result=(struct siftset*)malloc(sizeof(struct siftset));
if( img->nChannels > 1 )result->gray8 = cvCreateImage( cvGetSize(img), IPL_DEPTH_8U, 1 );
else result->gray8 = NULL;
result->gray32 = cvCreateImage( cvGetSize(img), IPL_DEPTH_32F, 1 );
if (img_dbl) result->init_img=cvCreateImage( cvSize( img->width*2, img->height*2 ),IPL_DEPTH_32F, 1 );
else result->init_img=result->gray32;
//Oktavenzahl berechnen
result->octvs = log( MIN( result->init_img->width, result->init_img->height ) ) / log(2) - 2;
result->intvls=intvls;
result->dbl_img=img_dbl;
result->channels=img->nChannels;
//gauss-pyramide bauen
int i,o;
result->gauss_pyr = calloc( result->octvs, sizeof( IplImage** ) );
  for( i = 0; i < result->octvs; i++ )
    result->gauss_pyr[i] = calloc( result->intvls + 3, sizeof( IplImage *) );
  for( o = 0; o < result->octvs; o++ )
    for( i = 0; i < result->intvls + 3; i++ )
      {
	if( o == 0  &&  i == 0 )
	  result->gauss_pyr[o][i] = result->init_img;

	/* base of new octvave is halved image from end of previous octave */
	else if( i == 0 )
	  result->gauss_pyr[o][i] = cvCreateImage( cvSize((result->gauss_pyr[o-1][intvls] )->width / 2, (result->gauss_pyr[o-1][intvls])->height / 2),
				     IPL_DEPTH_32F, 1 );	  
	/* blur the current octave's last image to create the next one */
	else
	  {
	    result->gauss_pyr[o][i] = cvCreateImage( cvGetSize(result->gauss_pyr[o][i-1]),IPL_DEPTH_32F, 1 );
	  }
      }
//dog pyr bauen
  result->dog_pyr = calloc( result->octvs, sizeof( IplImage** ) );
  for( i = 0; i < result->octvs; i++ )
    result->dog_pyr[i] = calloc( result->intvls + 2, sizeof(IplImage*) );
  for( o = 0; o < result->octvs; o++ )
    for( i = 0; i < result->intvls + 2; i++ )
      {
	result->dog_pyr[o][i] = cvCreateImage( cvGetSize(result->gauss_pyr[o][i]),
				       IPL_DEPTH_32F, 1 );
	}
return result;
}

void killsiftset (struct siftset* siftset){
  release_pyr( &(siftset->gauss_pyr), siftset->octvs, siftset->intvls + 3 );
  release_pyr( &(siftset->dog_pyr), siftset->octvs, siftset->intvls + 2 );
if (siftset->dbl_img) cvReleaseImage( &siftset->gray32);//normalerweise schon weg weils in der gausspyr drinne war
if( siftset->channels > 1 ) cvReleaseImage( &siftset->gray8);
free(siftset);
}

int _sift_features( IplImage* img, struct feature** feat, struct siftset* siftset, int intvls,
		    double sigma, double contr_thr, int curv_thr,
		    int img_dbl, int descr_width, int descr_hist_bins )
{
  //IplImage*** gauss_pyr, *** dog_pyr;
  CvMemStorage* storage;
  CvSeq* features;
  int octvs, i, n = 0;
  
  /* check arguments */
  if( ! img )
    g_print( "IMG is NULL\n");
  if( ! feat )
    g_print("Feat is NULL\n");

  /* build scale space pyramid; smallest dimension of top level is ~4 pixels */
  create_init_img( img,siftset, img_dbl, sigma );
  octvs = log( MIN( siftset->init_img->width, siftset->init_img->height ) ) / log(2) - 2;
  build_pyr( siftset, sigma );

  
  storage = cvCreateMemStorage( 0 );
  features = scale_space_extrema( siftset->dog_pyr, siftset->octvs, siftset->intvls, contr_thr,
				  curv_thr, storage );
  calc_feature_scales( features, sigma, siftset->intvls );
  if( img_dbl )
    adjust_for_img_dbl( features );
  calc_feature_oris( features, siftset->gauss_pyr );
  compute_descriptors( features, siftset->gauss_pyr, descr_width, descr_hist_bins );

  /* sort features by decreasing scale and move from CvSeq to array */
  cvSeqSort( features, (CvCmpFunc)feature_cmp, NULL );
  n = features->total;
  *feat = calloc( n, sizeof(struct feature) );
  *feat = cvCvtSeqToArray( features, *feat, CV_WHOLE_SEQ );
  for( i = 0; i < n; i++ )
    {
      free( (*feat)[i].feature_data );
      (*feat)[i].feature_data = NULL;
    }
  
  cvReleaseMemStorage( &storage );
  return n;
}


void create_init_img( IplImage* img, struct siftset* siftset, int img_dbl, double sigma )
{
  //IplImage* gray, * dbl;
  double sig_diff;

  convert_to_gray32( img, siftset );
  if( img_dbl )
    {
      sig_diff = sqrt( sigma * sigma - SIFT_INIT_SIGMA * SIFT_INIT_SIGMA * 4 );
      //dbl = cvCreateImage( cvSize( img->width*2, img->height*2 ),
	//		   IPL_DEPTH_32F, 1 );
      cvResize( siftset->gray32, siftset->init_img, CV_INTER_CUBIC );
      cvSmooth( siftset->init_img, siftset->init_img, CV_GAUSSIAN, 0, 0, sig_diff, sig_diff );
      //cvReleaseImage( &gray );
      //return dbl;
    }
  else
    {
      sig_diff = sqrt( sigma * sigma - SIFT_INIT_SIGMA * SIFT_INIT_SIGMA );
      cvSmooth( siftset->gray32, siftset->gray32, CV_GAUSSIAN, 0, 0, sig_diff, sig_diff );
      //return gray;
    }
return;
}



/*
  Converts an image to 32-bit grayscale

  @param img a 3-channel 8-bit color (BGR) or 8-bit gray image

  @return nothing as it writes the first 32-bit pic into the pyramid
*/
void convert_to_gray32( IplImage* img, struct siftset* siftset )
{
  if( img->nChannels == 1 )
    cvConvertScale( img, siftset->gray32, 1.0 / 255.0, 0 );
  else
    {
      //gray8 = cvCreateImage( cvGetSize(img), IPL_DEPTH_8U, 1 );
      cvCvtColor( img, siftset->gray8, CV_BGR2GRAY );
      cvConvertScale( siftset->gray8, siftset->gray32, 1.0 / 255.0, 0 );
    }
return;
}

void build_pyr( struct siftset* siftset, double sigma )
{
  //IplImage*** gauss_pyr;
  const int _intvls = siftset->intvls;
  double sig[_intvls+3], sig_total, sig_prev, k;
  int i, o;

  sig[0] = sigma;
  k = pow( 2.0, 1.0 / siftset->intvls );
  for( i = 1; i < siftset->intvls + 3; i++ )
    {
      sig_prev = pow( k, i - 1 ) * sigma;
      sig_total = sig_prev * k;
      sig[i] = sqrt( sig_total * sig_total - sig_prev * sig_prev );
    }
  for( o = 0; o < siftset->octvs; o++ )
    for( i = 0; i < siftset->intvls + 3; i++ )
      {
	if( o == 0  &&  i == 0 ){}

	/* base of new octvave is halved image from end of previous octave */
	else if( i == 0 )
	  //gauss_pyr[o][i] = downsample( gauss_pyr[o-1][intvls] );
	  cvResize(siftset->gauss_pyr[o-1][siftset->intvls],siftset->gauss_pyr[o][i],CV_INTER_NN );
	/* blur the current octave's last image to create the next one */
	else
	  {
	    cvSmooth( siftset->gauss_pyr[o][i-1], siftset->gauss_pyr[o][i],
		      CV_GAUSSIAN, 0, 0, sig[i], sig[i] );
	  }
      }

  //return gauss_pyr;
  //build dog pyr
  for( o = 0; o < siftset->octvs; o++ )
    for( i = 0; i < siftset->intvls + 2; i++ )
      {
	cvSub( siftset->gauss_pyr[o][i+1], siftset->gauss_pyr[o][i], siftset->dog_pyr[o][i], NULL );
      }
  return;
}


/*
  Detects features at extrema in DoG scale space.  Bad features are discarded
  based on contrast and ratio of principal curvatures.

  @param dog_pyr DoG scale space pyramid
  @param octvs octaves of scale space represented by dog_pyr
  @param intvls intervals per octave
  @param contr_thr low threshold on feature contrast
  @param curv_thr high threshold on feature ratio of principal curvatures
  @param storage memory storage in which to store detected features

  @return Returns an array of detected features whose scales, orientations,
    and descriptors are yet to be determined.
*/
CvSeq* scale_space_extrema( IplImage*** dog_pyr, int octvs, int intvls,
			    double contr_thr, int curv_thr,
			    CvMemStorage* storage )
{
  CvSeq* features;
  double prelim_contr_thr = 0.5 * contr_thr / intvls;
  struct feature* feat;
  struct detection_data* ddata;
  int o, i, r, c;

  features = cvCreateSeq( 0, sizeof(CvSeq), sizeof(struct feature), storage );
  for( o = 0; o < octvs; o++ )
    for( i = 1; i <= intvls; i++ )
      for(r = SIFT_IMG_BORDER; r < dog_pyr[o][0]->height-SIFT_IMG_BORDER; r++)
	for(c = SIFT_IMG_BORDER; c < dog_pyr[o][0]->width-SIFT_IMG_BORDER; c++)
	  /* perform preliminary check on contrast */
	  if( ABS( pixval32f( dog_pyr[o][i], r, c ) ) > prelim_contr_thr )
	    if( is_extremum( dog_pyr, o, i, r, c ) )
	      {
		feat = interp_extremum(dog_pyr, o, i, r, c, intvls, contr_thr);
		if( feat )
		  {
		    ddata = feat_detection_data( feat );
		    if( ! is_too_edge_like( dog_pyr[ddata->octv][ddata->intvl],
					    ddata->r, ddata->c, curv_thr ) )
		      {
			cvSeqPush( features, feat );
		      }
		    else
		      free( ddata );
		    free( feat );
		  }
	      }
  
  return features;
}



/*
  Determines whether a pixel is a scale-space extremum by comparing it to it's
  3x3x3 pixel neighborhood.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's scale space octave
  @param intvl pixel's within-octave interval
  @param r pixel's image row
  @param c pixel's image col

  @return Returns 1 if the specified pixel is an extremum (max or min) among
    it's 3x3x3 pixel neighborhood.
*/
int is_extremum( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  double val = pixval32f( dog_pyr[octv][intvl], r, c );
  int i, j, k;

  /* check for maximum */
  if( val > 0 )
    {
      for( i = -1; i <= 1; i++ )
	for( j = -1; j <= 1; j++ )
	  for( k = -1; k <= 1; k++ )
	    if( val < pixval32f( dog_pyr[octv][intvl+i], r + j, c + k ) )
	      return 0;
    }

  /* check for minimum */
  else
    {
      for( i = -1; i <= 1; i++ )
	for( j = -1; j <= 1; j++ )
	  for( k = -1; k <= 1; k++ )
	    if( val > pixval32f( dog_pyr[octv][intvl+i], r + j, c + k ) )
	      return 0;
    }

  return 1;
}



/*
  Interpolates a scale-space extremum's location and scale to subpixel
  accuracy to form an image feature.  Rejects features with low contrast.
  Based on Section 4 of Lowe's paper.  

  @param dog_pyr DoG scale space pyramid
  @param octv feature's octave of scale space
  @param intvl feature's within-octave interval
  @param r feature's image row
  @param c feature's image column
  @param intvls total intervals per octave
  @param contr_thr threshold on feature contrast

  @return Returns the feature resulting from interpolation of the given
    parameters or NULL if the given location could not be interpolated or
    if contrast at the interpolated loation was too low.  If a feature is
    returned, its scale, orientation, and descriptor are yet to be determined.
*/
struct feature* interp_extremum( IplImage*** dog_pyr, int octv, int intvl,
				 int r, int c, int intvls, double contr_thr )
{
  struct feature* feat;
  struct detection_data* ddata;
  double xi, xr, xc, contr;
  int i = 0;
  
  while( i < SIFT_MAX_INTERP_STEPS )
    {
      interp_step( dog_pyr, octv, intvl, r, c, &xi, &xr, &xc );
      if( ABS( xi ) < 0.5  &&  ABS( xr ) < 0.5  &&  ABS( xc ) < 0.5 )
	break;
      
      c += cvRound( xc );
      r += cvRound( xr );
      intvl += cvRound( xi );
      
      if( intvl < 1  ||
	  intvl > intvls  ||
	  c < SIFT_IMG_BORDER  ||
	  r < SIFT_IMG_BORDER  ||
	  c >= dog_pyr[octv][0]->width - SIFT_IMG_BORDER  ||
	  r >= dog_pyr[octv][0]->height - SIFT_IMG_BORDER )
	{
	  return NULL;
	}
      
      i++;
    }
  
  /* ensure convergence of interpolation */
  if( i >= SIFT_MAX_INTERP_STEPS )
    return NULL;
  
  contr = interp_contr( dog_pyr, octv, intvl, r, c, xi, xr, xc );
  if( ABS( contr ) < contr_thr / intvls )
    return NULL;

  feat = new_feature();
  ddata = feat_detection_data( feat );
  feat->img_pt.x = feat->x = ( c + xc ) * pow( 2.0, octv );
  feat->img_pt.y = feat->y = ( r + xr ) * pow( 2.0, octv );
  ddata->r = r;
  ddata->c = c;
  ddata->octv = octv;
  ddata->intvl = intvl;
  ddata->subintvl = xi;

  return feat;
}



/*
  Performs one step of extremum interpolation.  Based on Eqn. (3) in Lowe's
  paper.

  @param dog_pyr difference of Gaussians scale space pyramid
  @param octv octave of scale space
  @param intvl interval being interpolated
  @param r row being interpolated
  @param c column being interpolated
  @param xi output as interpolated subpixel increment to interval
  @param xr output as interpolated subpixel increment to row
  @param xc output as interpolated subpixel increment to col
*/

void interp_step( IplImage*** dog_pyr, int octv, int intvl, int r, int c,
		  double* xi, double* xr, double* xc )
{
  CvMat* dD, * H, * H_inv, X;
  double x[3] = { 0 };
  
  dD = deriv_3D( dog_pyr, octv, intvl, r, c );
  H = hessian_3D( dog_pyr, octv, intvl, r, c );
  H_inv = cvCreateMat( 3, 3, CV_64FC1 );
  cvInvert( H, H_inv, CV_SVD );
  cvInitMatHeader( &X, 3, 1, CV_64FC1, x, CV_AUTOSTEP );
  cvGEMM( H_inv, dD, -1, NULL, 0, &X, 0 );
  
  cvReleaseMat( &dD );
  cvReleaseMat( &H );
  cvReleaseMat( &H_inv );

  *xi = x[2];
  *xr = x[1];
  *xc = x[0];
}



/*
  Computes the partial derivatives in x, y, and scale of a pixel in the DoG
  scale space pyramid.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's octave in dog_pyr
  @param intvl pixel's interval in octv
  @param r pixel's image row
  @param c pixel's image col

  @return Returns the vector of partial derivatives for pixel I
    { dI/dx, dI/dy, dI/ds }^T as a CvMat*
*/
CvMat* deriv_3D( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  CvMat* dI;
  double dx, dy, ds;

  dx = ( pixval32f( dog_pyr[octv][intvl], r, c+1 ) -
	 pixval32f( dog_pyr[octv][intvl], r, c-1 ) ) / 2.0;
  dy = ( pixval32f( dog_pyr[octv][intvl], r+1, c ) -
	 pixval32f( dog_pyr[octv][intvl], r-1, c ) ) / 2.0;
  ds = ( pixval32f( dog_pyr[octv][intvl+1], r, c ) -
	 pixval32f( dog_pyr[octv][intvl-1], r, c ) ) / 2.0;
  
  dI = cvCreateMat( 3, 1, CV_64FC1 );
  cvmSet( dI, 0, 0, dx );
  cvmSet( dI, 1, 0, dy );
  cvmSet( dI, 2, 0, ds );

  return dI;
}



/*
  Computes the 3D Hessian matrix for a pixel in the DoG scale space pyramid.

  @param dog_pyr DoG scale space pyramid
  @param octv pixel's octave in dog_pyr
  @param intvl pixel's interval in octv
  @param r pixel's image row
  @param c pixel's image col

  @return Returns the Hessian matrix (below) for pixel I as a CvMat*

  / Ixx  Ixy  Ixs \ <BR>
  | Ixy  Iyy  Iys | <BR>
  \ Ixs  Iys  Iss /
*/
CvMat* hessian_3D( IplImage*** dog_pyr, int octv, int intvl, int r, int c )
{
  CvMat* H;
  double v, dxx, dyy, dss, dxy, dxs, dys;
  
  v = pixval32f( dog_pyr[octv][intvl], r, c );
  dxx = ( pixval32f( dog_pyr[octv][intvl], r, c+1 ) + 
	  pixval32f( dog_pyr[octv][intvl], r, c-1 ) - 2 * v );
  dyy = ( pixval32f( dog_pyr[octv][intvl], r+1, c ) +
	  pixval32f( dog_pyr[octv][intvl], r-1, c ) - 2 * v );
  dss = ( pixval32f( dog_pyr[octv][intvl+1], r, c ) +
	  pixval32f( dog_pyr[octv][intvl-1], r, c ) - 2 * v );
  dxy = ( pixval32f( dog_pyr[octv][intvl], r+1, c+1 ) -
	  pixval32f( dog_pyr[octv][intvl], r+1, c-1 ) -
	  pixval32f( dog_pyr[octv][intvl], r-1, c+1 ) +
	  pixval32f( dog_pyr[octv][intvl], r-1, c-1 ) ) / 4.0;
  dxs = ( pixval32f( dog_pyr[octv][intvl+1], r, c+1 ) -
	  pixval32f( dog_pyr[octv][intvl+1], r, c-1 ) -
	  pixval32f( dog_pyr[octv][intvl-1], r, c+1 ) +
	  pixval32f( dog_pyr[octv][intvl-1], r, c-1 ) ) / 4.0;
  dys = ( pixval32f( dog_pyr[octv][intvl+1], r+1, c ) -
	  pixval32f( dog_pyr[octv][intvl+1], r-1, c ) -
	  pixval32f( dog_pyr[octv][intvl-1], r+1, c ) +
	  pixval32f( dog_pyr[octv][intvl-1], r-1, c ) ) / 4.0;
  
  H = cvCreateMat( 3, 3, CV_64FC1 );
  cvmSet( H, 0, 0, dxx );
  cvmSet( H, 0, 1, dxy );
  cvmSet( H, 0, 2, dxs );
  cvmSet( H, 1, 0, dxy );
  cvmSet( H, 1, 1, dyy );
  cvmSet( H, 1, 2, dys );
  cvmSet( H, 2, 0, dxs );
  cvmSet( H, 2, 1, dys );
  cvmSet( H, 2, 2, dss );

  return H;
}



/*
  Calculates interpolated pixel contrast.  Based on Eqn. (3) in Lowe's
  paper.

  @param dog_pyr difference of Gaussians scale space pyramid
  @param octv octave of scale space
  @param intvl within-octave interval
  @param r pixel row
  @param c pixel column
  @param xi interpolated subpixel increment to interval
  @param xr interpolated subpixel increment to row
  @param xc interpolated subpixel increment to col

  @param Returns interpolated contrast.
*/
double interp_contr( IplImage*** dog_pyr, int octv, int intvl, int r,
		     int c, double xi, double xr, double xc )
{
  CvMat* dD, X, T;
  double t[1], x[3] = { xc, xr, xi };

  cvInitMatHeader( &X, 3, 1, CV_64FC1, x, CV_AUTOSTEP );
  cvInitMatHeader( &T, 1, 1, CV_64FC1, t, CV_AUTOSTEP );
  dD = deriv_3D( dog_pyr, octv, intvl, r, c );
  cvGEMM( dD, &X, 1, NULL, 0, &T,  CV_GEMM_A_T );
  cvReleaseMat( &dD );

  return pixval32f( dog_pyr[octv][intvl], r, c ) + t[0] * 0.5;
}



/*
  Allocates and initializes a new feature

  @return Returns a pointer to the new feature
*/
struct feature* new_feature( void )
{
  struct feature* feat;
  struct detection_data* ddata;

  feat = malloc( sizeof( struct feature ) );
  memset( feat, 0, sizeof( struct feature ) );
  ddata = malloc( sizeof( struct detection_data ) );
  memset( ddata, 0, sizeof( struct detection_data ) );
  feat->feature_data = ddata;
  feat->type = FEATURE_LOWE;

  return feat;
}



/*
  Determines whether a feature is too edge like to be stable by computing the
  ratio of principal curvatures at that feature.  Based on Section 4.1 of
  Lowe's paper.

  @param dog_img image from the DoG pyramid in which feature was detected
  @param r feature row
  @param c feature col
  @param curv_thr high threshold on ratio of principal curvatures

  @return Returns 0 if the feature at (r,c) in dog_img is sufficiently
    corner-like or 1 otherwise.
*/
int is_too_edge_like( IplImage* dog_img, int r, int c, int curv_thr )
{
  double d, dxx, dyy, dxy, tr, det;

  /* principal curvatures are computed using the trace and det of Hessian */
  d = pixval32f(dog_img, r, c);
  dxx = pixval32f( dog_img, r, c+1 ) + pixval32f( dog_img, r, c-1 ) - 2 * d;
  dyy = pixval32f( dog_img, r+1, c ) + pixval32f( dog_img, r-1, c ) - 2 * d;
  dxy = ( pixval32f(dog_img, r+1, c+1) - pixval32f(dog_img, r+1, c-1) -
	  pixval32f(dog_img, r-1, c+1) + pixval32f(dog_img, r-1, c-1) ) / 4.0;
  tr = dxx + dyy;
  det = dxx * dyy - dxy * dxy;

  /* negative determinant -> curvatures have different signs; reject feature */
  if( det <= 0 )
    return 1;

  if( tr * tr / det < ( curv_thr + 1.0 )*( curv_thr + 1.0 ) / curv_thr )
    return 0;
  return 1;
}



/*
  Calculates characteristic scale for each feature in an array.

  @param features array of features
  @param sigma amount of Gaussian smoothing per octave of scale space
  @param intvls intervals per octave of scale space
*/
void calc_feature_scales( CvSeq* features, double sigma, int intvls )
{
  struct feature* feat;
  struct detection_data* ddata;
  double intvl;
  int i, n;

  n = features->total;
  for( i = 0; i < n; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      ddata = feat_detection_data( feat );
      intvl = ddata->intvl + ddata->subintvl;
      feat->scl = sigma * pow( 2.0, ddata->octv + intvl / intvls );
      ddata->scl_octv = sigma * pow( 2.0, intvl / intvls );
    }
}



/*
  Halves feature coordinates and scale in case the input image was doubled
  prior to scale space construction.

  @param features array of features
*/
void adjust_for_img_dbl( CvSeq* features )
{
  struct feature* feat;
  int i, n;

  n = features->total;
  for( i = 0; i < n; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      feat->x /= 2.0;
      feat->y /= 2.0;
      feat->scl /= 2.0;
      feat->img_pt.x /= 2.0;
      feat->img_pt.y /= 2.0;
    }
}



/*
  Computes a canonical orientation for each image feature in an array.  Based
  on Section 5 of Lowe's paper.  This function adds features to the array when
  there is more than one dominant orientation at a given feature location.

  @param features an array of image features
  @param gauss_pyr Gaussian scale space pyramid
*/
void calc_feature_oris( CvSeq* features, IplImage*** gauss_pyr )
{
  struct feature* feat;
  struct detection_data* ddata;
  double* hist;
  double omax;
  int i, j, n = features->total;

  for( i = 0; i < n; i++ )
    {
      feat = malloc( sizeof( struct feature ) );
      cvSeqPopFront( features, feat );
      ddata = feat_detection_data( feat );
      hist = ori_hist( gauss_pyr[ddata->octv][ddata->intvl],
		       ddata->r, ddata->c, SIFT_ORI_HIST_BINS,
		       cvRound( SIFT_ORI_RADIUS * ddata->scl_octv ),
		       SIFT_ORI_SIG_FCTR * ddata->scl_octv );
      for( j = 0; j < SIFT_ORI_SMOOTH_PASSES; j++ )
	smooth_ori_hist( hist, SIFT_ORI_HIST_BINS );
      omax = dominant_ori( hist, SIFT_ORI_HIST_BINS );
      add_good_ori_features( features, hist, SIFT_ORI_HIST_BINS,
			     omax * SIFT_ORI_PEAK_RATIO, feat );
      free( ddata );
      free( feat );
      free( hist );
    }
}



/*
  Computes a gradient orientation histogram at a specified pixel.

  @param img image
  @param r pixel row
  @param c pixel col
  @param n number of histogram bins
  @param rad radius of region over which histogram is computed
  @param sigma std for Gaussian weighting of histogram entries

  @return Returns an n-element array containing an orientation histogram
    representing orientations between 0 and 2 PI.
*/
double* ori_hist( IplImage* img, int r, int c, int n, int rad, double sigma)
{
  double* hist;
  double mag, ori, w, exp_denom, PI2 = CV_PI * 2.0;
  int bin, i, j;

  hist = calloc( n, sizeof( double ) );
  exp_denom = 2.0 * sigma * sigma;
  for( i = -rad; i <= rad; i++ )
    for( j = -rad; j <= rad; j++ )
      if( calc_grad_mag_ori( img, r + i, c + j, &mag, &ori ) )
	{
	  w = exp( -( i*i + j*j ) / exp_denom );
	  bin = cvRound( n * ( ori + CV_PI ) / PI2 );
	  bin = ( bin < n )? bin : 0;
	  hist[bin] += w * mag;
	}

  return hist;
}



/*
  Calculates the gradient magnitude and orientation at a given pixel.

  @param img image
  @param r pixel row
  @param c pixel col
  @param mag output as gradient magnitude at pixel (r,c)
  @param ori output as gradient orientation at pixel (r,c)

  @return Returns 1 if the specified pixel is a valid one and sets mag and
    ori accordingly; otherwise returns 0
*/
int calc_grad_mag_ori( IplImage* img, int r, int c, double* mag, double* ori )
{
  double dx, dy;

  if( r > 0  &&  r < img->height - 1  &&  c > 0  &&  c < img->width - 1 )
    {
      dx = pixval32f( img, r, c+1 ) - pixval32f( img, r, c-1 );
      dy = pixval32f( img, r-1, c ) - pixval32f( img, r+1, c );
      *mag = sqrt( dx*dx + dy*dy );
      *ori = atan2( dy, dx );
      return 1;
    }

  else
    return 0;
}



/*
  Gaussian smooths an orientation histogram.

  @param hist an orientation histogram
  @param n number of bins
*/
void smooth_ori_hist( double* hist, int n )
{
  double prev, tmp, h0 = hist[0];
  int i;

  prev = hist[n-1];
  for( i = 0; i < n; i++ )
    {
      tmp = hist[i];
      hist[i] = 0.25 * prev + 0.5 * hist[i] + 
	0.25 * ( ( i+1 == n )? h0 : hist[i+1] );
      prev = tmp;
    }
}



/*
  Finds the magnitude of the dominant orientation in a histogram

  @param hist an orientation histogram
  @param n number of bins

  @return Returns the value of the largest bin in hist
*/
double dominant_ori( double* hist, int n )
{
  double omax;
  int maxbin, i;

  omax = hist[0];
  maxbin = 0;
  for( i = 1; i < n; i++ )
    if( hist[i] > omax )
      {
	omax = hist[i];
	maxbin = i;
      }
  return omax;
}



/*
  Interpolates a histogram peak from left, center, and right values
*/
#define interp_hist_peak( l, c, r ) ( 0.5 * ((l)-(r)) / ((l) - 2.0*(c) + (r)) )



/*
  Adds features to an array for every orientation in a histogram greater than
  a specified threshold.

  @param features new features are added to the end of this array
  @param hist orientation histogram
  @param n number of bins in hist
  @param mag_thr new features are added for entries in hist greater than this
  @param feat new features are clones of this with different orientations
*/
void add_good_ori_features( CvSeq* features, double* hist, int n,
			    double mag_thr, struct feature* feat )
{
  struct feature* new_feat;
  double bin, PI2 = CV_PI * 2.0;
  int l, r, i;

  for( i = 0; i < n; i++ )
    {
      l = ( i == 0 )? n - 1 : i-1;
      r = ( i + 1 ) % n;
      
      if( hist[i] > hist[l]  &&  hist[i] > hist[r]  &&  hist[i] >= mag_thr )
	{
	  bin = i + interp_hist_peak( hist[l], hist[i], hist[r] );
	  bin = ( bin < 0 )? n + bin : ( bin >= n )? bin - n : bin;
	  new_feat = clone_feature( feat );
	  new_feat->ori = ( ( PI2 * bin ) / n ) - CV_PI;
	  cvSeqPush( features, new_feat );
	  free( new_feat );
	}
    }
}



/*
  Makes a deep copy of a feature

  @param feat feature to be cloned

  @return Returns a deep copy of feat
*/
struct feature* clone_feature( struct feature* feat )
{
  struct feature* new_feat;
  struct detection_data* ddata;

  new_feat = new_feature();
  ddata = feat_detection_data( new_feat );
  memcpy( new_feat, feat, sizeof( struct feature ) );
  memcpy( ddata, feat_detection_data(feat), sizeof( struct detection_data ) );
  new_feat->feature_data = ddata;

  return new_feat;
}



/*
  Computes feature descriptors for features in an array.  Based on Section 6
  of Lowe's paper.

  @param features array of features
  @param gauss_pyr Gaussian scale space pyramid
  @param d width of 2D array of orientation histograms
  @param n number of bins per orientation histogram
*/
void compute_descriptors( CvSeq* features, IplImage*** gauss_pyr, int d, int n)
{
  struct feature* feat;
  struct detection_data* ddata;
  double*** hist;
  int i, k = features->total;

  for( i = 0; i < k; i++ )
    {
      feat = CV_GET_SEQ_ELEM( struct feature, features, i );
      ddata = feat_detection_data( feat );
      hist = descr_hist( gauss_pyr[ddata->octv][ddata->intvl], ddata->r,
			 ddata->c, feat->ori, ddata->scl_octv, d, n );
      hist_to_descr( hist, d, n, feat );
      release_descr_hist( &hist, d );
    }
}



/*
  Computes the 2D array of orientation histograms that form the feature
  descriptor.  Based on Section 6.1 of Lowe's paper.

  @param img image used in descriptor computation
  @param r row coord of center of orientation histogram array
  @param c column coord of center of orientation histogram array
  @param ori canonical orientation of feature whose descr is being computed
  @param scl scale relative to img of feature whose descr is being computed
  @param d width of 2d array of orientation histograms
  @param n bins per orientation histogram

  @return Returns a d x d array of n-bin orientation histograms.
*/
double*** descr_hist( IplImage* img, int r, int c, double ori,
		      double scl, int d, int n )
{
  double*** hist;
  double cos_t, sin_t, hist_width, exp_denom, r_rot, c_rot, grad_mag,
    grad_ori, w, rbin, cbin, obin, bins_per_rad, PI2 = 2.0 * CV_PI;
  int radius, i, j;

  hist = calloc( d, sizeof( double** ) );
  for( i = 0; i < d; i++ )
    {
      hist[i] = calloc( d, sizeof( double* ) );
      for( j = 0; j < d; j++ )
	hist[i][j] = calloc( n, sizeof( double ) );
    }
  
  cos_t = cos( ori );
  sin_t = sin( ori );
  bins_per_rad = n / PI2;
  exp_denom = d * d * 0.5;
  hist_width = SIFT_DESCR_SCL_FCTR * scl;
  radius = hist_width * sqrt(2) * ( d + 1.0 ) * 0.5 + 0.5;
  for( i = -radius; i <= radius; i++ )
    for( j = -radius; j <= radius; j++ )
      {
	/*
	  Calculate sample's histogram array coords rotated relative to ori.
	  Subtract 0.5 so samples that fall e.g. in the center of row 1 (i.e.
	  r_rot = 1.5) have full weight placed in row 1 after interpolation.
	*/
	c_rot = ( j * cos_t - i * sin_t ) / hist_width;
	r_rot = ( j * sin_t + i * cos_t ) / hist_width;
	rbin = r_rot + d / 2 - 0.5;
	cbin = c_rot + d / 2 - 0.5;
	
	if( rbin > -1.0  &&  rbin < d  &&  cbin > -1.0  &&  cbin < d )
	  if( calc_grad_mag_ori( img, r + i, c + j, &grad_mag, &grad_ori ))
	    {
	      grad_ori -= ori;
	      while( grad_ori < 0.0 )
		grad_ori += PI2;
	      while( grad_ori >= PI2 )
		grad_ori -= PI2;
	      
	      obin = grad_ori * bins_per_rad;
	      w = exp( -(c_rot * c_rot + r_rot * r_rot) / exp_denom );
	      interp_hist_entry( hist, rbin, cbin, obin, grad_mag * w, d, n );
	    }
      }

  return hist;
}



/*
  Interpolates an entry into the array of orientation histograms that form
  the feature descriptor.

  @param hist 2D array of orientation histograms
  @param rbin sub-bin row coordinate of entry
  @param cbin sub-bin column coordinate of entry
  @param obin sub-bin orientation coordinate of entry
  @param mag size of entry
  @param d width of 2D array of orientation histograms
  @param n number of bins per orientation histogram
*/
void interp_hist_entry( double*** hist, double rbin, double cbin,
			double obin, double mag, int d, int n )
{
  double d_r, d_c, d_o, v_r, v_c, v_o;
  double** row, * h;
  int r0, c0, o0, rb, cb, ob, r, c, o;

  r0 = cvFloor( rbin );
  c0 = cvFloor( cbin );
  o0 = cvFloor( obin );
  d_r = rbin - r0;
  d_c = cbin - c0;
  d_o = obin - o0;

  /*
    The entry is distributed into up to 8 bins.  Each entry into a bin
    is multiplied by a weight of 1 - d for each dimension, where d is the
    distance from the center value of the bin measured in bin units.
  */
  for( r = 0; r <= 1; r++ )
    {
      rb = r0 + r;
      if( rb >= 0  &&  rb < d )
	{
	  v_r = mag * ( ( r == 0 )? 1.0 - d_r : d_r );
	  row = hist[rb];
	  for( c = 0; c <= 1; c++ )
	    {
	      cb = c0 + c;
	      if( cb >= 0  &&  cb < d )
		{
		  v_c = v_r * ( ( c == 0 )? 1.0 - d_c : d_c );
		  h = row[cb];
		  for( o = 0; o <= 1; o++ )
		    {
		      ob = ( o0 + o ) % n;
		      v_o = v_c * ( ( o == 0 )? 1.0 - d_o : d_o );
		      h[ob] += v_o;
		    }
		}
	    }
	}
    }
}



/*
  Converts the 2D array of orientation histograms into a feature's descriptor
  vector.
  
  @param hist 2D array of orientation histograms
  @param d width of hist
  @param n bins per histogram
  @param feat feature into which to store descriptor
*/
void hist_to_descr( double*** hist, int d, int n, struct feature* feat )
{
  int int_val, i, r, c, o, k = 0;

  for( r = 0; r < d; r++ )
    for( c = 0; c < d; c++ )
      for( o = 0; o < n; o++ )
	feat->descr[k++] = hist[r][c][o];

  feat->d = k;
  normalize_descr( feat );
  for( i = 0; i < k; i++ )
    if( feat->descr[i] > SIFT_DESCR_MAG_THR )
      feat->descr[i] = SIFT_DESCR_MAG_THR;
  normalize_descr( feat );

  /* convert floating-point descriptor to integer valued descriptor */
  for( i = 0; i < k; i++ )
    {
      int_val = SIFT_INT_DESCR_FCTR * feat->descr[i];
      feat->descr[i] = MIN( 255, int_val );
    }
}



/*
  Normalizes a feature's descriptor vector to unitl length

  @param feat feature
*/
void normalize_descr( struct feature* feat )
{
  double cur, len_inv, len_sq = 0.0;
  int i, d = feat->d;

  for( i = 0; i < d; i++ )
    {
      cur = feat->descr[i];
      len_sq += cur*cur;
    }
  len_inv = 1.0 / sqrt( len_sq );
  for( i = 0; i < d; i++ )
    feat->descr[i] *= len_inv;
}



/*
  Compares features for a decreasing-scale ordering.  Intended for use with
  CvSeqSort

  @param feat1 first feature
  @param feat2 second feature
  @param param unused

  @return Returns 1 if feat1's scale is greater than feat2's, -1 if vice versa,
    and 0 if their scales are equal
*/
int feature_cmp( void* feat1, void* feat2, void* param )
{
  struct feature* f1 = (struct feature*) feat1;
  struct feature* f2 = (struct feature*) feat2;

  if( f1->scl < f2->scl )
    return 1;
  if( f1->scl > f2->scl )
    return -1;
  return 0;
}



/*
  De-allocates memory held by a descriptor histogram

  @param hist pointer to a 2D array of orientation histograms
  @param d width of hist
*/
void release_descr_hist( double**** hist, int d )
{
  int i, j;

  for( i = 0; i < d; i++)
    {
      for( j = 0; j < d; j++ )
	free( (*hist)[i][j] );
      free( (*hist)[i] );
    }
  free( *hist );
  *hist = NULL;
}

//GST-Stuff
// static GType
// gst_type_sifttool_op_mode (void)
// {
//   static GType sifttool_op_mode = 0;
//   static const GEnumValue functions[] = {
//     {MATCH, "matches two sets of features", "match"},
//     {MERGE, "merges two sets of features, set offset values!", "merge"},
//     {TRANSFORMPOINT,"transforms the point given by the offset values from the small image(2) to the big(1)!", "transformpoint"},
//     {TRANSFORM,"gives the transformation equation", "transform"}, 
//     {0, NULL, NULL}
//   };
// 
//   if (!sifttool_op_mode) {
//     sifttool_op_mode =
//         g_enum_register_static ("SiftToolOpMode", functions);
//   }
//   return sifttool_op_mode;
// }


/* GObject vmethod implementations */

static void
gst_siftfolder_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "siftfolder",
    "Gst-Plugins-Robotix",
    "Match Sift-Features with folder",
    "Hannes Bistry <<hannesb@gmx.de>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the siftfolder's class */
static void
gst_siftfolder_class_init (GstsiftfolderClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_siftfolder_set_property;
  gobject_class->get_property = gst_siftfolder_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM,
      g_param_spec_double ("maxsim", "Maxsim", "The second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.49, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MAXSIM2,
      g_param_spec_double ("maxsim2", "Maxsim2", "In case of a really good match (distance<gooddist), the second best match needs to be (1/(maxsim)) further away(1==can even be equal, 0=never match)",
          0,1,0.9, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_GOODDIST,
      g_param_spec_double ("gooddist", "Gooddist", "What is a good match, is square distance < than this value",
          0,1,0, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_THRESHOLD,
      g_param_spec_double ("threshold", "Threshold", "Threshold for Ransac",
          0,1,0.01, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_REMATCHDOUBLE,
      g_param_spec_boolean ("rematchdouble", "Rematchdouble", "Whether to kill double matches and rematch them in one direction.",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SHOWINLIERS,
      g_param_spec_boolean ("showinliers", "Showinliers", "Show only valid matches after Ransac.",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_DIRECTORY,
      g_param_spec_string ("directory", "Directory", "Use this directory to build models.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_NTOTAL,
      g_param_spec_int ("ntotal", "Ntotal", "Total Number of Instances on the same directory",
          -1,G_MAXINT,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_THISN,
      g_param_spec_int ("thisn", "Thisn", "In ntotal is set, set this number between 0 and ntotal-1",
          -1,G_MAXINT,-1, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_THRESH,
      g_param_spec_uint ("thresh", "Thresh", "Threshhold width/height where we scale down resolution",
          1,G_MAXUINT,G_MAXUINT, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, PROP_HFOV,
      g_param_spec_double ("hfov", "Hfov", "Horizontal field of view (<0 is ignored)",
          -1,G_MAXDOUBLE,-1, G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_UNDDIRECTORY,
      g_param_spec_string ("unddirectory", "Unddirectory", "Undistort directory, if enabled, it overwrites hfov setting with calibration filoes and takes distortion_coeffs.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_OTHERWAY,
      g_param_spec_boolean ("otherway", "Otherway", "Match from the Objects to Image instead...",
          FALSE, G_PARAM_READWRITE));  

  gstelement_class->change_state = gst_siftfolder_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static gboolean little_endian;

static void
gst_siftfolder_init (Gstsiftfolder * filter,
    GstsiftfolderClass * gclass)
{
  int x=1;
  little_endian=(gboolean) (*(char*)&x); 
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_siftfolder_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_siftfolder_get_sink_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_siftfolder_chain));
  //gst_pad_set_event_function (filter->sinkpad1,GST_DEBUG_FUNCPTR(gst_siftfolder_event));

  
  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_siftfolder_get_src_caps));
  gst_pad_set_setcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_siftfolder_set_caps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->mutex=g_mutex_new();
  filter->maxsim=0.49;
  filter->maxsim2=0.9;
  filter->threshold=0.01;
  filter->showinliers=FALSE;
  filter->directory=NULL;
  filter->siftheap=0;
  filter->double_image=1;
  filter->th=0;

  filter->ntotal=-1;
  filter->thisn=-1;
  filter->thresh=G_MAXUINT;
  filter->rematchdouble=TRUE;
  
  
  
  filter->c_x=-1;
  filter->c_y=-1;
  filter->f_x=-1; 
  filter->f_y=-1;
  filter->k_1=0;
  filter->k_2=0;
  filter->k_3=0;
  filter->k_4=0;
  filter->k_5=0;
  
  filter->hfov=0;
  CvMat* distortion_coeffs=NULL;
  CvMat* intrinsic_matrix=NULL;
  filter->unddirectory=NULL;
  filter->otherway=1;
}

gboolean gst_element_is_null(GstElement* element){
  if (element==NULL)return FALSE;
  if (!GST_IS_ELEMENT(element))return FALSE;
  GstState current;
  gst_element_get_state(element,&current,NULL,1);
  if (current == GST_STATE_NULL) return TRUE;
  else return FALSE;
}

static gboolean gst_undistort_init_cvstuff(Gstsiftfolder * filter){
  if (!filter->directory)return FALSE;
  gchar * this_object_name = (gchar *) malloc(strlen(filter->unddirectory)+20);
  strcpy(this_object_name, filter->unddirectory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Intrinsics.xml");
  //CvMat *intrinsic = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->intrinsic_matrix)filter->intrinsic_matrix= (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->intrinsic_matrix){
    g_print("Could not load %s\n",this_object_name);
    return FALSE;
  }
  strcpy(this_object_name, filter->unddirectory);
  if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
  strcat(this_object_name,"Distortion.xml");  
  //CvMat *distortion = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->distortion_coeffs)filter->distortion_coeffs = (CvMat*)cvLoad(this_object_name,NULL,NULL,NULL);
  if (!filter->distortion_coeffs){
    g_print("Could not load %s\n",this_object_name);    
    return FALSE;
  }
return TRUE;
}


static void
gst_siftfolder_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstsiftfolder *filter = GST_SIFTFOLDER (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_MAXSIM:
      filter->maxsim=g_value_get_double (value);
      break;
    case PROP_MAXSIM2:
      filter->maxsim2=g_value_get_double (value);
      break;
    case PROP_GOODDIST:
      filter->gooddist=g_value_get_double (value);
      break;
    case PROP_THRESHOLD:
      filter->threshold=g_value_get_double (value);
      break;
   case PROP_REMATCHDOUBLE:
      filter->rematchdouble = g_value_get_boolean (value);
      break;
   case PROP_SHOWINLIERS:
      filter->showinliers = g_value_get_boolean (value);
      break;
   case PROP_DIRECTORY:
      if(filter->directory)g_free (filter->directory);
      if (filter->th>0){
	pthread_cancel(filter->th);
	filter->th=0;
      }
      filter->directory = g_value_dup_string (value);
      /*if (!gst_siftfolder_load_gallery(filter)){
	if(filter->directory)g_free (filter->directory);
	filter->directory=NULL;
      }
      else{
	if (!gst_element_is_null(filter))pthread_create (&filter->th  , NULL, watchthread, filter);
	}*/
      break;
    case PROP_NTOTAL:
      filter->ntotal=g_value_get_int (value);
      break;
    case PROP_THISN:
      filter->thisn=g_value_get_int (value);
      break;
    case PROP_THRESH:
      filter->thresh=g_value_get_uint (value);
      break;
    case PROP_HFOV:
      filter->hfov=g_value_get_double (value);
      break;   
    case PROP_UNDDIRECTORY:
      if(filter->unddirectory)g_free (filter->unddirectory);   
      filter->unddirectory = g_value_dup_string (value);   
      gst_undistort_init_cvstuff(filter);
      break; 
   case PROP_OTHERWAY:
      filter->otherway = g_value_get_boolean (value);
      break;      
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_siftfolder_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstsiftfolder *filter = GST_SIFTFOLDER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_MAXSIM:
      g_value_set_double(value, filter->maxsim);
      break;
    case PROP_MAXSIM2:
      g_value_set_double(value, filter->maxsim2);
      break;
    case PROP_GOODDIST:
      g_value_set_double(value, filter->gooddist);
      break;
    case PROP_THRESHOLD:
      g_value_set_double(value, filter->threshold);
      break;
    case PROP_REMATCHDOUBLE:
      g_value_set_boolean (value, filter->rematchdouble);
      break;
    case PROP_SHOWINLIERS:
      g_value_set_boolean (value, filter->showinliers);
      break;
    case PROP_DIRECTORY:     
      g_value_set_string (value, filter->directory);
      break;
    case PROP_NTOTAL:
      g_value_set_int (value,filter->ntotal);
      break;
    case PROP_THISN:
      g_value_set_int (value,filter->thisn);
      break;
    case PROP_THRESH:
      g_value_set_uint (value,filter->thresh);
      break;
    case PROP_HFOV:
      g_value_set_double (value,filter->hfov);
      break; 
    case PROP_UNDDIRECTORY:     
      g_value_set_string (value, filter->unddirectory);
    break;    
    case PROP_OTHERWAY:
      g_value_set_boolean (value, filter->otherway);
    break; 
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}
//
GstBuffer* gst_siftfolder_buf_from_oxfd(char * input){
  struct feature* f;
  int i, j, n, d;
  double x, y, a, b, c, dv;
  char* temp=input;
  GstBuffer* result;
//get first two double values in d and n
  n=g_strtod (temp,&temp);
  d=g_strtod (temp,&temp);  

  if( d > FEATURE_MAX_D )
    {
      fprintf( stderr, "Warning: descriptor too long, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }
  
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;

  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      x=g_strtod (temp,&temp);
      y=g_strtod (temp,&temp); 
      a=g_strtod (temp,&temp);
      b=g_strtod (temp,&temp); 
      c=g_strtod (temp,&temp);        
      /**/
      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].a = a;
      f[i].b = b;
      f[i].c = c;
      f[i].d = d;
      f[i].type = FEATURE_OXFD;
      
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
	  dv=g_strtod (temp,&temp);   
	  f[i].descr[j] = dv;
	  g_print("%f",dv);
	}

      f[i].scl = f[i].ori = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }
  return result;
}

GstBuffer* gst_siftfolder_buf_from_lowe(char * input){
  struct feature* f;
  int i, j, n, d;
  double x, y, s, o, dv;
  char* temp=input;
  GstBuffer* result;

  n=(int)g_strtod (temp,&temp);
  d=(int)g_strtod (temp,&temp);  
  g_print("Processing %d features with each %d values",d,n);
  if( d > FEATURE_MAX_D )
    {
      fprintf( stderr, "Warning: descriptor too long, %s, line %d\n",
	       __FILE__, __LINE__ );
      return NULL;
    }
  
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;

  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      x=g_strtod (temp,&temp);
      y=g_strtod (temp,&temp); 
      s=g_strtod (temp,&temp);
      o=g_strtod (temp,&temp); 

      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].scl = s;
      f[i].ori = o;
      f[i].d = d;
      f[i].type = FEATURE_LOWE;
      g_print("Processing Nr. %d of %d and got values %f %f %f %f:\n",i,n,x,y,s,o);
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
          dv=g_strtod (temp,&temp); 
	  f[i].descr[j] = dv;
	  g_print("%f",dv);
	}

      f[i].a = f[i].b = f[i].c = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }

  return result;
}

GstBuffer* gst_siftfolder_buf_from_lowelight(GstBuffer * input){
  struct feature* f;
  struct lowelight* l;
  int i, j, n, d;
  double x, y, s, o, dv;
  GstBuffer* result;

  n=(input->size)/(sizeof(struct lowelight));
  d=128;
  result=gst_buffer_new_and_alloc (n*sizeof(struct feature));
  f = (struct feature*)result->data;
  l = (struct lowelight*)input->data;
  for( i = 0; i < n; i++ )
    {
      /* read affine region parameters */
      if (little_endian){
        x=l[i].x;
        y=l[i].y; 
        s=l[i].scl;
        o=l[i].ori; 
      }
      else{
        x=DoubleSwap(l[i].x);
        y=DoubleSwap(l[i].y); 
        s=DoubleSwap(l[i].scl);
        o=DoubleSwap(l[i].ori); 
      }

      f[i].img_pt.x = f[i].x = x;
      f[i].img_pt.y = f[i].y = y;
      f[i].scl = s;
      f[i].ori = o;
      f[i].d = d;
      f[i].type = FEATURE_LOWE;
      //g_print("Processing Nr. %d of %d and got values %f %f %f %f:\n",i,n,x,y,s,o);
      /* read descriptor */
      for( j = 0; j < d; j++ )
	{
          dv=l[i].data[j];
	  f[i].descr[j] = dv;
	  //g_print("%f",dv);
	}

      f[i].a = f[i].b = f[i].c = 0;
      f[i].category = 0;
      f[i].fwd_match = f[i].bck_match = f[i].mdl_match = NULL;
      f[i].mdl_pt.x = f[i].mdl_pt.y = -1;
      f[i].feature_data = NULL;
    }

  return result;
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_siftfolder_set_caps (GstPad * pad, GstCaps * caps)
{
  gchar* capschar=gst_caps_to_string (caps);
  g_print("Sifttools set caps with %s!!\n",capschar);
  g_free(capschar);
  Gstsiftfolder *filter;
  GstPad *otherpad;

  filter = GST_SIFTFOLDER (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);
  gboolean ret=TRUE;

  if (pad == filter->sinkpad){
     GstStructure *structure = gst_caps_get_structure (caps, 0);
     const gchar *mime;
     mime = gst_structure_get_name (structure);
     if (strcmp (mime, "sift/raw") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));
	filter->mode=0;
     }
     else if (strcmp (mime, "sift/lowe") == 0) filter->mode=1;
     else if (strcmp (mime, "sift/oxf") == 0) filter->mode=2;
     else if (strcmp (mime, "sift/lowelight") == 0){
	gst_structure_get_int(structure,"width",&(filter->width1));
	gst_structure_get_int(structure,"height",&(filter->height1));	
	filter->mode=3;
     }
     else ret=FALSE;
     gst_structure_get_double(structure,"c_x",&(filter->c_x));    
     gst_structure_get_double(structure,"c_y",&(filter->c_y));     
     gst_structure_get_double(structure,"f_x",&(filter->f_x));  
     gst_structure_get_double(structure,"f_x",&(filter->f_y));  
     gst_structure_get_double(structure,"k_1",&(filter->k_1));     
     gst_structure_get_double(structure,"k_2",&(filter->k_2));     
     gst_structure_get_double(structure,"k_3",&(filter->k_3)); 
     gst_structure_get_double(structure,"k_4",&(filter->k_4));     
     gst_structure_get_double(structure,"k_5",&(filter->k_5));     
	   
  }
  else ret=TRUE;
  //return gst_pad_set_caps (otherpad, caps);
  return ret;
}



static GstCaps *
gst_siftfolder_get_sink_caps (GstPad *pad){
  g_print("Start get sink caps!!\n");
  Gstsiftfolder *filter;
  filter = GST_SIFTFOLDER (GST_OBJECT_PARENT (pad));
  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  
}

static GstCaps *
gst_siftfolder_get_src_caps (GstPad *pad){
  g_print("Start get src caps!!\n");
  Gstsiftfolder *filter;

  filter = GST_SIFTFOLDER (GST_OBJECT_PARENT (pad));



  GstCaps* outcaps;
  outcaps=gst_caps_new_simple ("uber-coordinates",
		NULL);
  //outcaps=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  //g_print("Returning %s!!\n",gst_caps_to_string(outcaps));
  return outcaps;
}

gboolean addfile(Gstsiftfolder *filter, const gchar* object_name){//needs to be called with mutex
   gchar * this_object_name = (gchar *) malloc(strlen(filter->directory)+strlen(object_name) + 2);
   strcpy(this_object_name, filter->directory);
   if (this_object_name[strlen(this_object_name)-2]!='/')strcat(this_object_name,"/");
   strcat(this_object_name, object_name);
   g_print("Gstsiftfolder reading %s\n",this_object_name);
   if (g_file_test(this_object_name,G_FILE_TEST_IS_REGULAR)){//seems to be broken
	  //get CVimage current
     IplImage* process = 0;
     process=cvLoadImage(this_object_name,0);//load as grayscale
     if (process){
       g_print("Gstsiftfolder survived reading %s and threshold is %u\n",this_object_name,filter->thresh);
       if ((process->width>filter->thresh)||(process->height>filter->thresh)){
	 g_print("Resizing %s\n",this_object_name);
	 double factor;
	 if (process->width>process->height)factor=process->width/filter->thresh;
	 else factor=process->height/filter->thresh;
	  IplImage *destination = cvCreateImage
( cvSize((int)((double)(process->width)/factor) , (int)((double)(process->height)/factor )),
                                     8, 1 );
	 cvResize(process, destination,CV_INTER_LINEAR);
	 cvReleaseImage(&process);
	 process=destination;
	}
	    //
       struct siftset* newset;
       struct feature* newfeat=0;
       int num_features;
       newset=make_siftset(process,SIFT_INTVLS,filter->double_image);
       num_features=_sift_features(process,&newfeat ,newset, SIFT_INTVLS, SIFT_SIGMA, SIFT_CONTR_THR,
 			 SIFT_CURV_THR, filter->double_image, SIFT_DESCR_WIDTH,
			 SIFT_DESCR_HIST_BINS );
       if (!num_features){
	 g_print("Gstsiftfolder adding %s failed, no features!\n",this_object_name);
	 killsiftset(newset);
         cvReleaseImage(&process);
         free(this_object_name);
         return FALSE;
	}
       struct kd_node* node;
       node = kdtree_build( newfeat, num_features );
       //add facts to siftheap
       struct siftheap* current=(struct siftheap*)malloc(sizeof(struct siftheap));
       current->numfeat=num_features;
       current->feat=newfeat;
       current->node=node;
       current->width=process->width;
       current->height=process->height;
       strncpy(current->name,object_name,29);
       //if (strlen(object_name)<30)strcpy(current->name,object_name);
       //else memcpy(current->name,object_name,30);
       filter->siftheap=g_list_append(filter->siftheap,current);
       //tidy up
       killsiftset(newset);
       cvReleaseImage(&process);
       free(this_object_name);
       return TRUE;
      }
      else{ 
	g_print("Gstsiftfolder reading %s failed, OpenCV could not generate image\n",this_object_name);
	free(this_object_name);
      }
   }
   else g_print("Gstsiftfolder skipping %s, no regular file\n",this_object_name);
return FALSE;
}


void * watchthread (void* data)
{
  g_print("Start Watchthread!!\n");
  //sleep(10);
  Gstsiftfolder *filter;  
  filter = GST_SIFTFOLDER (data);
  char filename[512];
  filter->fd = inotify_init();
  if (filter->fd<0){
    printf("Siftfolder Watchthread: inotify init failed with error # %d, Errno %d\n", filter->fd, errno);
    return NULL;
  }
  filter->wd = inotify_add_watch( filter->fd,filter->directory , IN_CLOSE_WRITE | IN_MOVED_TO);
  //filter->wd = inotify_add_watch( filter->fd,filter->directory , IN_MOVED_TO);
  //filter->wd = inotify_add_watch( filter->fd,filter->directory , IN_CLOSE_WRITE);
  int length=0,i=0;
  while(42){
     if (i==0)length=read( filter->fd, filter->buffer, EVENT_BUF_LEN );
     if (i<0) printf("Siftfolder: inotify read returned Error %d\n", length);
     if (i<length){
       gboolean filename_valid=TRUE;
       int cntdots=0,j=0;	 
       struct inotify_event *pevent    = (struct inotify_event *) &filter->buffer[i];
       g_print("Siftfolder Watchthread: Event NR. %d, Filename: %s\n",i,pevent->name); 
       i += sizeof(struct inotify_event) + pevent->len;
      //strncpy(filename,filter->directory , 512);
      //if (filter->directory[strlen(filter->directory)-1]!='/') strncat(filename,"/",512);
       strncat(filename,pevent->name,512);
       g_print("Siftfolder Watchthread: Complete Name: %s\n",filename);
       for (j=0;j<strlen(filename);j++){
	  if (filename[j]=='.')cntdots++;
	  if ((filename[j]=='/')||(filename[j]=='\\'))cntdots=0;	  
       }
       if (cntdots!=1) filename_valid=FALSE;
       if (filename_valid){
         g_mutex_lock(filter->mutex);
         if((strstr(filename,".jpg"))||(strstr(filename,".png"))||(strstr(filename,".png"))||(strstr(filename,".JPG"))||(strstr(filename,".jpeg")))addfile(filter, filename);
         g_mutex_unlock(filter->mutex);
       }
       memset(filename,0,512);
       }
       if (i==length)i=0;
  }
  return NULL;
}


//function to load a gallery
static gboolean gst_siftfolder_load_gallery(Gstsiftfolder * filter)
{
  g_mutex_lock(filter->mutex);
  //unref old siftheap
  filter->siftheap=NULL;//:)
  //iterate through gallery, pack all SIFT-vectors in a list
  //save identities while we go
  
  if( !g_file_test(filter->directory, G_FILE_TEST_IS_DIR) )
    return FALSE;
  
  GDir * gallery_dir = g_dir_open (filter->directory,0, NULL);
  const gchar * object_name;
  for ( object_name = g_dir_read_name(gallery_dir);
	object_name != NULL;
	object_name = g_dir_read_name(gallery_dir) ) {
          addfile(filter, object_name);
        }
  g_dir_close(gallery_dir);
  g_mutex_unlock(filter->mutex);
  return TRUE;
}


/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_siftfolder_chain (GstPad * pad, GstBuffer * buf)
{
  g_print("Start matcher!!\n");
  buf=gst_buffer_make_writable(buf);
  Gstsiftfolder *filter;
  filter = GST_SIFTFOLDER (GST_OBJECT_PARENT (pad)); 
  
  if (filter->siftheap==NULL){
      if (!gst_siftfolder_load_gallery(filter)){
        g_print("Warning, Gstsiftfolder can not work because no objects to match, set directory path!!\n");	
	if(filter->directory)g_free (filter->directory);
	filter->directory=NULL;
      }
      else{
	if (!gst_element_is_null(filter))pthread_create (&filter->th  , NULL, watchthread, filter);
	}
      //g_mutex_unlock(filter->mutex);
      //return GST_FLOW_ERROR;
  }
  


 
  //get mutex
  g_mutex_lock(filter->mutex);


  //exchange buffer(unref old)
  switch(filter->mode){
    case 0:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=buf;
      filter->feat1=(struct feature*)buf->data;
      filter->n1=(buf->size)/(sizeof(struct feature));
      break;
    case 1:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_siftfolder_buf_from_lowe((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));     
      gst_buffer_unref(buf);  
      break;
    case 2:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_siftfolder_buf_from_oxfd((char*)buf->data);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));   
      gst_buffer_unref(buf);   
      break;
    case 3:
      if (filter->s1)gst_buffer_unref(filter->s1);
      filter->s1=gst_siftfolder_buf_from_lowelight(buf);
      filter->s1->timestamp=buf->timestamp;
      filter->s1->duration=buf->duration;
      filter->s1->offset=buf->offset;
      filter->s1->offset_end=buf->offset_end;
      filter->feat1=(struct feature*)filter->s1->data;  
      filter->n1=(filter->s1->size)/(sizeof(struct feature));    
      gst_buffer_unref(buf);   
      break;
    default:
      g_print("Serious Error!!\n");
      return GST_FLOW_ERROR; 
    }

    //g_print("Start Eval!!\n");

    
    
    GstFlowReturn ret;
    struct feature* feat;
    gint i,m=0,k;
    double d0, d1;
    struct feature** nbrs=NULL;
    CvMat* H=NULL;
    struct feature invalid;
    GList *temp;
    struct siftheap* current=NULL;
    struct kd_node* kd_root;

    if (filter->rematchdouble||filter->otherway)filter->kdback = kdtree_build( filter->feat1, filter->n1 );
    for(temp=filter->siftheap;(temp);temp=g_list_next(temp))if ((filter->ntotal==-1)||(g_list_position (filter->siftheap,temp)%filter->ntotal==filter->thisn)){
      current=((struct siftheap*)(temp->data));
	for( i = 0; i < current->numfeat; i++ ){
	  //current->feat->bck_match=NULL;
	  //current->feat->fwd_match=NULL;
	  current->feat[i].bck_match=NULL;
	  current->feat[i].fwd_match=NULL;	
	}
	//g_print("allocated %d for a maximum number of %d features!!\n",filter->n1*sizeof(struct coordinates),filter->n1);
	for( i = 0; i < filter->n1; i++ ){
		    /*g_print("Checking 1  %d, because %p points to %p.\n",
		    i,
		    &filter->feat1[i],
		    filter->feat1[i].fwd_match
		  );*/     
		  filter->feat1[i].fwd_match=NULL;
	}      
      if (filter->otherway){
	 for( i = 0; i < current->numfeat; i++ ){
	  feat = &current->feat[i];
	  k = kdtree_bbf_knn( filter->kdback, feat, 2, &nbrs, 200 );//todo 200 to parameter
	  if( k == 2 )
	    {    
	    d0 = descr_dist_sq( feat, nbrs[0] );
	    d1 = descr_dist_sq( feat, nbrs[1] );
	    if(( d0 < d1 * filter->maxsim )||((d0<filter->gooddist)&&(d0 < d1 * filter->maxsim2)))//0.49 to parameter
	      {
		m++;
		g_assert(current->feat[i].fwd_match == 0);
		current->feat[i].fwd_match = nbrs[0];
	      if (nbrs[0]->bck_match==NULL)
		  nbrs[0]->bck_match = &filter->feat1[i];
		else
		  nbrs[0]->bck_match=&invalid;	      
  // 	      g_print("feature %d x=%d y=%d corresponds to %p with x=%d y=%d and therefore the fwdmatch pointer is set to %p and backpointer is set to %p  / but should be %p!!\n",
  // 		      i,
  // 		      (int)filter->feat1[i].x,
  // 		      (int)filter->feat1[i].y,
  // 		      nbrs[0],
  // 		      (int)nbrs[0]->x,
  // 		      (int)nbrs[0]->y,
  // 		      filter->feat1[i].fwd_match,
  // 		      nbrs[0]->bck_match,
  // 		      &filter->feat1[i]
  // 		     );
		//g_print("1 address of fwd_match %p\n", filter->feat1[16].fwd_match);
	      }
	  }
	free( nbrs );	   
	   
	   
	   
	   
	 }
      }
      else{
	kd_root=current->node;
	//delete backwards matches:

	for( i = 0; i < filter->n1; i++ )
	  {
	  //g_print("In for!!\n");
	  //feat = filter->feat1 + i;
	  feat = &filter->feat1[i];
	  k = kdtree_bbf_knn( kd_root, feat, 2, &nbrs, 200 );//todo 200 to parameter
	  if( k == 2 )
	    {
	    
	    d0 = descr_dist_sq( feat, nbrs[0] );
	    d1 = descr_dist_sq( feat, nbrs[1] );
	    if(( d0 < d1 * filter->maxsim )||((d0<filter->gooddist)&&(d0 < d1 * filter->maxsim2)))//0.49 to parameter
	      {
		m++;
		g_assert(filter->feat1[i].fwd_match == 0);
		filter->feat1[i].fwd_match = nbrs[0];
	      if (nbrs[0]->bck_match==NULL)
		  nbrs[0]->bck_match = &filter->feat1[i];
		else
		  nbrs[0]->bck_match=&invalid;	      
  // 	      g_print("feature %d x=%d y=%d corresponds to %p with x=%d y=%d and therefore the fwdmatch pointer is set to %p and backpointer is set to %p  / but should be %p!!\n",
  // 		      i,
  // 		      (int)filter->feat1[i].x,
  // 		      (int)filter->feat1[i].y,
  // 		      nbrs[0],
  // 		      (int)nbrs[0]->x,
  // 		      (int)nbrs[0]->y,
  // 		      filter->feat1[i].fwd_match,
  // 		      nbrs[0]->bck_match,
  // 		      &filter->feat1[i]
  // 		     );
		//g_print("1 address of fwd_match %p\n", filter->feat1[16].fwd_match);
	      }
	  }
	free( nbrs );
      }
      g_print("Found %d matches for object name %s!\n",m,current->name);
      m=0; 
      
      //g_print("2 address of fwd_match %p\n", filter->feat1[16].fwd_match);
      
      //if (filter->rematchdouble){//now the masterpiece, first the normal ones
  //     for( i = 0; i < filter->n1; i++ )g_print("Checking 2 %d, because %p points to %p.\n",
  // 		  i,
  // 		  &filter->feat1[i],
  // 		  filter->feat1[i].fwd_match
  // 		 );
      if (filter->rematchdouble){
	//delete
	for( i = 0; i < filter->n1; i++ ){
	  if ((filter->feat1[i].fwd_match != NULL) && (filter->feat1[i].fwd_match->bck_match!=&filter->feat1[i])){
  // 	  g_print("deleting double %d, because %p points to %p and that points back to %p\n",
  // 		  i,
  // 		  &filter->feat1[i],
  // 		  filter->feat1[i].fwd_match,
  // 		  filter->feat1[i].fwd_match->bck_match
  // 		 );
	    filter->feat1[i].fwd_match=NULL;
	    }
	}
      //rematch
	struct kd_node* kdback=filter->kdback;
	//kdback = kdtree_build( filter->feat1, filter->n1 );
	for( i = 0; i < current->numfeat; i++ ){
	  if (current->feat[i].bck_match==&invalid){
	    k = kdtree_bbf_knn( kdback,&current->feat[i], 2, &nbrs, 200 );
	    if (k==2){
	      current->feat[i].bck_match=nbrs[0];
	      nbrs[0]->fwd_match=&current->feat[i];
	    } 
	  free( nbrs );	
	  }
	}    
      //kdtree_release( kdback );
      }
    }
    //Ransac transformation matrix generation
    int n_in;
    struct feature** inliers=NULL;
    if (!filter->otherway)H = ransac_xform( filter->feat1, filter->n1, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, &inliers, &n_in );
    else H = ransac_xform( current->feat, current->numfeat, FEATURE_FWD_MATCH, lsq_homog, 4, filter->threshold,
		      homog_xfer_err, 3.0, &inliers, &n_in );
    
    if (H){
        
        
        GstBuffer *outbuf;
        outbuf=gst_buffer_new_and_alloc(sizeof(struct uber_coordinates));
	struct uber_coordinates* bufdata=(struct uber_coordinates*)outbuf->data;
        outbuf->timestamp=filter->s1->timestamp;
	outbuf->duration=filter->s1->duration;
	outbuf->offset=filter->s1->offset;
	outbuf->offset_end=filter->s1->offset_end;
        gint x,y;
        gdouble bufscale;
	
	gint64 bfromname=-1;
         deserialize_int64 (&bfromname,current->name);
	if (bfromname>-1){//now we calculate the transformation
           gdouble elementsize=bfromname/1000.0;
	   g_print("We have %d inliers and now try to calculate the transformation for an Object with size of %f",n_in,elementsize);
	   CvMat* image_points      = cvCreateMat(n_in,2,CV_32FC1);
           CvMat* object_points     = cvCreateMat(n_in,3,CV_32FC1);   
           int i;
           for( i=0; i<n_in; ++i ) {
             CV_MAT_ELEM(*image_points, float,i,0) = inliers[i]->x;
             CV_MAT_ELEM(*image_points, float,i,1) = inliers[i]->y;
             //CV_MAT_ELEM(*image_points, float,i,0) = inliers[i]->x-(filter->width1/2.0);
             //CV_MAT_ELEM(*image_points, float,i,1) = inliers[i]->y-(filter->height1/2.0);
             //CV_MAT_ELEM(*object_points,float,i,0) = (inliers[i]->fwd_match->x-(current->width/2.0))*(elementsize/current->width);
             //CV_MAT_ELEM(*object_points,float,i,1) = (inliers[i]->fwd_match->y-(current->height/2.0))*(elementsize/current->width);
	    // CV_MAT_ELEM(*object_points,float,i,1) = (inliers[i]->fwd_match->y-(current->height))*(elementsize/current->width);
	     CV_MAT_ELEM(*object_points,float,i,0) = (inliers[i]->fwd_match->x)*(elementsize/current->width);
             CV_MAT_ELEM(*object_points,float,i,1) = (inliers[i]->fwd_match->y)*(elementsize/current->width);
             CV_MAT_ELEM(*object_points,float,i,2) = 0.0f;
           }    
           CvMat* rvec=cvCreateMat(3,1,CV_32FC1);
           CV_MAT_ELEM(*rvec, float,0,0) = 0;
           CV_MAT_ELEM(*rvec, float,1,0) = 0;
           CV_MAT_ELEM(*rvec, float,2,0) = 0;
           CvMat* tvec=cvCreateMat(3,1,CV_32FC1);   
           CV_MAT_ELEM(*tvec, float,0,0) = 0;
           CV_MAT_ELEM(*tvec, float,1,0) = 0;
           CV_MAT_ELEM(*tvec, float,2,0) = 0;
	   CvMat* intrinsic_matrix  = cvCreateMat(3,3,CV_32FC1);
	   double hfov=(3.14/180.0)*75;
	   if (filter->hfov>0) hfov=(3.14/180.0)*filter->hfov;
	   double f_x=((double)filter->width1)/(2.0*tan(hfov/2.0));
	   f_x=537;
	   CV_MAT_ELEM(*intrinsic_matrix, float,0,0) = f_x;
	   CV_MAT_ELEM(*intrinsic_matrix, float,1,1) = f_x;//same same
	   CV_MAT_ELEM(*intrinsic_matrix, float,2,2) = 1;
	   CV_MAT_ELEM(*intrinsic_matrix, float,0,2) = filter->width1/2.0-10;//c_x
	   CV_MAT_ELEM(*intrinsic_matrix, float,1,2) = filter->height1/2.0-10;//c_x
	   CV_MAT_ELEM(*intrinsic_matrix, float,0,1) =0;
	   CV_MAT_ELEM(*intrinsic_matrix, float,2,0) =0;
	   CV_MAT_ELEM(*intrinsic_matrix, float,2,1) =0;	  
	   CV_MAT_ELEM(*intrinsic_matrix, float,1,0) =0;		   
	   
	   
           if (filter->intrinsic_matrix && filter->distortion_coeffs)cvFindExtrinsicCameraParams2(object_points, image_points, filter->intrinsic_matrix, filter->distortion_coeffs, rvec, tvec,FALSE);
	   else cvFindExtrinsicCameraParams2(object_points, image_points, intrinsic_matrix, NULL, rvec, tvec,FALSE);
	   bufdata->rx=CV_MAT_ELEM(*tvec, float,0,0);
	   bufdata->ry=CV_MAT_ELEM(*tvec, float,1,0);
	   bufdata->rz=CV_MAT_ELEM(*tvec, float,2,0);
	   bufdata->rx=CV_MAT_ELEM(*rvec, float,0,0);
	   bufdata->ry=CV_MAT_ELEM(*rvec, float,1,0);
	   bufdata->rz=CV_MAT_ELEM(*rvec, float,2,0);	   
           if (filter->intrinsic_matrix && filter->distortion_coeffs)g_print ("Translation %.1f , %.1f , %.1f , Rotation %.1f , %.1f , %.1f with fx %f\n",100*CV_MAT_ELEM(*tvec, float,0,0),100*CV_MAT_ELEM(*tvec, float,1,0),100*CV_MAT_ELEM(*tvec, float,2,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,0,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,1,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,2,0),CV_MAT_ELEM(*filter->intrinsic_matrix, float, 1, 1 ));
	   else g_print ("Translation %.1f , %.1f , %.1f , Rotation %.1f , %.1f , %.1f with fx %f fy %f \n",100*CV_MAT_ELEM(*tvec, float,0,0),100*CV_MAT_ELEM(*tvec, float,1,0),100*CV_MAT_ELEM(*tvec, float,2,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,0,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,1,0),(180.0/3.14)*CV_MAT_ELEM(*rvec, float,2,0),f_x,CV_MAT_ELEM( *intrinsic_matrix, float, 1, 1 ));
	   CvMat* rotmat = cvCreateMat(3,3,CV_32FC1);
            CvMat* invrvec=cvCreateMat(3,1,CV_32FC1);
	   CvMat* invrotmat = cvCreateMat(3,3,CV_32FC1);	    
	   CvMat* homo_matrix  = cvCreateMat(4,4,CV_32FC1);
	   CvMat* inv_homo_matrix  = cvCreateMat(4,4,CV_32FC1);	
	   
	   cvRodrigues2(rvec, rotmat, NULL);
	   
	   CV_MAT_ELEM(*homo_matrix, float,0,3)=CV_MAT_ELEM(*tvec, float,0,0);
	   CV_MAT_ELEM(*homo_matrix, float,1,3)=CV_MAT_ELEM(*tvec, float,1,0);		   
	   CV_MAT_ELEM(*homo_matrix, float,2,3)=CV_MAT_ELEM(*tvec, float,2,0);
	   CV_MAT_ELEM(*homo_matrix, float,3,3)=1;	   
	   
	   CV_MAT_ELEM(*homo_matrix, float,0,0)=CV_MAT_ELEM(*rotmat, float,0,0);
	   CV_MAT_ELEM(*homo_matrix, float,0,1)=CV_MAT_ELEM(*rotmat, float,0,1);	   
	   CV_MAT_ELEM(*homo_matrix, float,0,2)=CV_MAT_ELEM(*rotmat, float,0,2);
	   
	   CV_MAT_ELEM(*homo_matrix, float,1,0)=CV_MAT_ELEM(*rotmat, float,1,0);
	   CV_MAT_ELEM(*homo_matrix, float,1,1)=CV_MAT_ELEM(*rotmat, float,1,1);	   
	   CV_MAT_ELEM(*homo_matrix, float,1,2)=CV_MAT_ELEM(*rotmat, float,1,2);	   
	   CV_MAT_ELEM(*homo_matrix, float,2,0)=CV_MAT_ELEM(*rotmat, float,2,0);
	   CV_MAT_ELEM(*homo_matrix, float,2,1)=CV_MAT_ELEM(*rotmat, float,2,1);	   
	   CV_MAT_ELEM(*homo_matrix, float,2,2)=CV_MAT_ELEM(*rotmat, float,2,2); 
	   
	   CV_MAT_ELEM(*homo_matrix, float,3,0)=0;
	   CV_MAT_ELEM(*homo_matrix, float,3,1)=0;	   
	   CV_MAT_ELEM(*homo_matrix, float,3,2)=0; 	   
	   
	   cvInvert( homo_matrix, inv_homo_matrix, CV_SVD );
	   
	   CV_MAT_ELEM(*invrotmat, float,0,0)=CV_MAT_ELEM(*inv_homo_matrix, float,0,0);
	   CV_MAT_ELEM(*invrotmat, float,0,1)=CV_MAT_ELEM(*inv_homo_matrix, float,0,1);	   
	   CV_MAT_ELEM(*invrotmat, float,0,2)=CV_MAT_ELEM(*inv_homo_matrix, float,0,2);
	   CV_MAT_ELEM(*invrotmat, float,1,0)=CV_MAT_ELEM(*inv_homo_matrix, float,1,0);
	   CV_MAT_ELEM(*invrotmat, float,1,1)=CV_MAT_ELEM(*inv_homo_matrix, float,1,1);	   
	   CV_MAT_ELEM(*invrotmat, float,1,2)=CV_MAT_ELEM(*inv_homo_matrix, float,1,2);	   
	   CV_MAT_ELEM(*invrotmat, float,2,0)=CV_MAT_ELEM(*inv_homo_matrix, float,2,0);
	   CV_MAT_ELEM(*invrotmat, float,2,1)=CV_MAT_ELEM(*inv_homo_matrix, float,2,1);	   
	   CV_MAT_ELEM(*invrotmat, float,2,2)=CV_MAT_ELEM(*inv_homo_matrix, float,2,2);	   
	   
	   cvRodrigues2(invrotmat, invrvec, NULL);
	   g_print ("Inverted Translation %.1f , %.1f , %.1f , Rotation %.1f , %.1f , %.1f and the last is %f \n",CV_MAT_ELEM(*inv_homo_matrix, float,0,3),CV_MAT_ELEM(*inv_homo_matrix, float,1,3),CV_MAT_ELEM(*inv_homo_matrix, float,2,3),(180.0/3.14)*CV_MAT_ELEM(*invrvec, float,0,0),(180.0/3.14)*CV_MAT_ELEM(*invrvec, float,1,0),(180.0/3.14)*CV_MAT_ELEM(*invrvec, float,2,0),CV_MAT_ELEM(*inv_homo_matrix, float,3,3));
	   
	   cvReleaseMat(&invrotmat);	  
	   cvReleaseMat(&rotmat);	   
	   cvReleaseMat(&invrvec);
	   cvReleaseMat(&inv_homo_matrix);
	   cvReleaseMat(&homo_matrix);
	   cvReleaseMat(&image_points);
           cvReleaseMat(&object_points);
           cvReleaseMat(&rvec);
           cvReleaseMat(&tvec);
	  cvReleaseMat(&intrinsic_matrix);
	  
	  
	  
	}
	
	
	
        //g_print("The point is x=%d and y=%d \n",(int)(filter->width2/2.0),(int)(filter->height2*0.9));
//  	printf("matrix data %f %f %f\n",cvmGet(H,0,0),cvmGet(H,0,1),cvmGet(H,0,2));
// 	printf("matrix data %f %f %f\n",cvmGet(H,1,0),cvmGet(H,1,1),cvmGet(H,1,2));
// 	printf("matrix data %f %f %f\n",cvmGet(H,2,0),cvmGet(H,2,1),cvmGet(H,2,2));

        double src_matrix[9], dst_matrix[9];
        CvMat A = cvMat( 3, 3, CV_64F, src_matrix ),
        invA = cvMat( 3, 3, CV_64F, dst_matrix );
        cvConvertScale( H, &A ,1,0);
        cvInvert( &A, &invA, CV_SVD );
	
//  	printf("inverted matrix data %f %f %f\n",cvmGet(&invA,0,0),cvmGet(&invA,0,1),cvmGet(&invA,0,2));
// 	printf("inverted matrix data %f %f %f\n",cvmGet(&invA,1,0),cvmGet(&invA,1,1),cvmGet(&invA,1,2));
// 	printf("inverted matrix data %f %f %f\n",cvmGet(&invA,2,0),cvmGet(&invA,2,1),cvmGet(&invA,2,2));
// 
 	//printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,0,0)/cvmGet(&invA,2,2),cvmGet(&invA,0,1)/cvmGet(&invA,2,2),cvmGet(&invA,0,2)/cvmGet(&invA,2,2));
 	//printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,1,0)/cvmGet(&invA,2,2),cvmGet(&invA,1,1)/cvmGet(&invA,2,2),cvmGet(&invA,1,2)/cvmGet(&invA,2,2));
 	//printf("inverted normalized matrix data %f %f %f\n",cvmGet(&invA,2,0)/cvmGet(&invA,2,2),cvmGet(&invA,2,1)/cvmGet(&invA,2,2),cvmGet(&invA,2,2)/cvmGet(&invA,2,2));

	bufdata->a00=cvmGet(&invA,0,0)/cvmGet(&invA,2,2);
	bufdata->a01=cvmGet(&invA,0,1)/cvmGet(&invA,2,2);
	bufdata->a02=cvmGet(&invA,0,2)/cvmGet(&invA,2,2);
	bufdata->a10=cvmGet(&invA,1,0)/cvmGet(&invA,2,2);
	bufdata->a11=cvmGet(&invA,1,1)/cvmGet(&invA,2,2);
	bufdata->a12=cvmGet(&invA,1,2)/cvmGet(&invA,2,2);
	bufdata->a20=cvmGet(&invA,2,0)/cvmGet(&invA,2,2);
	bufdata->a21=cvmGet(&invA,2,1)/cvmGet(&invA,2,2);
	bufdata->a22=cvmGet(&invA,2,2)/cvmGet(&invA,2,2);

	bufdata->x1=cvmGet(&invA,0,2)/cvmGet(&invA,2,2);
	bufdata->y1=cvmGet(&invA,1,2)/cvmGet(&invA,2,2);		
        bufscale=(cvmGet(&invA,2,0)*((gdouble)current->width)+cvmGet(&invA,2,1)*((float)current->height)+cvmGet(&invA,2,2));
	bufdata->x2=(cvmGet(&invA,0,0)*((gdouble)current->width)+cvmGet(&invA,0,1)*((float)current->height)+cvmGet(&invA,0,2))/bufscale;
	bufdata->y2=(cvmGet(&invA,1,0)*((gdouble)current->width)+cvmGet(&invA,1,1)*((float)current->height)+cvmGet(&invA,1,2))/bufscale;	
	bufdata->angle=atan2((cvmGet(&invA,0,0)/cvmGet(&invA,2,2)),(cvmGet(&invA,0,0)/cvmGet(&invA,2,2)));
	if (bufdata->x1>bufdata->x2){
	   gdouble tempx=bufdata->x1;
	   bufdata->x1=bufdata->x2;
	   bufdata->x2=tempx;
  	}
	if (bufdata->y1>bufdata->y2){
	   gdouble tempy=bufdata->y1;
	   bufdata->y1=bufdata->y2;
	   bufdata->y2=tempy;
  	}
	//bufdata->scale=(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))+(cvmGet(&invA,0,0)/cvmGet(&invA,2,2))*(cvmGet(&invA,0,0)/cvmGet(&invA,2,2));

	
	bufdata->imagewidth=(int)filter->width1;//size of the image where we found this 
	bufdata->imageheight=(int)filter->height1;//
	bufdata->mwidth=(int)current->width;//size of the model that we used to find this this   
	bufdata->mheight=(int)current->height;//
	bufdata->timestamp=outbuf->timestamp;
	bufdata->scale=sqrt(((bufdata->y2-bufdata->y1)*(bufdata->y2-bufdata->y1))+((bufdata->x2-bufdata->x1)*(bufdata->x2-bufdata->x1)))/sqrt((bufdata->mwidth*bufdata->mwidth)+(bufdata->mheight*bufdata->mheight));
	strncpy(bufdata->name,current->name,29);	
	//g_print("Found transformation for object name %s, starting Point x=%f y=%f, going to x=%f y=%f  scale  is %f!\n",current->name,bufdata->x1,bufdata->y1,bufdata->x2,bufdata->y2,bufdata->scale);
        //now we send it
        GstCaps *tmpcaps=gst_caps_new_simple ("uber-coordinates",NULL);
        gst_buffer_set_caps(outbuf,tmpcaps);	
	gst_caps_unref(tmpcaps);
	ret=gst_pad_push (filter->srcpad, outbuf);
	if (ret!=GST_FLOW_OK){
	  g_mutex_unlock(filter->mutex);
	  return ret;
	  }
	cvReleaseMat( &H ); 
    }
    else{
        //g_print("Sorry, could not find transformation!\n");
	GstBuffer *outbuf;
        outbuf=gst_buffer_new_and_alloc(sizeof(struct uber_coordinates));
	struct uber_coordinates* bufdata=(struct uber_coordinates*)outbuf->data;
        outbuf->timestamp=filter->s1->timestamp;
	outbuf->duration=filter->s1->duration;
	outbuf->offset=filter->s1->offset;
	outbuf->offset_end=filter->s1->offset_end;
        memset (bufdata,0,sizeof(struct uber_coordinates));
	strncpy(bufdata->name,current->name,29);
		bufdata->imagewidth=filter->width1;//size of the image where we found this 
	bufdata->imageheight=filter->height1;//
	bufdata->mwidth=current->width;//size of the model that we used to find this this   
	bufdata->mheight=current->height;//
	bufdata->timestamp=outbuf->timestamp;
        GstCaps *tmpcaps=gst_caps_new_simple ("uber-coordinates",NULL);
        gst_buffer_set_caps(outbuf,tmpcaps);	
	gst_caps_unref(tmpcaps);
        //now we send it
	ret=gst_pad_push (filter->srcpad, outbuf);
	if (ret!=GST_FLOW_OK){
	  g_mutex_unlock(filter->mutex);
	  return ret;
	  }
	}
       if (inliers)free(inliers);
    }//the for-loop that iterstes all known objects
    if (filter->rematchdouble||filter->otherway)kdtree_release(filter->kdback );
  g_mutex_unlock(filter->mutex);
  return GST_FLOW_OK; 
}

static GstStateChangeReturn
gst_siftfolder_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstsiftfolder *filter;
  filter = GST_SIFTFOLDER (element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      //if (filter->directory)pthread_create (&filter->th  , NULL, watchthread, filter);
      break;
      
      
      
    case GST_STATE_CHANGE_READY_TO_PAUSED:      
      if (filter->siftheap==NULL){
	if (!gst_siftfolder_load_gallery(filter)){
	  g_print("Warning, Gstsiftfolder can not work because no objects to match, set directory path!!\n");	
	  if(filter->directory)g_free (filter->directory);
	  filter->directory=NULL;
	}
	else{
	  if (!gst_element_is_null(filter))pthread_create (&filter->th  , NULL, watchthread, filter);
	  }
      }
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      if (filter->s1) gst_buffer_unref(filter->s1);
      filter->s1=NULL;
      if (filter->th>0){
	pthread_cancel(filter->th);
	filter->th=0;
      }
      break;
    default:
      break;
  }
 return ret;
}





static gboolean
gst_siftfolder_event (GstPad  *pad, GstEvent * event)
{
  Gstsiftfolder *filter;
  gboolean ret=TRUE;
  //GstPad *otherpad;

  filter = GST_SIFTFOLDER (gst_pad_get_parent (pad));

  switch (GST_EVENT_TYPE (event)) {
    //case GST_EVENT_EOS:
      //if (pad == filter->sinkpad1) filter->stopped1=TRUE;
      //if (pad == filter->sinkpad2) filter->stopped2=TRUE;
      ///if ((filter->stopped1)&&(filter->stopped2)) ret = gst_pad_push_event (filter->srcpad, event);
      //g_print("Event EOS!\n");
      //break;
    default:
      ret = gst_pad_push_event (filter->srcpad, event);
      break;
  }
  
  gst_object_unref (filter);
  return ret;
}





/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
siftfolder_init (GstPlugin * siftfolder)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template siftfolder' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_siftfolder_debug, "SiftFolder",
      0, "Template Sifttools");
  //g_print("Registering Sifttools with pointer %p\n",siftfolder);
  return gst_element_register (siftfolder, "siftfolder", GST_RANK_NONE,
      GST_TYPE_SIFTFOLDER);
}

/* gstreamer looks for this structure to register siftfolders
 *
 * exchange the string 'Template siftfolder' with your siftfolder description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "SiftFolder",
    "Matches SIFT-Vector to folder",
    siftfolder_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
