/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2008 bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef __GST_SIFTFOLDER_H__
#define __GST_SIFTFOLDER_H__

#include <gst/gst.h>

G_BEGIN_DECLS

/* #defines don't like whitespacey bits */
#define GST_TYPE_SIFTFOLDER \
  (gst_siftfolder_get_type())
#define GST_SIFTFOLDER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj),GST_TYPE_SIFTFOLDER,Gstsiftfolder))
#define GST_SIFTFOLDER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass),GST_TYPE_SIFTFOLDER,GstsiftfolderClass))
#define GST_IS_SIFTFOLDER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj),GST_TYPE_SIFTFOLDER))
#define GST_IS_SIFTFOLDER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass),GST_TYPE_SIFTFOLDER))

#define FEATURE_MAX_D 128
#define MINPQ_INIT_NALLOCD 512

/* RANSAC error tolerance in pixels */
#define RANSAC_ERR_TOL 3

/** pessimistic estimate of fraction of inlers for RANSAC */
#define RANSAC_INLIER_FRAC_EST 0.25

/** estimate of the probability that a correspondence supports a bad model */
#define RANSAC_PROB_BAD_SUPP 0.10

/** default number of sampled intervals per octave */
#define SIFT_INTVLS 3

/** default sigma for initial gaussian smoothing */
#define SIFT_SIGMA 1.6

/** default threshold on keypoint contrast |D(x)| */
#define SIFT_CONTR_THR 0.04

/** default threshold on keypoint ratio of principle curvatures */
#define SIFT_CURV_THR 10

/** double image size before pyramid construction? */
#define SIFT_IMG_DBL 1

/** default width of descriptor histogram array */
#define SIFT_DESCR_WIDTH 4

/** default number of bins per histogram in descriptor array */
#define SIFT_DESCR_HIST_BINS 8

/* assumed gaussian blur for input image */
#define SIFT_INIT_SIGMA 0.5

/* width of border in which to ignore keypoints */
#define SIFT_IMG_BORDER 5

/* maximum steps of keypoint interpolation before failure */
#define SIFT_MAX_INTERP_STEPS 5

/* default number of bins in histogram for orientation assignment */
#define SIFT_ORI_HIST_BINS 36

/* determines gaussian sigma for orientation assignment */
#define SIFT_ORI_SIG_FCTR 1.5

/* determines the radius of the region used in orientation assignment */
#define SIFT_ORI_RADIUS 3.0 * SIFT_ORI_SIG_FCTR

/* number of passes of orientation histogram smoothing */
#define SIFT_ORI_SMOOTH_PASSES 2

/* orientation magnitude relative to max that results in new feature */
#define SIFT_ORI_PEAK_RATIO 0.8

/* determines the size of a single descriptor orientation histogram */
#define SIFT_DESCR_SCL_FCTR 3.0

/* threshold on magnitude of elements of descriptor vector */
#define SIFT_DESCR_MAG_THR 0.2

/* factor used to convert floating-point descriptor to unsigned char */
#define SIFT_INT_DESCR_FCTR 512.0

/* returns a feature's detection data */
#define feat_detection_data(f) ( (struct detection_data*)(f->feature_data) )

/* colors in which to display different feature types */
#define FEATURE_OXFD_COLOR CV_RGB(255,255,0)
#define FEATURE_LOWE_COLOR CV_RGB(255,0,255)
/* extracts a feature's RANSAC data */
#define feat_ransac_data( feat ) ( (struct ransac_data*) (feat)->feature_data )

#define EVENT_SIZE  ( sizeof (struct inotify_event) )
#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )

typedef struct _Gstsiftfolder      Gstsiftfolder;
typedef struct _GstsiftfolderClass GstsiftfolderClass;

typedef enum {
  MATCH,
  MERGE,  //merge two sets of sift-features
  TRANSFORM,
  TRANSFORMPOINT  //calculate transformation of one poinnt
} SiftToolOpMode;

typedef CvMat* (*ransac_xform_fn)( CvPoint2D64f* pts, CvPoint2D64f* mpts,
				   int n );
typedef double (*ransac_err_fn)( CvPoint2D64f pt, CvPoint2D64f mpt, CvMat* T );


struct lowelight{
  double x;                      /**< x coord */
  double y;                      /**< y coord */
  double scl;                    /**< scale of a Lowe-style feature */
  double ori;                    /**< orientation of a Lowe-style feature */
  unsigned char data[128];	  //data that is btw only char values
};


struct detection_data
{
  int r;
  int c;
  int octv;
  int intvl;
  double subintvl;
  double scl_octv;
};

struct feature
{
  double x;                      /**< x coord */
  double y;                      /**< y coord */
  double a;                      /**< Oxford-type affine region parameter */
  double b;                      /**< Oxford-type affine region parameter */
  double c;                      /**< Oxford-type affine region parameter */
  double scl;                    /**< scale of a Lowe-style feature */
  double ori;                    /**< orientation of a Lowe-style feature */
  int d;                         /**< descriptor length */
  double descr[FEATURE_MAX_D];   /**< descriptor */
  int type;                      /**< feature type, OXFD or LOWE */
  int category;                  /**< all-purpose feature category */
  struct feature* fwd_match;     /**< matching feature from forward image */
  struct feature* bck_match;     /**< matching feature from backmward image */
  struct feature* mdl_match;     /**< matching feature from model */
  CvPoint2D64f img_pt;           /**< location in image */
  CvPoint2D64f mdl_pt;           /**< location in model */
  void* feature_data;            /**< user-definable data */
};

enum feature_type
  {
    FEATURE_OXFD,
    FEATURE_LOWE,
  };

enum feature_match_type
  {
    FEATURE_FWD_MATCH,
    FEATURE_BCK_MATCH,
    FEATURE_MDL_MATCH,
  };

struct kd_node
{
  int ki;                      /**< partition key index */
  double kv;                   /**< partition key value */
  int leaf;                    /**< 1 if node is a leaf, 0 otherwise */
  struct feature* features;    /**< features at this node */
  int n;                       /**< number of features */
  struct kd_node* kd_left;     /**< left child */
  struct kd_node* kd_right;    /**< right child */
};

struct bbf_data
{
  double d;
  void* old_data;
};

struct ransac_data
{
  void* orig_feat_data;
  int sampled;
};

struct pq_node
{
  void* data;
  int key;
};


/** a minimizing priority queue */
struct min_pq
{
  struct pq_node* pq_array;    /* array containing priority queue */
  int nallocd;                 /* number of elements allocated */
  int n;                       /**< number of elements in pq */
};

struct coordinates
{
  gdouble x1;
  gdouble y1;
  gdouble x2;
  gdouble y2;
};

struct transmat
{
  gdouble a00;
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
};

struct uber_coordinates
{
  int imagewidth;//size of the image where we found this 
  int imageheight;//
  int mwidth;//size of the model that we used to find this this   
  int mheight;//
  gdouble scale;//Rectangular Region where this is matched into
  gdouble angle;  
  gdouble x1;//Rectangular Region where this is matched into
  gdouble y1;
  gdouble x2;
  gdouble y2;
  gdouble a00;//transformation matrix 
  gdouble a01;
  gdouble a02;
  gdouble a10;
  gdouble a11;
  gdouble a12;
  gdouble a20;
  gdouble a21;
  gdouble a22;
  guint64 timestamp;
  char name[30];//name of found element
  gdouble tx;
  gdouble ty; 
  gdouble tz;
  gdouble rx;
  gdouble ry; 
  gdouble rz;  
};

struct siftset{
IplImage* init_img;
IplImage* gray8;//intermediate
IplImage* gray32;
IplImage*** gauss_pyr;
IplImage*** dog_pyr;
//next important for dereferencing
int intvls;
int octvs;
int dbl_img;
int channels;
};

struct siftheap{
int numfeat;
struct feature *feat;
struct kd_node* node;
int width;
int height;
gchar name[30]; 
};

struct _Gstsiftfolder
{
  GstElement element;

  GstPad *sinkpad, *srcpad;
  GList *siftheap;
  char* directory;//directory with image files for matching
  gboolean silent;
  gint width1,height1;
  gboolean rematchdouble,showinliers;//whether to put out coordinates in booka-style instead of matchlist
  int double_image;
  GstBuffer *s1;
  gint n1;
  struct feature* feat1;
  gint mode;//mode of the input (to parse)
  gint suggestedmodes1,suggestedmodes2;//suggestedmode from user input
  GMutex* mutex;
  gdouble maxsim,maxsim2,gooddist,threshold;
  //stuff for inotify
  int fd,wd;
  guint thresh;
  pthread_t th;
  char buffer[EVENT_BUF_LEN];
  int ntotal,thisn;
  gdouble c_x;
  gdouble c_y;
  gdouble f_x; 
  gdouble f_y;
  gdouble k_1;
  gdouble k_2;
  gdouble k_3;
  gdouble k_4;
  gdouble k_5;
  gdouble hfov;
  CvMat* distortion_coeffs;
  CvMat* intrinsic_matrix;
  char* unddirectory;
  struct kd_node* kdback;
  gboolean otherway;
};

struct _GstsiftfolderClass 
{
  GstElementClass parent_class;
};

GType gst_siftfolder_get_type (void);

int array_double( void** array, int n, int size )
{
  void* tmp;

  tmp = realloc( *array, 2 * n * size );
  if( ! tmp )
    {
      fprintf( stderr, "Warning: unable to allocate memory in array_double(),"
	       " %s line %d\n", __FILE__, __LINE__ );
      if( *array )
	free( *array );
      *array = NULL;
      return 0;
    }
  *array = tmp;
  return n*2;
}

void restore_minpq_order( struct pq_node*, int, int );
void decrease_pq_node_key( struct pq_node*, int, int );

static inline int parent( int i )
{
  return ( i - 1 ) / 2;
}


/* returns the array index of element i's right child */
static inline int right( int i )
{
  return 2 * i + 2;
}


/* returns the array index of element i's left child */
static inline int left( int i )
{
  return 2 * i + 1;
}


/********************** Functions prototyped in minpq.h **********************/


/*
  Creates a new minimizing priority queue.
*/
struct min_pq* minpq_init()
{
  struct min_pq* min_pq;

  min_pq = malloc( sizeof( struct min_pq ) );
  min_pq->pq_array = calloc( MINPQ_INIT_NALLOCD, sizeof( struct pq_node ) );
  min_pq->nallocd = MINPQ_INIT_NALLOCD;
  min_pq->n = 0;

  return min_pq;
}



/**
  Inserts an element into a minimizing priority queue.

  @param min_pq a minimizing priority queue
  @param data the data to be inserted
  @param key the key to be associated with \a data

  @return Returns 0 on success or 1 on failure.
*/
int minpq_insert( struct min_pq* min_pq, void* data, int key )
{
  int n = min_pq->n;

  /* double array allocation if necessary */
  if( min_pq->nallocd == n )
    {
      min_pq->nallocd = array_double( (void**)&min_pq->pq_array, (int)min_pq->nallocd,
				      sizeof( struct pq_node ) );
      if( ! min_pq->nallocd )
	{
	  fprintf( stderr, "Warning: unable to allocate memory, %s, line %d\n",
		   __FILE__, __LINE__ );
	  return 1;
	}
    }

  min_pq->pq_array[n].data = data;
  min_pq->pq_array[n].key = INT_MAX;
  decrease_pq_node_key( min_pq->pq_array, min_pq->n, key );
  min_pq->n++;

  return 0;
}



/*
  Returns the element of a minimizing priority queue with the smallest key
  without removing it from the queue.
  
  @param min_pq a minimizing priority queue
  
  @return Returns the element of \a min_pq with the smallest key or NULL
    if \a min_pq is empty
*/
void* minpq_get_min( struct min_pq* min_pq )
{
  if( min_pq->n < 1 )
    {
      fprintf( stderr, "Warning: PQ empty, %s line %d\n", __FILE__, __LINE__ );
      return NULL;
    }
  return min_pq->pq_array[0].data;
}



/*
  Removes and returns the element of a minimizing priority queue with the
  smallest key.
  
  @param min_pq a minimizing priority queue
  
  @return Returns the element of \a min_pq with the smallest key of NULL
    if \a min_pq is empty
*/
void* minpq_extract_min( struct min_pq* min_pq )
{
  void* data;

  if( min_pq->n < 1 )
    {
      fprintf( stderr, "Warning: PQ empty, %s line %d\n", __FILE__, __LINE__ );
      return NULL;
    }
  data = min_pq->pq_array[0].data;
  min_pq->n--;
  min_pq->pq_array[0] = min_pq->pq_array[min_pq->n];
  restore_minpq_order( min_pq->pq_array, 0, min_pq->n );

  return data;
}


/*
  De-allocates the memory held by a minimizing priorioty queue
  
  @param min_pq pointer to a minimizing priority queue
*/
void minpq_release( struct min_pq** min_pq )
{
  if( ! min_pq )
    {
      fprintf( stderr, "Warning: NULL pointer error, %s line %d\n", __FILE__,
	       __LINE__ );
      return;
    }
  if( *min_pq  &&  (*min_pq)->pq_array )
    {
      free( (*min_pq)->pq_array );
      free( *min_pq );
      *min_pq = NULL;
    }
}


/************************ Functions prototyped here **************************/

/*
  Decrease a minimizing pq element's key, rearranging the pq if necessary

  @param pq_array minimizing priority queue array
  @param i index of the element whose key is to be decreased
  @param key new value of element <EM>i</EM>'s key; if greater than current
    key, no action is taken
*/
void decrease_pq_node_key( struct pq_node* pq_array, int i, int key )
{
  struct pq_node tmp;

  if( key > pq_array[i].key )
    return;

  pq_array[i].key = key;
  while( i > 0  &&  pq_array[i].key < pq_array[parent(i)].key )
    {
      tmp = pq_array[parent(i)];
      pq_array[parent(i)] = pq_array[i];
      pq_array[i] = tmp;
      i = parent(i);
    }
}



/*
  Recursively restores correct priority queue order to a minimizing pq array

  @param pq_array a minimizing priority queue array
  @param i index at which to start reordering
  @param n number of elements in \a pq_array
*/
void restore_minpq_order( struct pq_node* pq_array, int i, int n )
{
  struct pq_node tmp;
  int l, r, min = i;

  l = left( i );
  r = right( i );
  if( l < n )
    if( pq_array[l].key < pq_array[i].key )
      min = l;
  if( r < n )
    if( pq_array[r].key < pq_array[min].key )
      min = r;

  if( min != i )
    {
      tmp = pq_array[min];
      pq_array[min] = pq_array[i];
      pq_array[i] = tmp;
      restore_minpq_order( pq_array, min, n );
    }
}

double dist_sq_2D( CvPoint2D64f p1, CvPoint2D64f p2 )
{
  double x_diff = p1.x - p2.x;
  double y_diff = p1.y - p2.y;

  return x_diff * x_diff + y_diff * y_diff;
}

gdouble DoubleSwap( double f )
{
  union
  {
    double f;
    unsigned char b[8];
  } dat1, dat2;

  dat1.f = f;
  dat2.b[0] = dat1.b[7];
  dat2.b[1] = dat1.b[6];
  dat2.b[2] = dat1.b[5];
  dat2.b[3] = dat1.b[4];
  dat2.b[4] = dat1.b[3];
  dat2.b[5] = dat1.b[2];
  dat2.b[6] = dat1.b[1];
  dat2.b[7] = dat1.b[0];
  return dat2.f;
}

char* deserialize_int64 (gint64* result,const char * s)//gives end of char
{
  if (!s)return NULL;
  while (!isdigit(*s)&&(*s!=0)&&(*s!=45))s++;
  if (*s==0){
    //*result=NAN;
    return NULL;
  }
  char *end;
  *result = g_ascii_strtoll (s, &end, 0);
  return end;
}
G_END_DECLS

#endif /* __GST_SIFTFOLDER_H__ */
