/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-tagcoords
 *
 * FIXME:Describe tagcoords here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! tagcoords ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gsttagcoords.h"
#include <string.h>
GST_DEBUG_CATEGORY_STATIC (gst_tagcoords_debug);
#define GST_CAT_DEFAULT gst_tagcoords_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_TAG,
  PROP_WIDTH,
  PROP_HEIGHT 
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "coordinates-list"
    )
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
      "uber-coordinates-list"
    )
    );

GST_BOILERPLATE (Gsttagcoords, gst_tagcoords, GstElement,
    GST_TYPE_ELEMENT);

static void gst_tagcoords_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_tagcoords_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_tagcoords_set_caps (GstPad * pad, GstCaps * caps);
static GstCaps * gst_tagcoords_get_caps (GstPad *pad);
static GstFlowReturn gst_tagcoords_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_tagcoords_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "tagcoords",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the tagcoords's class */
static void
gst_tagcoords_class_init (GsttagcoordsClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_tagcoords_set_property;
  gobject_class->get_property = gst_tagcoords_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_TAG,
      g_param_spec_string ("tag", "Tag", "How to tag the data.",
          NULL, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_WIDTH,
      g_param_spec_int ("width", "Width", "Tag width to value. -1=no",
          -1,G_MAXINT,-1, G_PARAM_READWRITE));  
  g_object_class_install_property (gobject_class, PROP_HEIGHT,
      g_param_spec_int ("height", "Height", "Tag height to value. -1=no",
          -1,G_MAXINT,-1, G_PARAM_READWRITE));  
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_tagcoords_init (Gsttagcoords * filter,
    GsttagcoordsClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_tagcoords_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_tagcoords_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_tagcoords_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_tagcoords_get_caps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->width=-1;
  filter->height=-1;
}

static void
gst_tagcoords_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gsttagcoords *filter = GST_TAGCOORDS (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case PROP_TAG:
      g_free (filter->tag);
      filter->tag = g_value_dup_string (value);
      break;
    case PROP_WIDTH:
      filter->width = g_value_get_int (value);
      break;
    case PROP_HEIGHT:
      filter->height = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_tagcoords_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gsttagcoords *filter = GST_TAGCOORDS (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_TAG:
      g_value_set_string (value, filter->tag);
      break;
    case PROP_WIDTH:
      g_value_set_int (value, filter->width);
      break;
    case PROP_HEIGHT:
      g_value_set_int (value, filter->height);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_tagcoords_set_caps (GstPad * pad, GstCaps * caps)
{
  //Gsttagcoords *filter;
  //GstPad *otherpad;

  //filter = GST_TAGCOORDS (gst_pad_get_parent (pad));
  //otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  //gst_object_unref (filter);

  //return gst_pad_set_caps (otherpad, caps);
  return TRUE;
}

static GstCaps *
gst_tagcoords_get_caps (GstPad *pad){
  //g_print("Tagcoords Start get caps!!\n");
  //Gstmultivis *filter;

  //filter = GST_MULTIVIS (GST_OBJECT_PARENT (pad));

  return gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_tagcoords_chain (GstPad * pad, GstBuffer * buf)
{
  Gsttagcoords *filter;

  filter = GST_TAGCOORDS (GST_OBJECT_PARENT (pad));

  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");

  struct coordinates* bufdata=(struct coordinates*)buf->data; 
  int ll,i;
  ll=buf->size/sizeof(struct coordinates);
  GstBuffer *outbuf;
  outbuf=gst_buffer_new_and_alloc(ll*sizeof(struct uber_coordinates));
  struct uber_coordinates* outbufdata=(struct uber_coordinates*)outbuf->data;
  outbuf->timestamp=buf->timestamp;
  outbuf->duration=buf->duration;
  outbuf->offset=buf->offset;
  outbuf->offset_end=buf->offset_end;
  
  for (i=0;i<ll;i++){
    outbufdata[i].x1=bufdata[i].x1;
    outbufdata[i].x2=bufdata[i].x2;
    outbufdata[i].y1=bufdata[i].y1;
    outbufdata[i].y2=bufdata[i].y2;
    if (strlen(filter->tag)<30)strcpy(outbufdata[i].name,filter->tag);
    else memcpy(outbufdata[i].name,filter->tag,30);
    if (filter->width>-1)outbufdata[i].imagewidth=filter->width;
    if (filter->height>-1)outbufdata[i].imageheight=filter->height;
  }
  gst_buffer_set_caps(outbuf,(GstCaps *)gst_pad_get_pad_template_caps(filter->srcpad));
  //gst_buffer_set_caps(outbuf,GST_STATIC_CAPS("uber-coordinates-list"));
  gst_buffer_unref(buf);
  /* just push out the incoming buffer without touching it */
  return gst_pad_push (filter->srcpad, outbuf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
tagcoords_init (GstPlugin * tagcoords)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template tagcoords' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_tagcoords_debug, "tagcoords",
      0, "Template tagcoords");

  return gst_element_register (tagcoords, "tagcoords", GST_RANK_NONE,
      GST_TYPE_TAGCOORDS);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirsttagcoords"
#endif

/* gstreamer looks for this structure to register tagcoordss
 *
 * exchange the string 'Template tagcoords' with your tagcoords description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "tagcoords",
    "Template tagcoords",
    tagcoords_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
