/*
 * GStreamer
 * Copyright 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 * Note from Hannes Bistry: The code for statistics is "somewhere from the internet:)"
 * 
 */
/*
 * NEue Auswertung: Es wird der Startwert für einen Offset angegeben und die Anzahl der Messungen, zusätzlich eine Datei für das Hineinschreiben der Werte...
 * 
 * */
/**
 * SECTION:element-plugin
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch -v -m audiotestsrc ! plugin ! fakesink silent=TRUE
 * </programlisting>
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/time.h>
#include <unistd.h>
#include <gst/gst.h>
#include <math.h>

#include "gsttimestamper.h"

GST_DEBUG_CATEGORY_STATIC (gst_timestamper_debug);
#define GST_CAT_DEFAULT gst_timestamper_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  ARG_0,
  ARG_SILENT,
  ARG_T1,
  ARG_T2,
  ARG_VALVE,
  ARG_VALVEWAIT,
  ARG_ADDOFFSET,
  ARG_RATE,
  ARG_OFFT,
  ARG_CODEJIFFIES,
  ARG_LATENCY,
  ARG_TIMEBETWEEN,
  ARG_DROPRATE,
  ARG_FWEOS,
  ARG_LASTOFFSET,
  ARG_FWOLD,
  ARG_STARTOFFSET,
  ARG_ENDOFFSET,
  ARG_STATFILE,
  ARG_BANDWIDTH,
  ARG_NUMPARTS,
  ARG_INFOEVERY,
  ARG_BANDWIDTH_METER,
  ARG_EVNUM
};

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gsttimestamper, gst_timestamper, GstElement,
    GST_TYPE_ELEMENT);

static void gst_timestamper_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_timestamper_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_timestamper_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_timestamper_chain (GstPad * pad, GstBuffer * buf);
static gboolean gst_timestamper_event (GstPad  *pad, GstEvent * event);
static GstStateChangeReturn
gst_timestamper_change_state (GstElement *element, GstStateChange transition);
static GstFlowReturn gst_timestamper_bufferalloc (GstPad * pad, guint64 offset,
    guint size, GstCaps * caps, GstBuffer ** buf);

static void
gst_timestamper_base_init (gpointer gclass)
{
  static GstElementDetails element_details = {
    "Timestamper",
    "Gst-Plugins-Robotix",
    "Timestamp Adjuster",
    "Hannes Bistry <bistry aaattt informatik.uni-hamburg.de>"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_timestamper_class_init (GsttimestamperClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_timestamper_set_property;
  gobject_class->get_property = gst_timestamper_get_property;

  g_object_class_install_property (gobject_class, ARG_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_T1,
      g_param_spec_int ("t1", "T1",
          "Type of sink timestamp?", -1,2,
          0, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_T2,
      g_param_spec_int ("t2", "T2",
          "Type of src timestamp?", -1,2,
          0, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_ADDOFFSET,
      g_param_spec_int64 ("addoffset", "Addoffset",
          "Delay stream output for x nanoseconds", G_MININT64,G_MAXINT64,
          0, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_VALVE,
      g_param_spec_uint64 ("valve", "Valve",
          "Block stream if buffers are x ms too fast", 0,G_MAXUINT64,
          G_MAXINT64, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_VALVEWAIT,
      g_param_spec_uint64 ("valvewait", "Valvewait",
          "Block stream x ns if valve limit reached(if not declared it will be valve/2)", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READWRITE));
 g_object_class_install_property (gobject_class, ARG_RATE,
      g_param_spec_double ("rate", "Rate",
          "Rate to delay buffers",G_MINDOUBLE ,G_MAXDOUBLE,
          1.00000, (GParamFlags)G_PARAM_READWRITE));
 g_object_class_install_property (gobject_class, ARG_OFFT,
      g_param_spec_int ("offsetformat", "Offsetformat",
          "Patch Offset? -1 no; 2 bytes; 3 streamtime in ns ...see GstFormat", -1,6,
          -1, (GParamFlags)G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_CODEJIFFIES,
      g_param_spec_boolean ("codejiffies", "Codejiffies", "Store/extract+compare clock() value in duration",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_LATENCY,
      g_param_spec_double ("latency", "Latency",
          "Average latency of each buffer", 0,G_MAXDOUBLE,
          0, (GParamFlags)G_PARAM_READABLE)); 
  g_object_class_install_property (gobject_class, ARG_TIMEBETWEEN,
      g_param_spec_double ("timebetween", "Timebetween",
          "Average time between two buffer", 0,G_MAXDOUBLE,
          0, (GParamFlags)G_PARAM_READABLE)); 
   g_object_class_install_property (gobject_class, ARG_DROPRATE,
      g_param_spec_double ("droprate", "Droprate",
          "Rate of dropped Buffers", 0,1,
          0, (GParamFlags)G_PARAM_READABLE)); 
  g_object_class_install_property (gobject_class, ARG_FWEOS,
      g_param_spec_boolean ("fweos", "Fweos", "Forward EOS Events or drop it",
          TRUE, G_PARAM_READWRITE));

  g_object_class_install_property (gobject_class, ARG_LASTOFFSET,
      g_param_spec_uint64 ("lastoffset", "Lastoffset",
          "Last offset of a passing buffer", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READABLE));
  g_object_class_install_property (gobject_class, ARG_FWOLD,
      g_param_spec_boolean ("fwold", "Fwold", "Forward outdated buffers",
          FALSE, G_PARAM_READWRITE));
   g_object_class_install_property (gobject_class, ARG_STARTOFFSET,
      g_param_spec_uint64 ("startoffset", "Startoffset",
          "StartOffset when to do extended Capture (GMAXUINT64=disable)", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READWRITE)); 
   g_object_class_install_property (gobject_class, ARG_ENDOFFSET,
      g_param_spec_uint64 ("endoffset", "Endoffset",
          "EndOffset before we end extended Capture", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READWRITE));  
   g_object_class_install_property (gobject_class, ARG_STATFILE,
      g_param_spec_string ("statfile", "statfile", "Filename where we store latency information",
          "/tmp/stat.txt", G_PARAM_READWRITE)); 
    g_object_class_install_property (gobject_class, ARG_BANDWIDTH,
      g_param_spec_uint64 ("bandwidthlimit", "Bandwidthlimit",
          "Limit Bandwith, in Bytes per Second, 0 is unlimited", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READWRITE));  
    g_object_class_install_property (gobject_class, ARG_NUMPARTS,
      g_param_spec_uint64 ("numparts", "Numparts",
          "If set to a N, we wait for N elements with the same offset and evaluate the last, 0 is disable", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READWRITE));      
    g_object_class_install_property (gobject_class, ARG_INFOEVERY,
      g_param_spec_uint64 ("infoevery", "Infoevery",
          "Output latency info every nth frame", 1,G_MAXUINT64,
          1, (GParamFlags)G_PARAM_READWRITE));  
    g_object_class_install_property (gobject_class, ARG_BANDWIDTH_METER,
      g_param_spec_uint64 ("bandwidth", "Bandwidth",
          "Current Used Bandwidth", 0,G_MAXUINT64,
          G_MAXUINT64, (GParamFlags)G_PARAM_READABLE));    
    g_object_class_install_property (gobject_class, ARG_EVNUM,
      g_param_spec_uint64 ("evnum", "Evnum",
          "Number of measurements to average", 1,G_MAXUINT64,
          10, (GParamFlags)G_PARAM_READWRITE));  
  gstelement_class->change_state = gst_timestamper_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_timestamper_init (Gsttimestamper * filter,
    GsttimestamperClass * gclass)
{
  GstElementClass *klass = GST_ELEMENT_GET_CLASS (filter);

  filter->sinkpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "sink"), "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_timestamper_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_bufferalloc_function (filter->sinkpad,
      GST_DEBUG_FUNCPTR (gst_timestamper_bufferalloc));

  filter->srcpad =
      gst_pad_new_from_template (gst_element_class_get_pad_template (klass,
          "src"), "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_timestamper_chain));
  gst_pad_set_event_function (filter->sinkpad,GST_DEBUG_FUNCPTR(gst_timestamper_event));
  gst_pad_set_event_function (filter->srcpad,GST_DEBUG_FUNCPTR(gst_timestamper_event));
  filter->silent = TRUE;
  filter->fweos = TRUE;
  if (filter->valve==0) filter->valve = G_MAXINT64;
  //filter->patchrate=TRUE;
  if ((filter->rate<0.0001)&&(filter->rate>-0.0001))filter->rate=1.0;
  filter->offsetformat=-1;
  filter->lastoffset=0;
  filter->fwold=FALSE;
  filter->startoffset=G_MAXUINT64;
  filter->endoffset=G_MAXUINT64; 
  filter->drops=0;
  filter->evaldone=FALSE;
  filter->filename=g_strdup_printf("/tmp/stat.txt"); 
  filter->latsum=0;
  filter->bandwidth=0;
  filter->numparts=0;
  filter->numparts_current=0;
  filter->infoevery=1;
  filter->evnum=10;
}

static void
gst_timestamper_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gsttimestamper *filter = GST_TIMESTAMPER (object);

  switch (prop_id) {
    case ARG_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    case ARG_T1:
      filter->t1 = g_value_get_int (value);
      break;
    case ARG_T2:
      filter->t2 = g_value_get_int (value);
      break;  
    case ARG_VALVE:
      filter->valve = g_value_get_uint64 (value);
      if (filter->valvewait==0){
	filter->valvewait = filter->valve/2;
	}
      break; 
    case ARG_VALVEWAIT:
      filter->valvewait = g_value_get_uint64 (value);
      break; 
  
    case ARG_ADDOFFSET:
      filter->addoffset = g_value_get_int64 (value);
      break; 
    case ARG_RATE:
      filter->rate = g_value_get_double (value);
      break; 
    case ARG_OFFT:
      filter->offsetformat = g_value_get_int (value);
      break; 
    case ARG_CODEJIFFIES:
      filter->codejiffies = g_value_get_boolean (value);
      break;
/*
    case ARG_LATENCY:
      filter->proplatency = g_value_get_uint64 (value);
      break; 
    case ARG_TIMEBETWEEN:
      filter->proptimebetweenframes= g_value_get_uint64 (value);
      break;
*/
    case ARG_FWEOS:
      filter->fweos = g_value_get_boolean (value);
      break; 
    case ARG_FWOLD:
      filter->fweos = g_value_get_boolean (value);
      break; 
    case ARG_STARTOFFSET:
      filter->startoffset = g_value_get_uint64 (value);
      break;
    case ARG_ENDOFFSET:
      filter->endoffset = g_value_get_uint64 (value);
      break;    
    case ARG_STATFILE:
      if (filter->filename)g_free (filter->filename);
      filter->filename = g_value_dup_string (value);
	//todo mutex,unadvertise,advertise,unlock
      break;
    case ARG_BANDWIDTH:
      filter->bandwidth = g_value_get_uint64 (value);
      break;  
    case ARG_NUMPARTS:
      filter->numparts = g_value_get_uint64 (value);
    break;  
    case ARG_INFOEVERY:
      filter->infoevery = g_value_get_uint64 (value);
    break;
    case ARG_EVNUM:
      filter->evnum = g_value_get_uint64 (value);
    break;  
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_timestamper_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gsttimestamper *filter = GST_TIMESTAMPER (object);

  switch (prop_id) {
    case ARG_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case ARG_T1:
      g_value_set_int (value, filter->t1);	
      break;
    case ARG_T2:
      g_value_set_int (value, filter->t2);	
      break;
    case ARG_VALVE:
      g_value_set_uint64(value, filter->valve);
      break; 
    case ARG_VALVEWAIT:
      g_value_set_uint64(value, filter->valvewait);
      break; 
    case ARG_ADDOFFSET:
      g_value_set_int64(value, filter->addoffset);
      break;
    case ARG_RATE:
      g_value_set_double(value, filter->rate);
      break;
    case ARG_OFFT:
      g_value_set_int (value, filter->offsetformat);
      break; 
    case ARG_CODEJIFFIES:
      g_value_set_boolean (value, filter->codejiffies);
      break;
    case ARG_LATENCY:
      g_value_set_double(value, filter->proplatency);
      break; 
    case ARG_TIMEBETWEEN:
      g_value_set_double(value, filter->proptimebetweenframes);
      break; 
    case ARG_DROPRATE:
      g_value_set_double(value, filter->droprate);
      break;
    case ARG_FWEOS:
      g_value_set_boolean (value, filter->fweos);
      break;
    case ARG_LASTOFFSET:
      g_value_set_uint64(value, filter->lastoffset);
      break;
    case ARG_FWOLD:
      g_value_set_boolean (value, filter->fwold);
      break;
    case ARG_STARTOFFSET:
      g_value_set_uint64 (value,filter->startoffset);
      break;
    case ARG_ENDOFFSET:
      g_value_set_uint64 (value,filter->endoffset);
      break;   
    case ARG_STATFILE:   
      g_value_set_string (value,filter->filename);    
      break; 
    case ARG_BANDWIDTH:
      g_value_set_uint64 (value,filter->bandwidth);
      break;        
    case ARG_NUMPARTS:
      g_value_set_uint64 (value, filter->numparts);
    break;  
    case ARG_INFOEVERY:
      g_value_set_uint64 (value, filter->infoevery);
    break; 
    case ARG_BANDWIDTH_METER:
      g_value_set_double (value,filter->bandwidthmeter);
      break;
    case ARG_EVNUM:
      g_value_set_uint64 (value, filter->evnum);
    break;  
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_timestamper_set_caps (GstPad * pad, GstCaps * caps)
{
  Gsttimestamper *filter;
  GstPad *otherpad;

  filter = GST_TIMESTAMPER (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

  return gst_pad_set_caps (otherpad, caps);
}

static GstStateChangeReturn
gst_timestamper_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gsttimestamper *filter =  GST_TIMESTAMPER(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
	g_print ("Start Playing now!!");
	struct timeval now;
  	gettimeofday(&now, NULL);
      	filter->timediff2=((now.tv_sec*1000000000ULL)+(now.tv_usec*1000ULL));
	g_print ("timediff2 is %llu!! Difference to timediff1: %llu\n Difference to Timediff0 is %llu\n",(long long unsigned int)filter->timediff2,(long long unsigned int)(filter->timediff2-filter->timediff1), (long long unsigned int)(filter->timediff2-filter->timediff0));
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      //gst_my_filter_free_memory (filter);
      break;
    default:
      break;
  }
 return ret;
}

 static GstFlowReturn
gst_timestamper_bufferalloc (GstPad * pad, guint64 offset, guint size, GstCaps * caps,
    GstBuffer ** buf)
{
  Gsttimestamper *filter;
  filter = GST_TIMESTAMPER (GST_PAD_PARENT (pad));
  GstFlowReturn result;
  if (!filter->silent){
	gchar* cs=gst_caps_to_string(caps);
	g_print ("Timestamper: Bufferalloc has been caled with Caps: %s\n",cs);
	g_free(cs);
  }


  /* Forward to src pad, without setting caps on the src pad */
  result = gst_pad_alloc_buffer (filter->srcpad, offset, size, caps, buf);

  if (!filter->silent){
	gchar* cs=gst_caps_to_string((*buf)->caps);
	g_print ("Timestamper: Got Buffer with Caps: %s\n",cs);
	g_free(cs);
  }
  
  
  return result;
}







static gboolean jiffiesinduration(guint64 * duration,guint32 clockval){//codes the 
 if (*duration>G_MAXUINT32){
	g_print("Warning, duration too high to encode value in it\n");
	return FALSE;
	}
 *duration=*duration+(((guint64)G_MAXUINT32)*((guint64)clockval));
 return TRUE; 
}
static guint32 getjiffiesback(guint64 * duration){
 if (*duration<G_MAXUINT32){
	g_print("Warning, duration too low to decode value from it: %llu\n",(long long unsigned int)*duration);
	return FALSE;
	}
 guint64 temp;
 temp=*duration;
 *duration=*duration%((guint64)G_MAXUINT32);
 return (temp/(guint64)G_MAXUINT32);
}

/* chain function
 * this function does the actual processing
 */

static int
mycmp(const void *p1, const void *p2)
{
    if ((*((guint64*)p1))==(*((guint64*)p2)))return 0;
    else if ((*((guint64*)p1))>(*((guint64*)p2)))return 1;
    else return -1;
}


static GstFlowReturn
gst_timestamper_chain (GstPad * pad, GstBuffer * buf)
{
  Gsttimestamper *filter;
  //sleep(1);
  filter = GST_TIMESTAMPER (GST_OBJECT_PARENT (pad));
  if (buf->offset==G_MAXUINT64){
    buf->offset=filter->index++;
    buf->offset_end=filter->index;
  }
  if (filter->bandwidth!=0){
    usleep((buf->size*1000000ULL)/filter->bandwidth);
  } 
  if ((!filter->fwold)&&(buf->offset<filter->lastoffset)){
    if (filter->silent == FALSE) g_print("Timestamper dropping %llu\n",(long long unsigned int)buf->offset);
    gst_buffer_unref(buf);
    return GST_FLOW_OK;
  }
  if (filter->numparts!=0){
    if ((filter->numparts_current==0)||(filter->numparts_offset!=buf->offset)){
      filter->numparts_current=filter->numparts;
      filter->numparts_offset=buf->offset;
    }
    filter->numparts_current--;
    if (filter->numparts_current>0){
      gst_buffer_unref(buf);
      return GST_FLOW_OK;
    }
  }


  filter->bufcnt++;
  //g_print(gst_caps_to_string(buf->caps));
  if(G_UNLIKELY(filter->timediff1==0)){
	filter->streambase1=buf->timestamp;
	filter->lastdrop=buf->timestamp;
	struct timeval now;
  	gettimeofday(&now, NULL);
      	filter->timediff1=((now.tv_sec*1000000000ULL)+(now.tv_usec*1000ULL));
	
	}


 if(filter->t2==-1){//set timestamp and duration to -1
	buf->timestamp=-1;
	buf->duration=-1;
	//buf->offset=546672;
	//buf->offset_end=43435;;
        filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);
}
 if(filter->t1==-1){//set timestamp to now...duration to 
 	struct timeval now;
	guint64 timenow;
  	gettimeofday(&now, NULL);
	timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);
	buf->timestamp=timenow+filter->addoffset;
	if ((buf->duration==-1)&&(filter->bufcnt!=1)) buf->duration=buf->timestamp-filter->lasttimestamp;//not the first
	filter->lasttimestamp=buf->timestamp;
        filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);
}

if((filter->t1==0)&&(filter->t2==1)){//from streamtime to 
 	struct timeval now;
	guint64 timenow;
  	gettimeofday(&now, NULL);
	timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);
	buf->timestamp=timenow+filter->addoffset;
	if (filter->codejiffies==TRUE){
		g_print("Will encode jiffies in Duration\n");
		jiffiesinduration(&buf->duration,(guint32)clock());
	}
	else g_print("Will not code jiffies in Duration, because filter->codejiffies not set\n");
        filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);	
}







  if((filter->t1==0)&&(filter->t2==2)){//from streamtime-displaytime to NTP-Displaytime//place element next to SRC
	guint64 tsbu;//timestampbackup
	tsbu=buf->timestamp;
	buf->timestamp+=filter->addoffset+filter->timediff0-filter->streambase0;
	//sleep(1);
	
	//wait if buffer is too fast
	struct timeval now;
	guint64 timenow,t;
	t=0;
  	gettimeofday(&now, NULL);
	timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);
	//g_print ("Will calculate with buf-timestamp %llu   timenow: %llu\n ",buf->timestamp,timenow);

	GstClock* ticktack;
	ticktack=gst_element_get_clock (gst_pad_get_parent_element(gst_pad_get_peer(filter->srcpad)));
	if(ticktack==NULL){
		g_print ("Ticktack gibbet nich !!!!!!!!!!!\n \n!\n");
	}
        else {	GstClockTime ourneighbour;
		ourneighbour=gst_clock_get_time(ticktack);
		if ((timenow-filter->oldntp)>10000000000ULL){
			;
			double peerdowndiff,peerupdiff;
			peerupdiff=(double)((buf->timestamp-filter->addoffset-filter->timediff0+filter->streambase0)-filter->oldpeerup)/(double)(timenow-filter->oldntp);
			peerdowndiff=(double)(ourneighbour-filter->oldpeerdown)/(double)(timenow-filter->oldntp);
			filter->oldntp=timenow;
			filter->oldpeerdown=ourneighbour;
			filter->oldpeerup=buf->timestamp-filter->addoffset-filter->timediff0+filter->streambase0;//
			g_print ("Timediff to sink %llu. Rate of downstream element: %f Rate of upstream element:%f\n",(long long unsigned int)(timenow-ourneighbour), peerdowndiff,peerupdiff);
		}
		
	}	
	//delay stream if it is too fast
	if((double)(tsbu-filter->streambase0)/filter->rate/*+filter->addoffset*/>(double)(timenow-filter->timediff0+filter->valve)){
		g_print ("Sleep ratebased  Buffer-Timestamp: %llu Streambase %llu Timenow: %llu Starttime0: %llu  !!!!!!!!!!!\n \n!\n",(long long unsigned int)tsbu,(long long unsigned int)filter->streambase0,(long long unsigned int)timenow,(long long unsigned int)filter->timediff0);
		usleep(filter->valvewait/1000);
	}
	else if (FALSE){
	//TODO	 Generate Qos
	}
	//
	//for testing
        g_print ("Will push buffer in mode 0-2 with timestamp %llu  (original %llu) The valve-value t: %llu  Threshold: %llu\n Time of our neighbour %llu\n",(long long unsigned int)buf->timestamp,(long long unsigned int)tsbu ,(long long unsigned int)t,(long long unsigned int)filter->valve,(long long unsigned int)filter->oldpeerdown);
        filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);
	}
  if((filter->t1==1)&&(filter->t2==1)){//Check Delay
	if (G_UNLIKELY(filter->bufcnt==1)){//initialize both statistical evaluation sets
		g_print ("Init evaluators\n");
		filter->latency=malloc(sizeof(struct avuint64));
		filter->timedistance=malloc(sizeof(struct avuint64));
		filter->dropdistance=malloc(sizeof(struct avuint64));
		filter->dropin=malloc(sizeof(struct avuint64));
		filter->dropout=malloc(sizeof(struct avuint64));
		filter->bw=malloc(sizeof(struct avuint64));
		initavuint64(filter->latency,64, 6);	
		initavuint64(filter->timedistance,64, 6);
		initavuint64(filter->dropdistance,8, 3);
		if (filter->codejiffies==TRUE){
			filter->jiffielatency=malloc(sizeof(struct avuint64));
			initavuint64(filter->jiffielatency,64, 6);			
		}
		initavuint64(filter->dropin,100, 0);
		initavuint64(filter->dropout,100, 0);	
		initavuint64(filter->bw,filter->evnum, 0);	
		if ((filter->startoffset<G_MAXUINT64)&&(filter->endoffset<G_MAXUINT64))filter->latarray=g_malloc((filter->endoffset-filter->startoffset)*sizeof(guint64));
		if ((filter->startoffset<G_MAXUINT64)&&(filter->endoffset<G_MAXUINT64))filter->latarrayp=g_malloc((filter->endoffset-filter->startoffset)*sizeof(guint64*));
		int i;
		for(i=0;i<(filter->endoffset-filter->startoffset);i++){
		  filter->latarray[i]=G_MAXUINT64;
		  filter->latarrayp[i]=&(filter->latarray[i]);
		}
	}

        struct timeval now;
	guint64 timenow;
  	gettimeofday(&now, NULL);
	timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);
	guint64 timesincelast,latency,numdrops;
//how many drops?
	numdrops=buf->offset-filter->lastoffset-1;
	avuint64app(filter->dropin,1);
	avuint64app(filter->dropout,numdrops);
	filter->droprate=(double)(filter->dropout->sum)/(double)(filter->dropin->sum+filter->dropout->sum);
        //g_print ("Detected %llu drops...\n",(long long unsigned int)numdrops);
	//while (numdrops>0){
	//  numdrops--;
	//  avuint64app(filter->dropdistance, (timenow-filter->lastdrop));
	 // filter->lastdrop=timenow;
	//}
	timesincelast=timenow-filter->lastarrive;
	gdouble fps;
	fps=0;
	if (timesincelast>0) fps=(gdouble)1000000000ULL/(gdouble)timesincelast;
        latency=timenow+filter->addoffset-buf->timestamp;
	//g_print ("Buffer age %llu. Last buffer arrived before %llu ns. Framerate %f\n",latency,timesincelast, (float)fps);
	avuint64app(filter->latency, latency);
	avuint64app(filter->timedistance, timesincelast);
	avuint64appt(filter->bw, buf->size, timenow);
	filter->proplatency=(gdouble)getavuint64(filter->latency)/1000000.0;
	filter->proptimebetweenframes=(gdouble)getavuint64(filter->timedistance)/1000000.0;
	if ((buf->offset>=filter->startoffset)&&(buf->offset<filter->endoffset)){
	  filter->latsum+=(latency/1000);
	  filter->latarray[buf->offset-filter->startoffset]=latency;
	  filter->drops+=numdrops;
	}
	if ((buf->offset>=filter->endoffset)&&(!filter->evaldone)){
	  int i;
	  Calcstat    res;
          CalcStatInit(&res);
	  for (i=0;i<(filter->endoffset-filter->startoffset);i++)if (G_MAXUINT64!=filter->latarray[i])CalcStat (((double)filter->latarray[i])/1000000.0 , &res); 
	  
	  
	  g_print ("Writing Evaluation...\n");
	  //sleep(1);
	  if (buf->offset>filter->endoffset)filter->drops+=(buf->offset-filter->lastoffset);
	  qsort((void*)filter->latarrayp[0], (size_t)(filter->endoffset-filter->startoffset),sizeof(guint64), mycmp);
	  FILE* gantfd=fopen(filter->filename,"w");  
	  fprintf(gantfd,"Anzahl              = %lf\n",res.ANZAHL); // Anzahl              = 3.000000
	  fprintf(gantfd,"Summe               = %lf\n",res.SUMME ); // Summe               = 6.000000
	  fprintf(gantfd,"Summe der Quadrate  = %lf\n",res.QUADSM); // Summe der Quadrate  = 14.000000
	  fprintf(gantfd,"Mittelwert          = %lf\n",res.MW    ); // Mittelwert      = 2.000000
	  fprintf(gantfd,"Quadrat der Mittelw.= %lf\n",res.QUDEMI); // Quadrat der Mittelw.= 12.000000
	  fprintf(gantfd,"Standartabweichnung = %lf\n",res.STDN  ); // Standartabweichnung = 0.816497
	  fprintf(gantfd,"Standartabw.(n-1)   = %lf\n",res.STDN1 ); // Standartabw.(n-1)   = 1.000000
	  fprintf(gantfd,"Varianz             = %lf\n",res.VAR   ); // Varianz             = 0.666667
	  fprintf(gantfd,"1 sigma             = %lf\n",res._1SIG ); // 1 sigma             = 0.500000
	  fprintf(gantfd,"2.sigma             = %lf\n",res._2SIG ); // 2.sigma             = 0.707107
	  fprintf(gantfd,"3.sigma             = %lf\n",res._3SIG ); // 3.sigma             = 1.414214
	  fprintf(gantfd,"RMS                 = %lf\n",res.RMS   ); // RMS             = 2.160247
	  fprintf(gantfd,"MAX                 = %lf\n",res.MAX   ); // Max                 = 3.000000
	  fprintf(gantfd,"MIN                 = %lf\n",res.MIN   ); // Min                 = 1.000000
	  fprintf(gantfd,"Maximum Betrag      = %lf\n",res.AMAX  ); // abs(Max)            = 3.000000
	  
	  
	  
	  
	  //fprintf(gantfd,"Overall %llu us\n",(long long unsigned int)filter->latsum);//write the sorted Array
	  for(i=0;i<(filter->endoffset-filter->startoffset);i++){
	    fprintf(gantfd,"%llu\n",(long long unsigned int)(*filter->latarrayp[i]));//write the sorted Array
	  }  
	  
	  fclose(gantfd);
	  filter->evaldone=TRUE;
	}
	filter->bandwidthmeter=getavuint64bw(filter->bw);
	if (((filter->endoffset==G_MAXUINT64)||(buf->offset>filter->endoffset))&&(G_UNLIKELY((filter->bufcnt+10)%filter->infoevery==0))){
		;g_print ("Trying Evaluation...\n");
		//g_print ("Evaluation, This frame: latency %llu  Overall latency %llu ms. Time betreen frames %llu ms Time between drops %llu ms\n",(long long unsigned int)(latency/1000000ULL),(long long unsigned int)getavuint64(filter->latency)/1000000,(long long unsigned int)getavuint64(filter->timedistance)/1000000,(long long unsigned int)getavuint64(filter->dropdistance)/1000000);
                g_print ("Timestamper %s, This frame(%llu): latency %.2f  Overall latency %.2f ms. Time betreen frames %.2f ms Passrate %.2f. Time between drops=%.2f ms FPS=%.2f Bandwidth %f\n",GST_ELEMENT_NAME((GstElement*)filter),(unsigned long long int)buf->offset,((float)latency/1000000.0),filter->proplatency,filter->proptimebetweenframes,(float)filter->rate,(float)getavuint64(filter->dropdistance)/1000000.0,fps,filter->bandwidthmeter);

	}
	buf->timestamp+=filter->addoffset;
	filter->lastarrive=timenow;

	if (filter->codejiffies==TRUE){
		guint64 time;
		time=(guint64 )(clock()-getjiffiesback(&buf->duration));
		avuint64app(filter->jiffielatency, time);
		if (G_UNLIKELY((filter->bufcnt+10)%1==0)) g_print ("Evaluation of jiffie based difference %llu, medium %llu \n",(long long unsigned int)time,(long long unsigned int)getavuint64(filter->jiffielatency));
	}
	filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);
}
  if((filter->t1==2)&&(filter->t2==2)){
        buf->timestamp+=filter->addoffset;
	buf->offset=buf->timestamp;
	buf->offset_end=buf->offset+buf->duration;

	//
	struct timeval now;
	guint64 timenow;
  	gettimeofday(&now, NULL);
	timenow=((guint64)now.tv_sec*1000000000ULL)+((guint64)now.tv_usec*1000ULL);
	gint64 lateness;//in us
	lateness=((gint64)(timenow/1000))-((gint64)(buf->timestamp/1000));
	g_print ("Buffer is %f s late. Timestamp offset: %llu\n",((gdouble)lateness/1000000.0),(long long unsigned int)(buf->timestamp-filter->lasttimestamp));
	//g_print ("Comparing timestamps: buf-timestamp %llu   timenow: %llu\n ",buf->timestamp,timenow);

	if ((filter->timediff2!=0)&&(42)){//TODO:add parameter patchinitialdelay
		buf->timestamp-=filter->timediff2-filter->timediff0;//substract initial starttime
	}
	filter->lasttimestamp=buf->timestamp;
        filter->lastoffset=buf->offset;
	return gst_pad_push (filter->srcpad, buf);
	}
  //if (buf->offset<10) buf->timestamp+=11000000000ULL;

  if (filter->silent == FALSE)
    g_print ("Doing nothing with the buffer....%llu.......t1=%d t2=%d\n",(long long unsigned int)buf->timestamp,filter->t1,filter->t2);

  /* just push out the incoming buffer without touching it */
  if (!filter->fweos)GST_BUFFER_FLAG_SET(buf,GST_BUFFER_FLAG_IN_CAPS);//just a dirty hack to keep multifdsink from dropping it
  if (0){
	GST_BUFFER_FLAG_UNSET(buf,GST_BUFFER_FLAG_DISCONT);
	g_print ("Warning, unsetting DISCONT flag!!\n");
  }
  filter->lastoffset=buf->offset;
  return gst_pad_push (filter->srcpad, buf);
}

//my event function...
static gboolean
gst_timestamper_event (GstPad  *pad, GstEvent * event)
{
  Gsttimestamper *filter;
  gboolean ret=TRUE;
  GstPad *otherpad;

  filter = GST_TIMESTAMPER (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;





  switch (GST_EVENT_TYPE (event)) {
    case GST_EVENT_NEWSEGMENT:
      /* maybe save and/or update the current segment (e.g. for output
       * clipping) or convert the event into one in a different format
       * (e.g. BYTES to TIME) or drop it and set a flag to send a newsegment
       * event in a different format later */
      g_print ("Got Newsegment Event ");
      //ret=TRUE;
        if ((filter->t1==0)&&(filter->t2==0)){
                gboolean update;
                gdouble rate;
                gdouble applied_rate;
                GstFormat format;
                gint64 start;
                gint64 stop;
                gint64 position;
		filter->index=0;
		gst_event_parse_new_segment_full    (event,&update,&rate,&applied_rate,&format,&start,&stop,&position);
		g_print ("old offset format: %d\n",format);
		filter->streambase0=start;
		GstEvent* newevent;
		//newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);
		format=GST_FORMAT_TIME;
		newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);            
		ret = gst_pad_push_event (otherpad, newevent);
		g_print ("Event sent with modified timestamp, format is %d\n",format);
		gst_event_unref(event);
	}
        if ((filter->t1==-1)&&(filter->t2==1)){
                gboolean update;
                gdouble rate;
                gdouble applied_rate;
                GstFormat format;
                gint64 start;
                gint64 stop;
                gint64 position;
		filter->index=0;
		gst_event_parse_new_segment_full    (event,&update,&rate,&applied_rate,&format,&start,&stop,&position);
		g_print ("old offset format: %d\n",format);
		filter->streambase0=start;
		GstEvent* newevent;
		//newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);
		format=GST_FORMAT_TIME;
		stop=G_MAXUINT64;
		newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);            
		ret = gst_pad_push_event (otherpad, newevent);
		g_print ("Event sent with modified stop, format is %d\n",format);
		gst_event_unref(event);
	}
	else if (filter->t2==2){//set clock of downstream element 
		//patch start-time
		//alles bullshit!!
		filter->streambase0=event->timestamp;
                gboolean update;
                gdouble rate;
                gdouble applied_rate;
                GstFormat format;
                gint64 start;
                gint64 stop;
                gint64 position;
		filter->index=0;
		gst_event_parse_new_segment_full    (event,&update,&rate,&applied_rate,&format,&start,&stop,&position);
		g_print ("old offset format: %d\n",format);
		filter->streambase0=start;
		GstEvent* newevent;
		struct timeval now;
  		gettimeofday(&now, NULL);
      		filter->timediff0=(((gint64)now.tv_sec*1000000000ULL)+((gint64)now.tv_usec*1000ULL));
		start=position=filter->timediff0;
		//newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);
		if (filter->offsetformat>-1) format=filter->offsetformat;
		newevent=gst_event_new_new_segment_full(update,rate,applied_rate,format,start,stop,position);
		newevent->timestamp=filter->timediff0;             
		ret = gst_pad_push_event (otherpad, newevent);
		g_print ("Event sent with modified timestamp, format is %d\n",format);
		gst_event_unref(event);
		

		//struct timeval now;
  		//gettimeofday(&now, NULL);
      		//filter->timediff0=event->timestamp=((now.tv_sec*1000000000ULL)+(now.tv_usec*1000ULL));

		ret = gst_pad_push_event (otherpad, event);
	}
      else ret = gst_pad_push_event (otherpad, event);
      break;
    case GST_EVENT_EOS:
      /* end-of-stream, we should close down all stream leftovers here */
      //gst_my_filter_stop_processing (filter);
      if(filter->fweos) ret = gst_pad_push_event (otherpad, event);
      else g_print ("Not forwarding EOS!!\n");
      break;
    case GST_EVENT_FLUSH_STOP:
      //gst_my_filter_clear_temporary_buffers (filter);
      g_print ("Flush not implemented!!\n");
      ret = gst_pad_push_event (otherpad, event);
      break;   
    //case GST_EVENT_LATENCY:
    //  g_print ("Event latency!!\n");
    //  ret = gst_pad_push_event (otherpad, event);
    //  break;  
    case GST_EVENT_QOS:
	;
	gdouble proportion;
        GstClockTimeDiff diff;
        GstClockTime timestamp;
        gst_event_parse_qos(event,&proportion,&diff,&timestamp);
      //g_print ("Event QOS!! Proportion:%f Diff: %lld Timestamp:%llu\n",proportion,(long long int)diff,(long long unsigned int)timestamp);
      ret = gst_pad_push_event (otherpad, event);
      break;  
    case GST_EVENT_NAVIGATION:
      g_print ("Event Navigation!!\n");
      ret = gst_pad_push_event (otherpad, event);
      break; 
    default:
      if (pad == filter->srcpad) g_print ("Unexpected upstream Event Name %s\n",GST_EVENT_TYPE_NAME(event));
	else g_print ("Unexpected downstream Event %s\n",GST_EVENT_TYPE_NAME(event));
        ret = gst_pad_event_default (pad, event);
      //ret=TRUE;
      break;
  }

  gst_object_unref (filter);
  return ret;
}



/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 *
 * exchange the string 'plugin' with your elemnt name
 */
static gboolean
plugin_init (GstPlugin * plugin)
{
  /* exchange the strings 'plugin' and 'Template plugin' with your
   * plugin name and description */
  GST_DEBUG_CATEGORY_INIT (gst_timestamper_debug, "timestamper",
      0, "Timestamp adjuster");

  return gst_element_register (plugin, "timestamper",
      GST_RANK_NONE, GST_TYPE_TIMESTAMPER);
}

/* this is the structure that gstreamer looks for to register plugins
 *
 * exchange the strings 'plugin' and 'Template plugin' with you plugin name and
 * description
 */
GST_PLUGIN_DEFINE (GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "timestamper",
    "Timestamp adjuster",
    plugin_init, VERSION, "LGPL", "GStreamer", "http://gstreamer.net/")
