/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2011 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-trigger
 *
 * FIXME:Describe trigger here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v videotestsrc ! trigger mode=2 rcvport=8888 name=dropper ! queue ! siftextractor ! trigger name=hugo mode=3 host=127.0.0.1 sendport=8888 ! fakesink
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gsttrigger.h"

GST_DEBUG_CATEGORY_STATIC (gst_trigger_debug);
#define GST_CAT_DEFAULT gst_trigger_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT,
  PROP_MODE,
  PROP_HOSTNAME,
  PROP_SPORT,
  PROP_RPORT,
  PROP_LASTDROP,
  PROP_PREVIEW,
  PROP_PREFIRE,
  PROP_BACKUP
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gsttrigger, gst_trigger, GstElement,
    GST_TYPE_ELEMENT);

static void gst_trigger_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_trigger_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_trigger_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_trigger_chain (GstPad * pad, GstBuffer * buf);
static GstStateChangeReturn gst_trigger_change_state (GstElement *element, GstStateChange transition);
/* GObject vmethod implementations */

static void
gst_trigger_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "trigger",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the trigger's class */
static void
gst_trigger_class_init (GsttriggerClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_trigger_set_property;
  gobject_class->get_property = gst_trigger_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_MODE,
      g_param_spec_int ("mode", "Mode", "Operation mode, 2=get triggered by TCP, 3=Trigger others by TCP!",0,4,3, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_SPORT,
      g_param_spec_int ("sendport", "Sendport", "To what port should we send Packets",0,65535,8888, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_RPORT,
      g_param_spec_int ("rcvport", "Rcvport", "To what port should we listen",0,65535,8888, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, PROP_HOSTNAME,
      g_param_spec_string ("host", "Host", "Host adress of Instance to trigger",
          "127.0.0.1", G_PARAM_READWRITE)); 
  g_object_class_install_property (gobject_class, PROP_LASTDROP,
      g_param_spec_boolean ("lastdrop", "Lastdrop", "Whether last buffer has been dropped",
          FALSE, G_PARAM_READABLE));
  g_object_class_install_property (gobject_class, PROP_PREVIEW,
      g_param_spec_uint ("current-level-buffers", "Current-level-buffers", "Number of buffers currently in work",0,G_MAXUINT,0, G_PARAM_READABLE)); 
    g_object_class_install_property (gobject_class, PROP_PREFIRE,
      g_param_spec_uint64 ("prefire", "Prefire", "Prefire in usec, do this in order to increase throughput",0,G_MAXUINT,0, G_PARAM_READWRITE)); 
    g_object_class_install_property (gobject_class, PROP_BACKUP,
      g_param_spec_uint64 ("backup", "Backup", "Backup Send intervall if not getting triggered in us",0,G_MAXUINT,1000000, G_PARAM_READWRITE));   
  
  
 gstelement_class->change_state = gst_trigger_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_trigger_init (Gsttrigger * filter,
    GsttriggerClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_trigger_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_trigger_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
  filter->lsd=-1;
  filter->csd=-1;
  filter->ssd=-1;
  filter->offset=-1;
  filter->mode=2;
  filter->backupsend=1000000;
  filter->remoteadress=NULL;
  filter->mutex=g_mutex_new();
  filter->preview=-1;

  filter->lastrcv=0; 
  filter->lastdur=G_MAXINT64; 
  filter->prefire=0;//in us  
  filter->prefired=FALSE;
  filter->lastsendprev=0;
}

static void
gst_trigger_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gsttrigger *filter = GST_TRIGGER (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
   case PROP_MODE:
      filter->mode = g_value_get_int (value);
      break;
   case PROP_SPORT:
      filter->sport = g_value_get_int (value);
      break;
    case PROP_RPORT:
      filter->rport = g_value_get_int (value);
      break;
    case PROP_HOSTNAME:
      if (filter->remoteadress)g_free (filter->remoteadress);
      filter->remoteadress = g_value_dup_string (value);
      //filter->firsttime=TRUE;
      break;
    case PROP_PREFIRE: 
      filter->prefire = g_value_get_uint64 (value);
      break; 
    case PROP_BACKUP: 
      filter->backupsend = g_value_get_uint64 (value);
      break;       
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_trigger_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gsttrigger *filter = GST_TRIGGER (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    case PROP_MODE:
      g_value_set_int (value, filter->mode);
      break;
    case PROP_SPORT:
      g_value_set_int (value, filter->sport);
      break;
    case PROP_RPORT:
      g_value_set_int (value, filter->rport);
      break;
    case PROP_HOSTNAME:
      g_value_set_string (value, filter->remoteadress);
      break;
    case PROP_LASTDROP:
      g_value_set_boolean (value, filter->lastdropped);
      break;
    case PROP_PREVIEW:
      g_mutex_lock(filter->mutex);
      if (filter->preview>-1)g_value_set_uint (value, (int)0);
      else if (filter->csd>-1){
	 unsigned char buffer;
	 int bytes;
	 bytes = recv(filter->csd, (char*)&buffer, 1, 0);
	 if ((bytes<0)&&(errno!=EWOULDBLOCK)&&(errno!=EAGAIN)){
	   g_value_set_uint (value, (int)0);
	   close(filter->csd);
	   filter->csd=-1;
	 }
	 else if (bytes<0) g_value_set_uint (value, (int)1);
	 else if ((bytes>0)&&(buffer==filter->offset)){
	   g_print("Trigger %s received offset %d previously in property request!\n",GST_ELEMENT_NAME(filter),buffer);
	   g_value_set_uint (value, (int)0);
	   filter->preview=buffer;
	 }
	 struct timeval now;
          guint64 time;
          gettimeofday(&now, NULL);
          time=((guint64)now.tv_sec*1000000ULL)+((guint64)now.tv_usec);
          //g_print("Trigger %s received has sent before %llu\n",GST_ELEMENT_NAME(filter),(long long unsigned int)(time-filter->lastsend));
          if ((time-filter->lastsend)>filter->backupsend)g_value_set_uint (value, (int)0);
 
      }
      else g_value_set_uint (value, (int)0);
      g_mutex_unlock(filter->mutex);      
      break;
    case PROP_PREFIRE: 
       g_value_set_uint64 (value,filter->prefire);
    break;  
    case PROP_BACKUP: 
       g_value_set_uint64 (value,filter->backupsend);
    break;     
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_trigger_set_caps (GstPad * pad, GstCaps * caps)
{
  Gsttrigger *filter;
  GstPad *otherpad;

  filter = GST_TRIGGER (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;
  gst_object_unref (filter);

  return gst_pad_set_caps (otherpad, caps);
}

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_trigger_chain (GstPad * pad, GstBuffer * buf)
{
  Gsttrigger *filter;
  gboolean drop=FALSE;
  filter = GST_TRIGGER (GST_OBJECT_PARENT (pad));
  g_print("Trigger %s received in Mode %d,  lsd=%d csd=%d ssd=%d!\n",GST_ELEMENT_NAME(filter),filter->mode,filter->lsd,filter->csd,filter->ssd);
  
  struct timeval now;
  guint64 time;
  gettimeofday(&now, NULL);
  time=((guint64)now.tv_sec*1000000ULL)+((guint64)now.tv_usec);
  
  if (filter->mode==0){
    filter->drop++;
    if (filter->drop%filter->passrate!=0){
      drop=TRUE;
    }
  }
  else if (filter->mode==1){//delay mode
      struct timeval now;
      guint64 time;
      gettimeofday(&now, NULL);
      time=((guint64)now.tv_sec*1000000ULL)+((guint64)now.tv_usec);
      if ((time-filter->lastsend)>filter->backupsend){
	drop=FALSE;
	}
     //sleep(usleep);
  }
  else if (filter->mode==2){//trigger mode
     drop=TRUE;
     if (filter->lsd==-1){
       filter->saddr.sin_family = AF_INET;
       filter->saddr.sin_port = htons(filter->rport);
       filter->saddr.sin_addr.s_addr = htons(INADDR_ANY);
       filter->lsd=socket(AF_INET, SOCK_STREAM, 0);  
       int yes=1;
       if (setsockopt(filter->lsd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1){
	 g_print("Trigger %s setsocketoption failed, errno: %d!\n",GST_ELEMENT_NAME(filter),filter->reportto ,errno);
       }
       fcntl(filter->lsd, F_SETFL, O_NONBLOCK); 
       if (0!=bind(filter->lsd, (struct sockaddr *) &filter->saddr,
              sizeof(filter->saddr))){
	 g_print("Trigger %s bind failed, errno: %d!\n",GST_ELEMENT_NAME(filter),filter->reportto ,errno);
          close(filter->lsd);
	 filter->lsd=-1;
       }
       else listen(filter->lsd, 1);
     }
     if ((filter->csd==-1)&&(filter->lsd>-1)){
       //if (0==listen(lsd, 1)){
	  socklen_t clen = sizeof(filter->clientaddr);
 	  g_print("Trigger %s listening for clients\n",GST_ELEMENT_NAME(filter));
	  filter->csd = accept(filter->lsd, (struct sockaddr*) &filter->clientaddr, &clen);
	  if ((!filter->csd)&&(errno!=EWOULDBLOCK)&&(errno!=EAGAIN)){
	    //how to close listen?
	    close(filter->lsd);
	    filter->lsd=-1;
	  }
          fcntl(filter->csd, F_SETFL, O_NONBLOCK); 
      }
      //now server should be up and running and client may be connected
      g_mutex_lock(filter->mutex);
      if (filter->preview>-1){
	drop=FALSE;
	filter->preview=-1;
      }
      else if (filter->csd>-1){
         unsigned char buffer;
	 int bytes=-1;
	 do{
	    bytes = recv(filter->csd, (char*)&buffer, 1, 0);
	    if ((bytes<0)&&(errno!=EWOULDBLOCK)&&(errno!=EAGAIN)){
		close(filter->csd);
		filter->csd=-1;
		}
	    //if ((bytes>0)&&(buffer==filter->offset)){
	    if ((bytes>0)){      
	      g_print("Trigger %s received offset %d that it has been waiting for!\n",GST_ELEMENT_NAME(filter),buffer);
	      filter->lastrcv=time;
	      if (! filter->prefired){
		filter->lastdur=time-filter->lastsend;	
		drop=FALSE;
	      }
	      else{
		filter->lastdur=time-filter->lastsendprev;
		filter->prefired=FALSE;
	      }
	    }
	    else if (bytes>0) g_print("Trigger received nonfitting\n");   
	    if (bytes==0){
		g_print("Trigger %s received ZERO bytes!\n",GST_ELEMENT_NAME(filter));
		close(filter->csd);
		filter->csd=-1;		
	    }
	 } while (drop && (bytes>0));
      }
      g_mutex_unlock(filter->mutex);
      //PRefire-Sendung
      if ((filter->prefire>0) && drop && !filter->prefired && (time>(filter->lastsend+filter->lastdur-filter->prefire))){
	if (!filter->silent)g_print("Trigger %s is doing prefire, because %llu overdue !\n",GST_ELEMENT_NAME(filter),(long long unsigned int)(time-(filter->lastsend+filter->lastdur-filter->prefire)));
	filter->prefired=TRUE;
	drop=FALSE;
      }
      //else g_print("Trigger %s is not doing prefire, because only passed %llu overdue , lastdur =%llu lastsend=%llu!\n",GST_ELEMENT_NAME(filter),(long long unsigned int)(time-(filter->lastsend)),
	//(long long unsigned int)filter->lastdur,(long long unsigned int)filter->lastsend);
      //Backup-Sendung

      g_print("Trigger %s received has sent before %llu\n",GST_ELEMENT_NAME(filter),(long long unsigned int)(time-filter->lastsend));
      if ((time-filter->lastsend)>filter->backupsend){
	filter->prefired=FALSE;
	drop=FALSE;
      }
    }

    
  
  
  else if(filter->mode==3){//write nonblocking
    if (filter->ssd==-1){//try to set up connection if inactive
	struct addrinfo *result, *rp;
 
	bzero(&filter->serveraddr, sizeof(filter->serveraddr));
	filter->reportto=(char*)malloc(100);
	sprintf(filter->reportto,filter->remoteadress);
	//if (inet_pton(AF_INET, filter->reportto, &(filter->serveraddr.sin_addr))<1) g_print("Trigger %s warning, could not parse %s!\n",GST_ELEMENT_NAME(filter),filter->reportto);
	//filter->serveraddr.sin_family = AF_INET;
	//filter->serveraddr.sin_port = htons(filter->sport);
	//bind(filter->ssd, (struct sockaddr*)&(filter->serveraddr), sizeof (struct sockaddr));
	char portchar[30];
	sprintf(portchar,"%u",filter->sport);
         if(getaddrinfo(filter->reportto,portchar,NULL, &result)!=0)g_print("Trigger %s warning, could not parse %s!\n",GST_ELEMENT_NAME(filter),filter->reportto);
	rp=result;
	//for (rp = result; rp != NULL; rp = rp->ai_next) g_print("Beep\n");
	//rp=rp->ai_next;	
	//rp=rp->ai_next;		
	filter->ssd= socket(AF_INET, SOCK_STREAM, 0);
	//filter->ssd= socket(rp->ai_family,rp->ai_socktype, rp->ai_protocol);
	//fcntl(filter->ssd, F_SETFL, O_NONBLOCK);
	//if(0!=connect(filter->ssd, (struct sockaddr*) &filter->serveraddr, sizeof(filter->serveraddr))){
	for (rp = result; rp != NULL; rp = rp->ai_next){  
	  if(0!=connect(filter->ssd, rp->ai_addr, rp->ai_addrlen)){ 
            g_print("Trigger %s connection to server %s failed, errno: %d!\n",GST_ELEMENT_NAME(filter),filter->reportto ,errno);
	    if (errno!=115){
              //close(filter->ssd);
	      //filter->ssd=-1;
	     }
	   else break;
	  }
	}
	freeaddrinfo(result);
    }
    if (filter->ssd!=-1){
        g_print("Trigger %s trying to send!\n",GST_ELEMENT_NAME(filter));
	unsigned char tosend=(unsigned char)(buf->offset%256);
	if (send(filter->ssd ,&tosend, 1, 0)!=1){
	  g_print("Trigger %s sending failed, errno: %d!\n",GST_ELEMENT_NAME(filter) ,errno);
	  if (errno!=115){
	   close(filter->ssd);
	   filter->ssd=-1;
          }
	}
    }	
  }

  /* just push out the incoming buffer without touching it */
  filter->lastdropped=drop;
  if (drop){
    gst_buffer_unref(buf);
    return GST_FLOW_OK;    
  }
  else {
    g_mutex_lock(filter->mutex);
    filter->preview=-1;
    g_mutex_unlock(filter->mutex); 
    filter->offset=(buf->offset%256);
    //gettimeofday
    struct timeval now;
    gettimeofday(&now, NULL);
    filter->lastsendprev=filter->lastsend;
    filter->lastsend=((guint64)now.tv_sec*1000000ULL)+((guint64)now.tv_usec);
    return gst_pad_push (filter->srcpad, buf);
  }
}

static GstStateChangeReturn gst_trigger_change_state (GstElement *element, GstStateChange transition){
    GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gsttrigger *filter;
  GError* err;
  filter =  GST_TRIGGER(element);

  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      //if (!filter->socketopen)setupsockets(filter);
      //filter->socketopen=TRUE;
      break;
    default:
      break;
  }

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      if (filter->csd>-1){
	 unsigned char buffer;
	 int bytes;
	 bytes = recv(filter->csd, (char*)&buffer, 1, 0);
	 while (recv(filter->csd, (char*)&buffer, 1, 0)>0){
	   g_print("Reading Rest\n");
	 }
      }
      if (filter->lsd)close(filter->lsd);
      filter->lsd=-1;
      if (filter->csd)close(filter->csd); 
      filter->csd=-1;
      if (filter->ssd){
	sleep(2);
	close(filter->ssd); 
         filter->ssd=-1;
      }
      //g_io_channel_close (filter->rchannel);
      //g_io_channel_unref(filter->rchannel);
      break;
    default:
      break;
  }
  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

 return ret;
  
  
  
  
}
/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
trigger_init (GstPlugin * trigger)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template trigger' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_trigger_debug, "trigger",
      0, "Template trigger");

  return gst_element_register (trigger, "trigger", GST_RANK_NONE,
      GST_TYPE_TRIGGER);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirsttrigger"
#endif

/* gstreamer looks for this structure to register triggers
 *
 * exchange the string 'Template trigger' with your trigger description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "trigger",
    "Template trigger",
    trigger_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
