/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2010 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-videojoin
 *
 * FIXME:Describe videojoin here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m videotestsrc ! videosplitter name=spl ! videojoin name=join ! ffmpegcolorspace ! ximagesink spl. ! join
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstvideojoin.h"
#include <string.h>

GST_DEBUG_CATEGORY_STATIC (gst_videojoin_debug);
#define GST_CAT_DEFAULT gst_videojoin_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink%d",
    GST_PAD_SINK,
    GST_PAD_REQUEST,
    GST_STATIC_CAPS ("ANY")
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("ANY")
    );

GST_BOILERPLATE (Gstvideojoin, gst_videojoin, GstElement,
    GST_TYPE_ELEMENT);

static void gst_videojoin_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_videojoin_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);
static GstCaps * gst_videojoinsrc_get_caps (GstPad *pad);
static gboolean gst_videojoin_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_videojoin_chain (GstPad * pad, GstBuffer * buf);

static GstPad *
gst_videojoin_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name);
static void
gst_videojoin_release_pad (GstElement * element, GstPad * pad);
static GstStateChangeReturn
gst_videojoin_change_state (GstElement *element, GstStateChange transition);
/* GObject vmethod implementations */

static void
gst_videojoin_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "videojoin",
    "Gst-Plugins-Robotix",
    "Joins Several Slices into one Video",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the videojoin's class */
static void
gst_videojoin_class_init (GstvideojoinClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_videojoin_set_property;
  gobject_class->get_property = gst_videojoin_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          TRUE, G_PARAM_READWRITE));
  gstelement_class->request_new_pad =
      GST_DEBUG_FUNCPTR (gst_videojoin_request_new_pad);
  gstelement_class->release_pad =
      GST_DEBUG_FUNCPTR (gst_videojoin_release_pad);
  gstelement_class->change_state=gst_videojoin_change_state;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_videojoin_init (Gstvideojoin * filter,
    GstvideojoinClass * gclass)
{
  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_videojoinsrc_get_caps));
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = TRUE;
  filter->mutex=g_mutex_new();
}

static void
gst_videojoin_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstvideojoin *filter = GST_VIDEOJOIN (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_videojoin_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstvideojoin *filter = GST_VIDEOJOIN (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

/* GstElement vmethod implementations */

gboolean gst_videojoin_modifycaps(GObject * object)
{
  Gstvideojoin *filter = GST_VIDEOJOIN (object);
  if (filter->srccaps) gst_caps_unref (filter->srccaps);
  filter->srccaps=NULL;
  if (GST_IS_CAPS(filter->sinkcaps))filter->srccaps=gst_caps_copy(filter->sinkcaps);
  else return FALSE;
  if (!filter->silent)g_print("Jideojoin: Before Src caps are %s.\n",gst_caps_to_string(filter->srccaps));
  GstStructure *structure=gst_caps_get_structure(filter->srccaps,0);
  if (!structure) g_print("Scheisse!\n");
  //gst_structure_fixate_field_nearest_int (structure,"width",filter->widthsmall);
  gst_structure_set(structure,"height",G_TYPE_INT,filter->height,NULL);
  gst_structure_set(structure,"width",G_TYPE_INT,filter->width,NULL);
  if (filter->vertical) gst_structure_set(structure,"overlap_v",G_TYPE_INT,0,NULL);//vertical will KEEP horizontal overlap 
  else gst_structure_set(structure,"overlap_h",G_TYPE_INT,0,NULL); //vice versa
  g_print("Now Source caps are %s.\n",gst_caps_to_string(filter->srccaps));
  return TRUE;
}




/* this function handles the link with other elements */
static gboolean
gst_videojoin_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstvideojoin *filter;
  //GstPad *otherpad;
  gboolean ret=TRUE;
  filter = GST_VIDEOJOIN (gst_pad_get_parent (pad));
  g_print("GstVideojoin set caps %s\n",gst_caps_to_string(caps));

  if (pad==filter->srcpad) return TRUE;
  g_mutex_lock(filter->mutex);
  GstStructure *structure = gst_caps_get_structure (caps, 0);
  gboolean bppset=FALSE;
  const gchar *mime;
  mime = gst_structure_get_name (structure);    
  if (gst_structure_get_int(structure,"bpp",&(filter->bpp))) bppset=TRUE;

  if (0 == strcmp (mime, "video/x-raw-yuv")){ 
	guint32 fourccval;
	if (!gst_structure_has_field(structure,"format")){
		g_print ("YUV No format field");
		ret=FALSE;
	}  
	if (!gst_structure_get_fourcc(structure,"format",&fourccval)){
		g_print ("FourCC could not be read");
		ret=FALSE;
	}
	if (fourccval==GST_MAKE_FOURCC ('Y', '8', '0', '0')){
	  if (!bppset) filter->bpp=8;
	  bppset=TRUE;
	  filter->unitw=filter->unith=1;
	}
	else if (fourccval==GST_MAKE_FOURCC ('Y', '4', '2', '2')){
	  if (!bppset) filter->bpp=16;
	  bppset=TRUE;
	  filter->unitw=2;
	  filter->unith=1;
	}
	else if (fourccval==GST_MAKE_FOURCC ('Y', 'U', 'Y', '2')){
	  if (!bppset) filter->bpp=16;
	  bppset=TRUE;
	  filter->unitw=2;
	  filter->unith=1;
	}
}
else if (0 == strcmp (mime, "video/x-raw-gray")){
filter->unitw=filter->unith=1;
}
else if (0 == strcmp (mime, "video/x-raw-rgb")){
filter->unitw=filter->unith=1;
}

  if (!gst_structure_get_int(structure,"width",&(filter->widthsmall))){
	g_print ("WARNING: Width not set!");
	ret=FALSE;
  }
  if (!gst_structure_get_int(structure,"height",&(filter->heightsmall))){
	g_print ("WARNING Height not set!");
	ret=FALSE;
  }
  if (!gst_structure_get_int(structure,"overlap_h",&(filter->overlap_h))){
	filter->overlap_h=0;
  }
  if (!gst_structure_get_int(structure,"overlap_v",&(filter->overlap_v))){
	filter->overlap_v=0;
  }
  if (filter->overlap_h>=filter->overlap_v) filter->vertical=FALSE;
  else filter->vertical=TRUE;

//calculate height of output
  if (filter->vertical){
     filter->height=filter->heightsmall;
     filter->width=(filter->num_sinkpads*filter->widthsmall)-((filter->num_sinkpads-1)*filter->overlap_v);
  }
  else{
     filter->width=filter->widthsmall;
     filter->height=(filter->num_sinkpads*filter->heightsmall)-((filter->num_sinkpads-1)*filter->overlap_h);          
  } 
  filter->sinkcaps=caps;
  gst_videojoin_modifycaps((GObject*)filter);  
  
  gst_object_unref (filter);

  ret=ret&&gst_pad_set_caps (filter->srcpad,filter->srccaps);
  g_mutex_unlock(filter->mutex);
  if (!ret)g_print("WARNING: Videojoin will return FALSE on setcaps!!!\n");
  return ret;
}

static GstCaps *
gst_videojoinsrc_get_caps (GstPad *pad){
  g_print("Videojoin src get caps!!\n");
  Gstvideojoin *filter;

  filter = GST_VIDEOJOIN (gst_pad_get_parent (pad));

  if (filter&&filter->srccaps)return gst_caps_copy(filter->srccaps);

  else return GST_CAPS_ANY;
}


static GstCaps *
gst_videojoinsink_get_caps (GstPad *pad){
  g_print("Videojoin sink get caps!!\n");
  Gstvideojoin *filter;
  GstCaps* result;
  filter = GST_VIDEOJOIN (gst_pad_get_parent (pad));

  //if (filter&&filter->sinkcaps)return gst_caps_copy(filter->sinkcaps);
  result=gst_pad_peer_get_caps (filter->srcpad);
  if (result) return result;
  else return GST_CAPS_ANY;
}
/*
static gboolean
intersect_caps_func (GstPad * pad, GValue * ret, GstPad * orig)
{

  if (pad != orig) {
    GstCaps *peercaps, *existing;

    existing = g_value_get_pointer (ret);
    peercaps = gst_pad_peer_get_caps (pad);
    if (peercaps == NULL)
      peercaps = gst_caps_new_any ();
    g_value_set_pointer (ret, gst_caps_intersect (existing, peercaps));
    gst_caps_unref (existing);
    gst_caps_unref (peercaps);
  }
  gst_object_unref (pad);
  return TRUE;
}

GstCaps *
gst_videojoinsink_get_caps (GstPad * pad)
{
  g_print("Videojoin get caps pad %s\n",gst_pad_get_name(pad));
  GstElement *element;
  GstCaps *caps, *intersected;
  GstIterator *iter;
  GstIteratorResult res;
  GValue ret = { 0, };

  g_return_val_if_fail (GST_IS_PAD (pad), NULL);

  //GST_CAT_DEBUG (GST_CAT_PADS, "proxying getcaps for %s:%s",
  //    GST_DEBUG_PAD_NAME (pad));

  element = gst_pad_get_parent_element (pad);
  if (element == NULL)
    return NULL;


  g_value_init (&ret, G_TYPE_POINTER);
  g_value_set_pointer (&ret, gst_caps_new_any ());

  iter = gst_element_iterate_pads (element);
  while (1) {
    res =
        gst_iterator_fold (iter, (GstIteratorFoldFunction) intersect_caps_func,
        &ret, pad);
    switch (res) {
      case GST_ITERATOR_RESYNC:

        if ((caps = g_value_get_pointer (&ret)))
          gst_caps_unref (caps);

        g_value_set_pointer (&ret, gst_caps_new_any ());
        gst_iterator_resync (iter);
        break;
      case GST_ITERATOR_DONE:

        goto done;
      default:

        if ((caps = g_value_get_pointer (&ret)))
          gst_caps_unref (caps);
        g_value_set_pointer (&ret, NULL);
        goto error;
    }
  }
done:
  gst_iterator_free (iter);

  gst_object_unref (element);

  caps = g_value_get_pointer (&ret);
  g_value_unset (&ret);

  intersected = gst_caps_intersect (caps, gst_pad_get_pad_template_caps (pad));
  gst_caps_unref (caps);
  g_print("Videojoin get caps returning %s\n",gst_caps_to_string(caps));
  return intersected;


error:
  {
    g_warning ("Pad list returned error on element %s",
        GST_ELEMENT_NAME (element));
    gst_iterator_free (iter);
    gst_object_unref (element);
    return NULL;
  }
}
*/

static GstPad *
gst_videojoin_request_new_pad (GstElement * element, GstPadTemplate * temp,
    const gchar * name)
{
  gchar* tmp;
  GstPad* tempPad;
  Gstvideojoin *filter;
  filter = GST_VIDEOJOIN (element);
  g_print("Videojoin: New pad with name %s requested\n",name);

  //if (!name) tmp = g_strdup_printf ("src%d", filter->num_srcpads);
  //else tmp=g_strdup_printf ("%s", name);
  tmp = g_strdup_printf ("sink%d", filter->num_sinkpads);
  tempPad = gst_pad_new_from_static_template (&sink_factory, tmp);
  g_free (tmp);

  filter->sinkpads=g_list_append(filter->sinkpads,tempPad);
  filter->num_sinkpads++;
  //gst_videosplitter_update_overlap(filter,filter->overlap);
  //gst_videosplitter_modifycaps((GObject*)filter);
  gst_pad_set_getcaps_function (tempPad,
      GST_DEBUG_FUNCPTR (gst_videojoinsink_get_caps));
  //gst_pad_set_getcaps_function (tempPad,
  //                              GST_DEBUG_FUNCPTR(gst_pad_proxy_getcaps));
  gst_pad_set_setcaps_function (tempPad,
                                GST_DEBUG_FUNCPTR(gst_videojoin_set_caps));
  gst_pad_set_chain_function (tempPad,
                              GST_DEBUG_FUNCPTR(gst_videojoin_chain));
  gst_element_add_pad (GST_ELEMENT (filter), tempPad);
  return tempPad;
}



static void
gst_videojoin_release_pad (GstElement * element, GstPad * pad)
{

g_print("releasing pads not supported yet:)\n");
for(;;);
}




/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_videojoin_chain (GstPad * pad, GstBuffer * buf)
{
  Gstvideojoin *filter;
  GstFlowReturn result=GST_FLOW_OK;
  GList *temp;
  gint i=0,index=-1;
  filter = GST_VIDEOJOIN (GST_OBJECT_PARENT (pad));
  g_mutex_unlock(filter->mutex);
  if (!filter->bufferinwork) {
    filter->bufferinwork=gst_buffer_new_and_alloc (filter->width*filter->height*filter->bpp/8);
    gst_buffer_set_caps(filter->bufferinwork,filter->srccaps);
  }
  //
  //calculate my index
  for(temp=filter->sinkpads;((temp)&&(index<0));temp=g_list_next(temp)){
    if (temp->data==pad) index=i;
    i++;
  }
  if G_UNLIKELY(index<0) {
    g_print ("Could not find Index.\n");
    return GST_FLOW_ERROR;
  }

  if (filter->silent == FALSE)
    g_print ("OK Wrinting part %d.\n",index);

  if (filter->vertical){
    for (i=0;i<filter->height;i++){
    memcpy(filter->bufferinwork->data+((i*(filter->width)+index*(filter->widthsmall-filter->overlap_v))*filter->bpp/8),buf->data+i*(filter->widthsmall*filter->bpp/8),filter->widthsmall*filter->bpp/8);    
    }
  }
  else{ //horizontal
    memcpy(filter->bufferinwork->data+((index*filter->bpp*filter->width*(filter->heightsmall-filter->overlap_h))/8),buf->data,filter->width*filter->heightsmall*filter->bpp/8);
  }

  filter->completed++;
  
  if (filter->completed==filter->num_sinkpads){
    if (!filter->silent)g_print("Videojoin now pushing buffer!!!\n");
    result=gst_pad_push (filter->srcpad, filter->bufferinwork);
    filter->bufferinwork=NULL;
    filter->completed=0;
  }
  gst_buffer_unref(buf);
  g_mutex_unlock(filter->mutex);
  return result;
}

static GstStateChangeReturn
gst_videojoin_change_state (GstElement *element, GstStateChange transition){
  GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
  Gstvideojoin *filter =  GST_VIDEOJOIN(element);
  
  switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:
      break;
    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
      break;
    default:
      break;
  }

  ret = GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
  if (ret == GST_STATE_CHANGE_FAILURE)
    return ret;

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_NULL:
      if GST_IS_BUFFER(filter->bufferinwork)gst_buffer_unref(filter->bufferinwork);
      filter->bufferinwork=NULL;
      filter->completed=0;     
      break;
    default:
      break;
  }
 return ret;
}
/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
videojoin_init (GstPlugin * videojoin)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template videojoin' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_videojoin_debug, "videojoin",
      0, "Template videojoin");

  return gst_element_register (videojoin, "videojoin", GST_RANK_NONE,
      GST_TYPE_VIDEOJOIN);
}



/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstvideojoin"
#endif

/* gstreamer looks for this structure to register videojoins
 *
 * exchange the string 'Template videojoin' with your videojoin description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "videojoin",
    "Template videojoin",
    videojoin_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
