/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2012 Hannes Bistry <<user@hostname.org>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-waterdetect
 *
 * FIXME:Describe waterdetect here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! waterdetect ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <gst/video/video.h>
#include <opencv/cv.h>
#include "gstwaterdetect.h"

GST_DEBUG_CATEGORY_STATIC (gst_waterdetect_debug);
#define GST_CAT_DEFAULT gst_waterdetect_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
  PROP_0,
  PROP_SILENT
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */
static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (GST_VIDEO_CAPS_BGRx)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ("coordinates-list")
    );

GST_BOILERPLATE (Gstwaterdetect, gst_waterdetect, GstElement,
    GST_TYPE_ELEMENT);

static void gst_waterdetect_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_waterdetect_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static GstCaps *
gst_waterdetect_get_caps (GstPad *pad);
static gboolean gst_waterdetect_set_caps (GstPad * pad, GstCaps * caps);
static GstFlowReturn gst_waterdetect_chain (GstPad * pad, GstBuffer * buf);

/* GObject vmethod implementations */

static void
gst_waterdetect_base_init (gpointer gclass)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);

  gst_element_class_set_details_simple(element_class,
    "waterdetect",
    "FIXME:Generic",
    "FIXME:Generic Template Element",
    "Hannes Bistry <<user@hostname.org>>");

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&sink_factory));
}

/* initialize the waterdetect's class */
static void
gst_waterdetect_class_init (GstwaterdetectClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;

  gobject_class->set_property = gst_waterdetect_set_property;
  gobject_class->get_property = gst_waterdetect_get_property;

  g_object_class_install_property (gobject_class, PROP_SILENT,
      g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
          FALSE, G_PARAM_READWRITE));
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_waterdetect_init (Gstwaterdetect * filter,
    GstwaterdetectClass * gclass)
{
  filter->sinkpad = gst_pad_new_from_static_template (&sink_factory, "sink");
  gst_pad_set_setcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_waterdetect_set_caps));
  gst_pad_set_getcaps_function (filter->sinkpad,
                                GST_DEBUG_FUNCPTR(gst_waterdetect_get_caps));
  gst_pad_set_chain_function (filter->sinkpad,
                              GST_DEBUG_FUNCPTR(gst_waterdetect_chain));

  filter->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
  gst_pad_set_getcaps_function (filter->srcpad,
                                GST_DEBUG_FUNCPTR(gst_waterdetect_get_caps));

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  filter->silent = FALSE;
}

static void
gst_waterdetect_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  Gstwaterdetect *filter = GST_WATERDETECT (object);

  switch (prop_id) {
    case PROP_SILENT:
      filter->silent = g_value_get_boolean (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_waterdetect_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  Gstwaterdetect *filter = GST_WATERDETECT (object);

  switch (prop_id) {
    case PROP_SILENT:
      g_value_set_boolean (value, filter->silent);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static GstCaps *
gst_waterdetect_get_caps (GstPad *pad){
  GstCaps * ret;
  Gstwaterdetect* filter;

  filter = GST_WATERDETECT (GST_OBJECT_PARENT (pad));
  ret=gst_caps_copy(gst_pad_get_pad_template_caps(pad));
  return ret; 
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_waterdetect_set_caps (GstPad * pad, GstCaps * caps)
{
  Gstwaterdetect *filter;
  gboolean ret=TRUE;

  filter = GST_WATERDETECT (gst_pad_get_parent (pad));
  if (pad == filter->srcpad)return ret;

  GstStructure *structure=gst_caps_get_structure(caps,0);
  if (!gst_structure_get_int(structure,"width",&filter->imagesize.width)) ret=FALSE;
  if (!gst_structure_get_int(structure,"height",&filter->imagesize.height)) ret=FALSE;
  if (!gst_structure_get_int(structure,"bpp",&filter->bpp)) ret=FALSE;
  gst_object_unref (filter);

  return ret;
}

int max(int a,int b)
{
 if(a>b)
 return a;
 else
 return b;
}

int min(int a,int b)
{
 if(a>b)
 return b;
 else
 return b;
}

/****BEGIN***The function which compute the structure content of a patch*****BEGIN********/

float comvar(IplImage *P){
	double s1=0,s2=0;
	double u=0,var=0;

	uchar* data=(uchar*)P->imageData;
	int step=P->widthStep/sizeof(uchar);
	int j=0;
	int i=0;
	for (j=0;j<P->height;j++){
			  for (i=0;i<P->width;i++){
				  s1+=data[j*step+i];
			  }
		  }
	u=s1/(P->width*P->height);

	data=(uchar*)P->imageData;
	for (j=0;j<P->height;j++){
		for (i=0;i<P->width;i++){
			  float a=data[j*step+i]-u;
			  s2+=a*a;
			  }
		  }

	return var=sqrt(s2/(P->width*P->height));

}

int UnkObj_height=20;
int UnkObj_width=30;
int maxiter=10;

/* chain function
 * this function does the actual processing
 */
static GstFlowReturn
gst_waterdetect_chain (GstPad * pad, GstBuffer * buf)
{
  Gstwaterdetect *filter;





  filter = GST_WATERDETECT (GST_OBJECT_PARENT (pad));

  if (filter->silent == FALSE)
    g_print ("I'm plugged, therefore I'm in.\n");
  //generate the IplImage...
  IplImage *process;
  process=cvCreateImageHeader(filter->imagesize, IPL_DEPTH_8U,(filter->bpp/8));
  process->imageData=(char*)buf->data;
  //now we have a the Image, and for example detect the cirvles...
  //CvMemStorage* storage = cvCreateMemStorage(0);
  //CvSeq* features=NULL;
	IplImage *image=cvCreateImage(
			cvSize(process->width,process->height),
			IPL_DEPTH_8U,1);
	cvCvtColor(process,image,CV_RGB2GRAY);

	/*------Detect area of the Image-------*/
	int d1=0,d2=0,d3=0,d4=0;
	d1=maxiter*UnkObj_height;
	d2=image->height-maxiter*UnkObj_height;

	d3=maxiter*UnkObj_width;
	d4=image->width-maxiter*UnkObj_width;


	int skip[4]={0};

	/*------num_rect:the number of rectangle-------*/
	int num_rect=0;

	/*------UnkObj_position:the vector contains the coordinates of rectangle-----
	 * -----the order is like:(x1,y1,x2,y2,x1,y1,x2,y2,....)--------------------
	 * ----so the size of UnkObj_position is 4*num_rect-------------------------*/
	int UnkObj_position[99]={0};

	/*------BEGIN of Detection-------*/
	int j=d1;
	int i=d3;

	for (j=d1;j<d2;j+=UnkObj_height){
		for(i=d3;i<d4;i+=UnkObj_width){
			if ((i>skip[0])&&(i<skip[2])&&(j>skip[1])&&(j<skip[3]))
				continue;

			CvRect box0;
			box0=cvRect(i,j,UnkObj_width,UnkObj_height);
			IplImage *p0=cvCreateImage(cvSize(UnkObj_width,UnkObj_height),IPL_DEPTH_8U,1);
			cvSetImageROI(image,box0);
			cvCopy(image,p0,NULL);
			cvResetImageROI(image);

			float var0;
			var0=comvar(p0);

			int iter=2;
			for (iter=2;iter<maxiter-2;iter++){

				float var[4]={0};
				int num_simi=0;

				CvRect box1;
				box1=cvRect(i,j-iter*UnkObj_height,UnkObj_width,UnkObj_height);
				IplImage *p1=cvCreateImage(cvSize(UnkObj_width,UnkObj_height),IPL_DEPTH_8U,1);
				cvSetImageROI(image,box1);
				cvCopy(image,p1,NULL);
				cvResetImageROI(image);

				CvRect box2;
				box2=cvRect(i+iter*UnkObj_width,j,UnkObj_width,UnkObj_height);
				IplImage *p2=cvCreateImage(cvSize(UnkObj_width,UnkObj_height),IPL_DEPTH_8U,1);
				cvSetImageROI(image,box2);
				cvCopy(image,p2,NULL);
				cvResetImageROI(image);

				CvRect box3;
				box3=cvRect(i,j+iter*UnkObj_height,UnkObj_width,UnkObj_height);
				IplImage *p3=cvCreateImage(cvSize(UnkObj_width,UnkObj_height),IPL_DEPTH_8U,1);
				cvSetImageROI(image,box3);
				cvCopy(image,p3,NULL);
				cvResetImageROI(image);

				CvRect box4;
				box4=cvRect(i-iter*UnkObj_width,j,UnkObj_width,UnkObj_height);
				IplImage *p4=cvCreateImage(cvSize(UnkObj_width,UnkObj_height),IPL_DEPTH_8U,1);
				cvSetImageROI(image,box4);
				cvCopy(image,p4,NULL);
				cvResetImageROI(image);

				var[0]=comvar(p1);
				var[1]=comvar(p2);
				var[2]=comvar(p3);
				var[3]=comvar(p4);

				int k=0;
				for (k=0;k<4;k++){
					if (var0-var[k]>30)
						num_simi=num_simi+1;
				}

				if (num_simi>3)
				{
					num_rect=num_rect+1;

					if (num_rect>24){
						return -1;
					}

					UnkObj_position[4*(num_rect-1)]=i-iter*UnkObj_width;
					UnkObj_position[4*(num_rect-1)+1]=j-iter*UnkObj_height;
					UnkObj_position[4*(num_rect-1)+2]=i+(iter+1)*UnkObj_width;
					UnkObj_position[4*(num_rect-1)+3]=j+(iter+1)*UnkObj_height;

					skip[0]=i-iter*UnkObj_width;
					skip[1]=j-iter*UnkObj_height;
					skip[2]=i+(iter+2)*UnkObj_width;
					skip[3]=j+(iter+2)*UnkObj_height;

					cvReleaseImage(&p0);
					cvReleaseImage(&p1);
					cvReleaseImage(&p2);
					cvReleaseImage(&p3);
					cvReleaseImage(&p4);

					break;
				}
				else{
					cvReleaseImage(&p0);
					cvReleaseImage(&p1);
					cvReleaseImage(&p2);
					cvReleaseImage(&p3);
					cvReleaseImage(&p4);
				}
			}

		}
	}

	/*------End of Detection-------*/

	/*---BEGIN----Detection of Overlap of the rectangle-------BEGIN------*/
	/*If there are overlap of the two rectangles, fuse the two rectangles into one big rectangle,
	 * which is stored in latter rectangle, and then set the former rectangle to zero*/
	int c=1;
	for (c=1;c<num_rect;c++){

		int x21=UnkObj_position[c*4];
		int x22=UnkObj_position[c*4+1];
		int x11=UnkObj_position[(c-1)*4+2];
		int x12=UnkObj_position[(c-1)*4+3];

		if ((x21<x11)&&(x22<x12)){

			UnkObj_position[c*4]=min(UnkObj_position[(c-1)*4],
					UnkObj_position[c*4]);
			UnkObj_position[c*4+1]=min(UnkObj_position[(c-1)*4+1],
					UnkObj_position[c*4+1]);
			UnkObj_position[c*4+2]=max(UnkObj_position[(c-1)*4+2],
					UnkObj_position[c*4+2]);
			UnkObj_position[c*4+3]=max(UnkObj_position[(c-1)*4+3],
					UnkObj_position[c*4+3]);

			UnkObj_position[(c-1)*4]=0;
			UnkObj_position[(c-1)*4+1]=0;
			UnkObj_position[(c-1)*4+2]=0;
			UnkObj_position[(c-1)*4+3]=0;

			x21=UnkObj_position[c*4];
			x22=UnkObj_position[c*4+1];
			x11=UnkObj_position[(c-2)*4+2];
			x12=UnkObj_position[(c-2)*4+3];


			if ((x21<x11)&&(x22<x12)){

				UnkObj_position[c*4]=min(UnkObj_position[(c-2)*4],
						UnkObj_position[c*4]);
				UnkObj_position[c*4+1]=min(UnkObj_position[(c-2)*4+1],
						UnkObj_position[c*4+1]);
				UnkObj_position[c*4+2]=max(UnkObj_position[(c-2)*4+2],
									UnkObj_position[c*4+2]);
				UnkObj_position[c*4+3]=max(UnkObj_position[(c-2)*4+3],
									UnkObj_position[c*4+3]);

				UnkObj_position[(c-2)*4]=0;
				UnkObj_position[(c-2)*4+1]=0;
				UnkObj_position[(c-2)*4+2]=0;
				UnkObj_position[(c-2)*4+3]=0;

			}
		}
	}
	/*---END----Detection of Overlap of the rectangle-------END------*/







  //prepare the output buffer
  int num_features=num_rect;
  GstBuffer *outcoord;
  if (num_features>0) outcoord=gst_buffer_new_and_alloc(sizeof(struct coordinates)*num_features);
  else outcoord=gst_buffer_new_and_alloc(0);
  for( i = 0; i < num_features; i++ ){
    ((struct coordinates*)(outcoord->data))[i].x1=(gdouble)UnkObj_position[i*4];
    ((struct coordinates*)(outcoord->data))[i].y1=(gdouble)UnkObj_position[i*4+1];
    ((struct coordinates*)(outcoord->data))[i].x2=(gdouble)UnkObj_position[i*4+2];
    ((struct coordinates*)(outcoord->data))[i].y2=(gdouble)UnkObj_position[i*4+3];
  }

  //cvReleaseMat(&image);
	/*-----END--------Draw the rectangle----------END------------*/

  




  GstCaps *tempcaps=gst_caps_copy(gst_pad_get_pad_template_caps(filter->srcpad));
  gst_buffer_set_caps(outcoord,tempcaps);
  gst_caps_unref(tempcaps);
  outcoord->timestamp=buf->timestamp;
  outcoord->duration=buf->duration;
  outcoord->offset=buf->offset;
  outcoord->offset_end=buf->offset_end;
  g_print("Pushing srccoord buffer in facemode with timestamp %lu\n",outcoord->offset);
  cvReleaseImageHeader(&process);
  cvReleaseImage(&image);
  gst_buffer_unref(buf);
  return gst_pad_push (filter->srcpad, outcoord);

}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
waterdetect_init (GstPlugin * waterdetect)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template waterdetect' with your description
   */
  GST_DEBUG_CATEGORY_INIT (gst_waterdetect_debug, "waterdetect",
      0, "Template waterdetect");

  return gst_element_register (waterdetect, "waterdetect", GST_RANK_NONE,
      GST_TYPE_WATERDETECT);
}

/* PACKAGE: this is usually set by autotools depending on some _INIT macro
 * in configure.ac and then written into and defined in config.h, but we can
 * just set it ourselves here in case someone doesn't use autotools to
 * compile this code. GST_PLUGIN_DEFINE needs PACKAGE to be defined.
 */
#ifndef PACKAGE
#define PACKAGE "myfirstwaterdetect"
#endif

/* gstreamer looks for this structure to register waterdetects
 *
 * exchange the string 'Template waterdetect' with your waterdetect description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "waterdetect",
    "Template waterdetect",
    waterdetect_init,
    VERSION,
    "LGPL",
    "GStreamer",
    "http://gstreamer.net/"
)
